diff --git a/NEWS b/NEWS
index 9108d6aa6..6e8946f29 100644
--- a/NEWS
+++ b/NEWS
@@ -123,6 +123,14 @@ Changes to the Mercury language
 
         :- pragma memo(predname/arity, [disable_warning_if_ignored])]).
 
+Changes to the Mercury compiler
+-------------------------------
+
+* If a command line first enables an optimization (say opt1), and then
+  sets the optimization level to N with -O<N>, then the compiler will now
+  keep opt1 enabled even if opt1 is not normally enabled at optimization
+  level N.
+
 NEWS for Mercury 20.06.1
 ========================
 
diff --git a/compiler/accumulator.m b/compiler/accumulator.m
index bc391ec11..e703b8c23 100644
--- a/compiler/accumulator.m
+++ b/compiler/accumulator.m
@@ -179,6 +179,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
@@ -246,12 +247,12 @@
 accu_transform_proc(proc(PredId, ProcId), PredInfo, !ProcInfo, !ModuleInfo,
         !Cookie) :-
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals,
-        optimize_constructor_last_call_accumulator, DoLCO),
+    globals.get_opt_tuple(Globals, OptTuple),
+    DoLCMC = OptTuple ^ ot_opt_lcmc_accumulator,
     globals.lookup_bool_option(Globals, fully_strict, FullyStrict),
     ( if
         should_attempt_accu_transform(!ModuleInfo, PredId, ProcId, PredInfo,
-            !ProcInfo, FullyStrict, DoLCO, Warnings)
+            !ProcInfo, FullyStrict, DoLCMC, Warnings)
     then
         globals.lookup_bool_option(Globals, very_verbose, VeryVerbose),
         (
@@ -360,10 +361,11 @@ generate_warning(ModuleInfo, VarSet, Warning, Msg) :-
     %
 :- pred should_attempt_accu_transform(module_info::in, module_info::out,
     pred_id::in, proc_id::in, pred_info::in, proc_info::in, proc_info::out,
-    bool::in, bool::in, list(accu_warning)::out) is semidet.
+    bool::in, maybe_opt_lcmc_accumulator::in,
+    list(accu_warning)::out) is semidet.
 
 should_attempt_accu_transform(!ModuleInfo, PredId, ProcId, PredInfo,
-        !ProcInfo, FullyStrict, DoLCO, Warnings) :-
+        !ProcInfo, FullyStrict, DoLCMC, Warnings) :-
     proc_info_get_goal(!.ProcInfo, Goal0),
     proc_info_get_headvars(!.ProcInfo, HeadVars),
     proc_info_get_initial_instmap(!.ProcInfo, !.ModuleInfo,
@@ -377,7 +379,7 @@ should_attempt_accu_transform(!ModuleInfo, PredId, ProcId, PredInfo,
     list.length(Rec, M),
 
     should_attempt_accu_transform_2(!ModuleInfo, PredId, PredInfo, !ProcInfo,
-        HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCO,
+        HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCMC,
         RecCallIds, C, M, Rec, Warnings).
 
     % should_attempt_accu_transform_2 takes a list of locations of the
@@ -393,18 +395,19 @@ should_attempt_accu_transform(!ModuleInfo, PredId, ProcId, PredInfo,
     %
 :- pred should_attempt_accu_transform_2(module_info::in, module_info::out,
     pred_id::in, pred_info::in, proc_info::in, proc_info::out,
-    list(prog_var)::in, instmap::in, top_level::in, bool::in, bool::in,
+    list(prog_var)::in, instmap::in, top_level::in, bool::in,
+    maybe_opt_lcmc_accumulator::in,
     list(accu_goal_id)::in, accu_goal_store::in, int::in, list(hlds_goal)::in,
     list(accu_warning)::out) is semidet.
 
 should_attempt_accu_transform_2(!ModuleInfo, PredId, PredInfo, !ProcInfo,
-        HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCO,
+        HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCMC,
         [Id | Ids], C, M, Rec, Warnings) :-
     proc_info_get_vartypes(!.ProcInfo, VarTypes0),
     identify_out_and_out_prime(!.ModuleInfo, VarTypes0, InitialInstMap,
         Id, Rec, HeadVars, Out, OutPrime, HeadToCallSubst, CallToHeadSubst),
     ( if
-        accu_stage1(!.ModuleInfo, VarTypes0, FullyStrict, DoLCO, Id, M, C,
+        accu_stage1(!.ModuleInfo, VarTypes0, FullyStrict, DoLCMC, Id, M, C,
             Sets),
         accu_stage2(!.ModuleInfo, !.ProcInfo, Id, C, Sets, OutPrime, Out,
             VarSet, VarTypes, Accs, BaseCase, BasePairs, Substs, CS,
@@ -416,7 +419,7 @@ should_attempt_accu_transform_2(!ModuleInfo, PredId, PredInfo, !ProcInfo,
         Warnings = WarningsPrime
     else
         should_attempt_accu_transform_2(!ModuleInfo, PredId, PredInfo,
-            !ProcInfo, HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCO,
+            !ProcInfo, HeadVars, InitialInstMap, TopLevel, FullyStrict, DoLCMC,
             Ids, C, M, Rec, Warnings)
     ).
 
@@ -681,10 +684,11 @@ identify_out_and_out_prime(ModuleInfo, VarTypes, InitialInstMap, GoalId,
     % Stage 1 is responsible for identifying which goals are associative,
     % which can be moved before the recursive call and so on.
     %
-:- pred accu_stage1(module_info::in, vartypes::in, bool::in, bool::in,
-    accu_goal_id::in, int::in, accu_goal_store::in, accu_sets::out) is semidet.
+:- pred accu_stage1(module_info::in, vartypes::in, bool::in,
+    maybe_opt_lcmc_accumulator::in, accu_goal_id::in, int::in,
+    accu_goal_store::in, accu_sets::out) is semidet.
 
-accu_stage1(ModuleInfo, VarTypes, FullyStrict, DoLCO, GoalId, M, GoalStore,
+accu_stage1(ModuleInfo, VarTypes, FullyStrict, DoLCMC, GoalId, M, GoalStore,
         Sets) :-
     GoalId = accu_goal_id(Case, K),
     NextGoalId = accu_goal_id(Case, K + 1),
@@ -706,13 +710,13 @@ accu_stage1(ModuleInfo, VarTypes, FullyStrict, DoLCO, GoalId, M, GoalStore,
         not set.is_empty(Update)
     ),
     (
-        DoLCO = no,
+        DoLCMC = do_not_opt_lcmc_accumulator,
         % If LCMC is not turned on, then there must be no construction
         % unifications after the recursive call.
         set.is_empty(Construct),
         set.is_empty(ConstructAssoc)
     ;
-        DoLCO = yes
+        DoLCMC = opt_lcmc_accumulator
     ).
 
     % For each goal after the recursive call decide which set
diff --git a/compiler/add_pragma_type_spec.m b/compiler/add_pragma_type_spec.m
index d12a75b3a..6f417d156 100644
--- a/compiler/add_pragma_type_spec.m
+++ b/compiler/add_pragma_type_spec.m
@@ -32,6 +32,7 @@
 :- import_module hlds.pred_table.
 :- import_module hlds.vartypes.
 :- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
@@ -103,8 +104,8 @@ add_pragma_type_spec_for_pred(TSInfo0, Context, PredId, !ModuleInfo, !QualInfo,
         map.map_values_only(reset_imported_structure_sharing_reuse,
             Procs1, Procs),
         module_info_get_globals(!.ModuleInfo, Globals),
-        globals.lookup_bool_option(Globals, user_guided_type_specialization,
-            DoTypeSpec),
+        globals.get_opt_tuple(Globals, OptTuple),
+        DoTypeSpec = OptTuple ^ ot_spec_types_user_guided,
         globals.lookup_bool_option(Globals, smart_recompilation, Smart),
         % XXX Should check whether smart recompilation has been disabled?
         ( if
@@ -118,7 +119,7 @@ add_pragma_type_spec_for_pred(TSInfo0, Context, PredId, !ModuleInfo, !QualInfo,
             % problems with differing output for the recompilation tests
             % in debugging grades.
 
-            ( DoTypeSpec = yes
+            ( DoTypeSpec = spec_types_user_guided
             ; not pred_info_is_imported(PredInfo0)
             ; Smart = yes
             )
diff --git a/compiler/add_trail_ops.m b/compiler/add_trail_ops.m
index 581f40e68..7333d39a8 100644
--- a/compiler/add_trail_ops.m
+++ b/compiler/add_trail_ops.m
@@ -56,13 +56,15 @@
 :- import_module hlds.
 :- import_module hlds.hlds_module.
 :- import_module hlds.hlds_pred.
+:- import_module libs.
+:- import_module libs.optimization_options.
 
 :- import_module bool.
 
 %-----------------------------------------------------------------------------%
 
-:- pred add_trail_ops(bool::in, bool::in, module_info::in,
-    proc_info::in, proc_info::out) is det.
+:- pred add_trail_ops(bool::in, maybe_gen_trail_ops_inline::in,
+    module_info::in, proc_info::in, proc_info::out) is det.
 
 %-----------------------------------------------------------------------------%
 %-----------------------------------------------------------------------------%
@@ -78,7 +80,6 @@
 :- import_module hlds.pred_table.
 :- import_module hlds.quantification.
 :- import_module hlds.vartypes.
-:- import_module libs.
 :- import_module libs.globals.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
@@ -113,7 +114,7 @@
                 trail_var_types     :: vartypes,
                 trail_module_info   :: module_info,
                 opt_trail_usage     :: bool,
-                inline_ops          :: bool
+                inline_ops          :: maybe_gen_trail_ops_inline
             ).
 
 add_trail_ops(OptTrailUsage, GenerateInline, ModuleInfo0, !Proc) :-
@@ -435,12 +436,12 @@ cases_add_trail_ops([Case0 | Cases0], [Case | Cases], !Info) :-
 gen_store_ticket(TicketVar, Context, SaveTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("store_ticket", detism_det, purity_impure,
             [TicketVar], instmap_delta_bind_var(TicketVar),
             Info ^ trail_module_info, Context, SaveTicketGoal)
     ;
-        GenerateInline =  yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(TicketVar,
             yes(foreign_arg_name_mode("Ticket", out_mode)),
             ticket_type, bp_native_if_possible)],
@@ -456,12 +457,12 @@ gen_store_ticket(TicketVar, Context, SaveTicketGoal, Info) :-
 gen_reset_ticket_undo(TicketVar, Context, ResetTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("reset_ticket_undo", detism_det, purity_impure,
             [TicketVar], instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, ResetTicketGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(TicketVar,
             yes(foreign_arg_name_mode("Ticket", in_mode)),
             ticket_type, bp_native_if_possible)],
@@ -477,12 +478,12 @@ gen_reset_ticket_undo(TicketVar, Context, ResetTicketGoal, Info) :-
 gen_reset_ticket_solve(TicketVar, Context, ResetTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("reset_ticket_solve", detism_det, purity_impure,
             [TicketVar], instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, ResetTicketGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(TicketVar,
             yes(foreign_arg_name_mode("Ticket", in_mode)),
             ticket_type, bp_native_if_possible)],
@@ -498,12 +499,12 @@ gen_reset_ticket_solve(TicketVar, Context, ResetTicketGoal, Info) :-
 gen_reset_ticket_commit(TicketVar, Context, ResetTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("reset_ticket_commit", detism_det, purity_impure,
             [TicketVar], instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, ResetTicketGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(TicketVar,
             yes(foreign_arg_name_mode("Ticket", in_mode)),
             ticket_type, bp_native_if_possible)],
@@ -519,12 +520,12 @@ gen_reset_ticket_commit(TicketVar, Context, ResetTicketGoal, Info) :-
 gen_prune_ticket(Context, PruneTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("prune_ticket", detism_det, purity_impure, [],
             instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, PruneTicketGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [],
         ForeignCode = "MR_prune_ticket();",
         trail_generate_foreign_proc("prune_ticket", purity_impure,
@@ -538,12 +539,12 @@ gen_prune_ticket(Context, PruneTicketGoal, Info) :-
 gen_discard_ticket(Context, DiscardTicketGoal, Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("discard_ticket", detism_det, purity_impure, [],
             instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, DiscardTicketGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [],
         ForeignCode = "MR_discard_ticket();",
         trail_generate_foreign_proc("discard_ticket", purity_impure,
@@ -558,13 +559,13 @@ gen_mark_ticket_stack(SavedTicketCounterVar, Context, MarkTicketStackGoal,
         Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("mark_ticket_stack", detism_det, purity_impure,
             [SavedTicketCounterVar],
             instmap_delta_bind_no_var, Info ^ trail_module_info,
             Context, MarkTicketStackGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(SavedTicketCounterVar,
             yes(foreign_arg_name_mode("TicketCounter", out_mode)),
             ticket_counter_type, bp_native_if_possible)],
@@ -581,12 +582,12 @@ gen_prune_tickets_to(SavedTicketCounterVar, Context, PruneTicketsToGoal,
         Info) :-
     GenerateInline = Info ^ inline_ops,
     (
-        GenerateInline = no,
+        GenerateInline = do_not_gen_trail_ops_inline,
         trail_generate_call("prune_tickets_to", detism_det, purity_impure,
             [SavedTicketCounterVar], instmap_delta_bind_no_var,
             Info ^ trail_module_info, Context, PruneTicketsToGoal)
     ;
-        GenerateInline = yes,
+        GenerateInline = gen_trail_ops_inline,
         Args = [foreign_arg(SavedTicketCounterVar,
             yes(foreign_arg_name_mode("TicketCounter", in_mode)),
             ticket_counter_type, bp_native_if_possible)],
diff --git a/compiler/code_info.m b/compiler/code_info.m
index 115b309ed..95f784440 100644
--- a/compiler/code_info.m
+++ b/compiler/code_info.m
@@ -37,6 +37,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module ll_backend.continuation_info.
 :- import_module ll_backend.global_data.
 :- import_module ll_backend.layout.
@@ -124,7 +125,7 @@
 :- pred get_emit_region_ops(code_info::in, add_region_ops::out) is det.
 :- pred get_opt_region_ops(code_info::in, bool::out) is det.
 :- pred get_auto_comments(code_info::in, bool::out) is det.
-:- pred get_lcmc_null(code_info::in, bool::out) is det.
+:- pred get_lcmc_null(code_info::in, maybe_opt_lcmc_null::out) is det.
 :- pred get_profile_memory(code_info::in, bool::out) is det.
 :- pred get_may_use_atomic_alloc(code_info::in,
     may_use_atomic_alloc::out) is det.
@@ -274,7 +275,7 @@
 
                 % The settings of --optimize-constructor-last-call-null,
                 % --profile-memory, and --use-atomic-cells.
-                cis_lcmc_null           :: bool,
+                cis_lcmc_null           :: maybe_opt_lcmc_null,
                 cis_profile_memory      :: bool,
                 cis_may_use_atomic_alloc :: may_use_atomic_alloc,
 
@@ -396,25 +397,25 @@ code_info_init(ModuleInfo, PredId, ProcId, PredInfo, ProcInfo,
         EmitTrailOps = do_not_add_trail_ops
     ),
     globals.lookup_bool_option(Globals, optimize_trail_usage, OptTrailOps),
-    globals.lookup_bool_option(Globals, region_analysis, UseRegions),
+    globals.get_opt_tuple(Globals, OptTuple),
+    UseRegions = OptTuple ^ ot_analyse_regions,
     (
-        UseRegions = yes,
+        UseRegions = analyse_regions,
         EmitRegionOps = add_region_ops
     ;
-        UseRegions = no,
+        UseRegions = do_not_analyse_regions,
         EmitRegionOps = do_not_add_region_ops
     ),
     globals.lookup_bool_option(Globals, optimize_region_ops, OptRegionOps),
     globals.lookup_bool_option(Globals, auto_comments, AutoComments),
-    globals.lookup_bool_option(Globals, optimize_constructor_last_call_null,
-        LCMCNull),
+    LCMCNull = OptTuple ^ ot_opt_lcmc_null,
     globals.lookup_bool_option(Globals, profile_memory, ProfileMemory),
-    globals.lookup_bool_option(Globals, use_atomic_cells, UseAtomicCells),
+    UseAtomicCells = OptTuple ^ ot_use_atomic_cells,
     (
-        UseAtomicCells = no,
+        UseAtomicCells = do_not_use_atomic_cells,
         InitMayUseAtomic = may_not_use_atomic_alloc
     ;
-        UseAtomicCells = yes,
+        UseAtomicCells = use_atomic_cells,
         InitMayUseAtomic = may_use_atomic_alloc
     ),
     globals.lookup_int_option(Globals, num_ptag_bits, NumPtagBitsInt),
@@ -512,12 +513,13 @@ init_exprn_opts(Globals) = ExprnOpts :-
         OptASM = no,
         ASM = do_not_have_asm_labels
     ),
-    globals.lookup_bool_option(Globals, static_ground_cells, OptSGCell),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptSGCell = OptTuple ^ ot_use_static_ground_cells,
     (
-        OptSGCell = yes,
+        OptSGCell = use_static_ground_cells,
         SGCell = have_static_ground_cells
     ;
-        OptSGCell = no,
+        OptSGCell = do_not_use_static_ground_cells,
         SGCell = do_not_have_static_ground_cells
     ),
     globals.lookup_bool_option(Globals, unboxed_float, OptUBF),
@@ -552,29 +554,28 @@ init_exprn_opts(Globals) = ExprnOpts :-
         OptUBI64s = no,
         UBI64s = do_not_have_unboxed_int64s
     ),
-    globals.lookup_bool_option(Globals, static_ground_floats, OptSGFloat),
+    OptSGFloat = OptTuple ^ ot_use_static_ground_floats,
     (
-        OptSGFloat = yes,
+        OptSGFloat = use_static_ground_floats,
         SGFloat = have_static_ground_floats
     ;
-        OptSGFloat = no,
+        OptSGFloat = do_not_use_static_ground_floats,
         SGFloat = do_not_have_static_ground_floats
     ),
-    globals.lookup_bool_option(Globals, static_ground_int64s, OptSGInt64s),
+    OptSGInt64s = OptTuple ^ ot_use_static_ground_int64s,
     (
-        OptSGInt64s = yes,
+        OptSGInt64s = use_static_ground_int64s,
         SGInt64s = have_static_ground_int64s
     ;
-        OptSGInt64s = no,
+        OptSGInt64s = do_not_use_static_ground_int64s,
         SGInt64s = do_not_have_static_ground_int64s
     ),
-    globals.lookup_bool_option(Globals, static_code_addresses,
-        OptStaticCodeAddr),
+    OptStaticCodeAddr = OptTuple ^ ot_use_static_code_addresses,
     (
-        OptStaticCodeAddr = yes,
+        OptStaticCodeAddr = use_static_code_addresses,
         StaticCodeAddrs = have_static_code_addresses
     ;
-        OptStaticCodeAddr = no,
+        OptStaticCodeAddr = do_not_use_static_code_addresses,
         StaticCodeAddrs = do_not_have_static_code_addresses
     ),
     ExprnOpts = exprn_opts(NLG, ASM, UBF, UseFloatRegs, DetStackFloatWidth,
@@ -982,7 +983,8 @@ make_proc_entry_label(CI, ModuleInfo, PredId, ProcId, Immed0) = CodeAddr :-
     ;
         Immed0 = yes,
         get_globals(CI, Globals),
-        globals.lookup_int_option(Globals, procs_per_c_function, ProcsPerFunc),
+        globals.get_opt_tuple(Globals, OptTuple),
+        ProcsPerFunc = OptTuple ^ ot_procs_per_c_function,
         get_pred_id(CI, CurPredId),
         get_proc_id(CI, CurProcId),
         Immed = yes(ProcsPerFunc - proc(CurPredId, CurProcId))
diff --git a/compiler/code_loc_dep.m b/compiler/code_loc_dep.m
index bb72a55d3..1331e3a05 100644
--- a/compiler/code_loc_dep.m
+++ b/compiler/code_loc_dep.m
@@ -42,6 +42,7 @@
 :- import_module hlds.hlds_pred.
 :- import_module hlds.instmap.
 :- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.code_info.
 :- import_module ll_backend.continuation_info.
@@ -220,16 +221,10 @@ code_loc_dep_init(FollowVars, ResumePoint, !CI, !:CLD) :-
     init_var_locn_state(ModuleInfo, ArgList, EffLiveness, VarSet, VarTypes,
         FloatRegType, StackSlots, FollowVars, VarLocnInfo),
     stack.init(ResumePoints),
-    globals.lookup_bool_option(Globals, allow_hijacks, AllowHijack),
-    (
-        AllowHijack = yes,
-        Hijack = allowed
-    ;
-        AllowHijack = no,
-        Hijack = not_allowed
-    ),
+    globals.get_opt_tuple(Globals, OptTuple),
+    AllowHijack = OptTuple ^ ot_allow_hijacks,
     DummyFailInfo = fail_info(ResumePoints, resume_point_unknown,
-        may_be_different, not_inside_non_condition, Hijack),
+        may_be_different, not_inside_non_condition, AllowHijack),
     set.init(TempsInUse),
     Zombies = set_of_var.init,
     NestedParConjDepth = 0,
@@ -506,12 +501,12 @@ generate_branch_end(StoreMap, MaybeEnd0, MaybeEnd, Code, !.CLD) :-
             CurfrMaxfr = may_be_different
         ),
         ( if
-            Hijack0 = allowed,
-            Hijack1 = allowed
+            Hijack0 = allow_hijacks,
+            Hijack1 = allow_hijacks
         then
-            Hijack = allowed
+            Hijack = allow_hijacks
         else
-            Hijack = not_allowed
+            Hijack = do_not_allow_hijacks
         ),
         expect(unify(CondEnv0, CondEnv1), $pred,
             "some but not all branches inside a non condition"),
@@ -775,7 +770,7 @@ save_hp_in_branch(Code, Slot, Pos0, Pos, !CI) :-
                 resume_point_known,
                 curfr_vs_maxfr,
                 condition_env,
-                hijack_allowed
+                maybe_allow_hijacks
             ).
 
     % A resumption point has one or two labels associated with it.
@@ -818,10 +813,6 @@ save_hp_in_branch(Code, Slot, Pos0, Pos, !CI) :-
     --->    inside_non_condition
     ;       not_inside_non_condition.
 
-:- type hijack_allowed
-    --->    allowed
-    ;       not_allowed.
-
 %---------------------------------------------------------------------------%
 
 :- type disj_hijack_info
@@ -865,7 +856,7 @@ prepare_for_disj_hijack(CodeModel, HijackInfo, Code, !CI, !CLD) :-
         ;
             CondEnv = not_inside_non_condition,
             (
-                Allow = not_allowed,
+                Allow = do_not_allow_hijacks,
                 ( if
                     CurfrMaxfr = must_be_equal,
                     ResumeKnown = resume_point_known(has_been_done),
@@ -884,7 +875,7 @@ prepare_for_disj_hijack(CodeModel, HijackInfo, Code, !CI, !CLD) :-
                         !CI, !CLD)
                 )
             ;
-                Allow = allowed,
+                Allow = allow_hijacks,
                 (
                     CurfrMaxfr = must_be_equal,
                     (
@@ -1055,7 +1046,7 @@ prepare_for_ite_hijack(CondCodeModel, MaybeEmbeddedFrameId, HijackInfo, Code,
     ;
         CondCodeModel = model_non,
         ( if
-            ( Allow = not_allowed
+            ( Allow = do_not_allow_hijacks
             ; CondEnv = inside_non_condition
             ; MaybeEmbeddedFrameId = yes(_)
             )
@@ -1416,7 +1407,7 @@ prepare_for_semi_commit(AddTrailOps, AddRegionOps, ForwardLiveVarsBeforeGoal,
     pick_stack_resume_point(NewResumePoint, _, StackLabel),
     StackLabelConst = const(llconst_code_addr(StackLabel)),
     ( if
-        ( Allow = not_allowed ; CondEnv = inside_non_condition )
+        ( Allow = do_not_allow_hijacks ; CondEnv = inside_non_condition )
     then
         acquire_temp_slot(slot_lval(maxfr), non_persistent_temp_slot,
             MaxfrSlot, !CI, !CLD),
diff --git a/compiler/compile_target_code.m b/compiler/compile_target_code.m
index f23d8bbe1..fce479ac7 100644
--- a/compiler/compile_target_code.m
+++ b/compiler/compile_target_code.m
@@ -238,6 +238,7 @@
 :- implementation.
 
 :- import_module libs.compute_grade.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module parse_tree.error_util.
@@ -395,13 +396,14 @@ gather_c_compiler_flags(Globals, PIC, AllCFlags) :-
         TypeLayoutOption = yes,
         TypeLayoutOpt = ""
     ),
-    globals.lookup_bool_option(Globals, c_optimize, C_optimize),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptimizeC = OptTuple ^ ot_opt_c,
     (
-        C_optimize = yes,
+        OptimizeC = opt_c,
         globals.lookup_string_option(Globals, cflags_for_optimization,
             OptimizeOpt)
     ;
-        C_optimize = no,
+        OptimizeC = do_not_opt_c,
         OptimizeOpt = ""
     ),
     globals.lookup_bool_option(Globals, ansi_c, Ansi),
@@ -412,14 +414,14 @@ gather_c_compiler_flags(Globals, PIC, AllCFlags) :-
         Ansi = no,
         AnsiOpt = ""
     ),
-    globals.lookup_bool_option(Globals, inline_alloc, InlineAlloc),
+    InlineAlloc = OptTuple ^ ot_inline_alloc,
     (
-        InlineAlloc = yes,
+        InlineAlloc = inline_alloc,
         % XXX disabled because inline allocation is broken in gc7.0 alpha6.
         % InlineAllocOpt = "-DMR_INLINE_ALLOC "
         InlineAllocOpt = ""
     ;
-        InlineAlloc = no,
+        InlineAlloc = do_not_inline_alloc,
         InlineAllocOpt = ""
     ),
     globals.lookup_bool_option(Globals, warn_target_code, Warn),
diff --git a/compiler/const_struct.m b/compiler/const_struct.m
index 3ab0f0016..97b061ff5 100644
--- a/compiler/const_struct.m
+++ b/compiler/const_struct.m
@@ -19,11 +19,11 @@
 
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module parse_tree.
 :- import_module parse_tree.prog_data.
 
 :- import_module assoc_list.
-:- import_module bool.
 :- import_module list.
 :- import_module map.
 
@@ -111,19 +111,19 @@
     % and simplify.m respectively.
     %
 :- pred const_struct_db_get_poly_enabled(const_struct_db::in,
-    bool::out) is det.
+    maybe_enable_const_struct::out) is det.
 :- pred const_struct_db_get_ground_term_enabled(const_struct_db::in,
-    bool::out) is det.
+    maybe_enable_const_struct::out) is det.
 
 %-----------------------------------------------------------------------------%
 
 :- implementation.
 
-:- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
 
+:- import_module bool.
 :- import_module int.
 :- import_module maybe.
 :- import_module pair.
@@ -137,13 +137,13 @@ const_struct_db_init(Globals, Db) :-
     ;
         Target = target_java,
         can_enable_const_struct(Globals, PolyEnabled, _GroundTermEnabled),
-        GroundTermEnabled = no
+        GroundTermEnabled = do_not_enable_const_struct
     ;
         ( Target = target_csharp
         ; Target = target_erlang
         ),
-        PolyEnabled = no,
-        GroundTermEnabled = no
+        PolyEnabled = do_not_enable_const_struct,
+        GroundTermEnabled = do_not_enable_const_struct
     ),
     Db = const_struct_db(PolyEnabled, GroundTermEnabled, 0,
         map.init, map.init, map.init, map.init).
@@ -154,19 +154,19 @@ const_struct_db_init(Globals, Db) :-
     % (const_struct_db_init/2) must also check if the compilation grade
     % supports constant structures.
     %
-:- pred can_enable_const_struct(globals::in, bool::out, bool::out) is det.
+:- pred can_enable_const_struct(globals::in,
+    maybe_enable_const_struct::out, maybe_enable_const_struct::out) is det.
 
 can_enable_const_struct(Globals, PolyEnabled, GroundTermEnabled) :-
-    globals.lookup_bool_option(Globals, enable_const_struct,
-        OptionEnabled),
-    PolyEnabled = OptionEnabled,
+    globals.get_opt_tuple(Globals, OptTuple),
+    PolyEnabled = OptTuple ^ ot_enable_const_struct,
 
     globals.get_trace_level(Globals, TraceLevel),
     globals.get_trace_suppress(Globals, TraceSuppress),
     Bodies = trace_needs_proc_body_reps(TraceLevel, TraceSuppress),
     (
         Bodies = no,
-        GroundTermEnabled = OptionEnabled
+        GroundTermEnabled = PolyEnabled
     ;
         Bodies = yes,
         % We generate representations of procedure bodies for the
@@ -178,16 +178,16 @@ can_enable_const_struct(Globals, PolyEnabled, GroundTermEnabled) :-
         % arising in the first place. (We never look for the original
         % sources of type infos and typeclass infos, so we can use constant
         % structures for them.)
-        GroundTermEnabled = no
+        GroundTermEnabled = do_not_enable_const_struct
     ).
 
 lookup_insert_const_struct(ConstStruct, ConstNum, !Db) :-
     const_struct_db_get_poly_enabled(!.Db, Enabled),
     (
-        Enabled = no,
+        Enabled = do_not_enable_const_struct,
         unexpected($pred, "not enabled")
     ;
-        Enabled = yes,
+        Enabled = enable_const_struct,
         ConstStruct = const_struct(ConsId, Args, Type, Inst),
         ( if ConsId = cons(SymName, _, _) then
             Name = unqualify_name(SymName),
@@ -286,8 +286,8 @@ const_struct_db_get_structs(Db, Structs) :-
 
 :- type const_struct_db
     --->    const_struct_db(
-                csdb_poly_enabled           :: bool,
-                csdb_ground_term_enabled    :: bool,
+                csdb_poly_enabled           :: maybe_enable_const_struct,
+                csdb_ground_term_enabled    :: maybe_enable_const_struct,
                 csdb_next_num               :: int,
                 csdb_cons_proxy_map         :: map(cons_proxy_struct, int),
                 csdb_other_struct_map       :: map(const_struct, int),
diff --git a/compiler/deforest.m b/compiler/deforest.m
index 9b4f8c988..e8860c077 100644
--- a/compiler/deforest.m
+++ b/compiler/deforest.m
@@ -66,6 +66,7 @@
 :- import_module libs.
 :- import_module libs.dependency_graph.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
@@ -132,9 +133,10 @@ deforestation(!ModuleInfo) :-
     map.keys(VersionIndex, Versions),
 
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, constraint_propagation, Constraints),
+    globals.get_opt_tuple(Globals, OptTuple),
+    PropConstraints = OptTuple ^ ot_prop_constraints,
     ( if
-        Constraints = yes,
+        PropConstraints = prop_constraints,
         Versions = [_ | _]
     then
         % We can sometimes improve efficiency by rerunning determinism
@@ -302,25 +304,24 @@ deforest_goal_expr(GoalExpr0, GoalExpr, !GoalInfo, !PDInfo) :-
                 NonLocals = goal_info_get_nonlocals(!.GoalInfo),
                 pd_info_get_module_info(!.PDInfo, ModuleInfo),
                 module_info_get_globals(ModuleInfo, Globals),
-                globals.lookup_bool_option(Globals, deforestation,
-                    Deforestation),
+                globals.get_opt_tuple(Globals, OptTuple),
+                Deforest = OptTuple ^ ot_deforest,
                 (
-                    Deforestation = yes,
+                    Deforest = deforest,
                     compute_goal_infos(!Goals, !PDInfo),
                     pd_info_set_instmap(InstMap0, !PDInfo),
                     deforest_conj(!.Goals, NonLocals, [], !:Goals, !PDInfo)
                 ;
-                    Deforestation = no
+                    Deforest = do_not_deforest
                 ),
-                globals.lookup_bool_option(Globals, constraint_propagation,
-                    Constraints),
+                PropConstraints = OptTuple ^ ot_prop_constraints,
                 pd_info_set_instmap(InstMap0, !PDInfo),
                 (
-                    Constraints = yes,
+                    PropConstraints = prop_constraints,
                     propagate_conj_constraints(!.Goals, NonLocals, [], !:Goals,
                         !PDInfo)
                 ;
-                    Constraints = no
+                    PropConstraints = do_not_prop_constraints
                 ),
                 pd_info_set_instmap(InstMap0, !PDInfo)
             ;
@@ -887,16 +888,17 @@ can_optimize_conj(EarlierGoal, BetweenGoals, MaybeLaterGoal, ShouldTry,
     pd_info_get_depth(!.PDInfo, Depth0),
     pd_info_get_module_info(!.PDInfo, ModuleInfo),
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_option(Globals, deforestation_depth_limit, DepthLimitOpt),
+    globals.get_opt_tuple(Globals, OptTuple),
+    MaxDepth = OptTuple ^ ot_deforestation_depth_limit,
     globals.lookup_bool_option(Globals, debug_pd, DebugPD),
     Depth = Depth0 + 1,
     pd_info_set_depth(Depth, !PDInfo),
-    globals.lookup_int_option(Globals, deforestation_size_threshold,
-        SizeLimit),
+    SizeLimit = OptTuple ^ ot_deforestation_size_threshold,
     globals.lookup_option(Globals, fully_strict, FullyStrictOp),
     ( if
-        DepthLimitOpt = int(MaxDepth),
         MaxDepth \= -1,     % no depth limit set
+        % XXX The *default* value of the depth_limit is 4,
+        % so -1 does NOT mean that there is no depth limit set.
         Depth0 >= MaxDepth
     then
         % The depth limit was exceeded. This should not occur too often in
@@ -1018,7 +1020,8 @@ is_improvement_worth_while(PDInfo, Optimized0, CostDelta0, SizeDelta0)
 
     pd_info_get_module_info(PDInfo, ModuleInfo),
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_int_option(Globals, deforestation_cost_factor, Factor),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Factor = OptTuple ^ ot_deforestation_cost_factor,
     globals.lookup_bool_option(Globals, debug_pd, DebugPD),
     ( if
         Optimized0 = yes,
@@ -1169,16 +1172,19 @@ create_deforest_goal(EarlierGoal, BetweenGoals, MaybeLaterGoal,
         MaybeGeneralised, MaybeCallGoal, !PDInfo) :-
     pd_info_get_module_info(!.PDInfo, ModuleInfo0),
     module_info_get_globals(ModuleInfo0, Globals),
-    globals.lookup_int_option(Globals, deforestation_vars_threshold, VarsOpt),
+    globals.get_opt_tuple(Globals, OptTuple),
+    VarsThreshold = OptTuple ^ ot_deforestation_vars_threshold,
     globals.lookup_bool_option(Globals, debug_pd, DebugPD),
     ( if
         EarlierGoal = hlds_goal(EarlierGoalExpr, _),
         EarlierGoalExpr = plain_call(PredId1, ProcId1, Args1, _, _, _),
         (
             % No threshold set.
-            VarsOpt = -1
+            VarsThreshold = -1
+            % XXX The *default* value of the vars threshold is 200,
+            % so -1 does NOT mean that there is no threshold set.
         ;
-            % Check that we're not creating a procedure with a massive number
+            % Check that we are not creating a procedure with a massive number
             % of variables. We assume that all the variables in the first
             % called goal are present in the final version. If the number
             % of variables in the first called goal plus the number of
@@ -1196,7 +1202,7 @@ create_deforest_goal(EarlierGoal, BetweenGoals, MaybeLaterGoal,
             list.length(GoalVarsList1, NumVars1),
             list.length(GoalVarsList2, NumVars2),
             NumVars = NumVars1 + NumVars2,
-            NumVars < VarsOpt
+            NumVars < VarsThreshold
         )
     then
         % Create the goal for the new predicate, unfolding the first call.
@@ -1916,8 +1922,8 @@ deforest_call(PredId, ProcId, Args, SymName, BuiltinState, Goal0, Goal,
     pd_info_get_local_term_info(!.PDInfo, LocalTermInfo0),
 
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_int_option(Globals, deforestation_size_threshold,
-        SizeThreshold),
+    globals.get_opt_tuple(Globals, OptTuple),
+    SizeThreshold = OptTuple ^ ot_deforestation_size_threshold,
     globals.lookup_bool_option(Globals, debug_pd, DebugPD),
     ( if
         % Check for extra information to the call.
@@ -1993,7 +1999,8 @@ unfold_call(CheckImprovement, CheckVars, PredId, ProcId, Args,
         Goal0, Goal, Optimized, !PDInfo) :-
     pd_info_get_module_info(!.PDInfo, ModuleInfo),
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_int_option(Globals, deforestation_vars_threshold, VarsOpt),
+    globals.get_opt_tuple(Globals, OptTuple),
+    VarsThreshold = OptTuple ^ ot_deforestation_vars_threshold,
     pd_info_get_proc_info(!.PDInfo, ProcInfo0),
     proc_info_get_varset(ProcInfo0, VarSet0),
     varset.vars(VarSet0, Vars),
@@ -2004,10 +2011,11 @@ unfold_call(CheckImprovement, CheckVars, PredId, ProcId, Args,
         (
             CheckVars = no
         ;
-            VarsOpt = -1
+            VarsThreshold = -1
+            % XXX The *default* value of the vars threshold is 200,
+            % so -1 does NOT mean that there is no threshold set.
         ;
-            VarsOpt = MaxVars,
-            NumVars < MaxVars
+            NumVars < VarsThreshold
         )
     then
         pd_info_get_pred_info(!.PDInfo, PredInfo0),
@@ -2085,7 +2093,7 @@ unfold_call(CheckImprovement, CheckVars, PredId, ProcId, Args,
         CostDelta = CostDelta1 - CostDelta0,
         goal_size(Goal4, GoalSize),
         SizeDelta = GoalSize - cost_of_call,
-        globals.lookup_int_option(Globals, deforestation_cost_factor, Factor),
+        Factor = OptTuple ^ ot_deforestation_cost_factor,
         ( if
             Optimized0 = yes,
             (
diff --git a/compiler/dep_par_conj.m b/compiler/dep_par_conj.m
index ce2671f72..3d8395c62 100644
--- a/compiler/dep_par_conj.m
+++ b/compiler/dep_par_conj.m
@@ -148,6 +148,7 @@
 :- import_module libs.
 :- import_module libs.dependency_graph.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
@@ -231,7 +232,8 @@ impl_dep_par_conjs_in_module(!ModuleInfo) :-
 
                 % The value of the --allow-some-paths-only-waits option.
                 % Read-only.
-                sync_allow_some_paths_only  :: bool,
+                sync_allow_some_paths_only  ::
+                                            maybe_allow_some_paths_only_waits,
 
                 % The varset and vartypes for the procedure being analysed.
                 % These fields are updated when we add new variables.
@@ -290,8 +292,8 @@ sync_dep_par_conjs_in_proc(PredId, ProcId, IgnoreVars, !ModuleInfo,
         proc_info_get_vartypes(!.ProcInfo, !:VarTypes),
         proc_info_get_initial_instmap(!.ProcInfo, !.ModuleInfo, InstMap0),
         module_info_get_globals(!.ModuleInfo, Globals),
-        globals.lookup_bool_option(Globals, allow_some_paths_only_waits,
-            AllowSomePathsOnly),
+        globals.get_opt_tuple(Globals, OptTuple),
+        AllowSomePathsOnly = OptTuple ^ ot_allow_some_paths_only_waits,
 
         % We rely on dependency information in order to determine which calls
         % are recursive. The information is stored within !ModuleInfo so
@@ -530,7 +532,8 @@ maybe_sync_dep_par_conj(Conjuncts, GoalInfo, NewGoal, InstMap, !SyncInfo) :-
     %           append(AB_10, A, ABA)
     %       ).
     %
-:- pred sync_dep_par_conj(module_info::in, bool::in, set_of_progvar::in,
+:- pred sync_dep_par_conj(module_info::in,
+    maybe_allow_some_paths_only_waits::in, set_of_progvar::in,
     list(hlds_goal)::in, hlds_goal_info::in, hlds_goal::out, instmap::in,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out,
     ts_string_table::in, ts_string_table::out) is det.
@@ -578,10 +581,10 @@ sync_dep_par_conj(ModuleInfo, AllowSomePathsOnly, SharedVars, Goals, GoalInfo,
     % procedure simply to maintain the above invariant. Can an unused argument
     % analysis prevent this situation?
     %
-:- pred sync_dep_par_proc_body(module_info::in, bool::in,
-    set_of_progvar::in, future_map::in, instmap::in,
-    hlds_goal::in, hlds_goal::out, prog_varset::in, prog_varset::out,
-    vartypes::in, vartypes::out) is det.
+:- pred sync_dep_par_proc_body(module_info::in,
+    maybe_allow_some_paths_only_waits::in, set_of_progvar::in,
+    future_map::in, instmap::in, hlds_goal::in, hlds_goal::out,
+    prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
 
 sync_dep_par_proc_body(ModuleInfo, AllowSomePathsOnly, SharedVars, FutureMap,
         InstMap, !Goal, !VarSet, !VarTypes) :-
@@ -618,10 +621,10 @@ sync_dep_par_proc_body(ModuleInfo, AllowSomePathsOnly, SharedVars, FutureMap,
             !Goal, !VarSet, !VarTypes)
     ).
 
-:- pred sync_dep_par_conjunct(module_info::in, bool::in,
-    set_of_progvar::in, future_map::in, hlds_goal::in, hlds_goal::out,
-    instmap::in, instmap::out, prog_varset::in, prog_varset::out,
-    vartypes::in, vartypes::out) is det.
+:- pred sync_dep_par_conjunct(module_info::in,
+    maybe_allow_some_paths_only_waits::in, set_of_progvar::in, future_map::in,
+    hlds_goal::in, hlds_goal::out, instmap::in, instmap::out,
+    prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
 
 sync_dep_par_conjunct(ModuleInfo, AllowSomePathsOnly, SharedVars, FutureMap,
         !Goal, !InstMap, !VarSet, !VarTypes) :-
@@ -636,8 +639,8 @@ sync_dep_par_conjunct(ModuleInfo, AllowSomePathsOnly, SharedVars, FutureMap,
             NonlocalSharedVars, ConsumedVarsList, ProducedVarsList),
 
         % Insert waits into the conjunct, as late as possible.
-        list.map_foldl3(insert_wait_in_goal(ModuleInfo, AllowSomePathsOnly,
-                FutureMap),
+        list.map_foldl3(
+            insert_wait_in_goal(ModuleInfo, AllowSomePathsOnly, FutureMap),
             ConsumedVarsList, _WaitedOnAllSuccessPaths,
             !Goal, !VarSet, !VarTypes),
 
@@ -671,8 +674,9 @@ consumed_and_produced_vars(ModuleInfo, InstMap, InstMapDelta, Vars,
     ConsumedVarsList = set_of_var.to_sorted_list(ConsumedVars),
     ProducedVarsList = set_of_var.to_sorted_list(ProducedVars).
 
-:- pred insert_wait_in_goal_for_proc(module_info::in, bool::in, future_map::in,
-    prog_var::in, hlds_goal::in, hlds_goal::out,
+:- pred insert_wait_in_goal_for_proc(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in, prog_var::in,
+    hlds_goal::in, hlds_goal::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
 
 insert_wait_in_goal_for_proc(ModuleInfo, AllowSomePathsOnly, FutureMap,
@@ -723,7 +727,8 @@ join_branches(not_waited_on_all_success_paths, not_waited_on_all_success_paths,
     % (We must ensure that if one branch of a branched goal inserts a wait
     % for a variable, then *all* branches of that goal insert a wait.)
     %
-:- pred insert_wait_in_goal(module_info::in, bool::in, future_map::in,
+:- pred insert_wait_in_goal(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in,
     prog_var::in, waited_on_all_success_paths::out,
     hlds_goal::in, hlds_goal::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
@@ -917,11 +922,11 @@ insert_wait_in_goal(ModuleInfo, AllowSomePathsOnly, FutureMap, ConsumedVar,
         ;
             WaitedOnAllSuccessPaths0 = not_waited_on_all_success_paths,
             (
-                AllowSomePathsOnly = yes,
+                AllowSomePathsOnly = allow_some_paths_only_waits,
                 WaitedOnAllSuccessPaths = WaitedOnAllSuccessPaths0,
                 Goal = Goal2
             ;
-                AllowSomePathsOnly = no,
+                AllowSomePathsOnly = do_not_allow_some_paths_only_waits,
                 (
                     InvariantEstablished = no,
                     WaitedOnAllSuccessPaths = waited_on_all_success_paths,
@@ -960,7 +965,8 @@ insert_wait_after_goal(ModuleInfo, FutureMap, ConsumedVar,
     % that references it. Any later conjuncts will get the waited-for variable
     % without having to call wait.
     %
-:- pred insert_wait_in_plain_conj(module_info::in, bool::in, future_map::in,
+:- pred insert_wait_in_plain_conj(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in,
     prog_var::in, waited_on_all_success_paths::out,
     list(hlds_goal)::in, list(hlds_goal)::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
@@ -1021,8 +1027,8 @@ insert_wait_in_plain_conj(ModuleInfo, AllowSomePathsOnly, FutureMap,
     % "earlier" conjuncts waited for, since those waits may not have finished
     % yet.
     %
-:- pred insert_wait_in_par_conj(module_info::in, bool::in,
-    future_map::in, prog_var::in,
+:- pred insert_wait_in_par_conj(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in, prog_var::in,
     waited_in_conjunct::in, waited_in_conjunct::out,
     list(hlds_goal)::in, list(hlds_goal)::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
@@ -1057,8 +1063,8 @@ insert_wait_in_par_conj(ModuleInfo, AllowSomePathsOnly, FutureMap, ConsumedVar,
     insert_wait_in_par_conj(ModuleInfo, AllowSomePathsOnly, FutureMap,
         ConsumedVar, !WaitedInConjunct, Goals0, Goals, !VarSet, !VarTypes).
 
-:- pred insert_wait_in_disj(module_info::in, bool::in,
-    future_map::in, prog_var::in,
+:- pred insert_wait_in_disj(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in, prog_var::in,
     waited_on_all_success_paths::in, waited_on_all_success_paths::out,
     list(hlds_goal)::in, list(hlds_goal)::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
@@ -1075,8 +1081,8 @@ insert_wait_in_disj(ModuleInfo, AllowSomePathsOnly, FutureMap, ConsumedVar,
     insert_wait_in_disj(ModuleInfo, AllowSomePathsOnly, FutureMap, ConsumedVar,
         !WaitedOnAllSuccessPaths, Goals0, Goals, !VarSet, !VarTypes).
 
-:- pred insert_wait_in_cases(module_info::in, bool::in,
-    future_map::in, prog_var::in,
+:- pred insert_wait_in_cases(module_info::in,
+    maybe_allow_some_paths_only_waits::in, future_map::in, prog_var::in,
     waited_on_all_success_paths::in, waited_on_all_success_paths::out,
     list(case)::in, list(case)::out,
     prog_varset::in, prog_varset::out, vartypes::in, vartypes::out) is det.
@@ -1658,8 +1664,8 @@ add_requested_specialized_par_proc(CallPattern, NewProc, !PendingParProcs,
         % Insert signals and waits into the procedure body. We treat the body
         % as it were a conjunct of a parallel conjunction, since it is.
         module_info_get_globals(InitialModuleInfo, Globals),
-        globals.lookup_bool_option(Globals, allow_some_paths_only_waits,
-            AllowSomePathsOnly),
+        globals.get_opt_tuple(Globals, OptTuple),
+        AllowSomePathsOnly = OptTuple ^ ot_allow_some_paths_only_waits,
         SharedVars = set_of_var.sorted_list_to_set(map.keys(FutureMap)),
         sync_dep_par_proc_body(!.ModuleInfo, AllowSomePathsOnly, SharedVars,
             FutureMap, InstMap0, Goal0, Goal, !VarSet, !VarTypes),
@@ -2367,13 +2373,13 @@ should_we_push(PredProcId, ArgPos, PushOp, IsWorthPushing, !SpecInfo) :-
     else
         InitialModuleInfo = !.SpecInfo ^ spec_initial_module,
         module_info_get_globals(InitialModuleInfo, Globals),
-        globals.lookup_bool_option(Globals, always_specialize_in_dep_par_conjs,
-            AlwaysSpecialize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        AlwaysSpecialize = OptTuple ^ ot_spec_in_all_dep_par_conjs,
         (
-            AlwaysSpecialize = yes,
+            AlwaysSpecialize = spec_in_all_dep_par_conjs,
             IsWorthPushing = worth_pushing
         ;
-            AlwaysSpecialize = no,
+            AlwaysSpecialize = do_not_spec_in_all_dep_par_conjs,
             should_we_push_test(PredProcId, ArgPos, PushOp, IsWorthPushing,
                 !.SpecInfo)
         ),
@@ -3126,13 +3132,13 @@ allocate_future(ModuleInfo, SharedVar, Goals, !VarSet, !VarTypes,
     Context = term.context_init,
     ShouldInline = should_inline_par_builtin_calls(ModuleInfo),
     (
-        ShouldInline = no,
+        ShouldInline = do_not_inline_par_builtins,
         ArgVars = [FutureNameVar, FutureVar],
         generate_simple_call(ModuleInfo, ModuleName, PredName, pf_predicate,
             only_mode, detism_det, purity_pure, ArgVars, Features,
             InstMapDelta, Context, AllocGoal)
     ;
-        ShouldInline = yes,
+        ShouldInline = inline_par_builtins,
         ForeignAttrs = par_builtin_foreign_proc_attributes(purity_pure, no),
         ArgName = foreign_arg(FutureNameVar,
             yes(foreign_arg_name_mode("Name", in_mode)),
@@ -3228,13 +3234,13 @@ make_wait_or_get(ModuleInfo, VarTypes, FutureVar, ConsumedVar, WaitOrGetPred,
     Context = term.context_init,
     ShouldInline = should_inline_par_builtin_calls(ModuleInfo),
     (
-        ShouldInline = no,
+        ShouldInline = do_not_inline_par_builtins,
         ArgVars = [FutureVar, ConsumedVar],
         generate_simple_call(ModuleInfo, ModuleName, PredName, pf_predicate,
             only_mode, detism_det, Purity, ArgVars, Features,
             InstMapDelta, Context, WaitGoal)
     ;
-        ShouldInline = yes,
+        ShouldInline = inline_par_builtins,
         ForeignAttrs = par_builtin_foreign_proc_attributes(Purity, no),
         lookup_var_type(VarTypes, FutureVar, FutureVarType),
         lookup_var_type(VarTypes, ConsumedVar, ConsumedVarType),
@@ -3263,13 +3269,13 @@ make_signal_goal(ModuleInfo, FutureMap, ProducedVar, VarTypes, SignalGoal) :-
     Context = term.context_init,
     ShouldInline = should_inline_par_builtin_calls(ModuleInfo),
     (
-        ShouldInline = no,
+        ShouldInline = do_not_inline_par_builtins,
         ArgVars = [FutureVar, ProducedVar],
         generate_simple_call(ModuleInfo, ModuleName, PredName, pf_predicate,
             only_mode, detism_det, purity_impure, ArgVars, Features,
             InstMapDelta, Context, SignalGoal)
     ;
-        ShouldInline = yes,
+        ShouldInline = inline_par_builtins,
         ForeignAttrs = par_builtin_foreign_proc_attributes(purity_impure,
             yes(needs_call_standard_output_registers)),
         lookup_var_type(VarTypes, FutureVar, FutureVarType),
@@ -3319,11 +3325,13 @@ new_future_code = "
     #endif
 ".
 
-:- func should_inline_par_builtin_calls(module_info) = bool.
+:- func should_inline_par_builtin_calls(module_info) =
+    maybe_inline_par_builtins.
 
 should_inline_par_builtin_calls(ModuleInfo) = ShouldInline :-
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, inline_par_builtins, ShouldInline).
+    globals.get_opt_tuple(Globals, OptTuple),
+    ShouldInline = OptTuple ^ ot_inline_par_builtins.
 
 :- func par_builtin_foreign_proc_attributes(purity,
     maybe(pragma_foreign_proc_extra_attribute))
diff --git a/compiler/disj_gen.m b/compiler/disj_gen.m
index e975e451b..26ec953c4 100644
--- a/compiler/disj_gen.m
+++ b/compiler/disj_gen.m
@@ -47,6 +47,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.code_gen.
 :- import_module ll_backend.continuation_info.
@@ -159,7 +160,8 @@ is_lookup_disj(AddTrailOps, AddRegionOps, ResumeVars, Disjuncts, DisjGoalInfo,
 
     % Lookup disjunctions rely on static ground terms to work.
     get_globals(!.CI, Globals),
-    globals.lookup_bool_option(Globals, static_ground_cells, yes),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptTuple ^ ot_use_static_ground_cells = use_static_ground_cells,
 
     % XXX The code to generate lookup disjunctions hasn't yet been updated
     % to handle region operations.
diff --git a/compiler/erl_code_gen.m b/compiler/erl_code_gen.m
index a9b9627d8..7299e7889 100644
--- a/compiler/erl_code_gen.m
+++ b/compiler/erl_code_gen.m
@@ -67,7 +67,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.prim_data.
 :- import_module mdbcomp.sym_name.
@@ -811,8 +811,8 @@ erl_gen_switch(Var, CanFail, CasesList, CodeModel, InstMap0, _Context,
         % enable the workaround unless the user specifically passes
         % --erlang-switch-on-strings-as-atoms.
         module_info_get_globals(ModuleInfo, Globals),
-        globals.lookup_bool_option(Globals, erlang_switch_on_strings_as_atoms,
-            yes),
+        globals.get_opt_tuple(Globals, OptTuple),
+        OptTuple ^ ot_switch_on_strings_as_atoms = switch_on_strings_as_atoms,
 
         list.length(CasesList) > switch_strings_as_atoms_limit,
 
diff --git a/compiler/format_call.m b/compiler/format_call.m
index c9f526e87..4ed69f45c 100644
--- a/compiler/format_call.m
+++ b/compiler/format_call.m
@@ -165,6 +165,7 @@
 :- import_module hlds.pred_table.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.builtin_modules.
 :- import_module mdbcomp.goal_path.
@@ -398,14 +399,15 @@ analyze_and_optimize_format_calls(ModuleInfo, GenImplicitStreamWarnings,
     format_call_traverse_goal(ModuleInfo, Goal1, _, [], FormatCallSites,
         Counter0, _Counter, ConjMaps0, ConjMaps, map.init, PredMap,
         set_of_var.init, _),
-    globals.lookup_bool_option(Globals, optimize_format_calls, OptFormatCalls),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptFormatCalls = OptTuple ^ ot_opt_format_calls,
     globals.lookup_bool_option(Globals, exec_trace, ExecTrace),
     % If users write a call to e.g. to string.format, they expect the debugger
     % to show them calls to string.format. Showing a sequence of calls to
     % lower level predicates instead of that higher level call would probably
     % confuse debugger users.
     ( if
-        OptFormatCalls = yes,
+        OptFormatCalls = opt_format_calls,
         ExecTrace = no
     then
         ShouldOptFormatCalls = opt_format_calls
@@ -439,10 +441,6 @@ analyze_and_optimize_format_calls(ModuleInfo, GenImplicitStreamWarnings,
         MaybeGoal = yes(Goal)
     ).
 
-:- type maybe_opt_format_calls
-    --->    do_not_opt_format_calls
-    ;       opt_format_calls.
-
 :- pred check_format_call_site(module_info::in,
     maybe_generate_implicit_stream_warnings::in, maybe_opt_format_calls::in,
     conj_maps::in, conj_pred_map::in, format_call_site::in,
diff --git a/compiler/global_data.m b/compiler/global_data.m
index b4da98712..8c241025d 100644
--- a/compiler/global_data.m
+++ b/compiler/global_data.m
@@ -20,6 +20,8 @@
 
 :- import_module hlds.
 :- import_module hlds.hlds_pred.
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module ll_backend.continuation_info.
 :- import_module ll_backend.layout.
 :- import_module ll_backend.llds.
@@ -28,7 +30,6 @@
 :- import_module parse_tree.
 :- import_module parse_tree.prog_data.
 
-:- import_module bool.
 :- import_module list.
 :- import_module map.
 :- import_module set_tree234.
@@ -94,7 +95,7 @@
 :- type static_cell_info.
 
 :- func init_static_cell_info(module_name, have_unboxed_floats,
-    have_unboxed_int64s, bool) = static_cell_info.
+    have_unboxed_int64s, maybe_use_common_data) = static_cell_info.
 
 :- pred add_scalar_static_cell_natural_types(list(rval)::in, data_id::out,
     static_cell_info::in, static_cell_info::out) is det.
@@ -349,7 +350,7 @@ make_alloc_id_map(AllocSite, Slot, Slot + 1, !Map) :-
                 scsi_module_name            :: module_name, % base file name
                 scsi_unbox_float            :: have_unboxed_floats,
                 scsi_unbox_int64s           :: have_unboxed_int64s,
-                scsi_common_data            :: bool
+                scsi_common_data            :: maybe_use_common_data
             ).
 
 :- type cell_type_bimap == bimap(common_cell_type, type_num).
@@ -434,7 +435,7 @@ do_add_scalar_static_cell(TypedArgs, CellType, CellValue, DataId, !Info) :-
 
         InsertCommonData = !.Info ^ sci_sub_info ^ scsi_common_data,
         (
-            InsertCommonData = yes,
+            InsertCommonData = use_common_data,
             MembersMap0 = !.CellGroup ^ scalar_cell_group_members,
             CellNumCounter0 = !.CellGroup ^ scalar_cell_counter,
             counter.allocate(CellNum, CellNumCounter0, CellNumCounter),
@@ -459,7 +460,7 @@ do_add_scalar_static_cell(TypedArgs, CellType, CellValue, DataId, !Info) :-
                 !Info ^ sci_scalar_cell_group_map := CellGroupMap
             )
         ;
-            InsertCommonData = no,
+            InsertCommonData = do_not_use_common_data,
             MembersMap0 = !.CellGroup ^ scalar_cell_group_members,
             ( if bimap.search(MembersMap0, Args, DataIdPrime) then
                 DataId = DataIdPrime
diff --git a/compiler/globals.m b/compiler/globals.m
index 12c1be79d..8eb56f55e 100644
--- a/compiler/globals.m
+++ b/compiler/globals.m
@@ -22,6 +22,7 @@
 :- interface.
 
 :- import_module libs.op_mode.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module mdbcomp.
@@ -279,7 +280,7 @@
 % Access predicates for the `globals' structure.
 %
 
-:- pred globals_init(option_table::in, op_mode::in,
+:- pred globals_init(option_table::in, opt_tuple::in, op_mode::in,
     compilation_target::in, gc_method::in,
     termination_norm::in, termination_norm::in,
     trace_level::in, trace_suppress_items::in, ssdb_trace_level::in,
@@ -289,6 +290,7 @@
     limit_error_contexts_map::in, globals::out) is det.
 
 :- pred get_options(globals::in, option_table::out) is det.
+:- pred get_opt_tuple(globals::in, opt_tuple::out) is det.
 :- pred get_op_mode(globals::in, op_mode::out) is det.
 :- pred get_target(globals::in, compilation_target::out) is det.
 :- pred get_gc_method(globals::in, gc_method::out) is det.
@@ -315,6 +317,7 @@
 :- pred set_option(option::in, option_data::in, globals::in, globals::out)
     is det.
 :- pred set_options(option_table::in, globals::in, globals::out) is det.
+:- pred set_opt_tuple(opt_tuple::in, globals::in, globals::out) is det.
 :- pred set_op_mode(op_mode::in, globals::in, globals::out) is det.
 :- pred set_gc_method(gc_method::in, globals::in, globals::out) is det.
 :- pred set_trace_level(trace_level::in, globals::in, globals::out) is det.
@@ -725,6 +728,7 @@ convert_line_number_range(RangeStr, line_number_range(MaybeMin, MaybeMax)) :-
 :- type globals
     --->    globals(
                 g_options                   :: option_table,
+                g_opt_tuple                 :: opt_tuple,
                 g_op_mode                   :: op_mode,
                 g_trace_suppress_items      :: trace_suppress_items,
                 g_reuse_strategy            :: reuse_strategy,
@@ -748,18 +752,19 @@ convert_line_number_range(RangeStr, line_number_range(MaybeMin, MaybeMax)) :-
                 g_target_env_type           :: env_type
             ).
 
-globals_init(Options, OpMode, Target, GC_Method,
+globals_init(Options, OptTuple, OpMode, Target, GC_Method,
         TerminationNorm, Termination2Norm, TraceLevel, TraceSuppress,
         SSTraceLevel, MaybeThreadSafe, C_CompilerType, CSharp_CompilerType,
         ReuseStrategy, MaybeFeedback, HostEnvType, SystemEnvType,
         TargetEnvType, FileInstallCmd, LimitErrorContextsMap, Globals) :-
-    Globals = globals(Options, OpMode, TraceSuppress,
+    Globals = globals(Options, OptTuple, OpMode, TraceSuppress,
         ReuseStrategy, MaybeFeedback, FileInstallCmd, LimitErrorContextsMap,
         C_CompilerType, CSharp_CompilerType, Target, GC_Method,
         TerminationNorm, Termination2Norm, TraceLevel, SSTraceLevel,
         MaybeThreadSafe, HostEnvType, SystemEnvType, TargetEnvType).
 
 get_options(Globals, Globals ^ g_options).
+get_opt_tuple(Globals, Globals ^ g_opt_tuple).
 get_op_mode(Globals, Globals ^ g_op_mode).
 get_target(Globals, Globals ^ g_target).
 get_gc_method(Globals, Globals ^ g_gc_method).
@@ -791,6 +796,9 @@ set_option(Option, OptionData, !Globals) :-
 set_options(Options, !Globals) :-
     !Globals ^ g_options := Options.
 
+set_opt_tuple(OptTuple, !Globals) :-
+    !Globals ^ g_opt_tuple := OptTuple.
+
 set_op_mode(OpMode, !Globals) :-
     !Globals ^ g_op_mode := OpMode.
 
@@ -1018,8 +1026,8 @@ double_width_floats_on_det_stack(Globals, FloatDwords) :-
 %---------------------------------------------------------------------------%
 
 globals_init_mutables(Globals, !IO) :-
-    globals.lookup_int_option(Globals, from_ground_term_threshold,
-        FromGroundTermThreshold),
+    globals.get_opt_tuple(Globals, OptTuple),
+    FromGroundTermThreshold = OptTuple ^ ot_from_ground_term_threshold,
     set_maybe_from_ground_term_threshold(yes(FromGroundTermThreshold), !IO).
 
 get_maybe_from_ground_term_threshold = MaybeThreshold :-
diff --git a/compiler/grab_modules.m b/compiler/grab_modules.m
index 8c94d7eda..c7b69eb6b 100644
--- a/compiler/grab_modules.m
+++ b/compiler/grab_modules.m
@@ -129,6 +129,7 @@
 
 :- implementation.
 
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.builtin_modules.
 :- import_module parse_tree.comp_unit_interface.    % undesirable dependency
@@ -1868,10 +1869,11 @@ grab_plain_opt_and_int_for_opt_files(Globals, FoundError,
     % condition reuse actually has none. But we have to maintain the interface
     % for modules that use the conditional reuse information from the `.opt'
     % file.
-    globals.lookup_bool_option(Globals, intermod_unused_args, UnusedArgs),
+    globals.get_opt_tuple(Globals, OptTuple),
+    UnusedArgs = OptTuple ^ ot_opt_unused_args_intermod,
     globals.lookup_bool_option(Globals, structure_reuse_analysis,
         StructureReuse),
-    ( if (UnusedArgs = yes ; StructureReuse = yes) then
+    ( if (UnusedArgs = opt_unused_args_intermod ; StructureReuse = yes) then
         read_plain_opt_file(Globals, VeryVerbose, ModuleName, OwnFileName,
             OwnParseTreePlainOpt0, OwnSpecs, OwnError, !IO),
         % XXX We should store the whole parse_tree, with a note next to it
@@ -1945,7 +1947,7 @@ grab_plain_opt_and_int_for_opt_files(Globals, FoundError,
     ).
 
 :- pred keep_only_unused_and_reuse_pragmas_in_parse_tree_plain_opt(
-    bool::in, bool::in,
+    maybe_opt_unused_args_intermod::in, bool::in,
     parse_tree_plain_opt::in, parse_tree_plain_opt::out) is det.
 
 keep_only_unused_and_reuse_pragmas_in_parse_tree_plain_opt(
@@ -1957,10 +1959,10 @@ keep_only_unused_and_reuse_pragmas_in_parse_tree_plain_opt(
         _MarkerPragmas, _TypeSpecs, UnusedArgs0, _TermInfos, _Term2Infos,
         _Exceptions, _Trailings, _MMTablings, _Sharings, Reuses0),
     (
-        KeepUnusedArgs = yes,
+        KeepUnusedArgs = opt_unused_args_intermod,
         UnusedArgs = UnusedArgs0
     ;
-        KeepUnusedArgs = no,
+        KeepUnusedArgs = do_not_opt_unused_args_intermod,
         UnusedArgs = []
     ),
     (
diff --git a/compiler/handle_options.m b/compiler/handle_options.m
index 19d99a6d4..2ca2b4dcd 100644
--- a/compiler/handle_options.m
+++ b/compiler/handle_options.m
@@ -73,12 +73,14 @@
 :- import_module libs.compiler_util.
 :- import_module libs.compute_grade.
 :- import_module libs.op_mode.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module mdbcomp.
 :- import_module mdbcomp.feedback.
 
 :- import_module bool.
+:- import_module cord.
 :- import_module dir.
 :- import_module getopt_io.
 :- import_module int.
@@ -100,7 +102,8 @@ handle_given_options(Args0, OptionArgs, Args, Specs, !:Globals, !IO) :-
         io.write_string("original arguments\n", !TIO),
         dump_arguments(Args0, !TIO)
     ),
-    process_given_options(Args0, OptionArgs, Args, MaybeOptionTable, !IO),
+    process_given_options(Args0, OptionArgs, Args, MaybeError, OptionTable,
+        OptOptions, !IO),
     trace [compile_time(flag("debug_handle_given_options")), io(!TIO)] (
         io.nl(!TIO),
         io.write_string("final option arguments\n", !TIO),
@@ -110,8 +113,8 @@ handle_given_options(Args0, OptionArgs, Args, Specs, !:Globals, !IO) :-
         io.write_string("final non-option arguments\n", !TIO),
         dump_arguments(Args, !TIO)
     ),
-    convert_option_table_result_to_globals(MaybeOptionTable, Specs,
-        !:Globals, !IO),
+    convert_option_table_result_to_globals(MaybeError, OptionTable, OptOptions,
+        Specs, !:Globals, !IO),
     (
         Specs = [_ | _]
         % Do NOT set the exit status. This predicate may be called before all
@@ -136,7 +139,7 @@ handle_given_options(Args0, OptionArgs, Args, Specs, !:Globals, !IO) :-
     ).
 
 separate_option_args(RawArgs, OptionArgs, NonOptionArgs, !IO) :-
-    process_given_options(RawArgs, OptionArgs, NonOptionArgs, _, !IO).
+    process_given_options(RawArgs, OptionArgs, NonOptionArgs, _, _, _, !IO).
 
     % process_given_options(Args, OptionArgs, NonOptionArgs, MaybeOptionTable,
     %   !IO):
@@ -145,15 +148,19 @@ separate_option_args(RawArgs, OptionArgs, NonOptionArgs, !IO) :-
     % useful for separating the list of arguments into option and non-option
     % arguments.
     %
-:- pred process_given_options(list(string)::in,
-    list(string)::out, list(string)::out, maybe_option_table_se(option)::out,
+:- pred process_given_options(list(string)::in, list(string)::out,
+    list(string)::out, maybe(option_error(option))::out,
+    option_table(option)::out, cord(optimization_option)::out,
     io::di, io::uo) is det.
 
-process_given_options(RawArgs, OptionArgs, NonOptionArgs, Result, !IO) :-
-    OptionOps = option_ops(short_option, long_option,
-        option_defaults, special_handler),
-    getopt_io.process_options_se(OptionOps, RawArgs,
-        OptionArgs, NonOptionArgs, Result, !IO).
+process_given_options(RawArgs, OptionArgs, NonOptionArgs, MaybeError,
+        OptionTable, OptOptionsCord, !IO) :-
+    OptionOps =
+        option_ops_userdata(short_option, long_option, special_handler),
+    getopt_io.init_option_table(option_defaults, OptionTable0),
+    getopt_io.process_options_userdata_se(OptionOps, RawArgs,
+        OptionArgs, NonOptionArgs, MaybeError, _OptionsSet,
+        OptionTable0, OptionTable, cord.init, OptOptionsCord, !IO).
 
 :- pred dump_arguments(list(string)::in, io::di, io::uo) is det.
 
@@ -169,13 +176,14 @@ dump_arguments([Arg | Args], !IO) :-
     % one option implies setting/unsetting another one).
     %
 :- pred convert_option_table_result_to_globals(
-    maybe_option_table_se(option)::in,
+    maybe(option_error(option))::in, option_table(option)::in,
+    cord(optimization_option)::in,
     list(error_spec)::out, globals::out, io::di, io::uo) is det.
 
-convert_option_table_result_to_globals(MaybeOptionTable0, !:Specs, Globals,
-        !IO) :-
+convert_option_table_result_to_globals(MaybeError, OptionTable0,
+        OptOptionsCord, !:Specs, Globals, !IO) :-
     (
-        MaybeOptionTable0 = error(Error),
+        MaybeError = yes(Error),
         ErrorMessage = option_error_to_string(Error),
         OptionTablePieces = [words(ErrorMessage)],
         OptionTableMsg = error_msg(no, do_not_treat_as_first, 0,
@@ -185,7 +193,9 @@ convert_option_table_result_to_globals(MaybeOptionTable0, !:Specs, Globals,
         !:Specs = [OptionTableSpec],
         generate_default_globals(Globals, !IO)
     ;
-        MaybeOptionTable0 = ok(OptionTable0),
+        MaybeError = no,
+        OptOptions = cord.list(OptOptionsCord),
+        process_optimization_options(OptionTable, OptOptions, OptTuple),
         check_option_values(OptionTable0, OptionTable, Target, GC_Method,
             TermNorm, Term2Norm, TraceLevel, TraceSuppress, SSTraceLevel,
             MaybeThreadSafe, C_CompilerType, CSharp_CompilerType,
@@ -206,7 +216,7 @@ convert_option_table_result_to_globals(MaybeOptionTable0, !:Specs, Globals,
         ),
         (
             !.Specs = [],
-            convert_options_to_globals(OptionTable, OpMode, Target,
+            convert_options_to_globals(OptionTable, OptTuple, OpMode, Target,
                 GC_Method, TermNorm, Term2Norm, TraceLevel, TraceSuppress,
                 SSTraceLevel, MaybeThreadSafe,
                 C_CompilerType, CSharp_CompilerType,
@@ -631,8 +641,8 @@ check_option_values(!OptionTable, Target, GC_Method, TermNorm, Term2Norm,
 
     % NOTE: each termination analyser has its own norm setting.
     %
-:- pred convert_options_to_globals(option_table::in, op_mode::in,
-    compilation_target::in, gc_method::in,
+:- pred convert_options_to_globals(option_table::in, opt_tuple::in,
+    op_mode::in, compilation_target::in, gc_method::in,
     termination_norm::in, termination_norm::in, trace_level::in,
     trace_suppress_items::in, ssdb_trace_level::in, may_be_thread_safe::in,
     c_compiler_type::in, csharp_compiler_type::in,
@@ -641,7 +651,7 @@ check_option_values(!OptionTable, Target, GC_Method, TermNorm, Term2Norm,
     list(error_spec)::in, list(error_spec)::out,
     globals::out, io::di, io::uo) is det.
 
-convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
+convert_options_to_globals(OptionTable0, !.OptTuple, OpMode, Target, GC_Method,
         TermNorm, Term2Norm, TraceLevel, TraceSuppress, SSTraceLevel,
         MaybeThreadSafe, C_CompilerType, CSharp_CompilerType,
         ReuseStrategy, MaybeFeedbackInfo,
@@ -657,7 +667,7 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         FileInstallCmd = install_cmd_user(InstallCmd, InstallCmdDirOption)
     ),
 
-    globals_init(OptionTable0, OpMode, Target, GC_Method,
+    globals_init(OptionTable0, !.OptTuple, OpMode, Target, GC_Method,
         TermNorm, Term2Norm, TraceLevel, TraceSuppress, SSTraceLevel,
         MaybeThreadSafe, C_CompilerType, CSharp_CompilerType,
         ReuseStrategy, MaybeFeedbackInfo,
@@ -901,7 +911,10 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         globals.set_option(can_compare_constants_as_ints, bool(no), !Globals),
         globals.set_option(can_compare_compound_values, bool(yes), !Globals),
         globals.set_option(order_constructors_for_erlang, bool(yes), !Globals),
-        globals.set_option(optimize_tailcalls, bool(no), !Globals),
+        % XXX opt_mlds_tailcalls is now in the opt_tuple.
+        % However, there is no need to disable it, since the elds backend
+        % pays not attention to its value.
+        % globals.set_option(optimize_mlds_tailcalls, bool(no), !Globals),
 
         % The following options do not directly affect the Erlang backend,
         % however we need to ensure they are set to values that are consistent
@@ -937,8 +950,14 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     option_implies(highlevel_code, gcc_global_registers, bool(no), !Globals),
     option_implies(highlevel_code, asm_labels, bool(no), !Globals),
 
-    % --no-mlds-optimize implies --no-optimize-tailcalls.
-    option_neg_implies(optimize, optimize_tailcalls, bool(no), !Globals),
+    Optimize = !.OptTuple ^ ot_optimize,
+    (
+        Optimize = do_not_optimize,
+        % --no-mlds-optimize implies --no-optimize-tailcalls.
+        !OptTuple ^ ot_opt_mlds_tailcalls := do_not_opt_mlds_tailcalls
+    ;
+        Optimize = optimize
+    ),
 
     % If no --lib-linkage option has been specified, default to the
     % set of all possible linkages.
@@ -1379,7 +1398,7 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     then
         % Switch off string hash switches until these backends implement
         % the hash operations.
-        globals.set_option(string_hash_switch_size, int(999999), !Globals)
+        !OptTuple ^ ot_string_binary_switch_size := 999999
     else
         true
     ),
@@ -1406,7 +1425,13 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     globals.lookup_bool_option(!.Globals, prof_optimized, ProfOptimized),
     (
         ProfOptimized = no,
-        option_implies(profile_deep, allow_inlining, bool(no), !Globals)
+        globals.lookup_bool_option(!.Globals, profile_deep, ProfDeep),
+        (
+            ProfDeep = no
+        ;
+            ProfDeep = yes,
+            !OptTuple ^ ot_allow_inlining := do_not_allow_inlining
+        )
     ;
         ProfOptimized = yes
     ),
@@ -1424,7 +1449,7 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     ( if ExpComp = "" then
         true
     else
-        globals.set_option(allow_inlining, bool(no), !Globals)
+        !OptTuple ^ ot_allow_inlining := do_not_allow_inlining
     ),
 
     % --decl-debug is an extension of --debug
@@ -1476,27 +1501,24 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
             % of the program, which makes it difficult to
             % relate the trace to the source code (although
             % it can be easily related to the transformed HLDS).
-            globals.set_option(allow_inlining, bool(no), !Globals),
-            globals.set_option(optimize_unused_args, bool(no), !Globals),
-            globals.set_option(optimize_higher_order, bool(no), !Globals),
-            globals.set_option(type_specialization, bool(no), !Globals),
-            globals.set_option(user_guided_type_specialization,
-                bool(no), !Globals),
-            globals.set_option(deforestation, bool(no), !Globals),
-            globals.set_option(constraint_propagation,
-                bool(no), !Globals),
-            globals.set_option(local_constraint_propagation,
-                bool(no), !Globals),
-            globals.set_option(optimize_duplicate_calls,
-                bool(no), !Globals),
-            globals.set_option(optimize_constructor_last_call,
-                bool(no), !Globals),
-            globals.set_option(optimize_saved_vars_cell,
-                bool(no), !Globals),
-            globals.set_option(loop_invariants, bool(no), !Globals),
-            globals.set_option(untuple, bool(no), !Globals),
-            globals.set_option(tuple, bool(no), !Globals),
-            globals.set_option(test_after_switch, bool(no), !Globals)
+            !OptTuple ^ ot_allow_inlining := do_not_allow_inlining,
+            !OptTuple ^ ot_opt_unused_args := do_not_opt_unused_args,
+            !OptTuple ^ ot_opt_higher_order := do_not_opt_higher_order,
+            !OptTuple ^ ot_spec_types := do_not_spec_types,
+            !OptTuple ^ ot_spec_types_user_guided
+                := do_not_spec_types_user_guided,
+            !OptTuple ^ ot_deforest := do_not_deforest,
+            !OptTuple ^ ot_prop_constraints := do_not_prop_constraints,
+            !OptTuple ^ ot_prop_local_constraints
+                := do_not_prop_local_constraints,
+            !OptTuple ^ ot_opt_dup_calls := do_not_opt_dup_calls,
+            !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc,
+            !OptTuple ^ ot_opt_svcell := do_not_opt_svcell,
+            !OptTuple ^ ot_opt_loop_invariants := do_not_opt_loop_invariants,
+            !OptTuple ^ ot_untuple := do_not_untuple,
+            !OptTuple ^ ot_tuple := do_not_tuple,
+            !OptTuple ^ ot_opt_test_after_switch
+                := do_not_opt_test_after_switch
         ;
             TraceOptimized = yes
         ),
@@ -1520,20 +1542,20 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         % more complex than simply disabling hijacks, it would be slower
         % as well, except in procedures that would have many nested
         % hijacks, and such code is extremely rare.
-        globals.set_option(allow_hijacks, bool(no), !Globals),
+        !OptTuple ^ ot_allow_hijacks := do_not_allow_hijacks,
 
         % The following option prevents useless variables from cluttering
         % the trace. Its explicit setting removes a source of variability
         % in the goal paths reported by tracing.
-        globals.set_option(excess_assign, bool(yes), !Globals),
+        !OptTuple ^ ot_elim_excess_assigns := do_not_elim_excess_assigns,
 
         % The explicit setting of the following option removes a source
         % of variability in the goal paths reported by tracing.
-        globals.set_option(follow_code, bool(yes), !Globals),
+        !OptTuple ^ ot_opt_follow_code := do_not_opt_follow_code,
 
         % The following option selects a special-case code generator
         % that cannot (yet) implement tracing.
-        globals.set_option(middle_rec, bool(no), !Globals),
+        !OptTuple ^ ot_opt_middle_rec := do_not_opt_middle_rec,
 
         % The following options cause the info required by tracing
         % to be generated.
@@ -1574,13 +1596,13 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
                 words("high level code."), nl],
             add_error(phase_options, DeepHLSpec, !Specs)
         ),
-        globals.set_option(optimize_constructor_last_call, bool(no), !Globals),
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc,
         globals.lookup_bool_option(!.Globals,
             use_lots_of_ho_specialization, LotsOfHOSpec),
         (
             LotsOfHOSpec = yes,
-            globals.set_option(optimize_higher_order, bool(yes), !Globals),
-            globals.set_option(higher_order_size_limit, int(999999), !Globals)
+            !OptTuple ^ ot_opt_higher_order := opt_higher_order,
+            !OptTuple ^ ot_higher_order_size_limit := 999999
         ;
             LotsOfHOSpec = no
         )
@@ -1604,7 +1626,7 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         ; RecordTermSizesAsCells = yes
         )
     then
-        globals.set_option(optimize_constructor_last_call, bool(no), !Globals),
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc,
         % Term size profiling breaks the assumption that even word offsets from
         % the start of the cell are double-word aligned memory addresses.
         %
@@ -1645,17 +1667,16 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         ( SSTraceLevel = shallow
         ; SSTraceLevel = deep
         ),
-        globals.set_option(allow_inlining, bool(no), !Globals),
-        globals.set_option(optimize_unused_args, bool(no), !Globals),
-        globals.set_option(optimize_higher_order, bool(no), !Globals),
-        globals.set_option(type_specialization, bool(no), !Globals),
-        globals.set_option(user_guided_type_specialization, bool(no),
-            !Globals),
-        globals.set_option(deforestation, bool(no), !Globals),
-        globals.set_option(constraint_propagation, bool(no), !Globals),
-        globals.set_option(local_constraint_propagation, bool(no), !Globals),
-        globals.set_option(optimize_duplicate_calls, bool(no), !Globals),
-        globals.set_option(optimize_constructor_last_call, bool(no), !Globals)
+        !OptTuple ^ ot_allow_inlining := do_not_allow_inlining,
+        !OptTuple ^ ot_opt_unused_args := do_not_opt_unused_args,
+        !OptTuple ^ ot_opt_higher_order := do_not_opt_higher_order,
+        !OptTuple ^ ot_spec_types := do_not_spec_types,
+        !OptTuple ^ ot_spec_types_user_guided := do_not_spec_types_user_guided,
+        !OptTuple ^ ot_deforest := do_not_deforest,
+        !OptTuple ^ ot_prop_constraints := do_not_prop_constraints,
+        !OptTuple ^ ot_prop_local_constraints := do_not_prop_local_constraints,
+        !OptTuple ^ ot_opt_dup_calls := do_not_opt_dup_calls,
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc
     ;
         SSTraceLevel = none
     ),
@@ -1665,8 +1686,13 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % XXX we don't pass -ansi to the C compiler anymore.
     option_implies(parallel, ansi_c, bool(no), !Globals),
 
-    option_neg_implies(inline_builtins, constant_propagation, bool(no),
-        !Globals),
+    InlineBuiltins = !.OptTuple ^ ot_inline_builtins,
+    (
+        InlineBuiltins = do_not_inline_builtins,
+        !OptTuple ^ ot_prop_constants := do_not_prop_constants
+    ;
+        InlineBuiltins = inline_builtins
+    ),
 
     % `--optimize-constant-propagation' effectively inlines builtins.
     %
@@ -1679,19 +1705,28 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % e.g. when tracing is enabled.
     (
         ProfileDeep = no,
-        option_neg_implies(allow_inlining, constant_propagation, bool(no),
-            !Globals)
+        AllowInlining = !.OptTuple ^ ot_allow_inlining,
+        (
+            AllowInlining = do_not_allow_inlining,
+            !OptTuple ^ ot_prop_constants := do_not_prop_constants
+        ;
+            AllowInlining = allow_inlining
+        )
     ;
         ProfileDeep = yes
     ),
 
     % --no-reorder-conj implies --no-deforestation,
     % --no-constraint-propagation and --no-local-constraint-propagation.
-    option_neg_implies(reorder_conj, deforestation, bool(no), !Globals),
-    option_neg_implies(reorder_conj, constraint_propagation, bool(no),
-        !Globals),
-    option_neg_implies(reorder_conj, local_constraint_propagation, bool(no),
-        !Globals),
+    globals.lookup_bool_option(!.Globals, reorder_conj, ReorderConj),
+    (
+        ReorderConj = no,
+        !OptTuple ^ ot_deforest := do_not_deforest,
+        !OptTuple ^ ot_prop_constraints := do_not_prop_constraints,
+        !OptTuple ^ ot_prop_local_constraints := do_not_prop_local_constraints
+    ;
+        ReorderConj = yes
+    ),
 
     % --stack-trace requires `procid' stack layouts
     option_implies(stack_trace, procid_stack_layout, bool(yes), !Globals),
@@ -1736,17 +1771,16 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         GC_Method = gc_accurate,
         globals.set_option(agc_stack_layout, bool(yes), !Globals),
         globals.set_option(body_typeinfo_liveness, bool(yes), !Globals),
-        globals.set_option(allow_hijacks, bool(no), !Globals),
-        globals.set_option(optimize_frames, bool(no), !Globals),
+        !OptTuple ^ ot_allow_hijacks := do_not_allow_hijacks,
+        !OptTuple ^ ot_opt_frames := do_not_opt_frames,
         globals.set_option(opt_no_return_calls, bool(no), !Globals),
-        globals.set_option(middle_rec, bool(no), !Globals),
-        globals.set_option(
-            reclaim_heap_on_semidet_failure, bool(no), !Globals),
-        globals.set_option(
-            reclaim_heap_on_nondet_failure, bool(no), !Globals),
-        globals.set_option(optimize_constructor_last_call, bool(no), !Globals),
-        globals.set_option(type_specialization, bool(no), !Globals),
-        globals.set_option(user_guided_type_specialization, bool(no), !Globals)
+        !OptTuple ^ ot_opt_middle_rec := do_not_opt_middle_rec,
+        globals.set_option(reclaim_heap_on_semidet_failure, bool(no),
+            !Globals),
+        globals.set_option(reclaim_heap_on_nondet_failure, bool(no), !Globals),
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc,
+        !OptTuple ^ ot_spec_types := do_not_spec_types,
+        !OptTuple ^ ot_spec_types_user_guided := do_not_spec_types_user_guided
     ;
         ( GC_Method = gc_automatic
         ; GC_Method = gc_none
@@ -1784,21 +1818,39 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % change the return address in a call to a different label whose code
     % is the same but which has a different label layout structure),
     % so we need to disable it when tracing.
-    option_implies(procid_stack_layout, optimize_dups, bool(no), !Globals),
+    globals.lookup_bool_option(!.Globals, procid_stack_layout,
+        ProcIdStackLayout),
+    globals.lookup_bool_option(!.Globals, agc_stack_layout, AgcStackLayout),
     % Likewise for accurate GC.
-    option_implies(agc_stack_layout, optimize_dups, bool(no), !Globals),
+    ( if (ProcIdStackLayout = yes ; AgcStackLayout = yes) then
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc
+    else
+        true
+    ),
 
     % stdlabel.m tries to perform operations that yield compiler aborts
     % if any stack layout information is present in the generated code.
-    option_implies(basic_stack_layout, standardize_labels, bool(no),
-        !Globals),
+    globals.lookup_bool_option(!.Globals, basic_stack_layout,
+        BasicStackLayout),
+    (
+        BasicStackLayout = yes,
+        !OptTuple ^ ot_standardize_labels := do_not_standardize_labels
+    ;
+        BasicStackLayout = no
+    ),
 
-    % XXX deforestation and constraint propagation do not perform
-    % folding on polymorphic predicates correctly with
-    % --body-typeinfo-liveness.
-    option_implies(body_typeinfo_liveness, deforestation, bool(no), !Globals),
-    option_implies(body_typeinfo_liveness, constraint_propagation, bool(no),
-        !Globals),
+    % XXX deforestation and constraint propagation do not perform folding
+    % on polymorphic predicates correctly with --body-typeinfo-liveness.
+    globals.lookup_bool_option(!.Globals, body_typeinfo_liveness,
+        BodyTypeInfoLiveness),
+    (
+        BodyTypeInfoLiveness = yes,
+        !OptTuple ^ ot_deforest := do_not_deforest,
+        !OptTuple ^ ot_prop_constraints := do_not_prop_constraints
+        % XXX What about ot_prop_local_constraints?
+    ;
+        BodyTypeInfoLiveness = no
+    ),
 
     % XXX If trailing is enabled, middle recursion optimization
     % can generate code which does not allocate a stack frame
@@ -1806,14 +1858,18 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % trail, if the code being optimized contains a construct which
     % might save/restore the trail state, i.e. an if-then-else,
     % negation, disjunction, or commit.
-    option_implies(use_trail, middle_rec, bool(no), !Globals),
-
+    %
     % The cut-down stack frames used by middle recursion optimization
     % don't include return addresses. Since stack extension arranges for
     % the return to the old stack segments by overriding the return address,
     % stack extension via stack segments and middle recursion optimization
     % are incompatible.
-    option_implies(stack_segments, middle_rec, bool(no), !Globals),
+    globals.lookup_bool_option(!.Globals, stack_segments, StackSegments),
+    ( if (UseTrail = yes ; StackSegments = yes) then
+        !OptTuple ^ ot_opt_middle_rec := do_not_opt_middle_rec
+    else
+        true
+    ),
 
     % Stack copy minimal model tabling needs to be able to rewrite all
     % the redoips in a given nondet stack segments. If we allow hijacks,
@@ -1822,8 +1878,12 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % Since we want to allow tabling in grades that do not have label
     % layout info, we disable hijacks instead.
     % XXX we should allow hijacks in table_builtin.m
-    option_implies(use_minimal_model_stack_copy, allow_hijacks, bool(no),
-        !Globals),
+    ( 
+        UseMinimalModelStackCopy = yes,
+        !OptTuple ^ ot_allow_hijacks := do_not_allow_hijacks
+    ;
+        UseMinimalModelStackCopy = no
+    ),
 
     % Stack copy minimal model tabling needs to generate extra code
     % at possibly negated contexts to handle the pneg stack and at commits
@@ -1876,31 +1936,51 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
 
     % If we are doing type-specialization, we may as well take advantage
     % of the declarations supplied by the programmer.
-    option_implies(type_specialization, user_guided_type_specialization,
-        bool(yes), !Globals),
+    TypeSpec = !.OptTuple ^ ot_spec_types,
+    (
+        TypeSpec = spec_types,
+        !OptTuple ^ ot_spec_types_user_guided := spec_types_user_guided
+    ;
+        TypeSpec = do_not_spec_types
+    ),
 
     % The local constraint propagation transformation (constraint.m)
     % is a required part of the constraint propagation transformation
     % performed by deforest.m.
-    option_implies(constraint_propagation, local_constraint_propagation,
-        bool(yes), !Globals),
+    ConstProp = !.OptTuple ^ ot_prop_constraints,
+    (
+        ConstProp = prop_constraints,
+        !OptTuple ^ ot_prop_local_constraints := prop_local_constraints
+    ;
+        ConstProp = do_not_prop_constraints
+    ),
 
     % --intermod-unused-args implies --intermodule-optimization and
     % --optimize-unused-args.
-    option_implies(intermod_unused_args, intermodule_optimization, bool(yes),
-        !Globals),
-    option_implies(intermod_unused_args, optimize_unused_args, bool(yes),
-        !Globals),
+    IntermodUnusedArgs = !.OptTuple ^ ot_opt_unused_args_intermod,
+    (
+        IntermodUnusedArgs = opt_unused_args_intermod,
+        !OptTuple ^ ot_opt_unused_args := opt_unused_args,
+        globals.set_option(intermodule_optimization, bool(yes), !Globals)
+    ;
+        IntermodUnusedArgs = do_not_opt_unused_args_intermod
+    ),
 
     % --introduce-accumulators implies --excess-assign and
     % --common-struct.
-    option_implies(introduce_accumulators, excess_assign, bool(yes), !Globals),
-    option_implies(introduce_accumulators, common_struct, bool(yes), !Globals),
+    IntroduceAccumulators = !.OptTuple ^ ot_introduce_accumulators,
+    (
+        IntroduceAccumulators = introduce_accumulators,
+        !OptTuple ^ ot_elim_excess_assigns := elim_excess_assigns,
+        !OptTuple ^ ot_opt_common_structs := opt_common_structs
+    ;
+        IntroduceAccumulators = do_not_introduce_accumulators
+    ),
 
     % Don't do the unused_args optimization when making the
     % optimization interface.
     ( if OpMode = opm_top_args(opma_augment(opmau_make_opt_int)) then
-        globals.set_option(optimize_unused_args, bool(no), !Globals)
+        !OptTuple ^ ot_opt_unused_args := do_not_opt_unused_args
     else
         true
     ),
@@ -2190,22 +2270,26 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
         ; WarnNonTailRecMutual = yes
         )
     then
-        globals.lookup_bool_option(!.Globals, pessimize_tailcalls,
-            PessimizeTailCalls),
-        globals.lookup_bool_option(!.Globals, optimize_tailcalls,
-            OptimizeTailCalls),
+        PessimizeTailCalls = !.OptTuple ^ ot_pessimize_tailcalls,
+        OptimizeTailCalls = !.OptTuple ^ ot_opt_mlds_tailcalls,
         (
-            PessimizeTailCalls = no
+            PessimizeTailCalls = do_not_pessimize_tailcalls
         ;
-            PessimizeTailCalls = yes,
+            PessimizeTailCalls = pessimize_tailcalls,
             PessimizeWords = "--warn-non-tail-recursion is incompatible" ++
                  " with --pessimize-tailcalls",
+            % XXX While these two options look diametrically opposed,
+            % they are actually compatible, because pessimize_tailcalls
+            % is implemented only for the LLDS backend, while the
+            % optimize_tailcalls option is only for the MLDS backend.
+            % (The LLDS backend's tail call optimization does NOT depend
+            % on the value of that option.)
             add_error(phase_options, [words(PessimizeWords)], !Specs)
         ),
         (
-            OptimizeTailCalls = yes
+            OptimizeTailCalls = opt_mlds_tailcalls
         ;
-            OptimizeTailCalls = no,
+            OptimizeTailCalls = do_not_opt_mlds_tailcalls,
             OptimizeWords =
                 "--warn-non-tail-recursion requires --optimize-tailcalls",
             add_error(phase_options, [words(OptimizeWords)], !Specs)
@@ -2234,7 +2318,7 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     ;
         Target = target_erlang,
         BackendForeignLanguages = ["erlang"],
-        set_option(optimize_constructor_last_call, bool(no), !Globals),
+        !OptTuple ^ ot_opt_lcmc := do_not_opt_lcmc,
         set_option(allow_multi_arm_switches, bool(no), !Globals)
     ),
 
@@ -2282,10 +2366,11 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
 
     (
         HighLevelCode = no,
-        postprocess_options_lowlevel(!Globals)
+        postprocess_options_lowlevel(!Globals, !OptTuple)
     ;
         HighLevelCode = yes
     ),
+    globals.set_opt_tuple(!.OptTuple, !Globals),
     postprocess_options_libgrades(!Globals, !Specs),
     globals_init_mutables(!.Globals, !IO).
 
@@ -2299,28 +2384,45 @@ convert_options_to_globals(OptionTable0, OpMode, Target, GC_Method,
     % the user interface unchanged, but would let us handle the implications
     % for the two backends separately.
     %
-:- pred postprocess_options_lowlevel(globals::in, globals::out) is det.
+:- pred postprocess_options_lowlevel(globals::in, globals::out,
+    opt_tuple::in, opt_tuple::out) is det.
 
-postprocess_options_lowlevel(!Globals) :-
+postprocess_options_lowlevel(!Globals, !OptTuple) :-
     % --optimize-saved-vars-cell requires --use-local-vars for
     % acceptable performance.
-    option_implies(optimize_saved_vars_cell, use_local_vars, bool(yes),
-        !Globals),
+    SavedVarsCell = !.OptTuple ^ ot_opt_svcell,
+    (
+        SavedVarsCell = opt_svcell,
+        !OptTuple ^ ot_use_local_vars := use_local_vars
+    ;
+        SavedVarsCell = do_not_opt_svcell
+    ),
 
     % --optimize-frames requires --optimize-labels and
     % --optimize-jumps
-    option_implies(optimize_frames, optimize_labels, bool(yes), !Globals),
-    option_implies(optimize_frames, optimize_jumps, bool(yes), !Globals),
+    OptFrames = !.OptTuple ^ ot_opt_frames,
+    (
+        OptFrames = opt_frames,
+        !OptTuple ^ ot_opt_labels := opt_labels,
+        !OptTuple ^ ot_opt_jumps := opt_jumps
+    ;
+        OptFrames = do_not_opt_frames
+    ),
 
     % --optimize-proc-dups is implemented only with --trad-passes.
-    option_implies(optimize_proc_dups, trad_passes, bool(yes), !Globals),
+    OptProcDups = !.OptTuple ^ ot_opt_proc_dups,
+    (
+        OptProcDups = opt_proc_dups,
+        globals.set_option(trad_passes, bool(yes), !Globals)
+    ;
+        OptProcDups = do_not_opt_proc_dups
+    ),
 
-    globals.lookup_bool_option(!.Globals, optimize_frames, OptFrames),
-    globals.lookup_bool_option(!.Globals, use_local_vars, OptLocalVars),
-    globals.lookup_int_option(!.Globals, optimize_repeat, OptRepeat),
+    UseLocalVars = !.OptTuple ^ ot_use_local_vars,
+    OptRepeat = !.OptTuple ^ ot_opt_repeat,
     ( if
-        ( OptFrames = yes
-        ; OptLocalVars = yes
+        ( OptFrames = opt_frames
+        ; UseLocalVars = use_local_vars
         ),
         OptRepeat < 1
     then
@@ -2328,7 +2430,7 @@ postprocess_options_lowlevel(!Globals) :-
         % the jump and label optimization having been done. They are turned
         % on above, but they still won't be executed unless optimize_repeat
         % is at least one.
-        globals.set_option(optimize_repeat, int(1), !Globals)
+        !OptTuple ^ ot_opt_repeat := 1
     else
         true
     ),
@@ -2338,34 +2440,30 @@ postprocess_options_lowlevel(!Globals) :-
     globals.lookup_bool_option(!.Globals, unboxed_float, UnboxedFloat),
     (
         UnboxedFloat = yes,
-        % If we're using unboxed (MR_Word-sized) floats, floating point values
+        % If we are using unboxed (MR_Word-sized) floats, floating point values
         % are always constants.
-        StaticGroundFloats = yes
+        !OptTuple ^ ot_use_static_ground_floats := use_static_ground_floats
     ;
         UnboxedFloat = no,
-        % If we're using boxed floats, then we can generate a static constant
+        % If we are using boxed floats, then we can generate a static constant
         % variable to hold a float constant, and gcc doesn't mind us converting
         % from its address to MR_Word in a static initializer. In theory,
         % we should do this with --static-ground-terms. However, the code
         % generator does not yet handle the dynamic creation of boxed float
         % constants, and assumes that binding a variable to a constant
         % generates no code.
-        StaticGroundFloats = yes
+        !OptTuple ^ ot_use_static_ground_floats := use_static_ground_floats
     ),
-    globals.set_option(static_ground_floats, bool(StaticGroundFloats),
-        !Globals),
 
     % Ditto for 64-bit integers.
     globals.lookup_bool_option(!.Globals, unboxed_int64s, UnboxedInt64s),
     (
         UnboxedInt64s = yes,
-        StaticGroundInt64s = yes
+        !OptTuple ^ ot_use_static_ground_int64s := use_static_ground_int64s
     ;
         UnboxedInt64s = no,
-        StaticGroundInt64s = yes
+        !OptTuple ^ ot_use_static_ground_int64s := use_static_ground_int64s
     ),
-    globals.set_option(static_ground_int64s, bool(StaticGroundInt64s),
-        !Globals),
 
     % The setting of static_code_addresses is governed only by the settings
     % of gcc_non_local_gotos and asm_labels.
@@ -2380,11 +2478,11 @@ postprocess_options_lowlevel(!Globals) :-
         % of these global variables is not constant (i.e. not computable at
         % load time), since they can't be initialized until we call
         % init_modules().
-        StaticCodeAddrs = no
+        !OptTuple ^ ot_use_static_code_addresses
+            := do_not_use_static_code_addresses
     else
-        StaticCodeAddrs = yes
-    ),
-    globals.set_option(static_code_addresses, bool(StaticCodeAddrs), !Globals).
+        !OptTuple ^ ot_use_static_code_addresses := use_static_code_addresses
+    ).
 
     % option_implies(SourceBoolOption, ImpliedOption, ImpliedOptionValue):
     % If the SourceBoolOption is set to yes, then the ImpliedOption is set
@@ -2408,6 +2506,7 @@ option_implies(SourceOption, ImpliedOption, ImpliedOptionValue, !Globals) :-
     %
 :- pred option_neg_implies(option::in, option::in, option_data::in,
     globals::in, globals::out) is det.
+:- pragma consider_used(option_neg_implies/5).
 
 option_neg_implies(SourceOption, ImpliedOption, ImpliedOptionValue,
         !Globals) :-
diff --git a/compiler/higher_order.m b/compiler/higher_order.m
index de8997e9c..93ba08d4d 100644
--- a/compiler/higher_order.m
+++ b/compiler/higher_order.m
@@ -71,6 +71,7 @@
 :- import_module libs.
 :- import_module libs.file_util.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
@@ -106,14 +107,14 @@ specialize_higher_order(!ModuleInfo, !IO) :-
     % requests remaining.
 
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, optimize_higher_order, HigherOrder),
-    globals.lookup_bool_option(Globals, type_specialization, TypeSpec),
-    globals.lookup_bool_option(Globals, user_guided_type_specialization,
-        UserTypeSpec),
-    globals.lookup_int_option(Globals, higher_order_size_limit, SizeLimit),
-    globals.lookup_int_option(Globals, higher_order_arg_limit, ArgLimit),
-    Params = ho_params(HigherOrder, TypeSpec, UserTypeSpec, SizeLimit,
-        ArgLimit),
+    globals.get_opt_tuple(Globals, OptTuple),
+    HigherOrder = OptTuple ^ ot_opt_higher_order,
+    TypeSpec = OptTuple ^ ot_spec_types,
+    UserTypeSpec = OptTuple ^ ot_spec_types_user_guided,
+    SizeLimit = OptTuple ^ ot_higher_order_size_limit,
+    ArgLimit = OptTuple ^ ot_higher_order_arg_limit,
+    Params =
+        ho_params(HigherOrder, TypeSpec, UserTypeSpec, SizeLimit, ArgLimit),
     map.init(NewPredMap0),
     map.init(GoalSizes0),
     set.init(Requests0),
@@ -143,22 +144,26 @@ specialize_higher_order(!ModuleInfo, !IO) :-
                 NonUserSpecPredIdSet),
             set.to_sorted_list(NonUserSpecPredIdSet, NonUserSpecPredIds),
 
-            !GlobalInfo ^ hogi_params ^ param_do_user_type_spec := yes,
+            !GlobalInfo ^ hogi_params ^ param_do_user_type_spec
+                := spec_types_user_guided,
             list.foldl(get_specialization_requests, UserSpecPredIds,
                 !GlobalInfo),
             process_ho_spec_requests(!GlobalInfo, !IO)
         ),
 
-        bool.or_list([HigherOrder, TypeSpec, UserTypeSpec], AnySpec),
-        (
-            AnySpec = yes,
+        ( if
+            ( HigherOrder = opt_higher_order
+            ; TypeSpec = spec_types
+            ; UserTypeSpec = spec_types_user_guided
+            )
+        then
             % Process all other specializations until no more requests
             % are generated.
             list.foldl(get_specialization_requests, NonUserSpecPredIds,
                 !GlobalInfo),
             recursively_process_ho_spec_requests(!GlobalInfo, !IO)
-        ;
-            AnySpec = no
+        else
+            true
         ),
 
         % Remove the predicates which were used to force the production of
@@ -352,13 +357,13 @@ recursively_process_ho_spec_requests(!GlobalInfo, !IO) :-
 :- type ho_params
     --->    ho_params(
                 % Propagate higher-order constants.
-                param_do_higher_order_spec  :: bool,
+                param_do_higher_order_spec  :: maybe_opt_higher_order,
 
                 % Propagate type-info constants.
-                param_do_type_spec          :: bool,
+                param_do_type_spec          :: maybe_spec_types,
 
                 % User-guided type specialization.
-                param_do_user_type_spec     :: bool,
+                param_do_user_type_spec     :: maybe_spec_types_user_guided,
 
                 % Size limit on requested version.
                 param_size_limit            :: int,
@@ -1013,10 +1018,24 @@ is_interesting_cons_id(Params, ConsId) = IsInteresting :-
         ; ConsId = type_info_const(_)
         ; ConsId = typeclass_info_const(_)
         ),
-        IsInteresting = Params ^ param_do_user_type_spec
+        UserTypeSpec = Params ^ param_do_user_type_spec,
+        (
+            UserTypeSpec = spec_types_user_guided,
+            IsInteresting = yes
+        ;
+            UserTypeSpec = do_not_spec_types_user_guided,
+            IsInteresting = no
+        )
     ;
         ConsId = closure_cons(_, _),
-        IsInteresting = Params ^ param_do_higher_order_spec
+        HigherOrder = Params ^ param_do_higher_order_spec,
+        (
+            HigherOrder = opt_higher_order,
+            IsInteresting = yes
+        ;
+            HigherOrder = do_not_opt_higher_order,
+            IsInteresting = no
+        )
     ).
 
     % Process a higher-order call to see if it could possibly be specialized.
@@ -1513,7 +1532,7 @@ maybe_specialize_ordinary_call(CanRequest, CalledPred, CalledProc,
             IsUserSpecProc = yes
         ;
             !.Info ^ hoi_global_info ^ hogi_params ^ param_do_user_type_spec
-                = yes,
+                = spec_types_user_guided,
             lookup_var_types(VarTypes, Args0, ArgTypes),
 
             % Check whether any typeclass constraints now match an instance.
@@ -1763,7 +1782,7 @@ find_matching_version(Info, CalledPred, CalledProc, Args0, Context,
         TypeSpec = Params ^ param_do_type_spec,
         UserTypeSpec = Params ^ param_do_user_type_spec,
         (
-            UserTypeSpec = yes,
+            UserTypeSpec = spec_types_user_guided,
             IsUserSpecProc = yes
         ;
             module_info_pred_info(ModuleInfo, CalledPred, CalledPredInfo),
@@ -1773,7 +1792,7 @@ find_matching_version(Info, CalledPred, CalledProc, Args0, Context,
                 % to call the class methods for a specific instance. Without
                 % this, user-specified specialized versions of class methods
                 % won't be called.
-                UserTypeSpec = yes,
+                UserTypeSpec = spec_types_user_guided,
                 pred_info_get_markers(CalledPredInfo, Markers),
                 (
                     check_marker(Markers, marker_class_method)
@@ -1781,11 +1800,11 @@ find_matching_version(Info, CalledPred, CalledProc, Args0, Context,
                     check_marker(Markers, marker_class_instance_method)
                 )
             ;
-                HigherOrder = yes,
+                HigherOrder = opt_higher_order,
                 list.member(HOArg, HigherOrderArgs),
                 HOArg ^ hoa_cons_id = closure_cons(_, _)
             ;
-                TypeSpec = yes
+                TypeSpec = spec_types
             )
         )
     then
@@ -1946,7 +1965,7 @@ version_matches(Params, ModuleInfo, Request, Version, Match) :-
         not check_marker(Markers, marker_class_method),
         not check_marker(Markers, marker_class_instance_method),
         (
-            Params ^ param_do_type_spec = no
+            Params ^ param_do_type_spec = do_not_spec_types
         ;
             pred_info_is_imported(CalleePredInfo)
         )
diff --git a/compiler/hlds_pred.m b/compiler/hlds_pred.m
index d73a3beba..61da2a0a2 100644
--- a/compiler/hlds_pred.m
+++ b/compiler/hlds_pred.m
@@ -883,6 +883,8 @@
 
 :- implementation.
 
+:- import_module libs.optimization_options.
+
 :- type pred_id
     --->    pred_id(int).
 
@@ -3883,13 +3885,10 @@ builtin_state(ModuleInfo, CallerPredId, PredId, ProcId) = BuiltinState :-
         is_inline_builtin(ModuleName, PredName, ProcId, Arity),
         (
             module_info_get_globals(ModuleInfo, Globals),
-            globals.lookup_bool_option(Globals, allow_inlining,
-                AllowInlining),
-            AllowInlining = yes,
-            globals.lookup_bool_option(Globals, inline_builtins,
-                InlineBuiltins),
+            globals.get_opt_tuple(Globals, OptTuple),
+            OptTuple ^ ot_allow_inlining = allow_inlining,
             (
-                InlineBuiltins = yes
+                OptTuple ^ ot_inline_builtins = inline_builtins
             ;
                 module_info_pred_info(ModuleInfo, PredId, PredInfo),
                 pred_info_get_name(PredInfo, PredName),
diff --git a/compiler/inlining.m b/compiler/inlining.m
index 240d75f26..b1729bb1a 100644
--- a/compiler/inlining.m
+++ b/compiler/inlining.m
@@ -170,6 +170,7 @@
 :- import_module libs.
 :- import_module libs.dependency_graph.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module mdbcomp.
@@ -198,9 +199,10 @@
     %
 :- type inline_params
     --->    inline_params(
-                ip_simple                       :: bool,
-                ip_single_use                   :: bool,
-                ip_linear_tail_rec              :: bool,
+                ip_simple                       :: maybe_inline_simple,
+                ip_single_use                   :: maybe_inline_single_use,
+                ip_linear_tail_rec              ::
+                                            maybe_inline_linear_tail_rec_sccs,
 
                 ip_highlevel_code               :: bool,
 
@@ -228,18 +230,16 @@ inline_in_module(!ModuleInfo) :-
     % - whether we're in an MLDS grade
 
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, inline_simple, Simple),
-    globals.lookup_bool_option(Globals, inline_single_use, SingleUse),
-    globals.lookup_bool_option(Globals, inline_linear_tail_rec_sccs,
-        LinearTailRec),
-    globals.lookup_int_option(Globals, inline_linear_tail_rec_sccs_max_extra,
-        LinearTailRecMaxExtra),
-    globals.lookup_int_option(Globals, inline_call_cost, CallCost),
-    globals.lookup_int_option(Globals, inline_compound_threshold,
-        CompoundThreshold),
-    globals.lookup_int_option(Globals, inline_simple_threshold,
-        SimpleThreshold),
-    globals.lookup_int_option(Globals, inline_vars_threshold, VarThreshold),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Simple = OptTuple ^ ot_inline_simple,
+    SingleUse = OptTuple ^ ot_inline_single_use,
+    LinearTailRec = OptTuple ^ ot_inline_linear_tail_rec_sccs,
+    LinearTailRecMaxExtra =
+        OptTuple ^ ot_inline_linear_tail_rec_sccs_max_extra,
+    CallCost = OptTuple ^ ot_inline_call_cost,
+    CompoundThreshold = OptTuple ^ ot_inline_compound_threshold,
+    SimpleThreshold = OptTuple ^ ot_inline_simple_threshold,
+    VarThreshold = OptTuple ^ ot_inline_vars_threshold,
     globals.lookup_bool_option(Globals, highlevel_code, HighLevelCode),
     globals.get_trace_level(Globals, TraceLevel),
     AnyTracing = bool.not(given_trace_level_is_none(TraceLevel)),
@@ -247,7 +247,7 @@ inline_in_module(!ModuleInfo) :-
     % Get the usage counts for predicates (but only if needed, i.e. only if
     % --inline-single-use or --inline-compound-threshold has been specified).
     ( if
-        ( SingleUse = yes
+        ( SingleUse = inline_single_use
         ; CompoundThreshold > 0
         )
     then
@@ -267,10 +267,10 @@ inline_in_module(!ModuleInfo) :-
     % calls where the callee is in the *same* SCC as the caller.
 
     (
-        LinearTailRec = no,
+        LinearTailRec = do_not_inline_linear_tail_rec_sccs,
         module_info_ensure_dependency_info(!ModuleInfo, DepInfo)
     ;
-        LinearTailRec = yes,
+        LinearTailRec = inline_linear_tail_rec_sccs,
         % For this, we need *accurate* information about SCCs.
         % I (zs) am not 100% certain that every pass before this one
         % that invalidates any existing dependency info also clobbers
@@ -319,11 +319,11 @@ inline_in_scc(Params, SCCEntryPoints, !ShouldInlineProcs, !ModuleInfo) :-
         SCCProcs = [_, _ | _],
         LinearTailRec = Params ^ ip_linear_tail_rec,
         (
-            LinearTailRec = no,
+            LinearTailRec = do_not_inline_linear_tail_rec_sccs,
             inline_in_simple_non_singleton_scc(Params, SCCProcs,
                 !ShouldInlineProcs, !ModuleInfo)
         ;
-            LinearTailRec = yes,
+            LinearTailRec = inline_linear_tail_rec_sccs,
             inline_in_maybe_linear_tail_rec_scc(Params,
                 SCCEntryPoints, SCCProcs, !ShouldInlineProcs, !ModuleInfo)
         )
@@ -499,8 +499,7 @@ should_proc_be_inlined(Params, ModuleInfo, PredProcId) :-
 
     % The heuristic represented by the following code could be improved.
     (
-        Simple = Params ^ ip_simple,
-        Simple = yes,
+        Params ^ ip_simple = inline_simple,
         SimpleThreshold = Params ^ ip_simple_goal_threshold,
         is_simple_goal(CalledGoal, SimpleThreshold)
     ;
@@ -516,8 +515,7 @@ should_proc_be_inlined(Params, ModuleInfo, PredProcId) :-
         CompoundThreshold = Params ^ ip_compound_size_threshold,
         (Size - CallCost) * NumUses =< CompoundThreshold
     ;
-        SingleUse = Params ^ ip_single_use,
-        SingleUse = yes,
+        Params ^ ip_single_use = inline_single_use,
         NeededMap = Params ^ ip_needed_map,
         map.search(NeededMap, Entity, Needed),
         Needed = maybe_eliminable(NumUses),
diff --git a/compiler/intermod.m b/compiler/intermod.m
index 3e0626aa9..9f61a522f 100644
--- a/compiler/intermod.m
+++ b/compiler/intermod.m
@@ -228,6 +228,7 @@
 :- import_module libs.file_util.
 :- import_module libs.globals.
 :- import_module libs.lp_rational.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.polyhedron.
 :- import_module mdbcomp.
@@ -285,11 +286,10 @@ write_initial_opt_file(TmpOptStream, ModuleInfo, IntermodInfo,
 
 decide_what_to_opt_export(ModuleInfo, !:IntermodInfo) :-
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_int_option(Globals, intermod_inline_simple_threshold,
-        Threshold),
-    globals.lookup_bool_option(Globals, deforestation, Deforestation),
-    globals.lookup_int_option(Globals, higher_order_size_limit,
-        HigherOrderSizeLimit),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Threshold = OptTuple ^ ot_intermod_inline_simple_threshold,
+    HigherOrderSizeLimit = OptTuple ^ ot_higher_order_size_limit,
+    Deforest = OptTuple ^ ot_deforest,
 
     module_info_get_valid_pred_ids(ModuleInfo, RealPredIds),
     module_info_get_assertion_table(ModuleInfo, AssertionTable),
@@ -297,34 +297,35 @@ decide_what_to_opt_export(ModuleInfo, !:IntermodInfo) :-
     PredIds = AssertPredIds ++ RealPredIds,
 
     init_intermod_info(ModuleInfo, !:IntermodInfo),
-    gather_opt_export_preds(PredIds, yes, Threshold,
-        HigherOrderSizeLimit, Deforestation, !IntermodInfo),
+    gather_opt_export_preds(PredIds, yes, Threshold, HigherOrderSizeLimit,
+        Deforest, !IntermodInfo),
     gather_opt_export_instances(!IntermodInfo),
     gather_opt_export_types(!IntermodInfo).
 
 %---------------------------------------------------------------------------%
 
 :- pred gather_opt_export_preds(list(pred_id)::in, bool::in, int::in, int::in,
-    bool::in, intermod_info::in, intermod_info::out) is det.
+    maybe_deforest::in, intermod_info::in, intermod_info::out) is det.
 
 gather_opt_export_preds(AllPredIds, CollectTypes,
-        InlineThreshold, HigherOrderSizeLimit, Deforestation, !IntermodInfo) :-
+        InlineThreshold, HigherOrderSizeLimit, Deforest, !IntermodInfo) :-
     % First gather exported preds.
     ProcessLocalPreds = no,
     gather_opt_export_preds_in_list(AllPredIds, ProcessLocalPreds,
-        CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforestation,
+        CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforest,
         !IntermodInfo),
 
     % Then gather preds used by exported preds (recursively).
     set.init(ExtraExportedPreds0),
     gather_opt_export_preds_fixpoint(ExtraExportedPreds0, CollectTypes,
-        InlineThreshold, HigherOrderSizeLimit, Deforestation, !IntermodInfo).
+        InlineThreshold, HigherOrderSizeLimit, Deforest, !IntermodInfo).
 
 :- pred gather_opt_export_preds_fixpoint(set(pred_id)::in, bool::in,
-    int::in, int::in, bool::in, intermod_info::in, intermod_info::out) is det.
+    int::in, int::in, maybe_deforest::in,
+    intermod_info::in, intermod_info::out) is det.
 
 gather_opt_export_preds_fixpoint(ExtraExportedPreds0, CollectTypes,
-        InlineThreshold, HigherOrderSizeLimit, Deforestation, !IntermodInfo) :-
+        InlineThreshold, HigherOrderSizeLimit, Deforest, !IntermodInfo) :-
     intermod_info_get_pred_decls(!.IntermodInfo, ExtraExportedPreds),
     NewlyExportedPreds = set.to_sorted_list(
         set.difference(ExtraExportedPreds, ExtraExportedPreds0)),
@@ -334,19 +335,20 @@ gather_opt_export_preds_fixpoint(ExtraExportedPreds0, CollectTypes,
         NewlyExportedPreds = [_ | _],
         ProcessLocalPreds = yes,
         gather_opt_export_preds_in_list(NewlyExportedPreds, ProcessLocalPreds,
-            CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforestation,
+            CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforest,
             !IntermodInfo),
         gather_opt_export_preds_fixpoint(ExtraExportedPreds, CollectTypes,
-            InlineThreshold, HigherOrderSizeLimit, Deforestation,
+            InlineThreshold, HigherOrderSizeLimit, Deforest,
             !IntermodInfo)
     ).
 
 :- pred gather_opt_export_preds_in_list(list(pred_id)::in, bool::in, bool::in,
-    int::in, int::in, bool::in, intermod_info::in, intermod_info::out) is det.
+    int::in, int::in, maybe_deforest::in,
+    intermod_info::in, intermod_info::out) is det.
 
 gather_opt_export_preds_in_list([], _, _, _, _, _, !IntermodInfo).
 gather_opt_export_preds_in_list([PredId | PredIds], ProcessLocalPreds,
-        CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforestation,
+        CollectTypes, InlineThreshold, HigherOrderSizeLimit, Deforest,
         !IntermodInfo) :-
     intermod_info_get_module_info(!.IntermodInfo, ModuleInfo),
     module_info_get_preds(ModuleInfo, PredTable),
@@ -359,7 +361,7 @@ gather_opt_export_preds_in_list([PredId | PredIds], ProcessLocalPreds,
         vartypes_is_empty(ExplicitVarTypes),
         should_opt_export_pred(ModuleInfo, PredId, PredInfo,
             ProcessLocalPreds, TypeSpecForcePreds, InlineThreshold,
-            HigherOrderSizeLimit, Deforestation)
+            HigherOrderSizeLimit, Deforest)
     then
         SavedIntermodInfo = !.IntermodInfo,
         % Write a declaration to the `.opt' file for
@@ -397,14 +399,15 @@ gather_opt_export_preds_in_list([PredId | PredIds], ProcessLocalPreds,
         true
     ),
     gather_opt_export_preds_in_list(PredIds, ProcessLocalPreds, CollectTypes,
-        InlineThreshold, HigherOrderSizeLimit, Deforestation, !IntermodInfo).
+        InlineThreshold, HigherOrderSizeLimit, Deforest, !IntermodInfo).
 
 :- pred should_opt_export_pred(module_info::in, pred_id::in, pred_info::in,
-    bool::in, set(pred_id)::in, int::in, int::in, bool::in) is semidet.
+    bool::in, set(pred_id)::in, int::in, int::in, maybe_deforest::in)
+    is semidet.
 
 should_opt_export_pred(ModuleInfo, PredId, PredInfo, ProcessLocalPreds,
         TypeSpecForcePreds, InlineThreshold, HigherOrderSizeLimit,
-        Deforestation) :-
+        Deforest) :-
     (
         ProcessLocalPreds = no,
         ( pred_info_is_exported(PredInfo)
@@ -421,14 +424,15 @@ should_opt_export_pred(ModuleInfo, PredId, PredInfo, ProcessLocalPreds,
     ;
         may_opt_export_pred(PredId, PredInfo, TypeSpecForcePreds),
         opt_exporting_pred_is_likely_worthwhile(ModuleInfo, PredId, PredInfo,
-            InlineThreshold, HigherOrderSizeLimit, Deforestation)
+            InlineThreshold, HigherOrderSizeLimit, Deforest)
     ).
 
 :- pred opt_exporting_pred_is_likely_worthwhile(module_info::in,
-    pred_id::in, pred_info::in, int::in, int::in, bool::in) is semidet.
+    pred_id::in, pred_info::in, int::in, int::in, maybe_deforest::in)
+    is semidet.
 
 opt_exporting_pred_is_likely_worthwhile(ModuleInfo, PredId, PredInfo,
-        InlineThreshold, HigherOrderSizeLimit, Deforestation) :-
+        InlineThreshold, HigherOrderSizeLimit, Deforest) :-
     pred_info_get_clauses_info(PredInfo, ClauseInfo),
     clauses_info_get_clauses_rep(ClauseInfo, ClausesRep, _ItemNumbers),
     get_clause_list_maybe_repeated(ClausesRep, Clauses),
@@ -449,7 +453,7 @@ opt_exporting_pred_is_likely_worthwhile(ModuleInfo, PredId, PredInfo,
         clause_list_size(Clauses, GoalSize),
         GoalSize =< HigherOrderSizeLimit + Arity
     ;
-        Deforestation = yes,
+        Deforest = deforest,
         % Double the inline-threshold since goals we want to deforest
         % will have at least two disjuncts. This allows one simple goal
         % in each disjunct. The disjunction adds one to the goal size,
diff --git a/compiler/ite_gen.m b/compiler/ite_gen.m
index ccd3595ee..6bfa10d41 100644
--- a/compiler/ite_gen.m
+++ b/compiler/ite_gen.m
@@ -50,6 +50,7 @@
 :- import_module hlds.instmap.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.code_gen.
 :- import_module ll_backend.continuation_info.
@@ -316,7 +317,8 @@ generate_negation(CodeModel, Goal0, NotGoalInfo, Code, !CI, !CLD) :-
         GoalExpr = unify(_, _, _, simple_test(L, R), _),
         failure_is_direct_branch(!.CLD, CodeAddr),
         get_globals(!.CI, Globals),
-        globals.lookup_bool_option(Globals, simple_neg, yes)
+        globals.get_opt_tuple(Globals, OptTuple),
+        OptTuple ^ ot_opt_simple_neg = opt_simple_neg
     then
         % Because we are generating the negated goal ourselves, we need to
         % apply the pre- and post-goal updates that would normally be applied
diff --git a/compiler/jumpopt.m b/compiler/jumpopt.m
index b393cd7c8..42e138a67 100644
--- a/compiler/jumpopt.m
+++ b/compiler/jumpopt.m
@@ -16,6 +16,8 @@
 :- module ll_backend.jumpopt.
 :- interface.
 
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module ll_backend.llds.
 :- import_module mdbcomp.
 :- import_module mdbcomp.prim_data.
@@ -55,7 +57,8 @@
     % by the optimization.
     %
 :- pred optimize_jumps_in_proc(set_tree234(label)::in, may_alter_rtti::in,
-    proc_label::in, bool::in, bool::in, bool::in, bool::in,
+    proc_label::in, maybe_opt_fulljumps::in, bool::in,
+    maybe_pessimize_tailcalls::in, maybe_opt_checked_nondet_tailcalls::in,
     counter::in, counter::out, list(instruction)::in, list(instruction)::out,
     bool::out) is det.
 
@@ -123,9 +126,9 @@ optimize_jumps_in_proc(LayoutLabels, MayAlterRtti, ProcLabel, Fulljumpopt,
             !ProcMap, !SdprocMap, !SuccMap),
         jump_opt_build_forkmap(!.Instrs, !.SdprocMap, map.init, !:ForkMap),
         (
-            PessimizeTailCalls = no
+            PessimizeTailCalls = do_not_pessimize_tailcalls
         ;
-            PessimizeTailCalls = yes,
+            PessimizeTailCalls = pessimize_tailcalls,
             !:ProcMap = map.init,
             !:SdprocMap = map.init,
             !:SuccMap = map.init,
@@ -135,7 +138,7 @@ optimize_jumps_in_proc(LayoutLabels, MayAlterRtti, ProcLabel, Fulljumpopt,
             !.ProcMap, !.SdprocMap, !.ForkMap, !.SuccMap, LayoutLabels,
             Fulljumpopt, MayAlterRtti),
         (
-            CheckedNondetTailCall = yes,
+            CheckedNondetTailCall = opt_checked_nondet_tailcalls,
             CheckedNondetTailCallInfo0 =
                 check_nondet_tailcalls(ProcLabel, !.LabelNumCounter),
             jump_opt_instr_list(!.Instrs, comment(""), JumpOptInfo,
@@ -149,7 +152,7 @@ optimize_jumps_in_proc(LayoutLabels, MayAlterRtti, ProcLabel, Fulljumpopt,
                 unexpected($pred, "lost the next label number")
             )
         ;
-            CheckedNondetTailCall = no,
+            CheckedNondetTailCall = do_not_opt_checked_nondet_tailcalls,
             CheckedNondetTailCallInfo0 = dont_check_nondet_tailcalls,
             jump_opt_instr_list(!.Instrs, comment(""), JumpOptInfo,
                 CheckedNondetTailCallInfo0, _, [], RevInstrs)
@@ -259,7 +262,7 @@ jump_opt_build_forkmap([llds_instr(Uinstr, _Comment) | Instrs], SdprocMap,
                 joi_fork_map        :: tailmap,
                 joi_succ_map        :: tailmap,
                 joi_layout_labels   :: set_tree234(label),
-                joi_full_jump_opt   :: bool,
+                joi_full_jump_opt   :: maybe_opt_fulljumps,
                 joi_may_alter_rtti  :: may_alter_rtti
             ).
 
@@ -681,7 +684,7 @@ jump_opt_goto(Uinstr0, Comment0, Instrs0, PrevInstr, JumpOptInfo,
             % is necessary because another optimization below eliminates
             % labels, which is correct only if jumps to those labels are
             % short-circuited everywhere.
-            JumpOptInfo ^ joi_full_jump_opt = yes,
+            JumpOptInfo ^ joi_full_jump_opt = opt_fulljumps,
             InstrMap = JumpOptInfo ^ joi_instr_map,
             map.search(InstrMap, TargetLabel, TargetInstr),
             final_dest(InstrMap, TargetLabel, DestLabel,
@@ -816,7 +819,7 @@ jump_opt_if_val(Uinstr0, Comment0, Instrs0, _PrevInstr, JumpOptInfo,
             % and get the code processed again starting after the if_val,
             % to get the recursive call to replace the goto to L1
             % with the code at L1.
-            Fulljumpopt = yes,
+            Fulljumpopt = opt_fulljumps,
             map.search(BlockMap, TargetLabel, _TargetBlock),
             opt_util.skip_comments(Instrs0, Instrs1),
             Instrs1 = [GotoInstr | AfterGoto],
diff --git a/compiler/libs.m b/compiler/libs.m
index 22542197c..815a6d6d7 100644
--- a/compiler/libs.m
+++ b/compiler/libs.m
@@ -16,6 +16,7 @@
 :- include_module globals.
 :- include_module options.
 :- include_module handle_options.
+:- include_module optimization_options.
 :- include_module op_mode.
 :- include_module compute_grade.
 :- include_module trace_params.
diff --git a/compiler/llds_out_code_addr.m b/compiler/llds_out_code_addr.m
index 886133704..050e37d24 100644
--- a/compiler/llds_out_code_addr.m
+++ b/compiler/llds_out_code_addr.m
@@ -80,6 +80,8 @@
 
 :- import_module backend_libs.
 :- import_module backend_libs.name_mangle.
+:- import_module libs.
+:- import_module libs.optimization_options.
 
 :- import_module int.
 :- import_module string.
@@ -149,10 +151,10 @@ need_code_addr_decls(Info, CodeAddr, Need) :-
         ),
         UseMacro = Info ^ lout_use_macro_for_redo_fail,
         (
-            UseMacro = yes,
+            UseMacro = use_macro_for_redo_fail,
             Need = no
         ;
-            UseMacro = no,
+            UseMacro = do_not_use_macro_for_redo_fail,
             Need = yes
         )
     ).
@@ -177,18 +179,18 @@ output_code_addr_decls(Info, CodeAddr, !IO) :-
         CodeAddr = do_redo,
         UseMacro = Info ^ lout_use_macro_for_redo_fail,
         (
-            UseMacro = yes
+            UseMacro = use_macro_for_redo_fail
         ;
-            UseMacro = no,
+            UseMacro = do_not_use_macro_for_redo_fail,
             io.write_string("MR_declare_entry(MR_do_redo);\n", !IO)
         )
     ;
         CodeAddr = do_fail,
         UseMacro = Info ^ lout_use_macro_for_redo_fail,
         (
-            UseMacro = yes
+            UseMacro = use_macro_for_redo_fail
         ;
-            UseMacro = no,
+            UseMacro = do_not_use_macro_for_redo_fail,
             io.write_string("MR_declare_entry(MR_do_fail);\n", !IO)
         )
     ;
diff --git a/compiler/llds_out_data.m b/compiler/llds_out_data.m
index fa14f4daa..fc8a11dcb 100644
--- a/compiler/llds_out_data.m
+++ b/compiler/llds_out_data.m
@@ -170,6 +170,8 @@
 :- import_module ll_backend.layout_out.
 :- import_module ll_backend.llds_out.llds_out_code_addr.
 :- import_module ll_backend.rtti_out.
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
 :- import_module mdbcomp.sym_name.
@@ -734,7 +736,7 @@ output_record_rval_decls_format(Info, Rval, FirstIndent, LaterIndent,
             StaticGroundFloats = Info ^ lout_static_ground_floats,
             ( if
                 UnboxedFloat = no,
-                StaticGroundFloats = yes
+                StaticGroundFloats = use_static_ground_floats
             then
                 float_literal_name(FloatVal, FloatName),
                 FloatLabel = decl_float_label(FloatName),
@@ -759,7 +761,7 @@ output_record_rval_decls_format(Info, Rval, FirstIndent, LaterIndent,
             StaticGroundInt64s = Info ^ lout_static_ground_int64s,
             ( if
                 UnboxedInt64s = no,
-                StaticGroundInt64s = yes
+                StaticGroundInt64s = use_static_ground_int64s
             then
                 Int64Label = decl_int64_label(Int64Val),
                 ( if decl_set_is_member(Int64Label, !.DeclSet) then
@@ -784,7 +786,7 @@ output_record_rval_decls_format(Info, Rval, FirstIndent, LaterIndent,
             StaticGroundInt64s = Info ^ lout_static_ground_int64s,
             ( if
                 UnboxedInt64s = no,
-                StaticGroundInt64s = yes
+                StaticGroundInt64s = use_static_ground_int64s
             then
                 UInt64Label = decl_uint64_label(UInt64Val),
                 ( if decl_set_is_member(UInt64Label, !.DeclSet) then
@@ -846,7 +848,7 @@ output_record_rval_decls_format(Info, Rval, FirstIndent, LaterIndent,
             StaticGroundFloats = Info ^ lout_static_ground_floats,
             ( if
                 UnboxFloat = no,
-                StaticGroundFloats = yes,
+                StaticGroundFloats = use_static_ground_floats,
                 float_const_binop_expr_name(Op, SubRvalA, SubRvalB, FloatName)
             then
                 FloatLabel = decl_float_label(FloatName),
@@ -1704,7 +1706,7 @@ output_float_rval(Info, Rval, IsPtr, !IO) :-
     StaticGroundFloats = Info ^ lout_static_ground_floats,
     ( if
         UnboxFloat = no,
-        StaticGroundFloats = yes,
+        StaticGroundFloats = use_static_ground_floats,
         float_const_expr_name(Rval, FloatName)
     then
         (
@@ -1758,7 +1760,7 @@ output_int64_rval(Info, Rval, IsPtr, !IO) :-
     StaticGroundInt64s = Info ^ lout_static_ground_int64s,
     ( if
         UnboxInt64s = no,
-        StaticGroundInt64s = yes,
+        StaticGroundInt64s = use_static_ground_int64s,
         int64_const_expr_name(Rval, Int64Name)
     then
         (
@@ -1812,7 +1814,7 @@ output_uint64_rval(Info, Rval, IsPtr, !IO) :-
     StaticGroundInt64s = Info ^ lout_static_ground_int64s,
     ( if
         UnboxInt64s = no,
-        StaticGroundInt64s = yes,
+        StaticGroundInt64s = use_static_ground_int64s,
         uint64_const_expr_name(Rval, UInt64Name)
     then
         (
diff --git a/compiler/llds_out_file.m b/compiler/llds_out_file.m
index 8b824426c..215c0fca1 100644
--- a/compiler/llds_out_file.m
+++ b/compiler/llds_out_file.m
@@ -77,7 +77,7 @@
 :- import_module hlds.hlds_pred.
 :- import_module libs.compiler_util.
 :- import_module libs.file_util.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module libs.trace_params.
 :- import_module ll_backend.exprn_aux.
 :- import_module ll_backend.layout.
@@ -351,11 +351,11 @@ annotate_c_procedure(Info, Proc, AnnotatedProc,
     find_cont_labels(Instrs, set_tree234.init, ContLabels),
     EmitCLoops = Info ^ lout_emit_c_loops,
     (
-        EmitCLoops = no,
+        EmitCLoops = do_not_emit_c_loops,
         WhileLabels = set_tree234.init,
         UndefWhileLabels = set_tree234.init
     ;
-        EmitCLoops = yes,
+        EmitCLoops = emit_c_loops,
         find_while_labels(Instrs, set_tree234.init, WhileLabels),
 
         % We compute UndefWhileLabels by starting with an overapproximation,
@@ -1293,26 +1293,26 @@ output_annotated_c_procedure(Info, AnnotatedProc, !IO) :-
 
     LocalThreadEngineBase = Info ^ lout_local_thread_engine_base,
     (
-        LocalThreadEngineBase = yes,
+        LocalThreadEngineBase = use_local_thread_engine_base,
         io.write_string("#ifdef MR_maybe_local_thread_engine_base\n", !IO),
         io.write_string("\t#undef MR_maybe_local_thread_engine_base\n", !IO),
         io.write_string("\t#define MR_maybe_local_thread_engine_base " ++
             "MR_local_thread_engine_base\n", !IO),
         io.write_string("#endif\n", !IO)
     ;
-        LocalThreadEngineBase = no
+        LocalThreadEngineBase = do_not_use_local_thread_engine_base
     ),
     output_instruction_list(Info, Instrs, LabelOutputInfo,
         not_after_layout_label, !IO),
     (
-        LocalThreadEngineBase = yes,
+        LocalThreadEngineBase = use_local_thread_engine_base,
         io.write_string("#ifdef MR_maybe_local_thread_engine_base\n", !IO),
         io.write_string("\t#undef MR_maybe_local_thread_engine_base\n", !IO),
         io.write_string("\t#define MR_maybe_local_thread_engine_base " ++
             "MR_thread_engine_base\n", !IO),
         io.write_string("#endif\n", !IO)
     ;
-        LocalThreadEngineBase = no
+        LocalThreadEngineBase = do_not_use_local_thread_engine_base
     ).
 
     % Find the entry label for the procedure, for use as the profiling
@@ -1699,7 +1699,8 @@ c_data_linkage_string(DefaultLinkage, BeingDefined) = LinkageStr :-
 c_data_const_string(Globals, InclCodeAddr) =
     ( if
         InclCodeAddr = yes,
-        globals.lookup_bool_option(Globals, static_code_addresses, no)
+        globals.get_opt_tuple(Globals, OptTuple),
+        OptTuple ^ ot_use_static_code_addresses = use_static_code_addresses
     then
         ""
     else
diff --git a/compiler/llds_out_instr.m b/compiler/llds_out_instr.m
index d4d14d0ad..af5c2e2dc 100644
--- a/compiler/llds_out_instr.m
+++ b/compiler/llds_out_instr.m
@@ -85,6 +85,8 @@
 :- import_module hlds.
 :- import_module hlds.hlds_data.
 :- import_module hlds.hlds_pred.
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module ll_backend.layout_out.
 :- import_module ll_backend.llds_out.llds_out_code_addr.
 :- import_module ll_backend.llds_out.llds_out_data.
@@ -645,10 +647,10 @@ output_instruction(Info, Instr, LabelOutputInfo, !IO) :-
         output_label_defn(Label, !IO),
         LocalThreadEngineBase = Info ^ lout_local_thread_engine_base,
         (
-            LocalThreadEngineBase = yes,
+            LocalThreadEngineBase = use_local_thread_engine_base,
             io.write_string("\tMR_MAYBE_INIT_LOCAL_THREAD_ENGINE_BASE\n", !IO)
         ;
-            LocalThreadEngineBase = no
+            LocalThreadEngineBase = do_not_use_local_thread_engine_base
         ),
         maybe_output_update_prof_counter(Info, Label, LabelOutputInfo, !IO)
     ;
@@ -1517,20 +1519,20 @@ output_goto(Info, Target, CallerLabel, !IO) :-
         Target = do_redo,
         UseMacro = Info ^ lout_use_macro_for_redo_fail,
         (
-            UseMacro = yes,
+            UseMacro = use_macro_for_redo_fail,
             io.write_string("MR_redo();\n", !IO)
         ;
-            UseMacro = no,
+            UseMacro = do_not_use_macro_for_redo_fail,
             io.write_string("MR_GOTO(MR_ENTRY(MR_do_redo));\n", !IO)
         )
     ;
         Target = do_fail,
         UseMacro = Info ^ lout_use_macro_for_redo_fail,
         (
-            UseMacro = yes,
+            UseMacro = use_macro_for_redo_fail,
             io.write_string("MR_fail();\n", !IO)
         ;
-            UseMacro = no,
+            UseMacro = do_not_use_macro_for_redo_fail,
             io.write_string("MR_GOTO(MR_ENTRY(MR_do_fail));\n", !IO)
         )
     ;
diff --git a/compiler/llds_out_util.m b/compiler/llds_out_util.m
index 0450bdccb..c221b37f4 100644
--- a/compiler/llds_out_util.m
+++ b/compiler/llds_out_util.m
@@ -21,6 +21,7 @@
 :- import_module hlds.
 :- import_module hlds.hlds_pred.
 :- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module libs.globals.
 :- import_module libs.trace_params.
 :- import_module ll_backend.layout.
@@ -51,19 +52,23 @@
                                                     layout_slot_name),
                 lout_auto_comments              :: maybe_auto_comments,
                 lout_foreign_line_numbers       :: bool,
-                lout_emit_c_loops               :: bool,
+                lout_emit_c_loops               :: maybe_emit_c_loops,
                 lout_generate_bytecode          :: bool,
-                lout_local_thread_engine_base   :: bool,
+                lout_local_thread_engine_base   ::
+                                            maybe_use_local_thread_engine_base,
                 lout_profile_calls              :: bool,
                 lout_profile_time               :: bool,
                 lout_profile_memory             :: bool,
                 lout_profile_deep               :: bool,
                 lout_unboxed_float              :: bool,
                 lout_det_stack_dword_alignment  :: bool,
-                lout_static_ground_floats       :: bool,
+                lout_static_ground_floats       ::
+                                            maybe_use_static_ground_floats,
                 lout_unboxed_int64s             :: bool,
-                lout_static_ground_int64s       :: bool,
-                lout_use_macro_for_redo_fail    :: bool,
+                lout_static_ground_int64s       ::
+                                            maybe_use_static_ground_int64s,
+                lout_use_macro_for_redo_fail    ::
+                                            maybe_use_macro_for_redo_fail,
                 lout_trace_level                :: trace_level,
                 lout_globals                    :: globals
             ).
@@ -136,24 +141,21 @@ init_llds_out_info(ModuleName, SourceFileName, Globals,
     ),
     globals.lookup_bool_option(Globals, line_numbers_around_foreign_code,
         ForeignLineNumbers),
-    globals.lookup_bool_option(Globals, emit_c_loops, EmitCLoops),
+    globals.get_opt_tuple(Globals, OptTuple),
+    EmitCLoops = OptTuple ^ ot_emit_c_loops,
     globals.lookup_bool_option(Globals, generate_bytecode, GenerateBytecode),
-    globals.lookup_bool_option(Globals, local_thread_engine_base,
-        LocalThreadEngineBase),
+    LocalThreadEngineBase = OptTuple ^ ot_use_local_thread_engine_base,
     globals.lookup_bool_option(Globals, profile_calls, ProfileCalls),
     globals.lookup_bool_option(Globals, profile_time, ProfileTime),
     globals.lookup_bool_option(Globals, profile_memory, ProfileMemory),
     globals.lookup_bool_option(Globals, profile_deep, ProfileDeep),
     globals.lookup_bool_option(Globals, unboxed_float, UnboxedFloat),
     double_width_floats_on_det_stack(Globals, DetStackDwordAligment),
-    globals.lookup_bool_option(Globals, static_ground_floats,
-        StaticGroundFloats),
+    StaticGroundFloats = OptTuple ^ ot_use_static_ground_floats,
     globals.lookup_bool_option(Globals, unboxed_int64s, UnboxedInt64s),
     % XXX ARG_PACK We probably need double_width_int64s_on_det_stack.
-    globals.lookup_bool_option(Globals, static_ground_int64s,
-        StaticGroundInt64s),
-    globals.lookup_bool_option(Globals, use_macro_for_redo_fail,
-        UseMacroForRedoFail),
+    StaticGroundInt64s = OptTuple ^ ot_use_static_ground_int64s,
+    UseMacroForRedoFail = OptTuple ^ ot_use_macro_for_redo_fail,
     globals.get_trace_level(Globals, TraceLevel),
     Info = llds_out_info(ModuleName, MangledModuleName, SourceFileName,
         InternalLabelToLayoutMap, EntryLabelToLayoutMap, TableIoEntryMap,
diff --git a/compiler/matching.m b/compiler/matching.m
index fb5c60dd2..da642677a 100644
--- a/compiler/matching.m
+++ b/compiler/matching.m
@@ -19,6 +19,8 @@
 :- module backend_libs.matching.
 :- interface.
 
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module parse_tree.
 :- import_module parse_tree.prog_data.
 :- import_module parse_tree.set_of_var.
@@ -55,7 +57,7 @@
                 field_var_load_cost     :: int,
                 one_path_op_ratio       :: int,
                 one_path_node_ratio     :: int,
-                include_all_candidates  :: bool
+                include_all_candidates  :: maybe_opt_svcell_all_candidates
             ).
 
 :- type benefit_node.
@@ -151,14 +153,14 @@ find_via_cell_vars(CellVar, CandidateFieldVars, CellVarFlushedLater,
         RealizedBenefitNodes, RealizedCostNodes, ViaCellVars) :-
     InclAllCand = MatchingParams ^ include_all_candidates,
     (
-        InclAllCand = no,
+        InclAllCand = do_not_opt_svcell_all_candidates,
         AllSegmentVars = set_of_var.union_list([BeforeFlush | AfterFlush]),
         set_of_var.intersect(CandidateFieldVars, AllSegmentVars,
             OccurringCandidateFieldVars),
         set_of_var.difference(CandidateFieldVars, OccurringCandidateFieldVars,
             NonOccurringCandidateFieldVars)
     ;
-        InclAllCand = yes,
+        InclAllCand = opt_svcell_all_candidates,
         OccurringCandidateFieldVars = CandidateFieldVars,
         NonOccurringCandidateFieldVars = set_of_var.init
     ),
diff --git a/compiler/mercury_compile_front_end.m b/compiler/mercury_compile_front_end.m
index 018038f58..377f5a32e 100644
--- a/compiler/mercury_compile_front_end.m
+++ b/compiler/mercury_compile_front_end.m
@@ -113,6 +113,7 @@
 :- import_module hlds.hlds_statistics.
 :- import_module libs.file_util.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
@@ -609,7 +610,8 @@ mark_entities_in_opt_file_as_opt_exported(IntermodAnalysis,
 :- pred need_middle_pass_for_opt_file(globals::in, bool::out) is det.
 
 need_middle_pass_for_opt_file(Globals, NeedMiddlePassForOptFile) :-
-    globals.lookup_bool_option(Globals, intermod_unused_args, IntermodArgs),
+    globals.get_opt_tuple(Globals, OptTuple),
+    IntermodArgs = OptTuple ^ ot_opt_unused_args_intermod,
     globals.lookup_bool_option(Globals, termination, Termination),
     globals.lookup_bool_option(Globals, termination2, Termination2),
     globals.lookup_bool_option(Globals, structure_sharing_analysis,
@@ -621,7 +623,7 @@ need_middle_pass_for_opt_file(Globals, NeedMiddlePassForOptFile) :-
     globals.lookup_bool_option(Globals, analyse_trail_usage, TrailingAnalysis),
     globals.lookup_bool_option(Globals, analyse_mm_tabling, TablingAnalysis),
     ( if
-        ( IntermodArgs = yes
+        ( IntermodArgs = opt_unused_args_intermod
         ; Termination = yes
         ; Termination2 = yes
         ; ExceptionAnalysis = yes
@@ -1246,15 +1248,15 @@ maybe_simplify(Warn, SimplifyPass, Verbose, Stats, !HLDS, !Specs, !IO) :-
             %
             % NOTE: Any changes made here may also need to be made
             % to the relevant parts of backend_pass_by_preds_4/12.
-            globals.lookup_bool_option(Globals, constant_propagation,
-                ConstProp),
+            globals.get_opt_tuple(Globals, OptTuple),
+            ConstProp = OptTuple ^ ot_prop_constants,
             globals.lookup_bool_option(Globals, profile_deep, DeepProf),
             globals.lookup_bool_option(Globals, record_term_sizes_as_words,
                 TSWProf),
             globals.lookup_bool_option(Globals, record_term_sizes_as_cells,
                 TSCProf),
             ( if
-                ConstProp = yes,
+                ConstProp = prop_constants,
                 DeepProf = no,
                 TSWProf = no,
                 TSCProf = no
diff --git a/compiler/mercury_compile_llds_back_end.m b/compiler/mercury_compile_llds_back_end.m
index cc54c4219..9e7252e19 100644
--- a/compiler/mercury_compile_llds_back_end.m
+++ b/compiler/mercury_compile_llds_back_end.m
@@ -69,6 +69,7 @@
 :- import_module libs.dependency_graph.
 :- import_module libs.file_util.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.continuation_info.
 :- import_module ll_backend.dupproc.
@@ -112,7 +113,8 @@ llds_backend_pass(!HLDS, !:GlobalData, LLDS, !DumpInfo, !IO) :-
     module_info_get_name(!.HLDS, ModuleName),
     module_info_get_globals(!.HLDS, Globals),
     globals.lookup_bool_option(Globals, unboxed_float, OptUnboxFloat),
-    globals.lookup_bool_option(Globals, common_data, DoCommonData),
+    globals.get_opt_tuple(Globals, OptTuple),
+    DoCommonData = OptTuple ^ ot_use_common_data,
     (
         OptUnboxFloat = yes,
         UnboxFloats = have_unboxed_floats
@@ -211,13 +213,14 @@ llds_backend_pass_by_phases(!HLDS, !:LLDS, !GlobalData, !Specs,
 llds_backend_pass_by_preds(!HLDS, LLDS, !GlobalData, !Specs) :-
     module_info_get_valid_pred_ids(!.HLDS, PredIds),
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_proc_dups, ProcDups),
+    globals.get_opt_tuple(Globals, OptTuple),
+    ProcDups = OptTuple ^ ot_opt_proc_dups,
     (
-        ProcDups = no,
+        ProcDups = do_not_opt_proc_dups,
         OrderedPredIds = PredIds,
         MaybeDupProcMap = no
     ;
-        ProcDups = yes,
+        ProcDups = opt_proc_dups,
         PredDepInfo = build_pred_dependency_graph(!.HLDS, PredIds,
             do_not_include_imported),
         OrderedPredIds = dependency_info_get_condensed_bottom_up_sccs(
@@ -355,33 +358,32 @@ llds_backend_pass_for_proc(!HLDS, ConstStructMap, SCCMap, PredId, PredInfo,
         ProcId, !.ProcInfo, CProc, !GlobalData, !Specs) :-
     PredProcId = proc(PredId, ProcId),
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_saved_vars_const,
-        SavedVarsConst),
+    globals.get_opt_tuple(Globals, OptTuple),
+    SavedVarsConst = OptTuple ^ ot_opt_saved_vars_const,
     (
-        SavedVarsConst = yes,
+        SavedVarsConst = opt_saved_vars_const,
         saved_vars_proc(proc(PredId, ProcId), !ProcInfo, !HLDS)
     ;
-        SavedVarsConst = no
+        SavedVarsConst = do_not_opt_saved_vars_const
     ),
-    globals.lookup_bool_option(Globals, optimize_saved_vars_cell,
-        SavedVarsCell),
+    SavedVarsCell = OptTuple ^ ot_opt_svcell,
     (
-        SavedVarsCell = yes,
+        SavedVarsCell = opt_svcell,
         stack_opt_cell(PredProcId, !ProcInfo, !HLDS)
     ;
-        SavedVarsCell = no
+        SavedVarsCell = do_not_opt_svcell
     ),
-    globals.lookup_bool_option(Globals, follow_code, FollowCode),
+    FollowCode = OptTuple ^ ot_opt_follow_code,
     (
-        FollowCode = yes,
+        FollowCode = opt_follow_code,
         move_follow_code_in_proc(PredProcId, !ProcInfo, !HLDS)
     ;
-        FollowCode = no
+        FollowCode = do_not_opt_follow_code
     ),
     find_simplify_tasks(no, Globals, SimplifyTasks0),
     SimpList0 = simplify_tasks_to_list(SimplifyTasks0),
 
-    globals.lookup_bool_option(Globals, constant_propagation, ConstProp),
+    ConstProp = OptTuple ^ ot_prop_constants,
     globals.lookup_bool_option(Globals, profile_deep, DeepProf),
     globals.lookup_bool_option(Globals, record_term_sizes_as_words, TSWProf),
     globals.lookup_bool_option(Globals, record_term_sizes_as_cells, TSCProf),
@@ -394,7 +396,7 @@ llds_backend_pass_for_proc(!HLDS, ConstStructMap, SCCMap, PredId, PredInfo,
     % mercury_compile.simplify.
 
     ( if
-        ConstProp = yes,
+        ConstProp = prop_constants,
         ProfTrans = no
     then
         list.cons(simptask_constant_prop, SimpList0, SimpList1)
@@ -440,12 +442,12 @@ llds_backend_pass_for_proc(!HLDS, ConstStructMap, SCCMap, PredId, PredInfo,
     ),
     generate_proc_code(!.HLDS, ConstStructMap, PredId, PredInfo,
          ProcId, !.ProcInfo, CProc0, !GlobalData),
-    globals.lookup_bool_option(Globals, optimize, Optimize),
+    Optimize = OptTuple ^ ot_optimize,
     (
-        Optimize = yes,
+        Optimize = optimize,
         optimize_proc(Globals, !.GlobalData, CProc0, CProc)
     ;
-        Optimize = no,
+        Optimize = do_not_optimize,
         CProc = CProc0
     ),
     trace [io(!IO)] (
@@ -473,9 +475,10 @@ map_args_to_regs(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_saved_vars(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_saved_vars_const, SavedVars),
+    globals.get_opt_tuple(Globals, OptTuple),
+    SavedVars = OptTuple ^ ot_opt_saved_vars_const,
     (
-        SavedVars = yes,
+        SavedVars = opt_saved_vars_const,
         maybe_write_string(Verbose,
             "% Minimizing variable saves using constants...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -483,7 +486,7 @@ maybe_saved_vars(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        SavedVars = no
+        SavedVars = do_not_opt_saved_vars_const
     ).
 
 :- pred maybe_stack_opt(bool::in, bool::in,
@@ -491,9 +494,10 @@ maybe_saved_vars(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_stack_opt(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_saved_vars_cell, SavedVars),
+    globals.get_opt_tuple(Globals, OptTuple),
+    SavedVars = OptTuple ^ ot_opt_svcell,
     (
-        SavedVars = yes,
+        SavedVars = opt_svcell,
         maybe_write_string(Verbose,
             "% Minimizing variable saves using cells...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -501,7 +505,7 @@ maybe_stack_opt(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        SavedVars = no
+        SavedVars = do_not_opt_svcell
     ).
 
 :- pred maybe_followcode(bool::in, bool::in,
@@ -509,9 +513,10 @@ maybe_stack_opt(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_followcode(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, follow_code, FollowCode),
+    globals.get_opt_tuple(Globals, OptTuple),
+    FollowCode = OptTuple ^ ot_opt_follow_code,
     (
-        FollowCode = yes,
+        FollowCode = opt_follow_code,
         maybe_write_string(Verbose, "% Migrating branch code...", !IO),
         maybe_flush_output(Verbose, !IO),
         process_valid_nonimported_procs(update_module(move_follow_code_in_proc),
@@ -519,7 +524,7 @@ maybe_followcode(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, " done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        FollowCode = no
+        FollowCode = do_not_opt_follow_code
     ).
 
 :- pred compute_liveness(bool::in, bool::in,
@@ -599,16 +604,17 @@ generate_llds_code_for_module(HLDS, Verbose, Stats, !GlobalData, LLDS, !IO) :-
 
 maybe_optimize_llds(HLDS, GlobalData, Verbose, Stats, !LLDS, !IO) :-
     module_info_get_globals(HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize, Optimize),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Optimize = OptTuple ^ ot_optimize,
     (
-        Optimize = yes,
+        Optimize = optimize,
         maybe_write_string(Verbose, "% Doing optimizations...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         optimize_procs(Globals, GlobalData, !LLDS),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        Optimize = no
+        Optimize = do_not_optimize
     ).
 
 :- pred maybe_generate_stack_layouts(module_info::in, list(c_procedure)::in,
@@ -680,7 +686,8 @@ llds_output_pass(OpModeCodeGen, HLDS, GlobalData0, Procs, ModuleName,
     CModuleName = MangledModuleName ++ "_module",
 
     % Split the code up into bite-size chunks for the C compiler.
-    globals.lookup_int_option(Globals, procs_per_c_function, ProcsPerFunc),
+    globals.get_opt_tuple(Globals, OptTuple),
+    ProcsPerFunc = OptTuple ^ ot_procs_per_c_function,
     ( if ProcsPerFunc = 0 then
         % ProcsPerFunc = 0 really means infinity -
         % we store all the procs in a single function.
diff --git a/compiler/mercury_compile_main.m b/compiler/mercury_compile_main.m
index f5ab74510..1cba99a59 100644
--- a/compiler/mercury_compile_main.m
+++ b/compiler/mercury_compile_main.m
@@ -59,6 +59,7 @@
 :- import_module libs.file_util.
 :- import_module libs.handle_options.
 :- import_module libs.op_mode.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.timestamp.
 :- import_module make.
@@ -1768,7 +1769,10 @@ process_augmented_module(Globals0, OpModeAugment, ModuleAndImports,
             globals.lookup_bool_option(Globals, warn_unused_args, UnusedArgs),
             (
                 UnusedArgs = yes,
-                globals.set_option(optimize_unused_args, bool(no),
+                globals.get_opt_tuple(Globals, OptTuple),
+                NoOptUnusedArgsOptTuple = OptTuple ^ ot_opt_unused_args
+                    := do_not_opt_unused_args,
+                globals.set_opt_tuple(NoOptUnusedArgsOptTuple,
                     Globals, NoOptUnusedArgsGlobals),
                 module_info_set_globals(NoOptUnusedArgsGlobals,
                     HLDS21, HLDS21a),
diff --git a/compiler/mercury_compile_middle_passes.m b/compiler/mercury_compile_middle_passes.m
index 8ac31f39c..9bb96b7e9 100644
--- a/compiler/mercury_compile_middle_passes.m
+++ b/compiler/mercury_compile_middle_passes.m
@@ -62,6 +62,7 @@
 :- import_module libs.
 :- import_module libs.file_util.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.
 :- import_module ll_backend.deep_profiling.
@@ -772,9 +773,10 @@ maybe_bytecodes(HLDS0, Verbose, Stats, !DumpInfo, !IO) :-
 
 maybe_untuple_arguments(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, untuple, Untuple),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Untuple = OptTuple ^ ot_untuple,
     (
-        Untuple = yes,
+        Untuple = untuple,
         maybe_write_string(Verbose, "% Untupling...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         untuple_arguments(!HLDS, !IO),
@@ -785,7 +787,7 @@ maybe_untuple_arguments(Verbose, Stats, !HLDS, !IO) :-
             "simplify has errors"),
         maybe_report_stats(Stats, !IO)
     ;
-        Untuple = no
+        Untuple = do_not_untuple
     ).
 
 %---------------------------------------------------------------------------%
@@ -795,16 +797,17 @@ maybe_untuple_arguments(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_tuple_arguments(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, tuple, Tuple),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Tuple = OptTuple ^ ot_tuple,
     (
-        Tuple = yes,
+        Tuple = tuple,
         maybe_write_string(Verbose, "% Tupling...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         tuple_arguments(!HLDS, !IO),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        Tuple = no
+        Tuple = do_not_tuple
     ).
 
 %---------------------------------------------------------------------------%
@@ -814,10 +817,10 @@ maybe_tuple_arguments(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_higher_order(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_higher_order, HigherOrder),
+    globals.get_opt_tuple(Globals, OptTuple),
+    HigherOrder = OptTuple ^ ot_opt_higher_order,
     % --type-specialization implies --user-guided-type-specialization.
-    globals.lookup_bool_option(Globals, user_guided_type_specialization,
-        Types),
+    UserTypeSpec = OptTuple ^ ot_spec_types_user_guided,
 
     % Always produce the specialized versions for which
     % `:- pragma type_spec' declarations exist, because
@@ -825,8 +828,8 @@ maybe_higher_order(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_type_spec_info(!.HLDS, TypeSpecInfo),
     TypeSpecInfo = type_spec_info(TypeSpecPreds, _, _, _),
     ( if
-        ( HigherOrder = yes
-        ; Types = yes
+        ( HigherOrder = opt_higher_order
+        ; UserTypeSpec = spec_types_user_guided
         ; set.is_non_empty(TypeSpecPreds)
         )
     then
@@ -901,9 +904,10 @@ maybe_implicit_parallelism(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_introduce_accumulators(Verbose, Stats, !HLDS, !Specs, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, introduce_accumulators, Optimize),
+    globals.get_opt_tuple(Globals, OptTuple),
+    IntroduceAccumulators = OptTuple ^ ot_introduce_accumulators,
     (
-        Optimize = yes,
+        IntroduceAccumulators = introduce_accumulators,
         maybe_write_string(Verbose,
             "% Attempting to introduce accumulators...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -921,7 +925,7 @@ maybe_introduce_accumulators(Verbose, Stats, !HLDS, !Specs, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        Optimize = no
+        IntroduceAccumulators = do_not_introduce_accumulators
     ).
 
 %---------------------------------------------------------------------------%
@@ -931,18 +935,18 @@ maybe_introduce_accumulators(Verbose, Stats, !HLDS, !Specs, !IO) :-
 
 maybe_do_inlining(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, allow_inlining, Allow),
-    globals.lookup_bool_option(Globals, inline_simple, Simple),
-    globals.lookup_bool_option(Globals, inline_single_use, SingleUse),
-    globals.lookup_int_option(Globals, inline_compound_threshold, Threshold),
-    globals.lookup_bool_option(Globals, inline_linear_tail_rec_sccs,
-        LinearRec),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Allow = OptTuple ^ ot_allow_inlining,
+    Simple = OptTuple ^ ot_inline_simple,
+    SingleUse = OptTuple ^ ot_inline_single_use,
+    Threshold = OptTuple ^ ot_inline_compound_threshold,
+    LinearRec = OptTuple ^ ot_inline_linear_tail_rec_sccs,
     ( if
-        Allow = yes,
-        ( Simple = yes
-        ; SingleUse = yes
+        Allow = allow_inlining,
+        ( Simple = inline_simple
+        ; SingleUse = inline_single_use
         ; Threshold > 0
-        ; LinearRec = yes
+        ; LinearRec = inline_linear_tail_rec_sccs
         )
     then
         maybe_write_string(Verbose, "% Inlining...\n", !IO),
@@ -962,9 +966,10 @@ maybe_do_inlining(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_loop_inv(Verbose, Stats, !HLDS, !DumpInfo, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, loop_invariants, LoopInv),
+    globals.get_opt_tuple(Globals, OptTuple),
+    LoopInv = OptTuple ^ ot_opt_loop_invariants,
     (
-        LoopInv = yes,
+        LoopInv = opt_loop_invariants,
         % We run the mark_static pass because we need the construct_how flag
         % to be valid.
         maybe_mark_static_terms(Verbose, Stats, !HLDS, !IO),
@@ -977,7 +982,7 @@ maybe_loop_inv(Verbose, Stats, !HLDS, !DumpInfo, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        LoopInv = no
+        LoopInv = do_not_opt_loop_invariants
     ).
 
 %---------------------------------------------------------------------------%
@@ -987,31 +992,26 @@ maybe_loop_inv(Verbose, Stats, !HLDS, !DumpInfo, !IO) :-
 
 maybe_deforestation(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, deforestation, Deforest),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Deforest = OptTuple ^ ot_deforest,
 
     % --constraint-propagation implies --local-constraint-propagation.
-    globals.lookup_bool_option(Globals, local_constraint_propagation,
-        Constraints),
-    ( if
-        ( Deforest = yes
-        ; Constraints = yes
-        )
-    then
+    PropLocalConstraints = OptTuple ^ ot_prop_local_constraints,
+    (
+        Deforest = do_not_deforest,
+        PropLocalConstraints = do_not_prop_local_constraints
+    ;
         (
-            Deforest = no,
-            Constraints = no,
-            unexpected($pred, "no no")
-        ;
-            Deforest = yes,
-            Constraints = yes,
+            Deforest = deforest,
+            PropLocalConstraints = prop_local_constraints,
             Msg = "% Deforestation and constraint propagation...\n"
         ;
-            Deforest = yes,
-            Constraints = no,
+            Deforest = deforest,
+            PropLocalConstraints = do_not_prop_local_constraints,
             Msg = "% Deforestation...\n"
         ;
-            Deforest = no,
-            Constraints = yes,
+            Deforest = do_not_deforest,
+            PropLocalConstraints = prop_local_constraints,
             Msg = "% Constraint propagation...\n"
         ),
         maybe_write_string(Verbose, Msg, !IO),
@@ -1019,8 +1019,6 @@ maybe_deforestation(Verbose, Stats, !HLDS, !IO) :-
         deforestation(!HLDS),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
-    else
-        true
     ).
 
 %---------------------------------------------------------------------------%
@@ -1030,9 +1028,10 @@ maybe_deforestation(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_delay_construct(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, delay_construct, DelayConstruct),
+    globals.get_opt_tuple(Globals, OptTuple),
+    DelayConstruct = OptTuple ^ ot_delay_constructs,
     (
-        DelayConstruct = yes,
+        DelayConstruct = delay_constructs,
         maybe_write_string(Verbose,
             "% Delaying construction unifications ...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -1041,7 +1040,7 @@ maybe_delay_construct(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        DelayConstruct = no
+        DelayConstruct = do_not_delay_constructs
     ).
 
 %---------------------------------------------------------------------------%
@@ -1089,13 +1088,14 @@ maybe_structure_reuse_analysis(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_unused_args(Verbose, Stats, UnusedArgsInfos, !HLDS, !Specs, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, intermod_unused_args, Intermod),
-    globals.lookup_bool_option(Globals, optimize_unused_args, Optimize),
-    globals.lookup_bool_option(Globals, warn_unused_args, Warn),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptUnusedArgs = OptTuple ^ ot_opt_unused_args,
+    OptUnusedArgsIntermod = OptTuple ^ ot_opt_unused_args_intermod,
+    globals.lookup_bool_option(Globals, warn_unused_args, WarnUnusedArgs),
     ( if
-        ( Optimize = yes
-        ; Warn = yes
-        ; Intermod = yes
+        ( OptUnusedArgs = opt_unused_args
+        ; OptUnusedArgsIntermod = opt_unused_args_intermod
+        ; WarnUnusedArgs = yes
         )
     then
         maybe_write_string(Verbose, "% Finding unused arguments ...\n", !IO),
@@ -1133,9 +1133,10 @@ maybe_analyse_trail_usage(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_unneeded_code(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, unneeded_code, UnneededCode),
+    globals.get_opt_tuple(Globals, OptTuple),
+    UnneededCode = OptTuple ^ ot_opt_unneeded_code,
     (
-        UnneededCode = yes,
+        UnneededCode = opt_unneeded_code,
         maybe_write_string(Verbose,
             "% Removing unneeded code from procedure bodies...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -1144,7 +1145,7 @@ maybe_unneeded_code(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        UnneededCode = no
+        UnneededCode = do_not_opt_unneeded_code
     ).
 
 %---------------------------------------------------------------------------%
@@ -1154,9 +1155,10 @@ maybe_unneeded_code(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_lco(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_constructor_last_call, LCO),
+    globals.get_opt_tuple(Globals, OptTuple),
+    LCMC = OptTuple ^ ot_opt_lcmc,
     (
-        LCO = yes,
+        LCMC = opt_lcmc,
         maybe_write_string(Verbose,
             "% Looking for LCO modulo constructor application ...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -1164,7 +1166,7 @@ maybe_lco(Verbose, Stats, !HLDS, !IO) :-
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        LCO = no
+        LCMC = do_not_opt_lcmc
     ).
 
 %---------------------------------------------------------------------------%
@@ -1433,16 +1435,17 @@ maybe_experimental_complexity(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_region_analysis(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, region_analysis, Analysis),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Analysis = OptTuple ^ ot_analyse_regions,
     (
-        Analysis = yes,
+        Analysis = analyse_regions,
         maybe_write_string(Verbose, "% Analysing regions ...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         do_region_analysis(!HLDS, !IO),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        Analysis = no
+        Analysis = do_not_analyse_regions
     ).
 
 %---------------------------------------------------------------------------%
@@ -1452,16 +1455,17 @@ maybe_region_analysis(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_eliminate_dead_procs(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_dead_procs, Dead),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptDeadProcs = OptTuple ^ ot_opt_dead_procs,
     (
-        Dead = yes,
+        OptDeadProcs = opt_dead_procs,
         maybe_write_string(Verbose, "% Eliminating dead procedures...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         dead_proc_elim(elim_opt_imported, !HLDS),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        Dead = no
+        OptDeadProcs = do_not_opt_dead_procs
     ).
 
 %---------------------------------------------------------------------------%
diff --git a/compiler/mercury_compile_mlds_back_end.m b/compiler/mercury_compile_mlds_back_end.m
index 544e020cc..6ea25d8be 100644
--- a/compiler/mercury_compile_mlds_back_end.m
+++ b/compiler/mercury_compile_mlds_back_end.m
@@ -59,6 +59,7 @@
 :- import_module hlds.mark_static_terms.            % HLDS -> HLDS
 :- import_module hlds.mark_tail_calls.              % HLDS -> HLDS
 :- import_module libs.file_util.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ml_backend.add_trail_ops.          % HLDS -> HLDS
 :- import_module ml_backend.add_heap_ops.           % HLDS -> HLDS
@@ -74,7 +75,6 @@
 :- import_module top_level.mercury_compile_front_end.
 :- import_module top_level.mercury_compile_llds_back_end.
 
-:- import_module getopt_io.
 :- import_module maybe.
 :- import_module pprint.
 :- import_module require.
@@ -150,16 +150,18 @@ mlds_backend(!HLDS, !:MLDS, !:Specs, !DumpInfo, !IO) :-
     %
     % However, we need to disable optimize_initializations, because
     % ml_elim_nested doesn't correctly handle code containing initializations.
-    globals.lookup_bool_option(Globals, optimize, Optimize),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Optimize = OptTuple ^ ot_optimize,
     (
-        Optimize = yes,
-        globals.set_option(optimize_initializations, bool(no),
-            Globals, NoInitOptGlobals),
+        Optimize = optimize,
+        NoInitOptTuple = OptTuple ^ ot_opt_initializations
+            := do_not_opt_initializations,
+        globals.set_opt_tuple(NoInitOptTuple, Globals, NoInitOptGlobals),
         maybe_write_string(Verbose, "% Optimizing MLDS...\n", !IO),
         mlds_optimize(NoInitOptGlobals, !MLDS),
         maybe_write_string(Verbose, "% done.\n", !IO)
     ;
-        Optimize = no
+        Optimize = do_not_optimize
     ),
     maybe_report_stats(Stats, !IO),
     maybe_dump_mlds(Globals, !.MLDS, 25, "optimize1", !IO),
@@ -199,12 +201,12 @@ mlds_backend(!HLDS, !:MLDS, !:Specs, !DumpInfo, !IO) :-
     % optimize_initializations. (It may also help pick up some additional
     % optimization opportunities for the other optimizations in this pass.)
     (
-        Optimize = yes,
+        Optimize = optimize,
         maybe_write_string(Verbose, "% Optimizing MLDS again...\n", !IO),
         mlds_optimize(Globals, !MLDS),
         maybe_write_string(Verbose, "% done.\n", !IO)
     ;
-        Optimize = no
+        Optimize = do_not_optimize
     ),
     maybe_report_stats(Stats, !IO),
     maybe_dump_mlds(Globals, !.MLDS, 40, "optimize2", !IO),
@@ -215,16 +217,17 @@ mlds_backend(!HLDS, !:MLDS, !:Specs, !DumpInfo, !IO) :-
 
 maybe_mark_static_terms(Verbose, Stats, !HLDS, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, static_ground_cells, SGCells),
+    globals.get_opt_tuple(Globals, OptTuple),
+    SGCells = OptTuple ^ ot_use_static_ground_cells,
     (
-        SGCells = yes,
+        SGCells = use_static_ground_cells,
         maybe_write_string(Verbose, "% Marking static ground terms...\n", !IO),
         maybe_flush_output(Verbose, !IO),
         process_valid_nonimported_procs(update_proc(mark_static_terms), !HLDS),
         maybe_write_string(Verbose, "% done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        SGCells = no
+        SGCells = do_not_use_static_ground_cells
     ).
 
 :- pred maybe_add_trail_ops(bool::in, bool::in,
@@ -254,8 +257,8 @@ maybe_add_trail_ops(Verbose, Stats, !HLDS, !IO) :-
         globals.get_target(Globals, Target),
         (
             Target = target_c,
-            globals.lookup_bool_option(Globals, generate_trail_ops_inline,
-                GenerateInline)
+            globals.get_opt_tuple(Globals, OptTuple),
+            GenerateInline = OptTuple ^ ot_gen_trail_ops_inline
         ;
             % XXX Currently, we can only generate trail ops inline for
             % the C backends.
@@ -264,7 +267,7 @@ maybe_add_trail_ops(Verbose, Stats, !HLDS, !IO) :-
             ; Target = target_java
             ; Target = target_erlang
             ),
-            GenerateInline = no
+            GenerateInline = do_not_gen_trail_ops_inline
         ),
         maybe_write_string(Verbose, "% Adding trailing operations...\n", !IO),
         maybe_flush_output(Verbose, !IO),
@@ -324,9 +327,10 @@ maybe_add_heap_ops(Verbose, Stats, !HLDS, !IO) :-
 
 maybe_mark_tail_rec_calls_hlds(Verbose, Stats, !HLDS, !Specs, !IO) :-
     module_info_get_globals(!.HLDS, Globals),
-    globals.lookup_bool_option(Globals, optimize_tailcalls, OptimizeTailCalls),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptimizeTailCalls = OptTuple ^ ot_opt_mlds_tailcalls,
     (
-        OptimizeTailCalls = yes,
+        OptimizeTailCalls = opt_mlds_tailcalls,
         maybe_write_string(Verbose, "% Marking tail recursive calls...", !IO),
         maybe_flush_output(Verbose, !IO),
         module_info_rebuild_dependency_info(!HLDS, DepInfo),
@@ -335,7 +339,7 @@ maybe_mark_tail_rec_calls_hlds(Verbose, Stats, !HLDS, !Specs, !IO) :-
         maybe_write_string(Verbose, " done.\n", !IO),
         maybe_report_stats(Stats, !IO)
     ;
-        OptimizeTailCalls = no
+        OptimizeTailCalls = do_not_opt_mlds_tailcalls
     ).
 
 :- pred mlds_gen_rtti_data(module_info::in, mlds_target_lang::in,
diff --git a/compiler/ml_disj_gen.m b/compiler/ml_disj_gen.m
index 93bb87193..2f0de4d27 100644
--- a/compiler/ml_disj_gen.m
+++ b/compiler/ml_disj_gen.m
@@ -129,7 +129,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module ml_backend.ml_code_gen.
 :- import_module ml_backend.ml_code_util.
 :- import_module ml_backend.ml_global_data.
@@ -167,9 +167,9 @@ ml_gen_disj(Disjuncts, GoalInfo, CodeModel, Context, Stmts, !Info) :-
 
                 ml_gen_info_get_module_info(!.Info, ModuleInfo),
                 module_info_get_globals(ModuleInfo, Globals),
-                globals.lookup_bool_option(Globals, static_ground_cells,
-                    StaticGroundCells),
-                StaticGroundCells = yes,
+                globals.get_opt_tuple(Globals, OptTuple),
+                OptTuple ^ ot_use_static_ground_cells =
+                    use_static_ground_cells,
 
                 DisjNonLocals = goal_info_get_nonlocals(GoalInfo),
                 all_disjuncts_are_conj_of_unify(DisjNonLocals, Disjuncts),
diff --git a/compiler/ml_gen_info.m b/compiler/ml_gen_info.m
index a2d09b609..7bbacfd49 100644
--- a/compiler/ml_gen_info.m
+++ b/compiler/ml_gen_info.m
@@ -25,6 +25,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module ml_backend.ml_global_data.
 :- import_module ml_backend.mlds.
 :- import_module parse_tree.
@@ -505,7 +506,8 @@
 :- pred ml_gen_info_get_gc(ml_gen_info::in, gc_method::out) is det.
 :- pred ml_gen_info_get_det_copy_out(ml_gen_info::in, bool::out) is det.
 :- pred ml_gen_info_get_nondet_copy_out(ml_gen_info::in, bool::out) is det.
-:- pred ml_gen_info_get_use_atomic_cells(ml_gen_info::in, bool::out) is det.
+:- pred ml_gen_info_get_use_atomic_cells(ml_gen_info::in,
+    maybe_use_atomic_cells::out) is det.
 :- pred ml_gen_info_get_profile_memory(ml_gen_info::in, bool::out) is det.
 :- pred ml_gen_info_get_num_ptag_bits(ml_gen_info::in, uint8::out) is det.
 :- pred ml_gen_info_get_const_struct_map(ml_gen_info::in,
@@ -774,7 +776,7 @@ generate_tail_rec_start_label(TsccKind, Id) = Label :-
 /*  - */        mgri_gc                 :: gc_method,
 /*  - */        mgri_det_copy_out       :: bool,
 /*  - */        mgri_nondet_copy_out    :: bool,
-/*  - */        mgri_use_atomic_cells   :: bool,
+/*  - */        mgri_use_atomic_cells   :: maybe_use_atomic_cells,
 /*  - */        mgri_profile_memory     :: bool,
 
 /*  - */        mgri_num_ptag_bits      :: uint8,
@@ -968,7 +970,8 @@ ml_gen_info_init(ModuleInfo, Target, ConstStructMap, PredProcId, ProcInfo,
     globals.get_gc_method(Globals, GC),
     globals.lookup_bool_option(Globals, det_copy_out, DetCopyOut),
     globals.lookup_bool_option(Globals, nondet_copy_out, NondetCopyOut),
-    globals.lookup_bool_option(Globals, use_atomic_cells, UseAtomicCells),
+    globals.get_opt_tuple(Globals, OptTuple),
+    UseAtomicCells = OptTuple ^ ot_use_atomic_cells,
     globals.lookup_bool_option(Globals, profile_memory, ProfileMemory),
     globals.lookup_int_option(Globals, num_ptag_bits, NumPtagBitsInt),
     NumPtagBits = uint8.det_from_int(NumPtagBitsInt),
diff --git a/compiler/ml_lookup_switch.m b/compiler/ml_lookup_switch.m
index 956b01bf9..20e030281 100644
--- a/compiler/ml_lookup_switch.m
+++ b/compiler/ml_lookup_switch.m
@@ -126,7 +126,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module ml_backend.ml_code_util.
 :- import_module ml_backend.ml_global_data.
 :- import_module ml_backend.ml_util.
@@ -151,10 +151,9 @@ ml_is_lookup_switch(SwitchVar, TaggedCases, GoalInfo, CodeModel,
     set_of_var.to_sorted_list(OtherNonLocals, OutVars),
     ml_gen_info_get_module_info(!.Info, ModuleInfo),
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, static_ground_cells,
-        StaticGroundCells),
+    globals.get_opt_tuple(Globals, OptTuple),
     ( if
-        StaticGroundCells = yes,
+        OptTuple ^ ot_use_static_ground_cells = use_static_ground_cells,
         ml_generate_constants_for_lookup_switch(CodeModel, OutVars,
             OtherNonLocals, TaggedCases, map.init, CaseSolnMap, !Info)
     then
diff --git a/compiler/ml_optimize.m b/compiler/ml_optimize.m
index a5e91b631..54fa7a94c 100644
--- a/compiler/ml_optimize.m
+++ b/compiler/ml_optimize.m
@@ -40,6 +40,7 @@
 
 :- import_module backend_libs.
 :- import_module backend_libs.builtin_ops.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
@@ -120,11 +121,12 @@ optimize_in_maybe_stmt(OptInfo, !MaybeStmt) :-
 optimize_in_stmts(OptInfo, !Stmts) :-
     list.map(optimize_in_stmt(OptInfo), !Stmts),
     Globals = OptInfo ^ oi_globals,
-    globals.lookup_bool_option(Globals, optimize_peep, OptPeep),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptPeep = OptTuple ^ ot_opt_peep,
     (
-        OptPeep = no
+        OptPeep = do_not_opt_peep
     ;
-        OptPeep = yes,
+        OptPeep = opt_peep,
         peephole_opt_statements(!Stmts)
     ).
 
@@ -706,12 +708,13 @@ cases_affect_lvals(Lvals, [Head | Tail], Affects) :-
 maybe_convert_assignments_into_initializers(OptInfo, !Defns, !Stmts) :-
     Globals = OptInfo ^ oi_globals,
     % Check if --optimize-initializations is enabled.
-    globals.lookup_bool_option(Globals, optimize_initializations, OptInit),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptInit = OptTuple ^ ot_opt_initializations,
     (
-        OptInit = yes,
+        OptInit = opt_initializations,
         convert_assignments_into_initializers(OptInfo, !Defns, !Stmts)
     ;
-        OptInit = no
+        OptInit = do_not_opt_initializations
     ).
 
 :- pred convert_assignments_into_initializers(opt_info::in,
@@ -807,13 +810,14 @@ find_this_var_defn(VarName, [LocalVarDefn | LocalVarDefns], !RevPrevDefns,
     list(mlds_stmt)::in, list(mlds_stmt)::out) is det.
 
 maybe_eliminate_locals(OptInfo, !LocalVarDefns, !FuncDefns, !Stmts) :-
-    globals.lookup_bool_option(OptInfo ^ oi_globals, eliminate_local_vars,
-        EliminateLocalVars),
+    Globals = OptInfo ^ oi_globals,
+    globals.get_opt_tuple(Globals, OptTuple),
+    EliminateLocalVars = OptTuple ^ ot_elim_local_vars,
     (
-        EliminateLocalVars = yes,
+        EliminateLocalVars = elim_local_vars,
         eliminate_locals(OptInfo, !LocalVarDefns, !FuncDefns, !Stmts)
     ;
-        EliminateLocalVars = no
+        EliminateLocalVars = do_not_elim_local_vars
     ).
 
 :- pred eliminate_locals(opt_info::in,
diff --git a/compiler/ml_proc_gen.m b/compiler/ml_proc_gen.m
index 9422e497e..a9ba671c3 100644
--- a/compiler/ml_proc_gen.m
+++ b/compiler/ml_proc_gen.m
@@ -51,7 +51,7 @@
 :- import_module libs.
 :- import_module libs.dependency_graph.
 :- import_module libs.globals.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module ml_backend.ml_args_util.
 :- import_module ml_backend.ml_code_gen.
 :- import_module ml_backend.ml_code_util.
@@ -95,12 +95,13 @@ ml_gen_preds(Target, ConstStructMap, FuncDefns,
 
     % Optimize tail calls only if asked.
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, optimize_tailcalls, TailCalls),
+    globals.get_opt_tuple(Globals, OptTuple),
+    TailCalls = OptTuple ^ ot_opt_mlds_tailcalls,
     (
-        TailCalls = yes,
+        TailCalls = opt_mlds_tailcalls,
         OptTailCalls = tail_call_opt_in_code_gen
     ;
-        TailCalls = no,
+        TailCalls = do_not_opt_mlds_tailcalls,
         OptTailCalls = no_tail_call_opt_in_code_gen
     ),
     get_default_warn_parms(Globals, DefaultWarnParams),
@@ -1690,15 +1691,15 @@ ml_gen_proc_body(CodeModel, ArgTuples, CopiedOutputVars, Goal, SeenAtLabelMap,
         LocalVarDefns1 = ConvLocalVarDefns ++ LocalVarDefns0
     ),
     ml_gen_info_get_globals(!.Info, Globals),
-    globals.lookup_bool_option(Globals, eliminate_unused_mlds_assigns,
-        EliminateUnusedAssigns),
+    globals.get_opt_tuple(Globals, OptTuple),
+    EliminateUnusedAssigns = OptTuple ^ ot_elim_unused_mlds_assigns,
     (
-        EliminateUnusedAssigns = no,
+        EliminateUnusedAssigns = do_not_elim_unused_mlds_assigns,
         LocalVarDefns = LocalVarDefns1,
         FuncDefns = FuncDefns0,
         Stmts = Stmts1
     ;
-        EliminateUnusedAssigns = yes,
+        EliminateUnusedAssigns = elim_unused_mlds_assigns,
         list.map((func(var_mvar_type_mode(_, LocalVar, _, _)) = LocalVar),
             ArgTuples) = ArgLocalVars,
         (
diff --git a/compiler/ml_simplify_switch.m b/compiler/ml_simplify_switch.m
index c7cf7b90f..ecf8af818 100644
--- a/compiler/ml_simplify_switch.m
+++ b/compiler/ml_simplify_switch.m
@@ -42,6 +42,7 @@
 :- import_module backend_libs.switch_util.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ml_backend.ml_code_util.
 :- import_module ml_backend.ml_target_util.
@@ -79,12 +80,12 @@ ml_simplify_switch(Stmt0, Stmt, !Info) :-
 
         % Is the switch big enough?
         list.length(Cases, NumCases),
-        globals.lookup_int_option(Globals, dense_switch_size, DenseSize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        DenseSize = OptTuple ^ ot_dense_switch_size,
         NumCases >= DenseSize,
 
         % ... and dense enough?
-        globals.lookup_int_option(Globals, dense_switch_req_density,
-            ReqDensity),
+        ReqDensity = OptTuple ^ ot_dense_switch_req_density,
         is_dense_switch(Cases, ReqDensity)
     then
         maybe_eliminate_default(Range, Cases, Default, ReqDensity,
diff --git a/compiler/ml_switch_gen.m b/compiler/ml_switch_gen.m
index 94d7d8e19..9680c831a 100644
--- a/compiler/ml_switch_gen.m
+++ b/compiler/ml_switch_gen.m
@@ -142,6 +142,7 @@
 :- import_module hlds.hlds_module.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ml_backend.ml_code_gen.
 :- import_module ml_backend.ml_code_util.
@@ -204,7 +205,8 @@ ml_gen_switch(SwitchVar, CanFail, Cases, CodeModel, Context, GoalInfo,
         ;
             SwitchCategory = tag_switch,
             num_cons_ids_in_tagged_cases(TaggedCases, NumConsIds, NumArms),
-            globals.lookup_int_option(Globals, tag_switch_size, TagSize),
+            globals.get_opt_tuple(Globals, OptTuple),
+            TagSize = OptTuple ^ ot_tag_switch_size,
             ( if
                 NumConsIds >= TagSize,
                 NumArms > 1,
@@ -246,12 +248,12 @@ ml_gen_smart_atomic_switch(SwitchVar, SwitchVarType, CanFail, TaggedCases,
         ml_gen_info_get_high_level_data(!.Info, no),
         MaybeIntSwitchInfo = int_switch(IntSwitchInfo),
         IntSwitchInfo = int_switch_info(LowerLimit, UpperLimit, NumValues),
-        globals.lookup_bool_option(Globals, static_ground_cells, yes),
-        globals.lookup_int_option(Globals, lookup_switch_size, LookupSize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        OptTuple ^ ot_use_static_ground_cells = use_static_ground_cells,
+        LookupSize = OptTuple ^ ot_lookup_switch_size,
         NumConsIds >= LookupSize,
         NumArms > 1,
-        globals.lookup_int_option(Globals, lookup_switch_req_density,
-            ReqDensity),
+        ReqDensity = OptTuple ^ ot_lookup_switch_req_density,
         filter_out_failing_cases_if_needed(CodeModel,
             TaggedCases, FilteredTaggedCases,
             CanFail, FilteredCanFail),
@@ -364,9 +366,9 @@ ml_gen_smart_string_switch(SwitchVar, CanFail, TaggedCases, CodeModel,
             CodeModel, MaybeLookupSwitchInfo, !Info),
         ml_gen_var(!.Info, SwitchVar, SwitchVarLval),
         SwitchVarRval = ml_lval(SwitchVarLval),
+        globals.get_opt_tuple(Globals, OptTuple),
         ( if
-            globals.lookup_int_option(Globals, string_trie_switch_size,
-                StringTrieSwitchSize),
+            StringTrieSwitchSize = OptTuple ^ ot_string_trie_switch_size,
             NumConsIds >= StringTrieSwitchSize,
             globals.get_target(Globals, target_c)
         then
@@ -383,8 +385,7 @@ ml_gen_smart_string_switch(SwitchVar, CanFail, TaggedCases, CodeModel,
             ),
             Decls = []
         else if
-            globals.lookup_int_option(Globals, string_hash_switch_size,
-                StringHashSwitchSize),
+            StringHashSwitchSize = OptTuple ^ ot_string_hash_switch_size,
             NumConsIds >= StringHashSwitchSize
         then
             (
@@ -399,8 +400,7 @@ ml_gen_smart_string_switch(SwitchVar, CanFail, TaggedCases, CodeModel,
                     EntryPackedWordMap, Context, Decls, Stmts, !Info)
             )
         else if
-            globals.lookup_int_option(Globals, string_binary_switch_size,
-                StringBinarySwitchSize),
+            StringBinarySwitchSize = OptTuple ^ ot_string_binary_switch_size,
             NumConsIds >= StringBinarySwitchSize
         then
             (
diff --git a/compiler/ml_unify_gen_construct.m b/compiler/ml_unify_gen_construct.m
index 108bfe951..c3b98f942 100644
--- a/compiler/ml_unify_gen_construct.m
+++ b/compiler/ml_unify_gen_construct.m
@@ -98,6 +98,8 @@
 :- import_module hlds.hlds_data.
 :- import_module hlds.hlds_pred.
 :- import_module hlds.vartypes.
+:- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
 :- import_module ml_backend.ml_closure_gen.
@@ -468,10 +470,10 @@ ml_gen_new_object_dynamically(MaybeConsId, MaybeCtorName, Ptag,
         Stmts, !Info) :-
     ml_gen_info_get_use_atomic_cells(!.Info, UseAtomicCells),
     (
-        UseAtomicCells = yes,
+        UseAtomicCells = use_atomic_cells,
         MayUseAtomic0 = may_use_atomic_alloc
     ;
-        UseAtomicCells = no,
+        UseAtomicCells = do_not_use_atomic_cells,
         MayUseAtomic0 = may_not_use_atomic_alloc
     ),
     ml_generate_and_pack_dynamic_construct_args(!.Info, RHSVarsTypesWidths,
diff --git a/compiler/optimization_options.m b/compiler/optimization_options.m
index e69de29bb..fece027bf 100644
--- a/compiler/optimization_options.m
+++ b/compiler/optimization_options.m
@@ -0,0 +1,1873 @@
+%---------------------------------------------------------------------------%
+% vim: ft=mercury ts=4 sw=4 et
+%---------------------------------------------------------------------------%
+% Copyright (C) 2020 The Mercury team.
+% This file may only be copied under the terms of the GNU General
+% Public License - see the file COPYING in the Mercury distribution.
+%---------------------------------------------------------------------------%
+%
+% File: optimization_options.m
+% Main author: zs, via tools/make_opt
+%
+% This module manages optimization options. It defines a type, the opt_tuple,
+% that allows quick access to them, sets their default values, and implements
+% the turning on of optimizations when a given optimization level is selected.
+%
+
+:- module libs.optimization_options.
+:- interface.
+
+:- import_module libs.options.
+
+:- import_module bool.
+:- import_module getopt_io.
+:- import_module list.
+:- import_module map.
+
+%---------------------------------------------------------------------------%
+
+:- type maybe_allow_inlining
+    --->    allow_inlining
+    ;       do_not_allow_inlining.
+:- type maybe_inline_simple
+    --->    inline_simple
+    ;       do_not_inline_simple.
+:- type maybe_inline_builtins
+    --->    inline_builtins
+    ;       do_not_inline_builtins.
+:- type maybe_inline_single_use
+    --->    inline_single_use
+    ;       do_not_inline_single_use.
+:- type maybe_inline_linear_tail_rec_sccs
+    --->    inline_linear_tail_rec_sccs
+    ;       do_not_inline_linear_tail_rec_sccs.
+:- type maybe_enable_const_struct
+    --->    enable_const_struct
+    ;       do_not_enable_const_struct.
+:- type maybe_opt_common_structs
+    --->    opt_common_structs
+    ;       do_not_opt_common_structs.
+:- type maybe_prop_constraints
+    --->    prop_constraints
+    ;       do_not_prop_constraints.
+:- type maybe_prop_local_constraints
+    --->    prop_local_constraints
+    ;       do_not_prop_local_constraints.
+:- type maybe_opt_dup_calls
+    --->    opt_dup_calls
+    ;       do_not_opt_dup_calls.
+:- type maybe_prop_constants
+    --->    prop_constants
+    ;       do_not_prop_constants.
+:- type maybe_elim_excess_assigns
+    --->    elim_excess_assigns
+    ;       do_not_elim_excess_assigns.
+:- type maybe_opt_test_after_switch
+    --->    opt_test_after_switch
+    ;       do_not_opt_test_after_switch.
+:- type maybe_opt_format_calls
+    --->    opt_format_calls
+    ;       do_not_opt_format_calls.
+:- type maybe_opt_loop_invariants
+    --->    opt_loop_invariants
+    ;       do_not_opt_loop_invariants.
+:- type maybe_opt_saved_vars_const
+    --->    opt_saved_vars_const
+    ;       do_not_opt_saved_vars_const.
+:- type maybe_opt_svcell
+    --->    opt_svcell
+    ;       do_not_opt_svcell.
+:- type maybe_opt_svcell_loop
+    --->    opt_svcell_loop
+    ;       do_not_opt_svcell_loop.
+:- type maybe_opt_svcell_full_path
+    --->    opt_svcell_full_path
+    ;       do_not_opt_svcell_full_path.
+:- type maybe_opt_svcell_on_stack
+    --->    opt_svcell_on_stack
+    ;       do_not_opt_svcell_on_stack.
+:- type maybe_opt_svcell_candidate_headvars
+    --->    opt_svcell_candidate_headvars
+    ;       do_not_opt_svcell_candidate_headvars.
+:- type maybe_opt_svcell_all_candidates
+    --->    opt_svcell_all_candidates
+    ;       do_not_opt_svcell_all_candidates.
+:- type maybe_delay_constructs
+    --->    delay_constructs
+    ;       do_not_delay_constructs.
+:- type maybe_opt_follow_code
+    --->    opt_follow_code
+    ;       do_not_opt_follow_code.
+:- type maybe_opt_unused_args
+    --->    opt_unused_args
+    ;       do_not_opt_unused_args.
+:- type maybe_opt_unused_args_intermod
+    --->    opt_unused_args_intermod
+    ;       do_not_opt_unused_args_intermod.
+:- type maybe_opt_higher_order
+    --->    opt_higher_order
+    ;       do_not_opt_higher_order.
+:- type maybe_opt_unneeded_code
+    --->    opt_unneeded_code
+    ;       do_not_opt_unneeded_code.
+:- type maybe_spec_types
+    --->    spec_types
+    ;       do_not_spec_types.
+:- type maybe_spec_types_user_guided
+    --->    spec_types_user_guided
+    ;       do_not_spec_types_user_guided.
+:- type maybe_introduce_accumulators
+    --->    introduce_accumulators
+    ;       do_not_introduce_accumulators.
+:- type maybe_opt_lcmc_accumulator
+    --->    opt_lcmc_accumulator
+    ;       do_not_opt_lcmc_accumulator.
+:- type maybe_opt_lcmc_null
+    --->    opt_lcmc_null
+    ;       do_not_opt_lcmc_null.
+:- type maybe_opt_lcmc
+    --->    opt_lcmc
+    ;       do_not_opt_lcmc.
+:- type maybe_opt_dead_procs
+    --->    opt_dead_procs
+    ;       do_not_opt_dead_procs.
+:- type maybe_deforest
+    --->    deforest
+    ;       do_not_deforest.
+:- type maybe_untuple
+    --->    untuple
+    ;       do_not_untuple.
+:- type maybe_tuple
+    --->    tuple
+    ;       do_not_tuple.
+:- type maybe_inline_par_builtins
+    --->    inline_par_builtins
+    ;       do_not_inline_par_builtins.
+:- type maybe_spec_in_all_dep_par_conjs
+    --->    spec_in_all_dep_par_conjs
+    ;       do_not_spec_in_all_dep_par_conjs.
+:- type maybe_allow_some_paths_only_waits
+    --->    allow_some_paths_only_waits
+    ;       do_not_allow_some_paths_only_waits.
+:- type maybe_analyse_regions
+    --->    analyse_regions
+    ;       do_not_analyse_regions.
+:- type maybe_use_smart_indexing
+    --->    use_smart_indexing
+    ;       do_not_use_smart_indexing.
+:- type maybe_use_smart_indexing_atomic
+    --->    use_smart_indexing_atomic
+    ;       do_not_use_smart_indexing_atomic.
+:- type maybe_use_smart_indexing_string
+    --->    use_smart_indexing_string
+    ;       do_not_use_smart_indexing_string.
+:- type maybe_use_smart_indexing_tag
+    --->    use_smart_indexing_tag
+    ;       do_not_use_smart_indexing_tag.
+:- type maybe_use_smart_indexing_float
+    --->    use_smart_indexing_float
+    ;       do_not_use_smart_indexing_float.
+:- type maybe_put_base_first_single_rec
+    --->    put_base_first_single_rec
+    ;       do_not_put_base_first_single_rec.
+:- type maybe_put_base_first_multi_rec
+    --->    put_base_first_multi_rec
+    ;       do_not_put_base_first_multi_rec.
+:- type maybe_use_static_ground_cells
+    --->    use_static_ground_cells
+    ;       do_not_use_static_ground_cells.
+:- type maybe_use_static_ground_floats
+    --->    use_static_ground_floats
+    ;       do_not_use_static_ground_floats.
+:- type maybe_use_static_ground_int64s
+    --->    use_static_ground_int64s
+    ;       do_not_use_static_ground_int64s.
+:- type maybe_use_static_code_addresses
+    --->    use_static_code_addresses
+    ;       do_not_use_static_code_addresses.
+:- type maybe_use_atomic_cells
+    --->    use_atomic_cells
+    ;       do_not_use_atomic_cells.
+:- type maybe_opt_middle_rec
+    --->    opt_middle_rec
+    ;       do_not_opt_middle_rec.
+:- type maybe_opt_simple_neg
+    --->    opt_simple_neg
+    ;       do_not_opt_simple_neg.
+:- type maybe_allow_hijacks
+    --->    allow_hijacks
+    ;       do_not_allow_hijacks.
+:- type maybe_opt_mlds_tailcalls
+    --->    opt_mlds_tailcalls
+    ;       do_not_opt_mlds_tailcalls.
+:- type maybe_opt_initializations
+    --->    opt_initializations
+    ;       do_not_opt_initializations.
+:- type maybe_elim_unused_mlds_assigns
+    --->    elim_unused_mlds_assigns
+    ;       do_not_elim_unused_mlds_assigns.
+:- type maybe_elim_local_vars
+    --->    elim_local_vars
+    ;       do_not_elim_local_vars.
+:- type maybe_gen_trail_ops_inline
+    --->    gen_trail_ops_inline
+    ;       do_not_gen_trail_ops_inline.
+:- type maybe_use_common_data
+    --->    use_common_data
+    ;       do_not_use_common_data.
+:- type maybe_use_common_layout_data
+    --->    use_common_layout_data
+    ;       do_not_use_common_layout_data.
+:- type maybe_optimize
+    --->    optimize
+    ;       do_not_optimize.
+:- type maybe_opt_peep
+    --->    opt_peep
+    ;       do_not_opt_peep.
+:- type maybe_opt_peep_mkword
+    --->    opt_peep_mkword
+    ;       do_not_opt_peep_mkword.
+:- type maybe_opt_jumps
+    --->    opt_jumps
+    ;       do_not_opt_jumps.
+:- type maybe_opt_fulljumps
+    --->    opt_fulljumps
+    ;       do_not_opt_fulljumps.
+:- type maybe_pessimize_tailcalls
+    --->    pessimize_tailcalls
+    ;       do_not_pessimize_tailcalls.
+:- type maybe_opt_checked_nondet_tailcalls
+    --->    opt_checked_nondet_tailcalls
+    ;       do_not_opt_checked_nondet_tailcalls.
+:- type maybe_use_local_vars
+    --->    use_local_vars
+    ;       do_not_use_local_vars.
+:- type maybe_standardize_labels
+    --->    standardize_labels
+    ;       do_not_standardize_labels.
+:- type maybe_opt_labels
+    --->    opt_labels
+    ;       do_not_opt_labels.
+:- type maybe_opt_dups
+    --->    opt_dups
+    ;       do_not_opt_dups.
+:- type maybe_opt_proc_dups
+    --->    opt_proc_dups
+    ;       do_not_opt_proc_dups.
+:- type maybe_opt_frames
+    --->    opt_frames
+    ;       do_not_opt_frames.
+:- type maybe_opt_delay_slot
+    --->    opt_delay_slot
+    ;       do_not_opt_delay_slot.
+:- type maybe_opt_reassign
+    --->    opt_reassign
+    ;       do_not_opt_reassign.
+:- type maybe_use_macro_for_redo_fail
+    --->    use_macro_for_redo_fail
+    ;       do_not_use_macro_for_redo_fail.
+:- type maybe_emit_c_loops
+    --->    emit_c_loops
+    ;       do_not_emit_c_loops.
+:- type maybe_use_local_thread_engine_base
+    --->    use_local_thread_engine_base
+    ;       do_not_use_local_thread_engine_base.
+:- type maybe_switch_on_strings_as_atoms
+    --->    switch_on_strings_as_atoms
+    ;       do_not_switch_on_strings_as_atoms.
+:- type maybe_inline_alloc
+    --->    inline_alloc
+    ;       do_not_inline_alloc.
+:- type maybe_opt_c
+    --->    opt_c
+    ;       do_not_opt_c.
+
+%---------------------%
+
+:- type optimization_option
+    --->    oo_allow_inlining(bool)
+    ;       oo_inline_simple(bool)
+    ;       oo_inline_builtins(bool)
+    ;       oo_inline_single_use(bool)
+    ;       oo_inline_linear_tail_rec_sccs(bool)
+    ;       oo_enable_const_struct(bool)
+    ;       oo_opt_common_structs(bool)
+    ;       oo_prop_constraints(bool)
+    ;       oo_prop_local_constraints(bool)
+    ;       oo_opt_dup_calls(bool)
+    ;       oo_prop_constants(bool)
+    ;       oo_elim_excess_assigns(bool)
+    ;       oo_opt_test_after_switch(bool)
+    ;       oo_opt_format_calls(bool)
+    ;       oo_opt_loop_invariants(bool)
+    ;       oo_opt_saved_vars_const(bool)
+    ;       oo_opt_svcell(bool)
+    ;       oo_opt_svcell_loop(bool)
+    ;       oo_opt_svcell_full_path(bool)
+    ;       oo_opt_svcell_on_stack(bool)
+    ;       oo_opt_svcell_candidate_headvars(bool)
+    ;       oo_opt_svcell_all_candidates(bool)
+    ;       oo_delay_constructs(bool)
+    ;       oo_opt_follow_code(bool)
+    ;       oo_opt_unused_args(bool)
+    ;       oo_opt_unused_args_intermod(bool)
+    ;       oo_opt_higher_order(bool)
+    ;       oo_opt_unneeded_code(bool)
+    ;       oo_spec_types(bool)
+    ;       oo_spec_types_user_guided(bool)
+    ;       oo_introduce_accumulators(bool)
+    ;       oo_opt_lcmc_accumulator(bool)
+    ;       oo_opt_lcmc_null(bool)
+    ;       oo_opt_lcmc(bool)
+    ;       oo_opt_dead_procs(bool)
+    ;       oo_deforest(bool)
+    ;       oo_untuple(bool)
+    ;       oo_tuple(bool)
+    ;       oo_inline_par_builtins(bool)
+    ;       oo_spec_in_all_dep_par_conjs(bool)
+    ;       oo_allow_some_paths_only_waits(bool)
+    ;       oo_analyse_regions(bool)
+    ;       oo_use_smart_indexing(bool)
+    ;       oo_use_smart_indexing_atomic(bool)
+    ;       oo_use_smart_indexing_string(bool)
+    ;       oo_use_smart_indexing_tag(bool)
+    ;       oo_use_smart_indexing_float(bool)
+    ;       oo_put_base_first_single_rec(bool)
+    ;       oo_put_base_first_multi_rec(bool)
+    ;       oo_use_static_ground_cells(bool)
+    ;       oo_use_static_ground_floats(bool)
+    ;       oo_use_static_ground_int64s(bool)
+    ;       oo_use_static_code_addresses(bool)
+    ;       oo_use_atomic_cells(bool)
+    ;       oo_opt_middle_rec(bool)
+    ;       oo_opt_simple_neg(bool)
+    ;       oo_allow_hijacks(bool)
+    ;       oo_opt_mlds_tailcalls(bool)
+    ;       oo_opt_initializations(bool)
+    ;       oo_elim_unused_mlds_assigns(bool)
+    ;       oo_elim_local_vars(bool)
+    ;       oo_gen_trail_ops_inline(bool)
+    ;       oo_use_common_data(bool)
+    ;       oo_use_common_layout_data(bool)
+    ;       oo_optimize(bool)
+    ;       oo_opt_peep(bool)
+    ;       oo_opt_peep_mkword(bool)
+    ;       oo_opt_jumps(bool)
+    ;       oo_opt_fulljumps(bool)
+    ;       oo_pessimize_tailcalls(bool)
+    ;       oo_opt_checked_nondet_tailcalls(bool)
+    ;       oo_use_local_vars(bool)
+    ;       oo_standardize_labels(bool)
+    ;       oo_opt_labels(bool)
+    ;       oo_opt_dups(bool)
+    ;       oo_opt_proc_dups(bool)
+    ;       oo_opt_frames(bool)
+    ;       oo_opt_delay_slot(bool)
+    ;       oo_opt_reassign(bool)
+    ;       oo_use_macro_for_redo_fail(bool)
+    ;       oo_emit_c_loops(bool)
+    ;       oo_use_local_thread_engine_base(bool)
+    ;       oo_switch_on_strings_as_atoms(bool)
+    ;       oo_inline_alloc(bool)
+    ;       oo_opt_c(bool)
+    ;       oo_inline_call_cost(int)
+    ;       oo_inline_compound_threshold(int)
+    ;       oo_inline_simple_threshold(int)
+    ;       oo_inline_vars_threshold(int)
+    ;       oo_intermod_inline_simple_threshold(int)
+    ;       oo_inline_linear_tail_rec_sccs_max_extra(int)
+    ;       oo_from_ground_term_threshold(int)
+    ;       oo_opt_svcell_cv_store_cost(int)
+    ;       oo_opt_svcell_cv_load_cost(int)
+    ;       oo_opt_svcell_fv_store_cost(int)
+    ;       oo_opt_svcell_fv_load_cost(int)
+    ;       oo_opt_svcell_op_ratio(int)
+    ;       oo_opt_svcell_node_ratio(int)
+    ;       oo_opt_svcell_all_path_node_ratio(int)
+    ;       oo_higher_order_size_limit(int)
+    ;       oo_higher_order_arg_limit(int)
+    ;       oo_opt_unneeded_code_copy_limit(int)
+    ;       oo_deforestation_depth_limit(int)
+    ;       oo_deforestation_cost_factor(int)
+    ;       oo_deforestation_vars_threshold(int)
+    ;       oo_deforestation_size_threshold(int)
+    ;       oo_tuple_costs_ratio(int)
+    ;       oo_tuple_min_args(int)
+    ;       oo_dense_switch_req_density(int)
+    ;       oo_lookup_switch_req_density(int)
+    ;       oo_dense_switch_size(int)
+    ;       oo_lookup_switch_size(int)
+    ;       oo_string_trie_switch_size(int)
+    ;       oo_string_hash_switch_size(int)
+    ;       oo_string_binary_switch_size(int)
+    ;       oo_tag_switch_size(int)
+    ;       oo_try_switch_size(int)
+    ;       oo_binary_switch_size(int)
+    ;       oo_local_var_access_threshold(int)
+    ;       oo_opt_repeat(int)
+    ;       oo_layout_compression_limit(int)
+    ;       oo_procs_per_c_function(int)
+    ;       oo_tuple_trace_counts_file(string)
+    ;       oo_opt_level(int)
+    ;       oo_opt_for_space.
+
+%---------------------%
+
+:- type opt_tuple
+    --->    opt_tuple(
+                ot_allow_inlining             :: maybe_allow_inlining,
+                ot_inline_simple              :: maybe_inline_simple,
+                ot_inline_builtins            :: maybe_inline_builtins,
+                ot_inline_single_use          :: maybe_inline_single_use,
+                ot_inline_linear_tail_rec_sccs :: maybe_inline_linear_tail_rec_sccs,
+                ot_enable_const_struct        :: maybe_enable_const_struct,
+                ot_opt_common_structs         :: maybe_opt_common_structs,
+                ot_prop_constraints           :: maybe_prop_constraints,
+                ot_prop_local_constraints     :: maybe_prop_local_constraints,
+                ot_opt_dup_calls              :: maybe_opt_dup_calls,
+                ot_prop_constants             :: maybe_prop_constants,
+                ot_elim_excess_assigns        :: maybe_elim_excess_assigns,
+                ot_opt_test_after_switch      :: maybe_opt_test_after_switch,
+                ot_opt_format_calls           :: maybe_opt_format_calls,
+                ot_opt_loop_invariants        :: maybe_opt_loop_invariants,
+                ot_opt_saved_vars_const       :: maybe_opt_saved_vars_const,
+                ot_opt_svcell                 :: maybe_opt_svcell,
+                ot_opt_svcell_loop            :: maybe_opt_svcell_loop,
+                ot_opt_svcell_full_path       :: maybe_opt_svcell_full_path,
+                ot_opt_svcell_on_stack        :: maybe_opt_svcell_on_stack,
+                ot_opt_svcell_candidate_headvars :: maybe_opt_svcell_candidate_headvars,
+                ot_opt_svcell_all_candidates  :: maybe_opt_svcell_all_candidates,
+                ot_delay_constructs           :: maybe_delay_constructs,
+                ot_opt_follow_code            :: maybe_opt_follow_code,
+                ot_opt_unused_args            :: maybe_opt_unused_args,
+                ot_opt_unused_args_intermod   :: maybe_opt_unused_args_intermod,
+                ot_opt_higher_order           :: maybe_opt_higher_order,
+                ot_opt_unneeded_code          :: maybe_opt_unneeded_code,
+                ot_spec_types                 :: maybe_spec_types,
+                ot_spec_types_user_guided     :: maybe_spec_types_user_guided,
+                ot_introduce_accumulators     :: maybe_introduce_accumulators,
+                ot_opt_lcmc_accumulator       :: maybe_opt_lcmc_accumulator,
+                ot_opt_lcmc_null              :: maybe_opt_lcmc_null,
+                ot_opt_lcmc                   :: maybe_opt_lcmc,
+                ot_opt_dead_procs             :: maybe_opt_dead_procs,
+                ot_deforest                   :: maybe_deforest,
+                ot_untuple                    :: maybe_untuple,
+                ot_tuple                      :: maybe_tuple,
+                ot_inline_par_builtins        :: maybe_inline_par_builtins,
+                ot_spec_in_all_dep_par_conjs  :: maybe_spec_in_all_dep_par_conjs,
+                ot_allow_some_paths_only_waits :: maybe_allow_some_paths_only_waits,
+                ot_analyse_regions            :: maybe_analyse_regions,
+                ot_use_smart_indexing         :: maybe_use_smart_indexing,
+                ot_use_smart_indexing_atomic  :: maybe_use_smart_indexing_atomic,
+                ot_use_smart_indexing_string  :: maybe_use_smart_indexing_string,
+                ot_use_smart_indexing_tag     :: maybe_use_smart_indexing_tag,
+                ot_use_smart_indexing_float   :: maybe_use_smart_indexing_float,
+                ot_put_base_first_single_rec  :: maybe_put_base_first_single_rec,
+                ot_put_base_first_multi_rec   :: maybe_put_base_first_multi_rec,
+                ot_use_static_ground_cells    :: maybe_use_static_ground_cells,
+                ot_use_static_ground_floats   :: maybe_use_static_ground_floats,
+                ot_use_static_ground_int64s   :: maybe_use_static_ground_int64s,
+                ot_use_static_code_addresses  :: maybe_use_static_code_addresses,
+                ot_use_atomic_cells           :: maybe_use_atomic_cells,
+                ot_opt_middle_rec             :: maybe_opt_middle_rec,
+                ot_opt_simple_neg             :: maybe_opt_simple_neg,
+                ot_allow_hijacks              :: maybe_allow_hijacks,
+                ot_opt_mlds_tailcalls         :: maybe_opt_mlds_tailcalls,
+                ot_opt_initializations        :: maybe_opt_initializations,
+                ot_elim_unused_mlds_assigns   :: maybe_elim_unused_mlds_assigns,
+                ot_elim_local_vars            :: maybe_elim_local_vars,
+                ot_gen_trail_ops_inline       :: maybe_gen_trail_ops_inline,
+                ot_use_common_data            :: maybe_use_common_data,
+                ot_use_common_layout_data     :: maybe_use_common_layout_data,
+                ot_optimize                   :: maybe_optimize,
+                ot_opt_peep                   :: maybe_opt_peep,
+                ot_opt_peep_mkword            :: maybe_opt_peep_mkword,
+                ot_opt_jumps                  :: maybe_opt_jumps,
+                ot_opt_fulljumps              :: maybe_opt_fulljumps,
+                ot_pessimize_tailcalls        :: maybe_pessimize_tailcalls,
+                ot_opt_checked_nondet_tailcalls :: maybe_opt_checked_nondet_tailcalls,
+                ot_use_local_vars             :: maybe_use_local_vars,
+                ot_standardize_labels         :: maybe_standardize_labels,
+                ot_opt_labels                 :: maybe_opt_labels,
+                ot_opt_dups                   :: maybe_opt_dups,
+                ot_opt_proc_dups              :: maybe_opt_proc_dups,
+                ot_opt_frames                 :: maybe_opt_frames,
+                ot_opt_delay_slot             :: maybe_opt_delay_slot,
+                ot_opt_reassign               :: maybe_opt_reassign,
+                ot_use_macro_for_redo_fail    :: maybe_use_macro_for_redo_fail,
+                ot_emit_c_loops               :: maybe_emit_c_loops,
+                ot_use_local_thread_engine_base :: maybe_use_local_thread_engine_base,
+                ot_switch_on_strings_as_atoms :: maybe_switch_on_strings_as_atoms,
+                ot_inline_alloc               :: maybe_inline_alloc,
+                ot_opt_c                      :: maybe_opt_c,
+                ot_inline_call_cost           :: int,
+                ot_inline_compound_threshold  :: int,
+                ot_inline_simple_threshold    :: int,
+                ot_inline_vars_threshold      :: int,
+                ot_intermod_inline_simple_threshold :: int,
+                ot_inline_linear_tail_rec_sccs_max_extra :: int,
+                ot_from_ground_term_threshold :: int,
+                ot_opt_svcell_cv_store_cost   :: int,
+                ot_opt_svcell_cv_load_cost    :: int,
+                ot_opt_svcell_fv_store_cost   :: int,
+                ot_opt_svcell_fv_load_cost    :: int,
+                ot_opt_svcell_op_ratio        :: int,
+                ot_opt_svcell_node_ratio      :: int,
+                ot_opt_svcell_all_path_node_ratio :: int,
+                ot_higher_order_size_limit    :: int,
+                ot_higher_order_arg_limit     :: int,
+                ot_opt_unneeded_code_copy_limit :: int,
+                ot_deforestation_depth_limit  :: int,
+                ot_deforestation_cost_factor  :: int,
+                ot_deforestation_vars_threshold :: int,
+                ot_deforestation_size_threshold :: int,
+                ot_tuple_costs_ratio          :: int,
+                ot_tuple_min_args             :: int,
+                ot_dense_switch_req_density   :: int,
+                ot_lookup_switch_req_density  :: int,
+                ot_dense_switch_size          :: int,
+                ot_lookup_switch_size         :: int,
+                ot_string_trie_switch_size    :: int,
+                ot_string_hash_switch_size    :: int,
+                ot_string_binary_switch_size  :: int,
+                ot_tag_switch_size            :: int,
+                ot_try_switch_size            :: int,
+                ot_binary_switch_size         :: int,
+                ot_local_var_access_threshold :: int,
+                ot_opt_repeat                 :: int,
+                ot_layout_compression_limit   :: int,
+                ot_procs_per_c_function       :: int,
+                ot_tuple_trace_counts_file    :: string
+            ).
+
+:- pred process_optimization_options(option_table::in,
+    list(optimization_option)::in, opt_tuple::out) is det.
+
+%---------------------------------------------------------------------------%
+
+:- implementation.
+
+:- import_module int.
+
+%---------------------%
+
+process_optimization_options(OptionTable, OptOptions, !:OptTuple) :-
+    !:OptTuple = init_opt_tuple,
+    list.foldl2(update_opt_tuple(OptionTable), OptOptions,
+        !OptTuple, not_seen_opt_level, MaybeSeenOptLevel),
+    (
+        MaybeSeenOptLevel = not_seen_opt_level,
+        set_opts_upto_level(OptionTable, 0, 2,
+            !OptTuple, MaybeSeenOptLevel, _)
+    ;
+        MaybeSeenOptLevel = seen_opt_level
+    ).
+
+%---------------------%
+
+:- func init_opt_tuple = opt_tuple.
+
+init_opt_tuple = opt_tuple(
+        allow_inlining,
+        do_not_inline_simple,
+        inline_builtins,
+        do_not_inline_single_use,
+        do_not_inline_linear_tail_rec_sccs,
+        enable_const_struct,
+        do_not_opt_common_structs,
+        do_not_prop_constraints,
+        do_not_prop_local_constraints,
+        do_not_opt_dup_calls,
+        do_not_prop_constants,
+        do_not_elim_excess_assigns,
+        do_not_opt_test_after_switch,
+        opt_format_calls,
+        do_not_opt_loop_invariants,
+        do_not_opt_saved_vars_const,
+        do_not_opt_svcell,
+        opt_svcell_loop,
+        opt_svcell_full_path,
+        opt_svcell_on_stack,
+        opt_svcell_candidate_headvars,
+        do_not_opt_svcell_all_candidates,
+        do_not_delay_constructs,
+        do_not_opt_follow_code,
+        do_not_opt_unused_args,
+        do_not_opt_unused_args_intermod,
+        do_not_opt_higher_order,
+        do_not_opt_unneeded_code,
+        do_not_spec_types,
+        do_not_spec_types_user_guided,
+        do_not_introduce_accumulators,
+        do_not_opt_lcmc_accumulator,
+        do_not_opt_lcmc_null,
+        do_not_opt_lcmc,
+        do_not_opt_dead_procs,
+        do_not_deforest,
+        do_not_untuple,
+        do_not_tuple,
+        do_not_inline_par_builtins,
+        do_not_spec_in_all_dep_par_conjs,
+        allow_some_paths_only_waits,
+        do_not_analyse_regions,
+        do_not_use_smart_indexing,
+        use_smart_indexing_atomic,
+        use_smart_indexing_string,
+        use_smart_indexing_tag,
+        use_smart_indexing_float,
+        do_not_put_base_first_single_rec,
+        put_base_first_multi_rec,
+        do_not_use_static_ground_cells,
+        do_not_use_static_ground_floats,
+        do_not_use_static_ground_int64s,
+        do_not_use_static_code_addresses,
+        do_not_use_atomic_cells,
+        do_not_opt_middle_rec,
+        do_not_opt_simple_neg,
+        allow_hijacks,
+        do_not_opt_mlds_tailcalls,
+        do_not_opt_initializations,
+        elim_unused_mlds_assigns,
+        do_not_elim_local_vars,
+        gen_trail_ops_inline,
+        do_not_use_common_data,
+        use_common_layout_data,
+        do_not_optimize,
+        do_not_opt_peep,
+        do_not_opt_peep_mkword,
+        do_not_opt_jumps,
+        do_not_opt_fulljumps,
+        do_not_pessimize_tailcalls,
+        do_not_opt_checked_nondet_tailcalls,
+        do_not_use_local_vars,
+        do_not_standardize_labels,
+        do_not_opt_labels,
+        do_not_opt_dups,
+        do_not_opt_proc_dups,
+        do_not_opt_frames,
+        do_not_opt_delay_slot,
+        do_not_opt_reassign,
+        do_not_use_macro_for_redo_fail,
+        do_not_emit_c_loops,
+        use_local_thread_engine_base,
+        do_not_switch_on_strings_as_atoms,
+        do_not_inline_alloc,
+        do_not_opt_c,
+        0,
+        0,
+        5,
+        100,
+        5,
+        0,
+        5,
+        3,
+        1,
+        1,
+        1,
+        100,
+        100,
+        100,
+        20,
+        10,
+        10,
+        4,
+        1000,
+        200,
+        15,
+        100,
+        4,
+        25,
+        25,
+        4,
+        4,
+        16,
+        8,
+        4,
+        3,
+        3,
+        4,
+        2,
+        0,
+        4000,
+        1,
+        ""
+    ).
+
+%---------------------%
+
+:- type maybe_seen_opt_level
+    --->    not_seen_opt_level
+    ;       seen_opt_level.
+
+:- pred update_opt_tuple(option_table::in, optimization_option::in,
+    opt_tuple::in, opt_tuple::out,
+    maybe_seen_opt_level::in, maybe_seen_opt_level::out) is det.
+update_opt_tuple(OptionTable, OptOption, !OptTuple, !MaybeSeenOptLevel) :-
+    require_complete_switch [OptOption]
+    (
+        OptOption = oo_allow_inlining(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_allow_inlining = do_not_allow_inlining
+        then
+            !OptTuple ^ ot_allow_inlining := allow_inlining
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_simple(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_simple = do_not_inline_simple
+        then
+            !OptTuple ^ ot_inline_simple := inline_simple
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_builtins(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_builtins = do_not_inline_builtins
+        then
+            !OptTuple ^ ot_inline_builtins := inline_builtins
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_single_use(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_single_use = do_not_inline_single_use
+        then
+            !OptTuple ^ ot_inline_single_use := inline_single_use
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_linear_tail_rec_sccs(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_linear_tail_rec_sccs = do_not_inline_linear_tail_rec_sccs
+        then
+            !OptTuple ^ ot_inline_linear_tail_rec_sccs := inline_linear_tail_rec_sccs
+        else
+            true
+        )
+    ;
+        OptOption = oo_enable_const_struct(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_enable_const_struct = do_not_enable_const_struct
+        then
+            !OptTuple ^ ot_enable_const_struct := enable_const_struct
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_common_structs(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_common_structs = do_not_opt_common_structs
+        then
+            !OptTuple ^ ot_opt_common_structs := opt_common_structs
+        else
+            true
+        )
+    ;
+        OptOption = oo_prop_constraints(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_prop_constraints = do_not_prop_constraints
+        then
+            !OptTuple ^ ot_prop_constraints := prop_constraints
+        else
+            true
+        )
+    ;
+        OptOption = oo_prop_local_constraints(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_prop_local_constraints = do_not_prop_local_constraints
+        then
+            !OptTuple ^ ot_prop_local_constraints := prop_local_constraints
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_dup_calls(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_dup_calls = do_not_opt_dup_calls
+        then
+            !OptTuple ^ ot_opt_dup_calls := opt_dup_calls
+        else
+            true
+        )
+    ;
+        OptOption = oo_prop_constants(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_prop_constants = do_not_prop_constants
+        then
+            !OptTuple ^ ot_prop_constants := prop_constants
+        else
+            true
+        )
+    ;
+        OptOption = oo_elim_excess_assigns(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_elim_excess_assigns = do_not_elim_excess_assigns
+        then
+            !OptTuple ^ ot_elim_excess_assigns := elim_excess_assigns
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_test_after_switch(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_test_after_switch = do_not_opt_test_after_switch
+        then
+            !OptTuple ^ ot_opt_test_after_switch := opt_test_after_switch
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_format_calls(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_format_calls = do_not_opt_format_calls
+        then
+            !OptTuple ^ ot_opt_format_calls := opt_format_calls
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_loop_invariants(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_loop_invariants = do_not_opt_loop_invariants
+        then
+            !OptTuple ^ ot_opt_loop_invariants := opt_loop_invariants
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_saved_vars_const(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_saved_vars_const = do_not_opt_saved_vars_const
+        then
+            !OptTuple ^ ot_opt_saved_vars_const := opt_saved_vars_const
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell = do_not_opt_svcell
+        then
+            !OptTuple ^ ot_opt_svcell := opt_svcell
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell_loop(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell_loop = do_not_opt_svcell_loop
+        then
+            !OptTuple ^ ot_opt_svcell_loop := opt_svcell_loop
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell_full_path(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell_full_path = do_not_opt_svcell_full_path
+        then
+            !OptTuple ^ ot_opt_svcell_full_path := opt_svcell_full_path
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell_on_stack(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell_on_stack = do_not_opt_svcell_on_stack
+        then
+            !OptTuple ^ ot_opt_svcell_on_stack := opt_svcell_on_stack
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell_candidate_headvars(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell_candidate_headvars = do_not_opt_svcell_candidate_headvars
+        then
+            !OptTuple ^ ot_opt_svcell_candidate_headvars := opt_svcell_candidate_headvars
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_svcell_all_candidates(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_svcell_all_candidates = do_not_opt_svcell_all_candidates
+        then
+            !OptTuple ^ ot_opt_svcell_all_candidates := opt_svcell_all_candidates
+        else
+            true
+        )
+    ;
+        OptOption = oo_delay_constructs(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_delay_constructs = do_not_delay_constructs
+        then
+            !OptTuple ^ ot_delay_constructs := delay_constructs
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_follow_code(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_follow_code = do_not_opt_follow_code
+        then
+            !OptTuple ^ ot_opt_follow_code := opt_follow_code
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_unused_args(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_unused_args = do_not_opt_unused_args
+        then
+            !OptTuple ^ ot_opt_unused_args := opt_unused_args
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_unused_args_intermod(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_unused_args_intermod = do_not_opt_unused_args_intermod
+        then
+            !OptTuple ^ ot_opt_unused_args_intermod := opt_unused_args_intermod
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_higher_order(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_higher_order = do_not_opt_higher_order
+        then
+            !OptTuple ^ ot_opt_higher_order := opt_higher_order
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_unneeded_code(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_unneeded_code = do_not_opt_unneeded_code
+        then
+            !OptTuple ^ ot_opt_unneeded_code := opt_unneeded_code
+        else
+            true
+        )
+    ;
+        OptOption = oo_spec_types(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_spec_types = do_not_spec_types
+        then
+            !OptTuple ^ ot_spec_types := spec_types
+        else
+            true
+        )
+    ;
+        OptOption = oo_spec_types_user_guided(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_spec_types_user_guided = do_not_spec_types_user_guided
+        then
+            !OptTuple ^ ot_spec_types_user_guided := spec_types_user_guided
+        else
+            true
+        )
+    ;
+        OptOption = oo_introduce_accumulators(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_introduce_accumulators = do_not_introduce_accumulators
+        then
+            !OptTuple ^ ot_introduce_accumulators := introduce_accumulators
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_lcmc_accumulator(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_lcmc_accumulator = do_not_opt_lcmc_accumulator
+        then
+            !OptTuple ^ ot_opt_lcmc_accumulator := opt_lcmc_accumulator
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_lcmc_null(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_lcmc_null = do_not_opt_lcmc_null
+        then
+            !OptTuple ^ ot_opt_lcmc_null := opt_lcmc_null
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_lcmc(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_lcmc = do_not_opt_lcmc
+        then
+            !OptTuple ^ ot_opt_lcmc := opt_lcmc
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_dead_procs(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_dead_procs = do_not_opt_dead_procs
+        then
+            !OptTuple ^ ot_opt_dead_procs := opt_dead_procs
+        else
+            true
+        )
+    ;
+        OptOption = oo_deforest(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_deforest = do_not_deforest
+        then
+            !OptTuple ^ ot_deforest := deforest
+        else
+            true
+        )
+    ;
+        OptOption = oo_untuple(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_untuple = do_not_untuple
+        then
+            !OptTuple ^ ot_untuple := untuple
+        else
+            true
+        )
+    ;
+        OptOption = oo_tuple(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_tuple = do_not_tuple
+        then
+            !OptTuple ^ ot_tuple := tuple
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_par_builtins(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_par_builtins = do_not_inline_par_builtins
+        then
+            !OptTuple ^ ot_inline_par_builtins := inline_par_builtins
+        else
+            true
+        )
+    ;
+        OptOption = oo_spec_in_all_dep_par_conjs(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_spec_in_all_dep_par_conjs = do_not_spec_in_all_dep_par_conjs
+        then
+            !OptTuple ^ ot_spec_in_all_dep_par_conjs := spec_in_all_dep_par_conjs
+        else
+            true
+        )
+    ;
+        OptOption = oo_allow_some_paths_only_waits(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_allow_some_paths_only_waits = do_not_allow_some_paths_only_waits
+        then
+            !OptTuple ^ ot_allow_some_paths_only_waits := allow_some_paths_only_waits
+        else
+            true
+        )
+    ;
+        OptOption = oo_analyse_regions(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_analyse_regions = do_not_analyse_regions
+        then
+            !OptTuple ^ ot_analyse_regions := analyse_regions
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_smart_indexing(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_smart_indexing = do_not_use_smart_indexing
+        then
+            !OptTuple ^ ot_use_smart_indexing := use_smart_indexing
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_smart_indexing_atomic(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_smart_indexing_atomic = do_not_use_smart_indexing_atomic
+        then
+            !OptTuple ^ ot_use_smart_indexing_atomic := use_smart_indexing_atomic
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_smart_indexing_string(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_smart_indexing_string = do_not_use_smart_indexing_string
+        then
+            !OptTuple ^ ot_use_smart_indexing_string := use_smart_indexing_string
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_smart_indexing_tag(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_smart_indexing_tag = do_not_use_smart_indexing_tag
+        then
+            !OptTuple ^ ot_use_smart_indexing_tag := use_smart_indexing_tag
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_smart_indexing_float(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_smart_indexing_float = do_not_use_smart_indexing_float
+        then
+            !OptTuple ^ ot_use_smart_indexing_float := use_smart_indexing_float
+        else
+            true
+        )
+    ;
+        OptOption = oo_put_base_first_single_rec(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_put_base_first_single_rec = do_not_put_base_first_single_rec
+        then
+            !OptTuple ^ ot_put_base_first_single_rec := put_base_first_single_rec
+        else
+            true
+        )
+    ;
+        OptOption = oo_put_base_first_multi_rec(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_put_base_first_multi_rec = do_not_put_base_first_multi_rec
+        then
+            !OptTuple ^ ot_put_base_first_multi_rec := put_base_first_multi_rec
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_static_ground_cells(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_static_ground_cells = do_not_use_static_ground_cells
+        then
+            !OptTuple ^ ot_use_static_ground_cells := use_static_ground_cells
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_static_ground_floats(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_static_ground_floats = do_not_use_static_ground_floats
+        then
+            !OptTuple ^ ot_use_static_ground_floats := use_static_ground_floats
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_static_ground_int64s(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_static_ground_int64s = do_not_use_static_ground_int64s
+        then
+            !OptTuple ^ ot_use_static_ground_int64s := use_static_ground_int64s
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_static_code_addresses(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_static_code_addresses = do_not_use_static_code_addresses
+        then
+            !OptTuple ^ ot_use_static_code_addresses := use_static_code_addresses
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_atomic_cells(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_atomic_cells = do_not_use_atomic_cells
+        then
+            !OptTuple ^ ot_use_atomic_cells := use_atomic_cells
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_middle_rec(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_middle_rec = do_not_opt_middle_rec
+        then
+            !OptTuple ^ ot_opt_middle_rec := opt_middle_rec
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_simple_neg(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_simple_neg = do_not_opt_simple_neg
+        then
+            !OptTuple ^ ot_opt_simple_neg := opt_simple_neg
+        else
+            true
+        )
+    ;
+        OptOption = oo_allow_hijacks(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_allow_hijacks = do_not_allow_hijacks
+        then
+            !OptTuple ^ ot_allow_hijacks := allow_hijacks
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_mlds_tailcalls(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_mlds_tailcalls = do_not_opt_mlds_tailcalls
+        then
+            !OptTuple ^ ot_opt_mlds_tailcalls := opt_mlds_tailcalls
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_initializations(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_initializations = do_not_opt_initializations
+        then
+            !OptTuple ^ ot_opt_initializations := opt_initializations
+        else
+            true
+        )
+    ;
+        OptOption = oo_elim_unused_mlds_assigns(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_elim_unused_mlds_assigns = do_not_elim_unused_mlds_assigns
+        then
+            !OptTuple ^ ot_elim_unused_mlds_assigns := elim_unused_mlds_assigns
+        else
+            true
+        )
+    ;
+        OptOption = oo_elim_local_vars(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_elim_local_vars = do_not_elim_local_vars
+        then
+            !OptTuple ^ ot_elim_local_vars := elim_local_vars
+        else
+            true
+        )
+    ;
+        OptOption = oo_gen_trail_ops_inline(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_gen_trail_ops_inline = do_not_gen_trail_ops_inline
+        then
+            !OptTuple ^ ot_gen_trail_ops_inline := gen_trail_ops_inline
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_common_data(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_common_data = do_not_use_common_data
+        then
+            !OptTuple ^ ot_use_common_data := use_common_data
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_common_layout_data(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_common_layout_data = do_not_use_common_layout_data
+        then
+            !OptTuple ^ ot_use_common_layout_data := use_common_layout_data
+        else
+            true
+        )
+    ;
+        OptOption = oo_optimize(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_optimize = do_not_optimize
+        then
+            !OptTuple ^ ot_optimize := optimize
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_peep(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_peep = do_not_opt_peep
+        then
+            !OptTuple ^ ot_opt_peep := opt_peep
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_peep_mkword(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_peep_mkword = do_not_opt_peep_mkword
+        then
+            !OptTuple ^ ot_opt_peep_mkword := opt_peep_mkword
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_jumps(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_jumps = do_not_opt_jumps
+        then
+            !OptTuple ^ ot_opt_jumps := opt_jumps
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_fulljumps(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_fulljumps = do_not_opt_fulljumps
+        then
+            !OptTuple ^ ot_opt_fulljumps := opt_fulljumps
+        else
+            true
+        )
+    ;
+        OptOption = oo_pessimize_tailcalls(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_pessimize_tailcalls = do_not_pessimize_tailcalls
+        then
+            !OptTuple ^ ot_pessimize_tailcalls := pessimize_tailcalls
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_checked_nondet_tailcalls(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_checked_nondet_tailcalls = do_not_opt_checked_nondet_tailcalls
+        then
+            !OptTuple ^ ot_opt_checked_nondet_tailcalls := opt_checked_nondet_tailcalls
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_local_vars(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_local_vars = do_not_use_local_vars
+        then
+            !OptTuple ^ ot_use_local_vars := use_local_vars
+        else
+            true
+        )
+    ;
+        OptOption = oo_standardize_labels(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_standardize_labels = do_not_standardize_labels
+        then
+            !OptTuple ^ ot_standardize_labels := standardize_labels
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_labels(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_labels = do_not_opt_labels
+        then
+            !OptTuple ^ ot_opt_labels := opt_labels
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_dups(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_dups = do_not_opt_dups
+        then
+            !OptTuple ^ ot_opt_dups := opt_dups
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_proc_dups(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_proc_dups = do_not_opt_proc_dups
+        then
+            !OptTuple ^ ot_opt_proc_dups := opt_proc_dups
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_frames(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_frames = do_not_opt_frames
+        then
+            !OptTuple ^ ot_opt_frames := opt_frames
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_delay_slot(Bool),
+        ( if
+            Bool = yes,
+            getopt_io.lookup_bool_option(OptionTable, have_delay_slot, yes),
+            !.OptTuple ^ ot_opt_delay_slot = do_not_opt_delay_slot
+        then
+            !OptTuple ^ ot_opt_delay_slot := opt_delay_slot
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_reassign(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_reassign = do_not_opt_reassign
+        then
+            !OptTuple ^ ot_opt_reassign := opt_reassign
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_macro_for_redo_fail(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_macro_for_redo_fail = do_not_use_macro_for_redo_fail
+        then
+            !OptTuple ^ ot_use_macro_for_redo_fail := use_macro_for_redo_fail
+        else
+            true
+        )
+    ;
+        OptOption = oo_emit_c_loops(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_emit_c_loops = do_not_emit_c_loops
+        then
+            !OptTuple ^ ot_emit_c_loops := emit_c_loops
+        else
+            true
+        )
+    ;
+        OptOption = oo_use_local_thread_engine_base(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_use_local_thread_engine_base = do_not_use_local_thread_engine_base
+        then
+            !OptTuple ^ ot_use_local_thread_engine_base := use_local_thread_engine_base
+        else
+            true
+        )
+    ;
+        OptOption = oo_switch_on_strings_as_atoms(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_switch_on_strings_as_atoms = do_not_switch_on_strings_as_atoms
+        then
+            !OptTuple ^ ot_switch_on_strings_as_atoms := switch_on_strings_as_atoms
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_alloc(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_inline_alloc = do_not_inline_alloc
+        then
+            !OptTuple ^ ot_inline_alloc := inline_alloc
+        else
+            true
+        )
+    ;
+        OptOption = oo_opt_c(Bool),
+        ( if
+            Bool = yes,
+            !.OptTuple ^ ot_opt_c = do_not_opt_c
+        then
+            !OptTuple ^ ot_opt_c := opt_c
+        else
+            true
+        )
+    ;
+        OptOption = oo_inline_call_cost(N),
+        OldN = !.OptTuple ^ ot_inline_call_cost,
+        !OptTuple ^ ot_inline_call_cost := int.max(OldN, N)
+    ;
+        OptOption = oo_inline_compound_threshold(N),
+        OldN = !.OptTuple ^ ot_inline_compound_threshold,
+        !OptTuple ^ ot_inline_compound_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_inline_simple_threshold(N),
+        OldN = !.OptTuple ^ ot_inline_simple_threshold,
+        !OptTuple ^ ot_inline_simple_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_inline_vars_threshold(N),
+        OldN = !.OptTuple ^ ot_inline_vars_threshold,
+        !OptTuple ^ ot_inline_vars_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_intermod_inline_simple_threshold(N),
+        OldN = !.OptTuple ^ ot_intermod_inline_simple_threshold,
+        !OptTuple ^ ot_intermod_inline_simple_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_inline_linear_tail_rec_sccs_max_extra(N),
+        OldN = !.OptTuple ^ ot_inline_linear_tail_rec_sccs_max_extra,
+        !OptTuple ^ ot_inline_linear_tail_rec_sccs_max_extra := int.max(OldN, N)
+    ;
+        OptOption = oo_from_ground_term_threshold(N),
+        OldN = !.OptTuple ^ ot_from_ground_term_threshold,
+        !OptTuple ^ ot_from_ground_term_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_cv_store_cost(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_cv_store_cost,
+        !OptTuple ^ ot_opt_svcell_cv_store_cost := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_cv_load_cost(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_cv_load_cost,
+        !OptTuple ^ ot_opt_svcell_cv_load_cost := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_fv_store_cost(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_fv_store_cost,
+        !OptTuple ^ ot_opt_svcell_fv_store_cost := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_fv_load_cost(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_fv_load_cost,
+        !OptTuple ^ ot_opt_svcell_fv_load_cost := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_op_ratio(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_op_ratio,
+        !OptTuple ^ ot_opt_svcell_op_ratio := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_node_ratio(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_node_ratio,
+        !OptTuple ^ ot_opt_svcell_node_ratio := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_svcell_all_path_node_ratio(N),
+        OldN = !.OptTuple ^ ot_opt_svcell_all_path_node_ratio,
+        !OptTuple ^ ot_opt_svcell_all_path_node_ratio := int.max(OldN, N)
+    ;
+        OptOption = oo_higher_order_size_limit(N),
+        OldN = !.OptTuple ^ ot_higher_order_size_limit,
+        !OptTuple ^ ot_higher_order_size_limit := int.max(OldN, N)
+    ;
+        OptOption = oo_higher_order_arg_limit(N),
+        OldN = !.OptTuple ^ ot_higher_order_arg_limit,
+        !OptTuple ^ ot_higher_order_arg_limit := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_unneeded_code_copy_limit(N),
+        OldN = !.OptTuple ^ ot_opt_unneeded_code_copy_limit,
+        !OptTuple ^ ot_opt_unneeded_code_copy_limit := int.max(OldN, N)
+    ;
+        OptOption = oo_deforestation_depth_limit(N),
+        OldN = !.OptTuple ^ ot_deforestation_depth_limit,
+        !OptTuple ^ ot_deforestation_depth_limit := int.max(OldN, N)
+    ;
+        OptOption = oo_deforestation_cost_factor(N),
+        OldN = !.OptTuple ^ ot_deforestation_cost_factor,
+        !OptTuple ^ ot_deforestation_cost_factor := int.max(OldN, N)
+    ;
+        OptOption = oo_deforestation_vars_threshold(N),
+        OldN = !.OptTuple ^ ot_deforestation_vars_threshold,
+        !OptTuple ^ ot_deforestation_vars_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_deforestation_size_threshold(N),
+        OldN = !.OptTuple ^ ot_deforestation_size_threshold,
+        !OptTuple ^ ot_deforestation_size_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_tuple_costs_ratio(N),
+        OldN = !.OptTuple ^ ot_tuple_costs_ratio,
+        !OptTuple ^ ot_tuple_costs_ratio := int.max(OldN, N)
+    ;
+        OptOption = oo_tuple_min_args(N),
+        OldN = !.OptTuple ^ ot_tuple_min_args,
+        !OptTuple ^ ot_tuple_min_args := int.max(OldN, N)
+    ;
+        OptOption = oo_dense_switch_req_density(N),
+        OldN = !.OptTuple ^ ot_dense_switch_req_density,
+        !OptTuple ^ ot_dense_switch_req_density := int.max(OldN, N)
+    ;
+        OptOption = oo_lookup_switch_req_density(N),
+        OldN = !.OptTuple ^ ot_lookup_switch_req_density,
+        !OptTuple ^ ot_lookup_switch_req_density := int.max(OldN, N)
+    ;
+        OptOption = oo_dense_switch_size(N),
+        OldN = !.OptTuple ^ ot_dense_switch_size,
+        !OptTuple ^ ot_dense_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_lookup_switch_size(N),
+        OldN = !.OptTuple ^ ot_lookup_switch_size,
+        !OptTuple ^ ot_lookup_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_string_trie_switch_size(N),
+        OldN = !.OptTuple ^ ot_string_trie_switch_size,
+        !OptTuple ^ ot_string_trie_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_string_hash_switch_size(N),
+        OldN = !.OptTuple ^ ot_string_hash_switch_size,
+        !OptTuple ^ ot_string_hash_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_string_binary_switch_size(N),
+        OldN = !.OptTuple ^ ot_string_binary_switch_size,
+        !OptTuple ^ ot_string_binary_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_tag_switch_size(N),
+        OldN = !.OptTuple ^ ot_tag_switch_size,
+        !OptTuple ^ ot_tag_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_try_switch_size(N),
+        OldN = !.OptTuple ^ ot_try_switch_size,
+        !OptTuple ^ ot_try_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_binary_switch_size(N),
+        OldN = !.OptTuple ^ ot_binary_switch_size,
+        !OptTuple ^ ot_binary_switch_size := int.max(OldN, N)
+    ;
+        OptOption = oo_local_var_access_threshold(N),
+        OldN = !.OptTuple ^ ot_local_var_access_threshold,
+        !OptTuple ^ ot_local_var_access_threshold := int.max(OldN, N)
+    ;
+        OptOption = oo_opt_repeat(N),
+        OldN = !.OptTuple ^ ot_opt_repeat,
+        !OptTuple ^ ot_opt_repeat := int.max(OldN, N)
+    ;
+        OptOption = oo_layout_compression_limit(N),
+        OldN = !.OptTuple ^ ot_layout_compression_limit,
+        !OptTuple ^ ot_layout_compression_limit := int.max(OldN, N)
+    ;
+        OptOption = oo_procs_per_c_function(N),
+        OldN = !.OptTuple ^ ot_procs_per_c_function,
+        !OptTuple ^ ot_procs_per_c_function := int.max(OldN, N)
+    ;
+        OptOption = oo_tuple_trace_counts_file(Str),
+        !OptTuple ^ ot_tuple_trace_counts_file := Str
+    ;
+        OptOption = oo_opt_level(OptLevel),
+        set_opts_upto_level(OptionTable, 0, OptLevel, !OptTuple, !.MaybeSeenOptLevel, _),
+        !:MaybeSeenOptLevel = seen_opt_level
+    ;
+        OptOption = oo_opt_for_space,
+        set_opts_for_space(!OptTuple)
+    ).
+
+:- pred set_opts_upto_level(option_table::in, int::in, int::in,
+    opt_tuple::in, opt_tuple::out,
+    maybe_seen_opt_level::in, maybe_seen_opt_level::out) is det.
+
+set_opts_upto_level(OptionTable, Cur, Max,
+        !OptTuple, !MaybeSeenOptLevel) :-
+    ( if
+        Cur =< Max,
+        opts_enabled_at_level(Cur, LevelOptOptions)
+    then
+        list.foldl2(update_opt_tuple(OptionTable), LevelOptOptions,
+            !OptTuple, !MaybeSeenOptLevel),
+        set_opts_upto_level(OptionTable, Cur + 1, Max,
+            !OptTuple, !MaybeSeenOptLevel)
+    else
+        true
+    ).
+
+:- pred set_opts_for_space(opt_tuple::in, opt_tuple::out) is det.
+
+set_opts_for_space(!OptTuple) :-
+    UnneededCopyLimit = !.OptTuple ^ ot_opt_unneeded_code_copy_limit,
+    !OptTuple ^ ot_opt_unneeded_code_copy_limit :=
+        int.min(UnneededCopyLimit, 1),
+    !OptTuple ^ ot_opt_dead_procs := opt_dead_procs,
+    !OptTuple ^ ot_opt_labels := opt_labels,
+    !OptTuple ^ ot_opt_dups := opt_dups,
+    !OptTuple ^ ot_opt_proc_dups := opt_proc_dups,
+    !OptTuple ^ ot_opt_fulljumps := opt_fulljumps,
+    !OptTuple ^ ot_opt_reassign := opt_reassign,
+    !OptTuple ^ ot_inline_alloc := inline_alloc,
+    !OptTuple ^ ot_use_macro_for_redo_fail := use_macro_for_redo_fail,
+    !OptTuple ^ ot_opt_loop_invariants := do_not_opt_loop_invariants.
+
+:- pred opts_enabled_at_level(int::in, list(optimization_option)::out)
+    is semidet.
+
+opts_enabled_at_level(0, [
+    % Optimization level 0: aim to minimize overall compilation time.
+    oo_use_common_data(yes),
+    oo_optimize(yes),
+    oo_opt_repeat(1),
+    oo_opt_peep(yes),
+    oo_opt_peep_mkword(yes),
+    oo_use_static_ground_cells(yes),
+    oo_use_smart_indexing(yes),
+    oo_opt_jumps(yes),
+    oo_opt_labels(yes),
+    oo_opt_dead_procs(yes),
+    oo_elim_excess_assigns(yes)
+]).
+opts_enabled_at_level(1, [
+    % Optimization level 1: apply optimizations which are cheap and
+    % have a good payoff while still keeping compilation time small.
+    oo_use_local_vars(yes),
+    oo_opt_c(yes),              % XXX We want `gcc -O1'
+    oo_opt_frames(yes),
+    oo_opt_delay_slot(yes),
+    oo_opt_middle_rec(yes),
+    oo_emit_c_loops(yes),
+    oo_opt_mlds_tailcalls(yes)
+]).
+opts_enabled_at_level(2, [
+    % Optimization level 2: apply optimizations which have a good payoff
+    % relative to their cost; but include optimizations which are
+    % more costly than with -O1.
+    oo_opt_fulljumps(yes),
+    oo_opt_repeat(3),
+    oo_opt_dups(yes),
+    oo_opt_follow_code(yes),
+    oo_inline_simple(yes),
+    oo_inline_single_use(yes),
+    oo_inline_compound_threshold(10),
+    oo_opt_common_structs(yes),
+    oo_spec_types_user_guided(yes),
+    oo_opt_simple_neg(yes),
+    oo_opt_test_after_switch(yes),
+    oo_opt_initializations(yes)
+]).
+opts_enabled_at_level(3, [
+    % Optimization level 3: apply optimizations which usually have a good
+    % payoff even if they increase compilation time quite a bit.
+    oo_opt_saved_vars_const(yes),
+    oo_opt_unused_args(yes),
+    oo_opt_higher_order(yes),
+    oo_deforest(yes),
+    oo_prop_constraints(yes),
+    oo_prop_local_constraints(yes),
+    oo_opt_reassign(yes),
+    oo_opt_repeat(4)
+]).
+opts_enabled_at_level(4, [
+    % Optimization level 4: apply optimizations which may have some payoff
+    % even if they increase compilation time quite a bit.
+    %
+    % Currently this enables the use of local variables
+    % and increases the inlining thresholds.
+    oo_inline_simple_threshold(8),
+    oo_inline_compound_threshold(20),
+    oo_higher_order_size_limit(30)
+]).
+opts_enabled_at_level(5, [
+    % Optimization level 5: apply optimizations which may have some
+    % payoff even if they increase compilation time a lot.
+    %
+    % Currently this enables the search for construction unifications that
+    % can be delayed past failing computations, allows more passes of the
+    % low-level optimizations, and increases the inlining thresholds
+    % still further. We also enable eliminate_local_vars only at
+    % this level, because that pass is implemented pretty inefficiently.
+    oo_opt_repeat(5),
+    oo_delay_constructs(yes),
+    oo_inline_compound_threshold(100),
+    oo_higher_order_size_limit(40),
+    oo_elim_local_vars(yes),
+    oo_opt_loop_invariants(yes)
+]).
+opts_enabled_at_level(6, [
+     % Optimization level 6: apply optimizations which may have any payoff
+    % even if they increase compilation time to completely unreasonable
+    % levels.
+    %
+    % Currently this sets `everything_in_one_c_function', which causes
+    % the compiler to put everything in the one C function and treat calls
+    % to predicates in the same module as local. We also enable inlining
+    % of GC_malloc(), redo(), and fail().
+    oo_procs_per_c_function(0),
+    oo_inline_alloc(yes),
+    oo_use_macro_for_redo_fail(yes)
+]).
+
+% The following optimization options are not enabled at any level:
+% XXX This comment is old. Since then, we have added many optimizations
+% that are not enabled automatically at any level :-(
+%
+%   checked_nondet_tailcalls:
+%       This is deliberate, because the transformation might make
+%       code run slower.
+%
+%   unneeded_code:
+%       Because it can cause slowdowns at high optimization levels;
+%       cause unknown
+%
+%   type_specialization:
+%       XXX why not?
+%
+%   introduce_accumulators:
+%       XXX Disabled until a bug in extras/trailed_update/var.m
+%       is resolved.
+%
+%   optimize_constructor_last_call:
+%       Not a speedup in general.
+
+%---------------------------------------------------------------------------%
+:- end_module libs.optimization_options.
+%---------------------------------------------------------------------------%
diff --git a/compiler/optimize.m b/compiler/optimize.m
index 24cc2bc3a..a53a5c2f8 100644
--- a/compiler/optimize.m
+++ b/compiler/optimize.m
@@ -40,6 +40,7 @@
 :- import_module hlds.code_model.
 :- import_module hlds.hlds_pred.
 :- import_module libs.file_util.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.continuation_info.
 :- import_module ll_backend.delay_slot.
@@ -324,7 +325,7 @@ optimize_initial(Info, LayoutLabelSet, ProcLabel, CodeModel, MayAlterRtti,
     LabelStr = opt_util.format_proc_label(ProcLabel),
     OptFrames = Info ^ lopt_opt_frames,
     ( if
-        OptFrames = yes,
+        OptFrames = opt_frames,
         MayAlterRtti = may_alter_rtti,
         CodeModel = model_non
     then
@@ -394,7 +395,7 @@ optimize_repeated(Info, Final, LayoutLabelSet, ProcLabel, MayAlterRtti,
     PessimizeTailCalls = Info ^ lopt_pes_tailcalls,
     CheckedNondetTailCalls = Info ^ lopt_checked_nondet_tailcalls,
     (
-        OptJump = yes,
+        OptJump = opt_jumps,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -411,12 +412,12 @@ optimize_repeated(Info, Final, LayoutLabelSet, ProcLabel, MayAlterRtti,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "jump", "after jump opt", ProcLabel, !OptDebugInfo)
     ;
-        OptJump = no,
+        OptJump = do_not_opt_jumps,
         Mod1 = no
     ),
     Peephole = Info ^ lopt_opt_peep,
     (
-        Peephole = yes,
+        Peephole = opt_peep,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -433,12 +434,12 @@ optimize_repeated(Info, Final, LayoutLabelSet, ProcLabel, MayAlterRtti,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "peep", "after peephole", ProcLabel, !OptDebugInfo)
     ;
-        Peephole = no,
+        Peephole = do_not_opt_peep,
         Mod2 = no
     ),
     OptLabels = Info ^ lopt_opt_labels,
     (
-        OptLabels = yes,
+        OptLabels = opt_labels,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -453,12 +454,12 @@ optimize_repeated(Info, Final, LayoutLabelSet, ProcLabel, MayAlterRtti,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "label", "after label opt", ProcLabel, !OptDebugInfo)
     ;
-        OptLabels = no,
+        OptLabels = do_not_opt_labels,
         Mod3 = no
     ),
     DupElim = Info ^ lopt_opt_dups,
     (
-        DupElim = yes,
+        DupElim = opt_dups,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -473,7 +474,7 @@ optimize_repeated(Info, Final, LayoutLabelSet, ProcLabel, MayAlterRtti,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "dup", "after duplicates", ProcLabel, !OptDebugInfo)
     ;
-        DupElim = no
+        DupElim = do_not_opt_dups
     ),
     ( if Mod1 = no, Mod2 = no, Mod3 = no, !.Instrs = InstrsAtStart then
         Mod = no
@@ -497,7 +498,7 @@ optimize_middle(Info, Final, LayoutLabelSet, ProcLabel, CodeModel,
 
     OptFrames = Info ^ lopt_opt_frames,
     (
-        OptFrames = yes,
+        OptFrames = opt_frames,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -531,7 +532,7 @@ optimize_middle(Info, Final, LayoutLabelSet, ProcLabel, CodeModel,
         PessimizeTailCalls = Info ^ lopt_pes_tailcalls,
         CheckedNondetTailCalls = Info ^ lopt_checked_nondet_tailcalls,
         ( if
-            ( OptFullJump = yes
+            ( OptFullJump = opt_fulljumps
             ; Mod1 = yes
             )
         then
@@ -592,11 +593,11 @@ optimize_middle(Info, Final, LayoutLabelSet, ProcLabel, CodeModel,
             Mod1 = no
         )
     ;
-        OptFrames = no
+        OptFrames = do_not_opt_frames
     ),
     UseLocalVars = Info ^ lopt_use_local_vars,
     (
-        UseLocalVars = yes,
+        UseLocalVars = use_local_vars,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -615,7 +616,7 @@ optimize_middle(Info, Final, LayoutLabelSet, ProcLabel, CodeModel,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter, "use_local",
             "after use_local_vars", ProcLabel, !OptDebugInfo)
     ;
-        UseLocalVars = no
+        UseLocalVars = do_not_use_local_vars
     ).
 
 :- pred optimize_last(llds_opt_info::in, set_tree234(label)::in,
@@ -632,10 +633,10 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
     UseLocalVars = Info ^ lopt_use_local_vars,
     StdLabels = Info ^ lopt_std_labels,
     ( if
-        ( Reassign = yes
-        ; DelaySlot = yes
-        ; UseLocalVars = yes
-        ; StdLabels = yes
+        ( Reassign = opt_reassign
+        ; DelaySlot = opt_delay_slot
+        ; UseLocalVars = use_local_vars
+        ; StdLabels = standardize_labels
         )
     then
         % We must get rid of any extra labels added by other passes,
@@ -657,7 +658,7 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
         true
     ),
     (
-        Reassign = yes,
+        Reassign = opt_reassign,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -672,10 +673,10 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "reassign", "after reassign", ProcLabel, !OptDebugInfo)
     ;
-        Reassign = no
+        Reassign = do_not_opt_reassign
     ),
     (
-        DelaySlot = yes,
+        DelaySlot = opt_delay_slot,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -690,7 +691,7 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "delay_slot", "after delay slots", ProcLabel, !OptDebugInfo)
     ;
-        DelaySlot = no
+        DelaySlot = do_not_opt_delay_slot
     ),
     (
         VeryVerbose = yes,
@@ -706,7 +707,7 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
     maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
         "decr_sp", "after combine decr_sp", ProcLabel, !OptDebugInfo),
     (
-        StdLabels = yes,
+        StdLabels = standardize_labels,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -721,10 +722,10 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "stdlabel", "after standard labels", ProcLabel, !OptDebugInfo)
     ;
-        StdLabels = no
+        StdLabels = do_not_standardize_labels
     ),
     (
-        UseLocalVars = yes,
+        UseLocalVars = use_local_vars,
         (
             VeryVerbose = yes,
             trace [io(!IO)] (
@@ -739,7 +740,7 @@ optimize_last(Info, LayoutLabelSet, ProcLabel,
         maybe_opt_debug(Info, !.Instrs, !.LabelNumCounter,
             "wrapblocks", "after wrap blocks", ProcLabel, !.OptDebugInfo, _)
     ;
-        UseLocalVars = no
+        UseLocalVars = do_not_use_local_vars
     ).
 
 %-----------------------------------------------------------------------------%
@@ -765,34 +766,35 @@ escape_dir_char(Char, !Str) :-
 
 :- type llds_opt_info
     --->    llds_opt_info(
-                lopt_debug_opt_pred_ids             :: list(string),
-                lopt_debug_opt_pred_names           :: list(string),
-                lopt_num_real_r_regs                :: int,
-                lopt_local_vars_access_threshold    :: int,
-                lopt_opt_repeat                     :: int,
-
-                lopt_gc_method                      :: gc_method,
-
-                lopt_debug_opt                      :: bool,
-
-                lopt_auto_comments                  :: maybe_auto_comments,
-                lopt_frameopt_comments              :: bool,
-                lopt_detailed_statistics            :: bool,
-                lopt_very_verbose                   :: bool,
-
-                lopt_checked_nondet_tailcalls       :: bool,
-                lopt_opt_delay_slots                :: bool,
-                lopt_opt_dups                       :: bool,
-                lopt_opt_frames                     :: bool,
-                lopt_opt_jumps                      :: bool,
-                lopt_opt_fulljumps                  :: bool,
-                lopt_opt_labels                     :: bool,
-                lopt_opt_peep                       :: bool,
-                lopt_opt_peep_mkword                :: bool,
-                lopt_opt_reassign                   :: bool,
-                lopt_pes_tailcalls                  :: bool,
-                lopt_std_labels                     :: bool,
-                lopt_use_local_vars                 :: bool
+                lopt_debug_opt_pred_ids         :: list(string),
+                lopt_debug_opt_pred_names       :: list(string),
+                lopt_num_real_r_regs            :: int,
+                lopt_local_vars_access_threshold :: int,
+                lopt_opt_repeat                 :: int,
+
+                lopt_gc_method                  :: gc_method,
+
+                lopt_debug_opt                  :: bool,
+
+                lopt_auto_comments              :: maybe_auto_comments,
+                lopt_frameopt_comments          :: bool,
+                lopt_detailed_statistics        :: bool,
+                lopt_very_verbose               :: bool,
+
+                lopt_checked_nondet_tailcalls   ::
+                                            maybe_opt_checked_nondet_tailcalls,
+                lopt_opt_delay_slots            :: maybe_opt_delay_slot,
+                lopt_opt_dups                   :: maybe_opt_dups,
+                lopt_opt_frames                 :: maybe_opt_frames,
+                lopt_opt_jumps                  :: maybe_opt_jumps,
+                lopt_opt_fulljumps              :: maybe_opt_fulljumps,
+                lopt_opt_labels                 :: maybe_opt_labels,
+                lopt_opt_peep                   :: maybe_opt_peep,
+                lopt_opt_peep_mkword            :: maybe_opt_peep_mkword,
+                lopt_opt_reassign               :: maybe_opt_reassign,
+                lopt_pes_tailcalls              :: maybe_pessimize_tailcalls,
+                lopt_std_labels                 :: maybe_standardize_labels,
+                lopt_use_local_vars             :: maybe_use_local_vars
             ).
 
 :- func init_llds_opt_info(globals) = llds_opt_info.
@@ -803,9 +805,9 @@ init_llds_opt_info(Globals) = Info :-
     globals.lookup_accumulating_option(Globals, debug_opt_pred_name,
         DebugOptPredNames),
     globals.lookup_int_option(Globals, num_real_r_regs, NumRealRRegs),
-    globals.lookup_int_option(Globals, local_var_access_threshold,
-        LocalVarAccessThreshold),
-    globals.lookup_int_option(Globals, optimize_repeat, OptRepeat),
+    globals.get_opt_tuple(Globals, OptTuple),
+    LocalVarAccessThreshold = OptTuple ^ ot_local_var_access_threshold,
+    OptRepeat = OptTuple ^ ot_opt_repeat,
 
     globals.get_gc_method(Globals, GCMethod),
 
@@ -820,21 +822,19 @@ init_llds_opt_info(Globals) = Info :-
         DetailedStatistics),
     globals.lookup_bool_option(Globals, very_verbose, VeryVerbose),
 
-    globals.lookup_bool_option(Globals, checked_nondet_tailcalls,
-        CheckedNondetTailCalls),
-    globals.lookup_bool_option(Globals, optimize_delay_slot, OptDelaySlots),
-    globals.lookup_bool_option(Globals, optimize_dups, OptDups),
-    globals.lookup_bool_option(Globals, optimize_frames, OptFrames),
-    globals.lookup_bool_option(Globals, optimize_jumps, OptJumps),
-    globals.lookup_bool_option(Globals, optimize_fulljumps, OptFullJumps),
-    globals.lookup_bool_option(Globals, optimize_labels, OptLabels),
-    globals.lookup_bool_option(Globals, optimize_peep, OptPeep),
-    globals.lookup_bool_option(Globals, optimize_peep_mkword, OptPeepMkword),
-    globals.lookup_bool_option(Globals, optimize_reassign, OptReassign),
-    globals.lookup_bool_option(Globals, pessimize_tailcalls,
-        PessimizeTailCalls),
-    globals.lookup_bool_option(Globals, standardize_labels, StdLabels),
-    globals.lookup_bool_option(Globals, use_local_vars, UseLocalVars),
+    CheckedNondetTailCalls = OptTuple ^ ot_opt_checked_nondet_tailcalls,
+    OptDelaySlots = OptTuple ^ ot_opt_delay_slot,
+    OptDups = OptTuple ^ ot_opt_dups,
+    OptFrames = OptTuple ^ ot_opt_frames,
+    OptJumps = OptTuple ^ ot_opt_jumps,
+    OptFullJumps = OptTuple ^ ot_opt_fulljumps,
+    OptLabels = OptTuple ^ ot_opt_labels,
+    OptPeep = OptTuple ^ ot_opt_peep,
+    OptPeepMkword = OptTuple ^ ot_opt_peep_mkword,
+    OptReassign = OptTuple ^ ot_opt_reassign,
+    PessimizeTailCalls = OptTuple ^ ot_pessimize_tailcalls,
+    StdLabels = OptTuple ^ ot_standardize_labels,
+    UseLocalVars = OptTuple ^ ot_use_local_vars,
 
     Info = llds_opt_info(DebugOptPredIdStrs, DebugOptPredNames,
         NumRealRRegs, LocalVarAccessThreshold, OptRepeat, GCMethod,
diff --git a/compiler/options.m b/compiler/options.m
index b7dc9d113..647700c45 100644
--- a/compiler/options.m
+++ b/compiler/options.m
@@ -63,7 +63,10 @@
 :- module libs.options.
 :- interface.
 
+:- import_module libs.optimization_options.
+
 :- import_module char.
+:- import_module cord.
 :- import_module getopt_io.
 :- import_module io.
 :- import_module list.
@@ -88,8 +91,9 @@
     % The MaybeOptionTableOut may either be ok(OptionTableOut), or it may
     % be error(ErrorString).
     %
-:- pred special_handler(option::in, special_data::in, option_table::in,
-    maybe_option_table::out) is semidet.
+:- pred special_handler(option::in, special_data::in,
+    option_table::in, maybe_option_table::out,
+    cord(optimization_option)::in, cord(optimization_option)::out) is semidet.
 
     % Return the style and non-style warning options.
     %
@@ -147,6 +151,19 @@
     % The documentation can go either next to the option definition
     % here, or as commented-out code in the appropriate subroutine
     % of options_help/2.
+    %
+    % NOTE: We should ensure that the order of the options
+    % is consistent in all the following places:
+    %
+    % - in this definition of the `option' type
+    % - in the clauses of the `optdef' predicate
+    % - in the clauses of `long_option' predicate
+    %   (not in the clauses of the `short_option' predicate, since
+    %   that is sorted alphabetically on the option letter)
+    % - in the special_handler predicate, if they appear there
+    % - in the help predicates.
+    %
+    % At the moment, there are considerable deviations from such consistency.
 :- type option
 
     % Warning options
@@ -370,6 +387,9 @@
     ;       compute_goal_modes
     ;       benchmark_modes
     ;       benchmark_modes_repeat
+    ;       unneeded_code_debug
+    ;       unneeded_code_debug_pred_name
+    ;       common_struct_preds
 
     % Language semantics options
     ;       reorder_conj
@@ -699,12 +719,155 @@
     ;       prefer_switch
     ;       prefer_while_loop_over_jump_self
     ;       prefer_while_loop_over_jump_mutual
-    ;       opt_no_return_calls
+    ;       opt_no_return_calls                     % XXX should be oc_opt
     ;       debug_class_init
 
     % Optimization Options
+
+    %   - HLDS
+    ;       optopt_allow_inlining
+    ;       inlining
+    ;       optopt_inline_simple
+    ;       optopt_inline_builtins
+    ;       optopt_inline_single_use
+    ;       optopt_inline_call_cost
+    ;       optopt_inline_compound_threshold
+    ;       optopt_inline_simple_threshold
+    ;       optopt_inline_vars_threshold
+    ;       optopt_intermod_inline_simple_threshold
+    ;       optopt_inline_linear_tail_rec_sccs
+    ;       optopt_inline_linear_tail_rec_sccs_max_extra
+    ;       optopt_from_ground_term_threshold
+    ;       optopt_enable_const_struct
+    ;       optopt_common_struct
+    ;       optopt_constraint_propagation
+    ;       optopt_local_constraint_propagation
+    ;       optopt_optimize_unused_args
+    ;       optopt_intermod_unused_args
+    ;       optopt_optimize_higher_order
+    ;       optopt_higher_order_size_limit
+    ;       optopt_higher_order_arg_limit
+    ;       optopt_unneeded_code
+    ;       optopt_unneeded_code_copy_limit
+    ;       optopt_type_specialization
+    ;       optopt_user_guided_type_specialization
+    ;       optopt_introduce_accumulators
+    ;       optopt_optimize_constructor_last_call_accumulator
+    ;       optopt_optimize_constructor_last_call_null
+    ;       optopt_optimize_constructor_last_call
+    ;       optopt_optimize_duplicate_calls
+    ;       optopt_constant_propagation
+    ;       optopt_excess_assign
+    ;       optopt_test_after_switch
+    ;       optopt_optimize_format_calls
+    ;       optopt_optimize_saved_vars_const
+    ;       optopt_optimize_saved_vars_cell
+    ;       optopt_optimize_saved_vars_cell_loop
+    ;       optopt_optimize_saved_vars_cell_full_path
+    ;       optopt_optimize_saved_vars_cell_on_stack
+    ;       optopt_optimize_saved_vars_cell_candidate_headvars
+    ;       optopt_optimize_saved_vars_cell_cv_store_cost
+    ;       optopt_optimize_saved_vars_cell_cv_load_cost
+    ;       optopt_optimize_saved_vars_cell_fv_store_cost
+    ;       optopt_optimize_saved_vars_cell_fv_load_cost
+    ;       optopt_optimize_saved_vars_cell_op_ratio
+    ;       optopt_optimize_saved_vars_cell_node_ratio
+    ;       optopt_optimize_saved_vars_cell_all_path_node_ratio
+    ;       optopt_optimize_saved_vars_cell_include_all_candidates
+    ;       optimize_saved_vars
+    ;       optopt_loop_invariants
+    ;       optopt_delay_construct
+    ;       optopt_follow_code
+    ;       optopt_optimize_dead_procs
+    ;       optopt_deforestation
+    ;       optopt_deforestation_depth_limit
+    ;       optopt_deforestation_cost_factor
+    ;       optopt_deforestation_vars_threshold
+    ;       optopt_deforestation_size_threshold
+    ;       optopt_untuple
+    ;       optopt_tuple
+    ;       optopt_tuple_trace_counts_file
+    ;       optopt_tuple_costs_ratio
+    ;       optopt_tuple_min_args
+    ;       optopt_inline_par_builtins
+    ;       optopt_always_specialize_in_dep_par_conjs
+    ;       optopt_allow_some_paths_only_waits
+    ;       optopt_region_analysis
+
+    %   - HLDS->LLDS
+    ;       optopt_smart_indexing
+    ;         optopt_dense_switch_req_density
+    ;         optopt_lookup_switch_req_density
+    ;         optopt_dense_switch_size
+    ;         optopt_lookup_switch_size
+    ;         optopt_string_trie_switch_size
+    ;         optopt_string_hash_switch_size
+    ;         optopt_string_binary_switch_size
+    ;         optopt_tag_switch_size
+    ;         optopt_try_switch_size
+    ;         optopt_binary_switch_size
+    ;         optopt_switch_single_rec_base_first
+    ;         optopt_switch_multi_rec_base_first
+
+    ;         optopt_smart_atomic_indexing
+    ;         optopt_smart_string_indexing
+    ;         optopt_smart_tag_indexing
+    ;         optopt_smart_float_indexing
+
+    ;       optopt_static_ground_cells
+    ;       optopt_static_ground_floats
+    ;       optopt_static_ground_int64s
+    ;       optopt_static_code_addresses
+
+    ;       optopt_use_atomic_cells
+    ;       optopt_middle_rec
+    ;       optopt_simple_neg
+    ;       optopt_allow_hijacks
+
+    %   - MLDS
+    ;       optopt_optimize_mlds_tailcalls
+    ;       optopt_optimize_initializations
+    ;       optopt_eliminate_unused_mlds_assigns
+    ;       optopt_eliminate_local_vars
+    ;       optopt_generate_trail_ops_inline
+
+    %   - LLDS
+    ;       optopt_common_data
+    ;       optopt_common_layout_data
+    ;       optopt_optimize            % Also used for MLDS->MLDS optimizations.
+    ;       optopt_optimize_peep
+    ;       optopt_optimize_peep_mkword
+    ;       optopt_optimize_jumps
+    ;       optopt_optimize_fulljumps
+    ;       optopt_pessimize_tailcalls
+    ;       optopt_checked_nondet_tailcalls
+    ;       optopt_use_local_vars
+    ;       optopt_local_var_access_threshold
+    ;       optopt_standardize_labels
+    ;       optopt_optimize_labels
+    ;       optopt_optimize_dups
+    ;       optopt_optimize_proc_dups
+    ;       optopt_optimize_frames
+    ;       optopt_optimize_delay_slot
+    ;       optopt_optimize_reassign
+    ;       optopt_optimize_repeat
+    ;       optopt_layout_compression_limit
+
+    %   - C
+    ;       optopt_use_macro_for_redo_fail
+    ;       optopt_emit_c_loops
+    ;       optopt_procs_per_c_function
+    ;       everything_in_one_c_function
+    ;       optopt_local_thread_engine_base
+    ;       optopt_inline_alloc
+    ;       optopt_c_optimize
+
+    %   - Erlang
+    ;       optopt_erlang_switch_on_strings_as_atoms
+
+    % Special optimization options.
+
     ;       opt_level
-    ;       opt_level_number
     ;       opt_space                   % Default is to optimize time.
     ;       intermodule_optimization
     ;       read_opt_files_transitively
@@ -715,82 +878,10 @@
     ;       analysis_repeat
     ;       analysis_file_cache
 
-    %   - HLDS
-    ;       allow_inlining
-    ;       inlining
-    ;       inline_simple
-    ;       inline_builtins
-    ;       inline_single_use
-    ;       inline_call_cost
-    ;       inline_compound_threshold
-    ;       inline_simple_threshold
-    ;       inline_vars_threshold
-    ;       intermod_inline_simple_threshold
-    ;       inline_linear_tail_rec_sccs
-    ;       inline_linear_tail_rec_sccs_max_extra
-    ;       from_ground_term_threshold
-    ;       enable_const_struct
-    ;       common_struct
-    ;       common_struct_preds
-    ;       constraint_propagation
-    ;       local_constraint_propagation
-    ;       optimize_unused_args
-    ;       intermod_unused_args
-    ;       optimize_higher_order
-    ;       higher_order_size_limit
-    ;       higher_order_arg_limit
-    ;       unneeded_code
-    ;       unneeded_code_copy_limit
-    ;       unneeded_code_debug
-    ;       unneeded_code_debug_pred_name
-    ;       type_specialization
-    ;       user_guided_type_specialization
-    ;       introduce_accumulators
-    ;       optimize_constructor_last_call_accumulator
-    ;       optimize_constructor_last_call_null
-    ;       optimize_constructor_last_call
-    ;       optimize_duplicate_calls
-    ;       constant_propagation
-    ;       excess_assign
-    ;       test_after_switch
-    ;       optimize_format_calls
-    ;       optimize_saved_vars_const
-    ;       optimize_saved_vars_cell
-    ;       optimize_saved_vars_cell_loop
-    ;       optimize_saved_vars_cell_full_path
-    ;       optimize_saved_vars_cell_on_stack
-    ;       optimize_saved_vars_cell_candidate_headvars
-    ;       optimize_saved_vars_cell_cv_store_cost
-    ;       optimize_saved_vars_cell_cv_load_cost
-    ;       optimize_saved_vars_cell_fv_store_cost
-    ;       optimize_saved_vars_cell_fv_load_cost
-    ;       optimize_saved_vars_cell_op_ratio
-    ;       optimize_saved_vars_cell_node_ratio
-    ;       optimize_saved_vars_cell_all_path_node_ratio
-    ;       optimize_saved_vars_cell_include_all_candidates
-    ;       optimize_saved_vars
-    ;       loop_invariants
-    ;       delay_construct
-    ;       follow_code
-    ;       optimize_dead_procs
-    ;       deforestation
-    ;       deforestation_depth_limit
-    ;       deforestation_cost_factor
-    ;       deforestation_vars_threshold
-    ;       deforestation_size_threshold
     ;       analyse_trail_usage
     ;       optimize_trail_usage
     ;       optimize_region_ops
     ;       analyse_mm_tabling
-    ;       untuple
-    ;       tuple
-    ;       tuple_trace_counts_file
-    ;       tuple_costs_ratio
-    ;       tuple_min_args
-    ;       inline_par_builtins
-    ;       always_specialize_in_dep_par_conjs
-    ;       allow_some_paths_only_waits
-    ;       region_analysis
 
     % Stuff for the CTGC system (structure sharing / structure reuse).
     ;       structure_sharing_analysis
@@ -823,75 +914,6 @@
     ;       analyse_exceptions
     ;       analyse_closures
 
-    %   - HLDS->LLDS
-    ;       smart_indexing
-    ;         dense_switch_req_density
-    ;         lookup_switch_req_density
-    ;         dense_switch_size
-    ;         lookup_switch_size
-    ;         string_trie_switch_size
-    ;         string_hash_switch_size
-    ;         string_binary_switch_size
-    ;         tag_switch_size
-    ;         try_switch_size
-    ;         binary_switch_size
-    ;         switch_single_rec_base_first
-    ;         switch_multi_rec_base_first
-
-    ;         smart_atomic_indexing
-    ;         smart_string_indexing
-    ;         smart_tag_indexing
-    ;         smart_float_indexing
-
-    ;       static_ground_cells
-    ;       static_ground_floats
-    ;       static_ground_int64s
-    ;       static_code_addresses
-
-    ;       use_atomic_cells
-    ;       middle_rec
-    ;       simple_neg
-    ;       allow_hijacks
-
-    %   - MLDS
-    ;       optimize_tailcalls
-    ;       optimize_initializations
-    ;       eliminate_unused_mlds_assigns
-    ;       eliminate_local_vars
-    ;       generate_trail_ops_inline
-
-    %   - LLDS
-    ;       common_data
-    ;       common_layout_data
-    ;       optimize            % Also used for MLDS->MLDS optimizations.
-    ;       optimize_peep
-    ;       optimize_peep_mkword
-    ;       optimize_jumps
-    ;       optimize_fulljumps
-    ;       pessimize_tailcalls
-    ;       checked_nondet_tailcalls
-    ;       use_local_vars
-    ;       local_var_access_threshold
-    ;       standardize_labels
-    ;       optimize_labels
-    ;       optimize_dups
-    ;       optimize_proc_dups
-    ;       optimize_frames
-    ;       optimize_delay_slot
-    ;       optimize_reassign
-    ;       optimize_repeat
-    ;       layout_compression_limit
-
-    %   - C
-    ;       use_macro_for_redo_fail
-    ;       emit_c_loops
-    ;       procs_per_c_function
-    ;       everything_in_one_c_function
-    ;       local_thread_engine_base
-
-    %   - Erlang
-    ;       erlang_switch_on_strings_as_atoms
-
     % Target code compilation options
     ;       target_debug
 
@@ -900,9 +922,7 @@
     ;       cflags
     ;       quoted_cflag
     ;       c_include_directory
-    ;       c_optimize
     ;       ansi_c
-    ;       inline_alloc
 
     % Flags for specific C compilers.
     ;       gcc_flags
@@ -1141,11 +1161,9 @@
 :- import_module assoc_list.
 :- import_module bool.
 :- import_module dir.
-:- import_module int.
 :- import_module map.
 :- import_module maybe.
 :- import_module pair.
-:- import_module require.
 :- import_module solutions.
 :- import_module string.
 
@@ -1411,6 +1429,9 @@ optdef(oc_aux_output, prop_mode_constraints,            bool(no)).
 optdef(oc_aux_output, compute_goal_modes,               bool(no)).
 optdef(oc_aux_output, benchmark_modes,                  bool(no)).
 optdef(oc_aux_output, benchmark_modes_repeat,           int(1)).
+optdef(oc_aux_output, unneeded_code_debug,              bool(no)).
+optdef(oc_aux_output, unneeded_code_debug_pred_name,    accumulating([])).
+optdef(oc_aux_output, common_struct_preds,              string("")).
 
     % Language semantics options.
 
@@ -1627,7 +1648,6 @@ optdef(oc_codegen, debug_class_init,                    bool(no)).
     % These ones are not affected by `-O<n>'.
 
 optdef(oc_spec_opt, opt_level,                          int_special).
-optdef(oc_spec_opt, opt_level_number,                   int(-2)).
 optdef(oc_spec_opt, opt_space,                          special).
 optdef(oc_spec_opt, intermodule_optimization,           bool(no)).
 optdef(oc_spec_opt, read_opt_files_transitively,        bool(yes)).
@@ -1675,147 +1695,146 @@ optdef(oc_spec_opt, analyse_mm_tabling,                 bool(no)).
     % in the opt_level table.
 
     % HLDS
-optdef(oc_opt, allow_inlining,                          bool(yes)).
+optdef(oc_opt, optopt_allow_inlining,                   bool_special).
 optdef(oc_opt, inlining,                                bool_special).
-optdef(oc_opt, inline_simple,                           bool(no)).
-optdef(oc_opt, inline_builtins,                         bool(yes)).
-optdef(oc_opt, inline_single_use,                       bool(no)).
-optdef(oc_opt, inline_call_cost,                        int(0)).
-optdef(oc_opt, inline_compound_threshold,               int(0)).
-optdef(oc_opt, inline_simple_threshold,                 int(5)).
+optdef(oc_opt, optopt_inline_simple,                    bool_special).
+optdef(oc_opt, optopt_inline_builtins,                  bool_special).
+optdef(oc_opt, optopt_inline_single_use,                bool_special).
+optdef(oc_opt, optopt_inline_call_cost,                 int_special).
+optdef(oc_opt, optopt_inline_compound_threshold,        int_special).
+optdef(oc_opt, optopt_inline_simple_threshold,          int_special).
     % Has no effect until --inline-simple is enabled.
-optdef(oc_opt, inline_vars_threshold,                   int(100)).
-optdef(oc_opt, intermod_inline_simple_threshold,        int(5)).
+optdef(oc_opt, optopt_inline_vars_threshold,            int_special).
+optdef(oc_opt, optopt_intermod_inline_simple_threshold, int_special).
     % Has no effect until --intermodule-optimization.
-optdef(oc_opt, inline_linear_tail_rec_sccs,             bool(no)).
-optdef(oc_opt, inline_linear_tail_rec_sccs_max_extra,   int(0)).
-optdef(oc_opt, from_ground_term_threshold,              int(5)).
-optdef(oc_opt, enable_const_struct,                     bool(yes)).
-optdef(oc_opt, common_struct,                           bool(no)).
-optdef(oc_opt, common_struct_preds,                     string("")).
-optdef(oc_opt, constraint_propagation,                  bool(no)).
-optdef(oc_opt, local_constraint_propagation,            bool(no)).
-optdef(oc_opt, optimize_duplicate_calls,                bool(no)).
-optdef(oc_opt, constant_propagation,                    bool(no)).
-optdef(oc_opt, excess_assign,                           bool(no)).
-optdef(oc_opt, test_after_switch,                       bool(no)).
-optdef(oc_opt, optimize_format_calls,                   bool(yes)).
-optdef(oc_opt, loop_invariants,                         bool(no)).
-optdef(oc_opt, optimize_saved_vars_const,               bool(no)).
-optdef(oc_opt, optimize_saved_vars_cell,                bool(no)).
-optdef(oc_opt, optimize_saved_vars_cell_loop,           bool(yes)).
-optdef(oc_opt, optimize_saved_vars_cell_full_path,      bool(yes)).
-optdef(oc_opt, optimize_saved_vars_cell_on_stack,       bool(yes)).
-optdef(oc_opt, optimize_saved_vars_cell_candidate_headvars, bool(yes)).
-optdef(oc_opt, optimize_saved_vars_cell_cv_store_cost,  int(3)).
-optdef(oc_opt, optimize_saved_vars_cell_cv_load_cost,   int(1)).
-optdef(oc_opt, optimize_saved_vars_cell_fv_store_cost,  int(1)).
-optdef(oc_opt, optimize_saved_vars_cell_fv_load_cost,   int(1)).
-optdef(oc_opt, optimize_saved_vars_cell_op_ratio,       int(100)).
-optdef(oc_opt, optimize_saved_vars_cell_node_ratio,     int(100)).
-optdef(oc_opt, optimize_saved_vars_cell_all_path_node_ratio, int(100)).
-optdef(oc_opt, optimize_saved_vars_cell_include_all_candidates, bool(yes)).
+optdef(oc_opt, optopt_inline_linear_tail_rec_sccs,      bool_special).
+optdef(oc_opt, optopt_inline_linear_tail_rec_sccs_max_extra, int_special).
+optdef(oc_opt, optopt_from_ground_term_threshold,       int_special).
+optdef(oc_opt, optopt_enable_const_struct,              bool_special).
+optdef(oc_opt, optopt_common_struct,                    bool_special).
+optdef(oc_opt, optopt_constraint_propagation,           bool_special).
+optdef(oc_opt, optopt_local_constraint_propagation,     bool_special).
+optdef(oc_opt, optopt_optimize_duplicate_calls,         bool_special).
+optdef(oc_opt, optopt_constant_propagation,             bool_special).
+optdef(oc_opt, optopt_excess_assign,                    bool_special).
+optdef(oc_opt, optopt_test_after_switch,                bool_special).
+optdef(oc_opt, optopt_optimize_format_calls,            bool_special).
+optdef(oc_opt, optopt_loop_invariants,                  bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_const,        bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell,         bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_loop,    bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_full_path, bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_on_stack, bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_candidate_headvars, bool_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_cv_store_cost, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_cv_load_cost, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_fv_store_cost, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_fv_load_cost, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_op_ratio, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_node_ratio, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_all_path_node_ratio, int_special).
+optdef(oc_opt, optopt_optimize_saved_vars_cell_include_all_candidates, bool_special).
 optdef(oc_opt, optimize_saved_vars,                     bool_special).
-optdef(oc_opt, delay_construct,                         bool(no)).
-optdef(oc_opt, follow_code,                             bool(no)).
-optdef(oc_opt, optimize_unused_args,                    bool(no)).
-optdef(oc_opt, intermod_unused_args,                    bool(no)).
-optdef(oc_opt, optimize_higher_order,                   bool(no)).
-optdef(oc_opt, higher_order_size_limit,                 int(20)).
-optdef(oc_opt, higher_order_arg_limit,                  int(10)).
-optdef(oc_opt, unneeded_code,                           bool(no)).
-optdef(oc_opt, unneeded_code_copy_limit,                int(10)).
-optdef(oc_opt, unneeded_code_debug,                     bool(no)).
-optdef(oc_opt, unneeded_code_debug_pred_name,           accumulating([])).
-optdef(oc_opt, type_specialization,                     bool(no)).
-optdef(oc_opt, user_guided_type_specialization,         bool(no)).
-optdef(oc_opt, introduce_accumulators,                  bool(no)).
-optdef(oc_opt, optimize_constructor_last_call_accumulator, bool(no)).
-optdef(oc_opt, optimize_constructor_last_call_null,     bool(no)).
-optdef(oc_opt, optimize_constructor_last_call,          bool(no)).
-optdef(oc_opt, optimize_dead_procs,                     bool(no)).
-optdef(oc_opt, deforestation,                           bool(no)).
-optdef(oc_opt, deforestation_depth_limit,               int(4)).
-optdef(oc_opt, deforestation_cost_factor,               int(1000)).
-optdef(oc_opt, deforestation_vars_threshold,            int(200)).
-optdef(oc_opt, deforestation_size_threshold,            int(15)).
-optdef(oc_opt, untuple,                                 bool(no)).
-optdef(oc_opt, tuple,                                   bool(no)).
-optdef(oc_opt, tuple_trace_counts_file,                 string("")).
-optdef(oc_opt, tuple_costs_ratio,                       int(100)).
-optdef(oc_opt, tuple_min_args,                          int(4)).
-optdef(oc_opt, inline_par_builtins,                     bool(no)).
-optdef(oc_opt, always_specialize_in_dep_par_conjs,      bool(no)).
-optdef(oc_opt, allow_some_paths_only_waits,             bool(yes)).
-optdef(oc_opt, region_analysis,                         bool(no)).
+optdef(oc_opt, optopt_delay_construct,                  bool_special).
+optdef(oc_opt, optopt_follow_code,                      bool_special).
+optdef(oc_opt, optopt_optimize_unused_args,             bool_special).
+optdef(oc_opt, optopt_intermod_unused_args,             bool_special).
+optdef(oc_opt, optopt_optimize_higher_order,            bool_special).
+optdef(oc_opt, optopt_higher_order_size_limit,          int_special).
+optdef(oc_opt, optopt_higher_order_arg_limit,           int_special).
+optdef(oc_opt, optopt_unneeded_code,                    bool_special).
+optdef(oc_opt, optopt_unneeded_code_copy_limit,         int_special).
+optdef(oc_opt, optopt_type_specialization,              bool_special).
+optdef(oc_opt, optopt_user_guided_type_specialization,  bool_special).
+optdef(oc_opt, optopt_introduce_accumulators,           bool_special).
+optdef(oc_opt, optopt_optimize_constructor_last_call_accumulator, bool_special).
+optdef(oc_opt, optopt_optimize_constructor_last_call_null, bool_special).
+optdef(oc_opt, optopt_optimize_constructor_last_call,   bool_special).
+optdef(oc_opt, optopt_optimize_dead_procs,              bool_special).
+optdef(oc_opt, optopt_deforestation,                    bool_special).
+optdef(oc_opt, optopt_deforestation_depth_limit,        int_special).
+optdef(oc_opt, optopt_deforestation_cost_factor,        int_special).
+optdef(oc_opt, optopt_deforestation_vars_threshold,     int_special).
+optdef(oc_opt, optopt_deforestation_size_threshold,     int_special).
+optdef(oc_opt, optopt_untuple,                          bool_special).
+optdef(oc_opt, optopt_tuple,                            bool_special).
+optdef(oc_opt, optopt_tuple_trace_counts_file,          string_special).
+optdef(oc_opt, optopt_tuple_costs_ratio,                int_special).
+optdef(oc_opt, optopt_tuple_min_args,                   int_special).
+optdef(oc_opt, optopt_inline_par_builtins,              bool_special).
+optdef(oc_opt, optopt_always_specialize_in_dep_par_conjs, bool_special).
+optdef(oc_opt, optopt_allow_some_paths_only_waits,      bool_special).
+optdef(oc_opt, optopt_region_analysis,                  bool_special).
 
     % HLDS -> LLDS
-optdef(oc_opt, smart_indexing,                          bool(no)).
-optdef(oc_opt, smart_atomic_indexing,                   bool(yes)).
-optdef(oc_opt, smart_string_indexing,                   bool(yes)).
-optdef(oc_opt, smart_tag_indexing,                      bool(yes)).
-optdef(oc_opt, smart_float_indexing,                    bool(yes)).
-optdef(oc_opt, dense_switch_req_density,                int(25)).
+optdef(oc_opt, optopt_smart_indexing,                   bool_special).
+optdef(oc_opt, optopt_smart_atomic_indexing,            bool_special).
+optdef(oc_opt, optopt_smart_string_indexing,            bool_special).
+optdef(oc_opt, optopt_smart_tag_indexing,               bool_special).
+optdef(oc_opt, optopt_smart_float_indexing,             bool_special).
+optdef(oc_opt, optopt_dense_switch_req_density,         int_special).
     % Minimum density before using a dense switch.
-optdef(oc_opt, lookup_switch_req_density,               int(25)).
+optdef(oc_opt, optopt_lookup_switch_req_density,        int_special).
     % Minimum density before using a lookup switch.
-optdef(oc_opt, dense_switch_size,                       int(4)).
-optdef(oc_opt, lookup_switch_size,                      int(4)).
-optdef(oc_opt, string_trie_switch_size,                 int(16)).
-optdef(oc_opt, string_hash_switch_size,                 int(8)).
-optdef(oc_opt, string_binary_switch_size,               int(4)).
-optdef(oc_opt, tag_switch_size,                         int(3)).
-optdef(oc_opt, try_switch_size,                         int(3)).
-optdef(oc_opt, binary_switch_size,                      int(4)).
-optdef(oc_opt, switch_single_rec_base_first,            bool(no)).
-optdef(oc_opt, switch_multi_rec_base_first,             bool(yes)).
-optdef(oc_opt, static_ground_cells,                     bool(no)).
-optdef(oc_opt, static_ground_floats,                    bool(no)).
-optdef(oc_opt, static_ground_int64s,                    bool(no)).
-optdef(oc_opt, static_code_addresses,                   bool(no)).
-optdef(oc_opt, use_atomic_cells,                        bool(no)).
-optdef(oc_opt, middle_rec,                              bool(no)).
-optdef(oc_opt, simple_neg,                              bool(no)).
-optdef(oc_opt, allow_hijacks,                           bool(yes)).
+optdef(oc_opt, optopt_dense_switch_size,                int_special).
+optdef(oc_opt, optopt_lookup_switch_size,               int_special).
+optdef(oc_opt, optopt_string_trie_switch_size,          int_special).
+optdef(oc_opt, optopt_string_hash_switch_size,          int_special).
+optdef(oc_opt, optopt_string_binary_switch_size,        int_special).
+optdef(oc_opt, optopt_tag_switch_size,                  int_special).
+optdef(oc_opt, optopt_try_switch_size,                  int_special).
+optdef(oc_opt, optopt_binary_switch_size,               int_special).
+optdef(oc_opt, optopt_switch_single_rec_base_first,     bool_special).
+optdef(oc_opt, optopt_switch_multi_rec_base_first,      bool_special).
+optdef(oc_opt, optopt_static_ground_cells,              bool_special).
+optdef(oc_opt, optopt_static_ground_floats,             bool_special).
+optdef(oc_opt, optopt_static_ground_int64s,             bool_special).
+optdef(oc_opt, optopt_static_code_addresses,            bool_special).
+optdef(oc_opt, optopt_use_atomic_cells,                 bool_special).
+optdef(oc_opt, optopt_middle_rec,                       bool_special).
+optdef(oc_opt, optopt_simple_neg,                       bool_special).
+optdef(oc_opt, optopt_allow_hijacks,                    bool_special).
 
     % MLDS
-optdef(oc_opt, optimize_tailcalls,                      bool(no)).
-optdef(oc_opt, optimize_initializations,                bool(no)).
-optdef(oc_opt, eliminate_unused_mlds_assigns,           bool(yes)).
-optdef(oc_opt, eliminate_local_vars,                    bool(no)).
-optdef(oc_opt, generate_trail_ops_inline,               bool(yes)).
+optdef(oc_opt, optopt_optimize_mlds_tailcalls,          bool_special).
+optdef(oc_opt, optopt_optimize_initializations,         bool_special).
+optdef(oc_opt, optopt_eliminate_unused_mlds_assigns,    bool_special).
+optdef(oc_opt, optopt_eliminate_local_vars,             bool_special).
+optdef(oc_opt, optopt_generate_trail_ops_inline,        bool_special).
 
     % LLDS
-optdef(oc_opt, common_data,                             bool(no)).
-optdef(oc_opt, common_layout_data,                      bool(yes)).
-optdef(oc_opt, optimize,                                bool(no)).
-optdef(oc_opt, optimize_peep,                           bool(no)).
-optdef(oc_opt, optimize_peep_mkword,                    bool(no)).
-optdef(oc_opt, optimize_jumps,                          bool(no)).
-optdef(oc_opt, optimize_fulljumps,                      bool(no)).
-optdef(oc_opt, pessimize_tailcalls,                     bool(no)).
-optdef(oc_opt, checked_nondet_tailcalls,                bool(no)).
-optdef(oc_opt, use_local_vars,                          bool(no)).
-optdef(oc_opt, local_var_access_threshold,              int(2)).
-optdef(oc_opt, standardize_labels,                      bool(no)).
-optdef(oc_opt, optimize_labels,                         bool(no)).
-optdef(oc_opt, optimize_dups,                           bool(no)).
-optdef(oc_opt, optimize_proc_dups,                      bool(no)).
-optdef(oc_opt, optimize_frames,                         bool(no)).
-optdef(oc_opt, optimize_delay_slot,                     bool(no)).
-optdef(oc_opt, optimize_reassign,                       bool(no)).
-optdef(oc_opt, optimize_repeat,                         int(0)).
-optdef(oc_opt, layout_compression_limit,                int(4000)).
+optdef(oc_opt, optopt_common_data,                      bool_special).
+optdef(oc_opt, optopt_common_layout_data,               bool_special).
+optdef(oc_opt, optopt_optimize,                         bool_special).
+optdef(oc_opt, optopt_optimize_peep,                    bool_special).
+optdef(oc_opt, optopt_optimize_peep_mkword,             bool_special).
+optdef(oc_opt, optopt_optimize_jumps,                   bool_special).
+optdef(oc_opt, optopt_optimize_fulljumps,               bool_special).
+optdef(oc_opt, optopt_pessimize_tailcalls,              bool_special).
+optdef(oc_opt, optopt_checked_nondet_tailcalls,         bool_special).
+optdef(oc_opt, optopt_use_local_vars,                   bool_special).
+optdef(oc_opt, optopt_local_var_access_threshold,       int_special).
+optdef(oc_opt, optopt_standardize_labels,               bool_special).
+optdef(oc_opt, optopt_optimize_labels,                  bool_special).
+optdef(oc_opt, optopt_optimize_dups,                    bool_special).
+optdef(oc_opt, optopt_optimize_proc_dups,               bool_special).
+optdef(oc_opt, optopt_optimize_frames,                  bool_special).
+optdef(oc_opt, optopt_optimize_delay_slot,              bool_special).
+optdef(oc_opt, optopt_optimize_reassign,                bool_special).
+optdef(oc_opt, optopt_optimize_repeat,                  int_special).
+optdef(oc_opt, optopt_layout_compression_limit,         int_special).
 
     % LLDS -> C
-optdef(oc_opt, use_macro_for_redo_fail,                 bool(no)).
-optdef(oc_opt, emit_c_loops,                            bool(no)).
-optdef(oc_opt, procs_per_c_function,                    int(1)).
+optdef(oc_opt, optopt_use_macro_for_redo_fail,          bool_special).
+optdef(oc_opt, optopt_emit_c_loops,                     bool_special).
+optdef(oc_opt, optopt_procs_per_c_function,             int_special).
 optdef(oc_opt, everything_in_one_c_function,            special).
-optdef(oc_opt, local_thread_engine_base,                bool(yes)).
+optdef(oc_opt, optopt_local_thread_engine_base,         bool_special).
+optdef(oc_opt, optopt_inline_alloc,                     bool_special).
+optdef(oc_opt, optopt_c_optimize,                       bool_special).
 
     % Erlang
-optdef(oc_opt, erlang_switch_on_strings_as_atoms,       bool(no)).
+optdef(oc_opt, optopt_erlang_switch_on_strings_as_atoms, bool_special).
 
     % Target code compilation options.
 
@@ -1828,9 +1847,7 @@ optdef(oc_target_comp, cc,                              string("gcc")).
 optdef(oc_target_comp, c_include_directory,             accumulating([])).
     % The `mmc' script will override the default with a value
     % determined at configuration time.
-optdef(oc_target_comp, c_optimize,                      bool(no)).
 optdef(oc_target_comp, ansi_c,                          bool(yes)).
-optdef(oc_target_comp, inline_alloc,                    bool(no)).
 optdef(oc_target_comp, cflags,                          accumulating([])).
 optdef(oc_target_comp, quoted_cflag,                    string_special).
 optdef(oc_target_comp, gcc_flags,                       accumulating([])).
@@ -2360,6 +2377,9 @@ long_option("compute-goal-modes",       compute_goal_modes).
 long_option("propagate-mode-constraints",   prop_mode_constraints).
 long_option("benchmark-modes",          benchmark_modes).
 long_option("benchmark-modes-repeat",   benchmark_modes_repeat).
+long_option("unneeded-code-debug",      unneeded_code_debug).
+long_option("unneeded-code-debug-pred-name",  unneeded_code_debug_pred_name).
+long_option("common-struct-preds",      common_struct_preds).
 
 % language semantics options
 long_option("reorder-conj",         reorder_conj).
@@ -2444,7 +2464,6 @@ long_option("deep-profile-tail-recursion",
 long_option("record-term-sizes-as-words", record_term_sizes_as_words).
 long_option("record-term-sizes-as-cells", record_term_sizes_as_cells).
 long_option("experimental-complexity",  experimental_complexity).
-long_option("region-analysis",      region_analysis).
 % (c) miscellaneous optional features
 long_option("gc",                   gc).
 long_option("garbage-collection",   gc).
@@ -2599,101 +2618,208 @@ long_option("analysis-file-cache",  analysis_file_cache).
 
 % HLDS->HLDS optimizations
 long_option("inlining",             inlining).
-long_option("inline-simple",        inline_simple).
-long_option("inline-builtins",      inline_builtins).
-long_option("inline-single-use",    inline_single_use).
-long_option("inline-call-cost",     inline_call_cost).
-long_option("inline-compound-threshold",    inline_compound_threshold).
-long_option("inline-simple-threshold",      inline_simple_threshold).
+long_option("inline-simple",        optopt_inline_simple).
+long_option("inline-builtins",      optopt_inline_builtins).
+long_option("inline-single-use",    optopt_inline_single_use).
+long_option("inline-call-cost",     optopt_inline_call_cost).
+long_option("inline-compound-threshold",    optopt_inline_compound_threshold).
+long_option("inline-simple-threshold",      optopt_inline_simple_threshold).
 long_option("intermod-inline-simple-threshold",
-                                    intermod_inline_simple_threshold).
-long_option("inline-linear-tail-rec-sccs",  inline_linear_tail_rec_sccs).
+                                    optopt_intermod_inline_simple_threshold).
+long_option("inline-linear-tail-rec-sccs",  optopt_inline_linear_tail_rec_sccs).
 long_option("inline-linear-tail-rec-sccs-max-extra",
-                                    inline_linear_tail_rec_sccs_max_extra).
+                                    optopt_inline_linear_tail_rec_sccs_max_extra).
 long_option("from-ground-term-threshold",
-                                    from_ground_term_threshold).
-long_option("inline-vars-threshold",        inline_vars_threshold).
-long_option("const-struct",         enable_const_struct).
-long_option("common-struct",        common_struct).
-long_option("common-struct-preds",  common_struct_preds).
-long_option("excess-assign",        excess_assign).
-long_option("test-after-switch",    test_after_switch).
-long_option("optimize-format-calls",         optimize_format_calls).
-long_option("optimize-duplicate-calls", optimize_duplicate_calls).
-long_option("optimise-duplicate-calls", optimize_duplicate_calls).
-long_option("optimise-constant-propagation", constant_propagation).
-long_option("optimize-constant-propagation", constant_propagation).
+                                    optopt_from_ground_term_threshold).
+long_option("inline-vars-threshold",        optopt_inline_vars_threshold).
+long_option("const-struct",         optopt_enable_const_struct).
+long_option("common-struct",        optopt_common_struct).
+long_option("excess-assign",        optopt_excess_assign).
+long_option("test-after-switch",    optopt_test_after_switch).
+long_option("optimize-format-calls",         optopt_optimize_format_calls).
+long_option("optimize-duplicate-calls", optopt_optimize_duplicate_calls).
+long_option("optimise-duplicate-calls", optopt_optimize_duplicate_calls).
+long_option("optimise-constant-propagation", optopt_constant_propagation).
+long_option("optimize-constant-propagation", optopt_constant_propagation).
 long_option("optimize-saved-vars",  optimize_saved_vars).
 long_option("optimise-saved-vars",  optimize_saved_vars).
-long_option("loop-invariants",      loop_invariants).
-long_option("optimize-saved-vars-const",    optimize_saved_vars_const).
-long_option("optimise-saved-vars-const",    optimize_saved_vars_const).
-long_option("optimize-saved-vars-cell", optimize_saved_vars_cell).
-long_option("optimise-saved-vars-cell", optimize_saved_vars_cell).
-long_option("osv-loop",             optimize_saved_vars_cell_loop).
-long_option("osv-full-path",        optimize_saved_vars_cell_full_path).
-long_option("osv-on-stack",         optimize_saved_vars_cell_on_stack).
+long_option("loop-invariants",      optopt_loop_invariants).
+long_option("optimize-saved-vars-const",    optopt_optimize_saved_vars_const).
+long_option("optimise-saved-vars-const",    optopt_optimize_saved_vars_const).
+long_option("optimize-saved-vars-cell", optopt_optimize_saved_vars_cell).
+long_option("optimise-saved-vars-cell", optopt_optimize_saved_vars_cell).
+long_option("osv-loop",             optopt_optimize_saved_vars_cell_loop).
+long_option("osv-full-path",        optopt_optimize_saved_vars_cell_full_path).
+long_option("osv-on-stack",         optopt_optimize_saved_vars_cell_on_stack).
 long_option("osv-cand-head",
-                            optimize_saved_vars_cell_candidate_headvars).
+                            optopt_optimize_saved_vars_cell_candidate_headvars).
 % The next four options are used by tupling.m as well; changes to them
 % may require changes there as well.
-long_option("osv-cvstore-cost",     optimize_saved_vars_cell_cv_store_cost).
-long_option("osv-cvload-cost",      optimize_saved_vars_cell_cv_load_cost).
-long_option("osv-fvstore-cost",     optimize_saved_vars_cell_fv_store_cost).
-long_option("osv-fvload-cost",      optimize_saved_vars_cell_fv_load_cost).
-long_option("osv-op-ratio",         optimize_saved_vars_cell_op_ratio).
-long_option("osv-node-ratio",       optimize_saved_vars_cell_node_ratio).
+long_option("osv-cvstore-cost",     optopt_optimize_saved_vars_cell_cv_store_cost).
+long_option("osv-cvload-cost",      optopt_optimize_saved_vars_cell_cv_load_cost).
+long_option("osv-fvstore-cost",     optopt_optimize_saved_vars_cell_fv_store_cost).
+long_option("osv-fvload-cost",      optopt_optimize_saved_vars_cell_fv_load_cost).
+long_option("osv-op-ratio",         optopt_optimize_saved_vars_cell_op_ratio).
+long_option("osv-node-ratio",       optopt_optimize_saved_vars_cell_node_ratio).
 long_option("osv-allpath-node-ratio",
-                            optimize_saved_vars_cell_all_path_node_ratio).
+                            optopt_optimize_saved_vars_cell_all_path_node_ratio).
 long_option("osv-all-cand",
-                            optimize_saved_vars_cell_include_all_candidates).
-long_option("delay-construct",      delay_construct).
-long_option("delay-constructs",     delay_construct).
-long_option("follow-code",          follow_code).
-long_option("constraint-propagation",   constraint_propagation).
-long_option("local-constraint-propagation", local_constraint_propagation).
-long_option("optimize-unused-args", optimize_unused_args).
-long_option("optimise-unused-args", optimize_unused_args).
-long_option("intermod-unused-args", intermod_unused_args).
-long_option("optimize-higher-order",    optimize_higher_order).
-long_option("optimise-higher-order",    optimize_higher_order).
-long_option("higher-order-size-limit",  higher_order_size_limit).
-long_option("higher-order-arg-limit",   higher_order_arg_limit).
-long_option("unneeded-code",        unneeded_code).
-long_option("unneeded-code-copy-limit", unneeded_code_copy_limit).
-long_option("unneeded-code-debug",  unneeded_code_debug).
-long_option("unneeded-code-debug-pred-name",  unneeded_code_debug_pred_name).
-long_option("type-specialization",  type_specialization).
-long_option("type-specialisation",  type_specialization).
+                            optopt_optimize_saved_vars_cell_include_all_candidates).
+long_option("delay-construct",      optopt_delay_construct).
+long_option("delay-constructs",     optopt_delay_construct).
+long_option("follow-code",          optopt_follow_code).
+long_option("constraint-propagation",   optopt_constraint_propagation).
+long_option("local-constraint-propagation", optopt_local_constraint_propagation).
+long_option("optimize-unused-args", optopt_optimize_unused_args).
+long_option("optimise-unused-args", optopt_optimize_unused_args).
+long_option("intermod-unused-args", optopt_intermod_unused_args).
+long_option("optimize-higher-order",    optopt_optimize_higher_order).
+long_option("optimise-higher-order",    optopt_optimize_higher_order).
+long_option("higher-order-size-limit",  optopt_higher_order_size_limit).
+long_option("higher-order-arg-limit",   optopt_higher_order_arg_limit).
+long_option("unneeded-code",        optopt_unneeded_code).
+long_option("unneeded-code-copy-limit", optopt_unneeded_code_copy_limit).
+long_option("type-specialization",  optopt_type_specialization).
+long_option("type-specialisation",  optopt_type_specialization).
 long_option("user-guided-type-specialization",
-                    user_guided_type_specialization).
+                    optopt_user_guided_type_specialization).
 long_option("user-guided-type-specialisation",
-                    user_guided_type_specialization).
+                    optopt_user_guided_type_specialization).
 % This option is for use in configure.in to test for some bug-fixes for
 % type-specialization which are needed to compile the library. It's not
 % documented, and should eventually be removed.
 long_option("fixed-user-guided-type-specialization",
-                    user_guided_type_specialization).
-long_option("introduce-accumulators",   introduce_accumulators).
+                    optopt_user_guided_type_specialization).
+long_option("introduce-accumulators",   optopt_introduce_accumulators).
 long_option("optimise-constructor-last-call-accumulator",
-                    optimize_constructor_last_call_accumulator).
+                    optopt_optimize_constructor_last_call_accumulator).
 long_option("optimize-constructor-last-call-accumulator",
-                    optimize_constructor_last_call_accumulator).
+                    optopt_optimize_constructor_last_call_accumulator).
 long_option("optimise-constructor-last-call-null",
-                    optimize_constructor_last_call_null).
+                    optopt_optimize_constructor_last_call_null).
 long_option("optimize-constructor-last-call-null",
-                    optimize_constructor_last_call_null).
+                    optopt_optimize_constructor_last_call_null).
 long_option("optimise-constructor-last-call",
-                    optimize_constructor_last_call).
+                    optopt_optimize_constructor_last_call).
 long_option("optimize-constructor-last-call",
-                    optimize_constructor_last_call).
-long_option("optimize-dead-procs",  optimize_dead_procs).
-long_option("optimise-dead-procs",  optimize_dead_procs).
-long_option("deforestation",        deforestation).
-long_option("deforestation-depth-limit",    deforestation_depth_limit).
-long_option("deforestation-cost-factor",    deforestation_cost_factor).
-long_option("deforestation-vars-threshold", deforestation_vars_threshold).
-long_option("deforestation-size-threshold", deforestation_size_threshold).
+                    optopt_optimize_constructor_last_call).
+long_option("optimize-dead-procs",  optopt_optimize_dead_procs).
+long_option("optimise-dead-procs",  optopt_optimize_dead_procs).
+long_option("deforestation",        optopt_deforestation).
+long_option("deforestation-depth-limit",    optopt_deforestation_depth_limit).
+long_option("deforestation-cost-factor",    optopt_deforestation_cost_factor).
+long_option("deforestation-vars-threshold", optopt_deforestation_vars_threshold).
+long_option("deforestation-size-threshold", optopt_deforestation_size_threshold).
+
+long_option("untuple",              optopt_untuple).
+long_option("tuple",                optopt_tuple).
+long_option("tuple-trace-counts-file",  optopt_tuple_trace_counts_file).
+long_option("tuple-costs-ratio",    optopt_tuple_costs_ratio).
+long_option("tuple-min-args",       optopt_tuple_min_args).
+long_option("inline-par-builtins",  optopt_inline_par_builtins).
+long_option("always-specialize-in-dep-par-conjs",
+                                    optopt_always_specialize_in_dep_par_conjs).
+long_option("allow-some-paths-only-waits",
+                                    optopt_allow_some_paths_only_waits).
+long_option("region-analysis",      optopt_region_analysis).
+
+% HLDS->LLDS optimizations
+long_option("smart-indexing",       optopt_smart_indexing).
+long_option("smart-atomic-indexing", optopt_smart_atomic_indexing).
+long_option("smart-string-indexing", optopt_smart_string_indexing).
+long_option("smart-tag-indexing",    optopt_smart_tag_indexing).
+long_option("smart-float-indexing",  optopt_smart_float_indexing).
+long_option("dense-switch-req-density", optopt_dense_switch_req_density).
+long_option("lookup-switch-req-density",optopt_lookup_switch_req_density).
+long_option("dense-switch-size",    optopt_dense_switch_size).
+long_option("lookup-switch-size",   optopt_lookup_switch_size).
+long_option("string-switch-size",   optopt_string_hash_switch_size).
+long_option("string-trie-size",     optopt_string_trie_switch_size).
+long_option("string-trie-switch-size",      optopt_string_trie_switch_size).
+long_option("string-hash-switch-size",      optopt_string_hash_switch_size).
+long_option("string-binary-switch-size",    optopt_string_binary_switch_size).
+long_option("tag-switch-size",      optopt_tag_switch_size).
+long_option("try-switch-size",      optopt_try_switch_size).
+long_option("binary-switch-size",   optopt_binary_switch_size).
+long_option("switch-single-rec-base-first", optopt_switch_single_rec_base_first).
+long_option("switch-multi-rec-base-first",  optopt_switch_multi_rec_base_first).
+long_option("static-ground-terms",  optopt_static_ground_cells).
+% static_ground_floats should be set only in handle_options.m.
+% long_option("static-ground-floats", static_ground_floats).
+% static_ground_int64s should be set only in handle_options.m.
+% long_option("static-ground-int64s", static_ground_int64s).
+% static_code_addresses should be set only in handle_options.m.
+% long_option("static-code-addresses", static_code_addresses).
+long_option("use-atomic-cells",     optopt_use_atomic_cells).
+long_option("middle-rec",           optopt_middle_rec).
+long_option("simple-neg",           optopt_simple_neg).
+long_option("allow-hijacks",        optopt_allow_hijacks).
+
+% MLDS optimizations
+% Option `optimize' is used for both MLDS and LLDS optimizations, but since
+% you can't use both at the same time it doesn't really matter.
+long_option("mlds-optimize",        optopt_optimize).
+long_option("mlds-optimise",        optopt_optimize).
+long_option("mlds-peephole",        optopt_optimize_peep).
+long_option("optimize-tailcalls",   optopt_optimize_mlds_tailcalls).
+long_option("optimise-tailcalls",   optopt_optimize_mlds_tailcalls).
+long_option("optimize-initializations", optopt_optimize_initializations).
+long_option("optimise-initializations", optopt_optimize_initializations).
+long_option("eliminate-unused-mlds-assigns", optopt_eliminate_unused_mlds_assigns).
+long_option("eliminate-local-vars", optopt_eliminate_local_vars).
+long_option("generate-trail-ops-inline", optopt_generate_trail_ops_inline).
+
+% LLDS optimizations
+long_option("common-data",          optopt_common_data).
+long_option("common-layout-data",   optopt_common_layout_data).
+long_option("llds-optimize",        optopt_optimize).
+long_option("llds-optimise",        optopt_optimize).
+long_option("optimize-peep",        optopt_optimize_peep).
+long_option("optimise-peep",        optopt_optimize_peep).
+long_option("optimize-peep-mkword", optopt_optimize_peep_mkword).
+long_option("optimise-peep-mkword", optopt_optimize_peep_mkword).
+long_option("optimize-jumps",       optopt_optimize_jumps).
+long_option("optimise-jumps",       optopt_optimize_jumps).
+long_option("optimize-fulljumps",   optopt_optimize_fulljumps).
+long_option("optimise-fulljumps",   optopt_optimize_fulljumps).
+long_option("pessimize-tailcalls",  optopt_pessimize_tailcalls).
+long_option("checked-nondet-tailcalls", optopt_checked_nondet_tailcalls).
+long_option("use-local-vars",       optopt_use_local_vars).
+long_option("local-var-access-threshold", optopt_local_var_access_threshold).
+long_option("standardise-labels",   optopt_standardize_labels).
+long_option("standardize-labels",   optopt_standardize_labels).
+long_option("optimize-labels",      optopt_optimize_labels).
+long_option("optimise-labels",      optopt_optimize_labels).
+long_option("optimize-dups",        optopt_optimize_dups).
+long_option("optimise-dups",        optopt_optimize_dups).
+long_option("optimize-proc-dups",   optopt_optimize_proc_dups).
+long_option("optimise-proc-dups",   optopt_optimize_proc_dups).
+%%% long_option("optimize-copyprop",    optimize_copyprop).
+%%% long_option("optimise-copyprop",    optimize_copyprop).
+long_option("optimize-frames",      optopt_optimize_frames).
+long_option("optimise-frames",      optopt_optimize_frames).
+long_option("optimize-delay-slot",  optopt_optimize_delay_slot).
+long_option("optimise-delay-slot",  optopt_optimize_delay_slot).
+long_option("optimize-reassign",    optopt_optimize_reassign).
+long_option("optimise-reassign",    optopt_optimize_reassign).
+long_option("optimize-repeat",      optopt_optimize_repeat).
+long_option("optimise-repeat",      optopt_optimize_repeat).
+long_option("layout-compression-limit",      optopt_layout_compression_limit).
+
+% LLDS->C optimizations
+long_option("use-macro-for-redo-fail",  optopt_use_macro_for_redo_fail).
+long_option("emit-c-loops",         optopt_emit_c_loops).
+long_option("procs-per-c-function", optopt_procs_per_c_function).
+long_option("procs-per-C-function", optopt_procs_per_c_function).
+long_option("everything-in-one-c-function", everything_in_one_c_function).
+long_option("everything-in-one-C-function", everything_in_one_c_function).
+long_option("inline-alloc",         optopt_inline_alloc).
+long_option("local-thread-engine-base", optopt_local_thread_engine_base).
+
+% Erlang
+long_option("erlang-switch-on-strings-as-atoms",
+                optopt_erlang_switch_on_strings_as_atoms).
+
 long_option("enable-termination",   termination).
 long_option("enable-term",          termination).
 long_option("check-termination",    termination_check).
@@ -2738,16 +2864,6 @@ long_option("analyse-trail-usage",  analyse_trail_usage).
 long_option("optimize-trail-usage", optimize_trail_usage).
 long_option("optimize-region-ops",  optimize_region_ops).
 long_option("analyse-mm-tabling",   analyse_mm_tabling).
-long_option("untuple",              untuple).
-long_option("tuple",                tuple).
-long_option("tuple-trace-counts-file",  tuple_trace_counts_file).
-long_option("tuple-costs-ratio",    tuple_costs_ratio).
-long_option("tuple-min-args",       tuple_min_args).
-long_option("inline-par-builtins",  inline_par_builtins).
-long_option("always-specialize-in-dep-par-conjs",
-                                    always_specialize_in_dep_par_conjs).
-long_option("allow-some-paths-only-waits",
-                                    allow_some_paths_only_waits).
 
 % CTGC related options.
 long_option("structure-sharing",    structure_sharing_analysis).
@@ -2762,109 +2878,12 @@ long_option("structure-reuse-max-conditions", structure_reuse_max_conditions).
 long_option("structure-reuse-repeat", structure_reuse_repeat).
 long_option("structure-reuse-free-cells", structure_reuse_free_cells).
 
-% HLDS->LLDS optimizations
-long_option("smart-indexing",       smart_indexing).
-long_option("smart-atomic-indexing", smart_atomic_indexing).
-long_option("smart-string-indexing", smart_string_indexing).
-long_option("smart-tag-indexing",    smart_tag_indexing).
-long_option("smart-float-indexing",  smart_float_indexing).
-long_option("dense-switch-req-density", dense_switch_req_density).
-long_option("lookup-switch-req-density",lookup_switch_req_density).
-long_option("dense-switch-size",    dense_switch_size).
-long_option("lookup-switch-size",   lookup_switch_size).
-long_option("string-switch-size",   string_hash_switch_size).
-long_option("string-trie-size",     string_trie_switch_size).
-long_option("string-trie-switch-size",      string_trie_switch_size).
-long_option("string-hash-switch-size",      string_hash_switch_size).
-long_option("string-binary-switch-size",    string_binary_switch_size).
-long_option("tag-switch-size",      tag_switch_size).
-long_option("try-switch-size",      try_switch_size).
-long_option("binary-switch-size",   binary_switch_size).
-long_option("switch-single-rec-base-first", switch_single_rec_base_first).
-long_option("switch-multi-rec-base-first",  switch_multi_rec_base_first).
-long_option("static-ground-terms",  static_ground_cells).
-% static_ground_floats should be set only in handle_options.m.
-% long_option("static-ground-floats", static_ground_floats).
-% static_ground_int64s should be set only in handle_options.m.
-% long_option("static-ground-int64s", static_ground_int64s).
-% static_code_addresses should be set only in handle_options.m.
-% long_option("static-code-addresses", static_code_addresses).
-long_option("use-atomic-cells",     use_atomic_cells).
-long_option("middle-rec",           middle_rec).
-long_option("simple-neg",           simple_neg).
-long_option("allow-hijacks",        allow_hijacks).
-
-% MLDS optimizations
-% Option `optimize' is used for both MLDS and LLDS optimizations, but since
-% you can't use both at the same time it doesn't really matter.
-long_option("mlds-optimize",        optimize).
-long_option("mlds-optimise",        optimize).
-long_option("mlds-peephole",        optimize_peep).
-long_option("optimize-tailcalls",   optimize_tailcalls).
-long_option("optimise-tailcalls",   optimize_tailcalls).
-long_option("optimize-initializations", optimize_initializations).
-long_option("optimise-initializations", optimize_initializations).
-long_option("eliminate-unused-mlds-assigns", eliminate_unused_mlds_assigns).
-long_option("eliminate-local-vars", eliminate_local_vars).
-long_option("generate-trail-ops-inline", generate_trail_ops_inline).
-
-% LLDS optimizations
-long_option("common-data",          common_data).
-long_option("common-layout-data",   common_layout_data).
-long_option("llds-optimize",        optimize).
-long_option("llds-optimise",        optimize).
-long_option("optimize-peep",        optimize_peep).
-long_option("optimise-peep",        optimize_peep).
-long_option("optimize-peep-mkword", optimize_peep_mkword).
-long_option("optimise-peep-mkword", optimize_peep_mkword).
-long_option("optimize-jumps",       optimize_jumps).
-long_option("optimise-jumps",       optimize_jumps).
-long_option("optimize-fulljumps",   optimize_fulljumps).
-long_option("optimise-fulljumps",   optimize_fulljumps).
-long_option("pessimize-tailcalls",  pessimize_tailcalls).
-long_option("checked-nondet-tailcalls", checked_nondet_tailcalls).
-long_option("use-local-vars",       use_local_vars).
-long_option("local-var-access-threshold", local_var_access_threshold).
-long_option("standardise-labels",   standardize_labels).
-long_option("standardize-labels",   standardize_labels).
-long_option("optimize-labels",      optimize_labels).
-long_option("optimise-labels",      optimize_labels).
-long_option("optimize-dups",        optimize_dups).
-long_option("optimise-dups",        optimize_dups).
-long_option("optimize-proc-dups",   optimize_proc_dups).
-long_option("optimise-proc-dups",   optimize_proc_dups).
-%%% long_option("optimize-copyprop",    optimize_copyprop).
-%%% long_option("optimise-copyprop",    optimize_copyprop).
-long_option("optimize-frames",      optimize_frames).
-long_option("optimise-frames",      optimize_frames).
-long_option("optimize-delay-slot",  optimize_delay_slot).
-long_option("optimise-delay-slot",  optimize_delay_slot).
-long_option("optimize-reassign",    optimize_reassign).
-long_option("optimise-reassign",    optimize_reassign).
-long_option("optimize-repeat",      optimize_repeat).
-long_option("optimise-repeat",      optimize_repeat).
-long_option("layout-compression-limit",      layout_compression_limit).
-
-% LLDS->C optimizations
-long_option("use-macro-for-redo-fail",  use_macro_for_redo_fail).
-long_option("emit-c-loops",         emit_c_loops).
-long_option("procs-per-c-function", procs_per_c_function).
-long_option("procs-per-C-function", procs_per_c_function).
-long_option("everything-in-one-c-function", everything_in_one_c_function).
-long_option("everything-in-one-C-function", everything_in_one_c_function).
-long_option("inline-alloc",         inline_alloc).
-long_option("local-thread-engine-base", local_thread_engine_base).
-
-% Erlang
-long_option("erlang-switch-on-strings-as-atoms",
-                erlang_switch_on_strings_as_atoms).
-
 % Target code compilation options
 long_option("target-debug",         target_debug).
 
 long_option("cc",                   cc).
-long_option("c-optimise",           c_optimize).
-long_option("c-optimize",           c_optimize).
+long_option("c-optimise",           optopt_c_optimize).
+long_option("c-optimize",           optopt_c_optimize).
 % XXX we should consider the relationship between c_debug and target_debug
 % more carefully. Perhaps target_debug could imply C debug if the target is C.
 % However for the moment they are just synonyms.
@@ -3162,7 +3181,7 @@ long_option("par-loop-control-preserve-tail-recursion",
 
 %---------------------------------------------------------------------------%
 
-special_handler(Option, SpecialData, !.OptionTable, Result) :-
+special_handler(Option, SpecialData, !.OptionTable, Result, !OptOptions) :-
     require_switch_arms_semidet [Option]
     (
         (
@@ -3259,20 +3278,24 @@ special_handler(Option, SpecialData, !.OptionTable, Result) :-
         ;
             Option = inlining,
             SpecialData = bool(Inline),
-            map.set(inline_simple, bool(Inline), !OptionTable),
-            map.set(inline_builtins, bool(Inline), !OptionTable),
-            map.set(inline_single_use, bool(Inline), !OptionTable),
             (
                 Inline = yes,
-                map.set(inline_compound_threshold, int(10), !OptionTable)
+                CompoundThreshold = 10
             ;
                 Inline = no,
-                map.set(inline_compound_threshold, int(0), !OptionTable)
-            )
+                CompoundThreshold = 0
+            ),
+            InlineOpts = cord.from_list([
+                oo_inline_simple(Inline),
+                oo_inline_builtins(Inline),
+                oo_inline_single_use(Inline),
+                oo_inline_compound_threshold(CompoundThreshold)
+            ]),
+            !:OptOptions = !.OptOptions ++ InlineOpts
         ;
             Option = everything_in_one_c_function,
             SpecialData = none,
-            map.set(procs_per_c_function, int(0), !OptionTable)
+            cord.snoc(oo_procs_per_c_function(0), !OptOptions)
         ;
             Option = reclaim_heap_on_failure,
             SpecialData = bool(Reclaim),
@@ -3341,28 +3364,14 @@ special_handler(Option, SpecialData, !.OptionTable, Result) :-
                     infer_modes                     -   bool(Infer),
                     infer_det                       -   bool(Infer)
                 ], !OptionTable)
-        ;
-            Option = opt_space,
-            SpecialData = none,
-            opt_space(OptionSettingsList),
-            override_options(OptionSettingsList, !OptionTable)
-        ;
-            Option = opt_level,
-            SpecialData = int(SpecifiedLevel),
-            ( if SpecifiedLevel > 6 then
-                EffectiveLevel = 6
-            else if SpecifiedLevel < -1 then
-                EffectiveLevel = -1
-            else
-                EffectiveLevel = SpecifiedLevel
-            ),
-            map.set(opt_level_number, int(EffectiveLevel), !OptionTable),
-            set_opt_level(EffectiveLevel, !OptionTable)
         ;
             Option = optimize_saved_vars,
             SpecialData = bool(Optimize),
-            map.set(optimize_saved_vars_const, bool(Optimize), !OptionTable),
-            map.set(optimize_saved_vars_cell, bool(Optimize), !OptionTable)
+            SavedVarsOpts = cord.from_list([
+                oo_opt_saved_vars_const(Optimize),
+                oo_opt_svcell(Optimize)
+            ]),
+            !:OptOptions = !.OptOptions ++ SavedVarsOpts
         ;
             Option = mercury_library_directory_special,
             SpecialData = string(Dir),
@@ -3445,6 +3454,510 @@ special_handler(Option, SpecialData, !.OptionTable, Result) :-
                 ], !OptionTable)
         ),
         Result = ok(!.OptionTable)
+    ;
+        (
+            Option = optopt_allow_inlining,
+            SpecialData = bool(Bool),
+            OptOption = oo_allow_inlining(Bool)
+        ;
+            Option = optopt_inline_simple,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_simple(Bool)
+        ;
+            Option = optopt_inline_builtins,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_builtins(Bool)
+        ;
+            Option = optopt_inline_single_use,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_single_use(Bool)
+        ;
+            Option = optopt_inline_linear_tail_rec_sccs,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_linear_tail_rec_sccs(Bool)
+        ;
+            Option = optopt_enable_const_struct,
+            SpecialData = bool(Bool),
+            OptOption = oo_enable_const_struct(Bool)
+        ;
+            Option = optopt_common_struct,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_common_structs(Bool)
+        ;
+            Option = optopt_constraint_propagation,
+            SpecialData = bool(Bool),
+            OptOption = oo_prop_constraints(Bool)
+        ;
+            Option = optopt_local_constraint_propagation,
+            SpecialData = bool(Bool),
+            OptOption = oo_prop_local_constraints(Bool)
+        ;
+            Option = optopt_optimize_duplicate_calls,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_dup_calls(Bool)
+        ;
+            Option = optopt_constant_propagation,
+            SpecialData = bool(Bool),
+            OptOption = oo_prop_constants(Bool)
+        ;
+            Option = optopt_excess_assign,
+            SpecialData = bool(Bool),
+            OptOption = oo_elim_excess_assigns(Bool)
+        ;
+            Option = optopt_test_after_switch,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_test_after_switch(Bool)
+        ;
+            Option = optopt_optimize_format_calls,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_format_calls(Bool)
+        ;
+            Option = optopt_loop_invariants,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_loop_invariants(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_const,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_saved_vars_const(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell_loop,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell_loop(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell_full_path,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell_full_path(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell_on_stack,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell_on_stack(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell_candidate_headvars,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell_candidate_headvars(Bool)
+        ;
+            Option = optopt_optimize_saved_vars_cell_include_all_candidates,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_svcell_all_candidates(Bool)
+        ;
+            Option = optopt_delay_construct,
+            SpecialData = bool(Bool),
+            OptOption = oo_delay_constructs(Bool)
+        ;
+            Option = optopt_follow_code,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_follow_code(Bool)
+        ;
+            Option = optopt_optimize_unused_args,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_unused_args(Bool)
+        ;
+            Option = optopt_intermod_unused_args,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_unused_args_intermod(Bool)
+        ;
+            Option = optopt_optimize_higher_order,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_higher_order(Bool)
+        ;
+            Option = optopt_unneeded_code,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_unneeded_code(Bool)
+        ;
+            Option = optopt_type_specialization,
+            SpecialData = bool(Bool),
+            OptOption = oo_spec_types(Bool)
+        ;
+            Option = optopt_user_guided_type_specialization,
+            SpecialData = bool(Bool),
+            OptOption = oo_spec_types_user_guided(Bool)
+        ;
+            Option = optopt_introduce_accumulators,
+            SpecialData = bool(Bool),
+            OptOption = oo_introduce_accumulators(Bool)
+        ;
+            Option = optopt_optimize_constructor_last_call_accumulator,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_lcmc_accumulator(Bool)
+        ;
+            Option = optopt_optimize_constructor_last_call_null,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_lcmc_null(Bool)
+        ;
+            Option = optopt_optimize_constructor_last_call,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_lcmc(Bool)
+        ;
+            Option = optopt_optimize_dead_procs,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_dead_procs(Bool)
+        ;
+            Option = optopt_deforestation,
+            SpecialData = bool(Bool),
+            OptOption = oo_deforest(Bool)
+        ;
+            Option = optopt_untuple,
+            SpecialData = bool(Bool),
+            OptOption = oo_untuple(Bool)
+        ;
+            Option = optopt_tuple,
+            SpecialData = bool(Bool),
+            OptOption = oo_tuple(Bool)
+        ;
+            Option = optopt_inline_par_builtins,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_par_builtins(Bool)
+        ;
+            Option = optopt_always_specialize_in_dep_par_conjs,
+            SpecialData = bool(Bool),
+            OptOption = oo_spec_in_all_dep_par_conjs(Bool)
+        ;
+            Option = optopt_allow_some_paths_only_waits,
+            SpecialData = bool(Bool),
+            OptOption = oo_allow_some_paths_only_waits(Bool)
+        ;
+            Option = optopt_region_analysis,
+            SpecialData = bool(Bool),
+            OptOption = oo_analyse_regions(Bool)
+        ;
+            Option = optopt_smart_indexing,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_smart_indexing(Bool)
+        ;
+            Option = optopt_smart_atomic_indexing,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_smart_indexing_atomic(Bool)
+        ;
+            Option = optopt_smart_string_indexing,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_smart_indexing_string(Bool)
+        ;
+            Option = optopt_smart_tag_indexing,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_smart_indexing_tag(Bool)
+        ;
+            Option = optopt_smart_float_indexing,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_smart_indexing_float(Bool)
+        ;
+            Option = optopt_switch_single_rec_base_first,
+            SpecialData = bool(Bool),
+            OptOption = oo_put_base_first_single_rec(Bool)
+        ;
+            Option = optopt_switch_multi_rec_base_first,
+            SpecialData = bool(Bool),
+            OptOption = oo_put_base_first_multi_rec(Bool)
+        ;
+            Option = optopt_static_ground_cells,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_static_ground_cells(Bool)
+        ;
+            Option = optopt_static_ground_floats,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_static_ground_floats(Bool)
+        ;
+            Option = optopt_static_ground_int64s,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_static_ground_int64s(Bool)
+        ;
+            Option = optopt_static_code_addresses,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_static_code_addresses(Bool)
+        ;
+            Option = optopt_use_atomic_cells,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_atomic_cells(Bool)
+        ;
+            Option = optopt_middle_rec,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_middle_rec(Bool)
+        ;
+            Option = optopt_simple_neg,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_simple_neg(Bool)
+        ;
+            Option = optopt_allow_hijacks,
+            SpecialData = bool(Bool),
+            OptOption = oo_allow_hijacks(Bool)
+        ;
+            Option = optopt_optimize_mlds_tailcalls,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_mlds_tailcalls(Bool)
+        ;
+            Option = optopt_optimize_initializations,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_initializations(Bool)
+        ;
+            Option = optopt_eliminate_unused_mlds_assigns,
+            SpecialData = bool(Bool),
+            OptOption = oo_elim_unused_mlds_assigns(Bool)
+        ;
+            Option = optopt_eliminate_local_vars,
+            SpecialData = bool(Bool),
+            OptOption = oo_elim_local_vars(Bool)
+        ;
+            Option = optopt_generate_trail_ops_inline,
+            SpecialData = bool(Bool),
+            OptOption = oo_gen_trail_ops_inline(Bool)
+        ;
+            Option = optopt_common_data,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_common_data(Bool)
+        ;
+            Option = optopt_common_layout_data,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_common_layout_data(Bool)
+        ;
+            Option = optopt_optimize,
+            SpecialData = bool(Bool),
+            OptOption = oo_optimize(Bool)
+        ;
+            Option = optopt_optimize_peep,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_peep(Bool)
+        ;
+            Option = optopt_optimize_peep_mkword,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_peep_mkword(Bool)
+        ;
+            Option = optopt_optimize_jumps,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_jumps(Bool)
+        ;
+            Option = optopt_optimize_fulljumps,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_fulljumps(Bool)
+        ;
+            Option = optopt_pessimize_tailcalls,
+            SpecialData = bool(Bool),
+            OptOption = oo_pessimize_tailcalls(Bool)
+        ;
+            Option = optopt_checked_nondet_tailcalls,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_checked_nondet_tailcalls(Bool)
+        ;
+            Option = optopt_use_local_vars,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_local_vars(Bool)
+        ;
+            Option = optopt_standardize_labels,
+            SpecialData = bool(Bool),
+            OptOption = oo_standardize_labels(Bool)
+        ;
+            Option = optopt_optimize_labels,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_labels(Bool)
+        ;
+            Option = optopt_optimize_dups,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_dups(Bool)
+        ;
+            Option = optopt_optimize_proc_dups,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_proc_dups(Bool)
+        ;
+            Option = optopt_optimize_frames,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_frames(Bool)
+        ;
+            Option = optopt_optimize_delay_slot,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_delay_slot(Bool)
+        ;
+            Option = optopt_optimize_reassign,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_reassign(Bool)
+        ;
+            Option = optopt_use_macro_for_redo_fail,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_macro_for_redo_fail(Bool)
+        ;
+            Option = optopt_emit_c_loops,
+            SpecialData = bool(Bool),
+            OptOption = oo_emit_c_loops(Bool)
+        ;
+            Option = optopt_local_thread_engine_base,
+            SpecialData = bool(Bool),
+            OptOption = oo_use_local_thread_engine_base(Bool)
+        ;
+            Option = optopt_erlang_switch_on_strings_as_atoms,
+            SpecialData = bool(Bool),
+            OptOption = oo_switch_on_strings_as_atoms(Bool)
+        ;
+            Option = optopt_inline_alloc,
+            SpecialData = bool(Bool),
+            OptOption = oo_inline_alloc(Bool)
+        ;
+            Option = optopt_c_optimize,
+            SpecialData = bool(Bool),
+            OptOption = oo_opt_c(Bool)
+        ;
+            Option = optopt_inline_call_cost,
+            SpecialData = int(N),
+            OptOption = oo_inline_call_cost(N)
+        ;
+            Option = optopt_inline_compound_threshold,
+            SpecialData = int(N),
+            OptOption = oo_inline_compound_threshold(N)
+        ;
+            Option = optopt_inline_simple_threshold,
+            SpecialData = int(N),
+            OptOption = oo_inline_simple_threshold(N)
+        ;
+            Option = optopt_inline_vars_threshold,
+            SpecialData = int(N),
+            OptOption = oo_inline_vars_threshold(N)
+        ;
+            Option = optopt_intermod_inline_simple_threshold,
+            SpecialData = int(N),
+            OptOption = oo_intermod_inline_simple_threshold(N)
+        ;
+            Option = optopt_inline_linear_tail_rec_sccs_max_extra,
+            SpecialData = int(N),
+            OptOption = oo_inline_linear_tail_rec_sccs_max_extra(N)
+        ;
+            Option = optopt_from_ground_term_threshold,
+            SpecialData = int(N),
+            OptOption = oo_from_ground_term_threshold(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_cv_store_cost,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_cv_store_cost(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_cv_load_cost,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_cv_load_cost(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_fv_store_cost,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_fv_store_cost(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_fv_load_cost,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_fv_load_cost(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_op_ratio,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_op_ratio(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_node_ratio,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_node_ratio(N)
+        ;
+            Option = optopt_optimize_saved_vars_cell_all_path_node_ratio,
+            SpecialData = int(N),
+            OptOption = oo_opt_svcell_all_path_node_ratio(N)
+        ;
+            Option = optopt_higher_order_size_limit,
+            SpecialData = int(N),
+            OptOption = oo_higher_order_size_limit(N)
+        ;
+            Option = optopt_higher_order_arg_limit,
+            SpecialData = int(N),
+            OptOption = oo_higher_order_arg_limit(N)
+        ;
+            Option = optopt_unneeded_code_copy_limit,
+            SpecialData = int(N),
+            OptOption = oo_opt_unneeded_code_copy_limit(N)
+        ;
+            Option = optopt_deforestation_depth_limit,
+            SpecialData = int(N),
+            OptOption = oo_deforestation_depth_limit(N)
+        ;
+            Option = optopt_deforestation_cost_factor,
+            SpecialData = int(N),
+            OptOption = oo_deforestation_cost_factor(N)
+        ;
+            Option = optopt_deforestation_vars_threshold,
+            SpecialData = int(N),
+            OptOption = oo_deforestation_vars_threshold(N)
+        ;
+            Option = optopt_deforestation_size_threshold,
+            SpecialData = int(N),
+            OptOption = oo_deforestation_size_threshold(N)
+        ;
+            Option = optopt_tuple_costs_ratio,
+            SpecialData = int(N),
+            OptOption = oo_tuple_costs_ratio(N)
+        ;
+            Option = optopt_tuple_min_args,
+            SpecialData = int(N),
+            OptOption = oo_tuple_min_args(N)
+        ;
+            Option = optopt_dense_switch_req_density,
+            SpecialData = int(N),
+            OptOption = oo_dense_switch_req_density(N)
+        ;
+            Option = optopt_lookup_switch_req_density,
+            SpecialData = int(N),
+            OptOption = oo_lookup_switch_req_density(N)
+        ;
+            Option = optopt_dense_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_dense_switch_size(N)
+        ;
+            Option = optopt_lookup_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_lookup_switch_size(N)
+        ;
+            Option = optopt_string_trie_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_string_trie_switch_size(N)
+        ;
+            Option = optopt_string_hash_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_string_hash_switch_size(N)
+        ;
+            Option = optopt_string_binary_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_string_binary_switch_size(N)
+        ;
+            Option = optopt_tag_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_tag_switch_size(N)
+        ;
+            Option = optopt_try_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_try_switch_size(N)
+        ;
+            Option = optopt_binary_switch_size,
+            SpecialData = int(N),
+            OptOption = oo_binary_switch_size(N)
+        ;
+            Option = optopt_local_var_access_threshold,
+            SpecialData = int(N),
+            OptOption = oo_local_var_access_threshold(N)
+        ;
+            Option = optopt_optimize_repeat,
+            SpecialData = int(N),
+            OptOption = oo_opt_repeat(N)
+        ;
+            Option = optopt_layout_compression_limit,
+            SpecialData = int(N),
+            OptOption = oo_layout_compression_limit(N)
+        ;
+            Option = optopt_procs_per_c_function,
+            SpecialData = int(N),
+            OptOption = oo_procs_per_c_function(N)
+        ;
+            Option = optopt_tuple_trace_counts_file,
+            SpecialData = string(Str),
+            OptOption = oo_tuple_trace_counts_file(Str)
+        ;
+            Option = opt_level,
+            SpecialData = int(N),
+            OptOption = oo_opt_level(N)
+        ;
+            Option = opt_space,
+            SpecialData = none,
+            OptOption = oo_opt_for_space
+        ),
+        cord.snoc(OptOption, !OptOptions),
+        Result = ok(!.OptionTable)
     ).
 
 style_warning_options = [
@@ -3531,30 +4044,6 @@ append_to_accumulating_option(Option - Value, !OptionTable) :-
     Values = Values0 ++ [Value],
     map.set(Option, accumulating(Values), !OptionTable).
 
-:- pred set_opt_level(int::in, option_table::in, option_table::out) is det.
-
-set_opt_level(N, !OptionTable) :-
-    % First reset all optimizations to their default
-    % (the default should be all optimizations off).
-    option_defaults_2(oc_opt, OptimizationDefaults),
-    override_options(OptimizationDefaults, !OptionTable),
-
-    % Next enable the optimization levels from 0 up to N.
-    enable_opt_levels(0, N, !OptionTable).
-
-:- pred enable_opt_levels(int::in, int::in,
-    option_table::in, option_table::out) is det.
-
-enable_opt_levels(Cur, Max, !OptionTable) :-
-    ( if Cur > Max then
-        true
-    else if opt_level(Cur, !.OptionTable, OptionSettingsList) then
-        override_options(OptionSettingsList, !OptionTable),
-        enable_opt_levels(Cur + 1, Max, !OptionTable)
-    else
-        unexpected($pred, "unknown optimization level")
-    ).
-
 :- pred override_options(list(pair(option, option_data))::in,
     option_table::in, option_table::out) is det.
 
@@ -3570,178 +4059,6 @@ set_all_options_to([Option | Options], Value, !OptionTable) :-
 
 %---------------------------------------------------------------------------%
 
-:- pred opt_space(list(pair(option, option_data))::out) is det.
-
-opt_space([
-    unneeded_code_copy_limit    -   int(1),
-    optimize_dead_procs         -   bool(yes),
-    optimize_labels             -   bool(yes),
-    optimize_dups               -   bool(yes),
-    optimize_proc_dups          -   bool(yes),
-    optimize_fulljumps          -   bool(no),
-    optimize_reassign           -   bool(yes),
-    inline_alloc                -   bool(no),
-    use_macro_for_redo_fail     -   bool(no),
-    loop_invariants             -   bool(no)
-]).
-
-%---------------------------------------------------------------------------%
-
-:- pred opt_level(int::in, option_table::in,
-    list(pair(option, option_data))::out) is semidet.
-
-opt_level(OptLevel, OptionTable, Settings) :-
-    % If the optimization level is not set, we use the default values of
-    % all the optimization options (HLDS->HLDS, HLDS->LLDS, LLDS->LLDS,
-    % LLDS->C, and C->object, and their equivalents for the other backends
-    % and targets.
-    %
-    % However, some optimizations are not controlled by options
-    % and thus can't be disabled.
-
-    (
-        OptLevel = 0,
-        % Optimization level 0: aim to minimize overall compilation time.
-        % XXX I just guessed. We should run lots of experiments.
-        Settings = [
-            common_data                 -   bool(yes),
-            optimize                    -   bool(yes),
-            optimize_repeat             -   int(1),
-            optimize_peep               -   bool(yes),
-            optimize_peep_mkword        -   bool(yes),
-            static_ground_cells         -   bool(yes),
-            smart_indexing              -   bool(yes),
-            optimize_jumps              -   bool(yes),
-            optimize_labels             -   bool(yes),
-            optimize_dead_procs         -   bool(yes),
-            excess_assign               -   bool(yes)   % ???
-        ]
-    ;
-        OptLevel = 1,
-        % Optimization level 1: apply optimizations which are cheap and
-        % have a good payoff while still keeping compilation time small.
-        getopt_io.lookup_bool_option(OptionTable, have_delay_slot, DelaySlot),
-        Settings = [
-            use_local_vars              -   bool(yes),
-            c_optimize                  -   bool(yes),  % XXX we want `gcc -O1'
-            optimize_frames             -   bool(yes),
-            optimize_delay_slot         -   bool(DelaySlot),
-            middle_rec                  -   bool(yes),
-            emit_c_loops                -   bool(yes),
-            optimize_tailcalls          -   bool(yes)
-            % dups?
-        ]
-    ;
-        OptLevel = 2,
-        % Optimization level 2: apply optimizations which have a good payoff
-        % relative to their cost; but include optimizations which are
-        % more costly than with -O1.
-        Settings = [
-            optimize_fulljumps          -   bool(yes),
-            optimize_repeat             -   int(3),
-            optimize_dups               -   bool(yes),
-            follow_code                 -   bool(yes),
-            inline_simple               -   bool(yes),
-            inline_single_use           -   bool(yes),
-            inline_compound_threshold   -   int(10),
-            common_struct               -   bool(yes),
-            user_guided_type_specialization -   bool(yes),
-            % XXX While inst `array' is defined as `ground', we can't optimize
-            % duplicate calls (we might combine calls to `array.init').
-            % optimize_duplicate_calls  -   bool(yes),
-            simple_neg                  -   bool(yes),
-            test_after_switch           -   bool(yes),
-
-            optimize_initializations    -  bool(yes)
-        ]
-    ;
-        OptLevel = 3,
-        % Optimization level 3: apply optimizations which usually have a good
-        % payoff even if they increase compilation time quite a bit.
-        Settings = [
-            optimize_saved_vars_const   - bool(yes),
-            optimize_unused_args        -   bool(yes),
-            optimize_higher_order       -   bool(yes),
-            deforestation               -   bool(yes),
-            local_constraint_propagation -  bool(yes),
-            constant_propagation        -   bool(yes),
-            optimize_reassign           -   bool(yes),
-            % Disabled until a bug in extras/trailed_update/var.m is resolved.
-            % introduce_accumulators    -   bool(yes),
-            optimize_repeat             -   int(4)
-        ]
-    ;
-        OptLevel = 4,
-        % Optimization level 4: apply optimizations which may have some payoff
-        % even if they increase compilation time quite a bit.
-        %
-        % Currently this enables the use of local variables
-        % and increases the inlining thresholds.
-        Settings = [
-            inline_simple_threshold     -   int(8),
-            inline_compound_threshold   -   int(20),
-            higher_order_size_limit     -   int(30)
-        ]
-    ;
-        OptLevel = 5,
-        % Optimization level 5: apply optimizations which may have some
-        % payoff even if they increase compilation time a lot.
-        %
-        % Currently this enables the search for construction unifications that
-        % can be delayed past failing computations, allows more passes of the
-        % low-level optimizations, and increases the inlining thresholds
-        % still further. We also enable eliminate_local_vars only at
-        % this level, because that pass is implemented pretty inefficiently.
-        Settings = [
-            optimize_repeat             -   int(5),
-            delay_construct             -   bool(yes),
-            inline_compound_threshold   -   int(100),
-            higher_order_size_limit     -   int(40),
-            eliminate_local_vars        -   bool(yes),
-            loop_invariants             -   bool(yes)
-        ]
-    ;
-        OptLevel = 6,
-        % Optimization level 6: apply optimizations which may have any payoff
-        % even if they increase compilation time to completely unreasonable
-        % levels.
-        %
-        % Currently this sets `everything_in_one_c_function', which causes
-        % the compiler to put everything in the one C function and treat calls
-        % to predicates in the same module as local. We also enable inlining
-        % of GC_malloc(), redo(), and fail().
-        Settings = [
-            procs_per_c_function        -   int(0),
-            inline_alloc                -   bool(yes),
-            use_macro_for_redo_fail     -   bool(yes)
-        ]
-
-        % The following optimization options are not enabled at any level:
-        %
-        %   checked_nondet_tailcalls:
-        %       This is deliberate, because the transformation
-        %       might make code run slower.
-        %
-        %   constraint_propagation:
-        %       I think this is deliberate, because the transformation
-        %       might make code run slower?
-        %
-        %   unneeded_code:
-        %       Because it can cause slowdowns at high optimization levels;
-        %       cause unknown
-        %   type_specialization:
-        %       XXX why not?
-        %
-        %   introduce_accumulators:
-        %       XXX Disabled until a bug in extras/trailed_update/var.m
-        %       is resolved.
-        %
-        %   optimize_constructor_last_call:
-        %       Not a speedup in general.
-    ).
-
-%---------------------------------------------------------------------------%
-
 :- pred handle_quoted_flag(option::in, string::in,
     option_table::in, option_table::out) is det.
 
@@ -4580,6 +4897,15 @@ options_help_aux_output -->
 %       "\tmode analysis.",
 %       "--compute-goal-modes",
 %       "\tCompute goal modes.",
+% These options are only intended to be used for debugging the compiler.
+%       "--unneeded-code-debug",
+%       "\tPrint progress messages during the unneeded code elimination",
+%       "\tpasses.",
+%       "--unneeded-code-debug-pred-name <predname>",
+%       "\tPrint the definition of <predname> at the start of each pass",
+%       "\tof the unneeded code elimination algorithm.",
+%       "--common-struct-preds <predids>",
+%       "\tLimit common struct optimization to the preds with the given ids.",
     ]).
 
 :- pred options_help_semantics(io::di, io::uo) is det.
@@ -5501,8 +5827,6 @@ options_help_hlds_hlds_optimization -->
         "\ttable.",
         "--no-common-struct",
         "\tDisable optimization of common term structures.",
-%       "--common-struct-preds <predids>",
-%       "\tLimit common struct optimization to the preds with the given ids.",
 
 %       Common goal optimization should not be turned off, since it can
 %       break programs that would otherwise compile properly (e.g.,
@@ -5589,12 +5913,6 @@ options_help_hlds_hlds_optimization -->
         "\tnot needed. A value of zero forbids goal movement and allows",
         "\tonly goal deletion; a value of one prevents any increase in the",
         "\tsize of the code.",
-%       "--unneeded-code-debug",
-%       "\tPrint progress messages during the unneeded code elimination",
-%       "\tpasses.",
-%       "--unneeded-code-debug-pred-name <predname>",
-%       "\tPrint the definition of <predname> at the start of each pass",
-%       "\tof the unneeded code elimination algorithm.",
         "--introduce-accumulators",
         "\tAttempt to introduce accumulating variables into",
         "\tprocedures, so as to make them tail recursive.",
diff --git a/compiler/pd_util.m b/compiler/pd_util.m
index 10c4b2a68..91ae6f788 100644
--- a/compiler/pd_util.m
+++ b/compiler/pd_util.m
@@ -175,6 +175,7 @@
 :- import_module hlds.quantification.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module parse_tree.error_util.
 :- import_module parse_tree.prog_type.
@@ -202,11 +203,11 @@ goal_get_calls(Goal0, CalledPreds) :-
 propagate_constraints(!Goal, !PDInfo) :-
     pd_info_get_module_info(!.PDInfo, ModuleInfo0),
     module_info_get_globals(ModuleInfo0, Globals),
-    globals.lookup_bool_option(Globals, local_constraint_propagation,
-        ConstraintProp),
+    globals.get_opt_tuple(Globals, OptTuple),
+    ConstraintProp = OptTuple ^ ot_prop_local_constraints,
     globals.lookup_bool_option(Globals, debug_pd, DebugPD),
     (
-        ConstraintProp = yes,
+        ConstraintProp = prop_local_constraints,
         Goal0 = !.Goal,
         trace [io(!IO)] (
             pd_debug_message(DebugPD, "%% Propagating constraints\n", [], !IO),
@@ -246,7 +247,7 @@ propagate_constraints(!Goal, !PDInfo) :-
             !:Goal = Goal0
         )
     ;
-        ConstraintProp = no
+        ConstraintProp = do_not_prop_local_constraints
     ).
 
 %-----------------------------------------------------------------------------%
diff --git a/compiler/peephole.m b/compiler/peephole.m
index cb67e190e..ade4d372f 100644
--- a/compiler/peephole.m
+++ b/compiler/peephole.m
@@ -16,17 +16,18 @@
 :- module ll_backend.peephole.
 :- interface.
 
-:- import_module ll_backend.llds.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
+:- import_module ll_backend.llds.
 
 :- import_module bool.
 :- import_module list.
 
     % Peephole optimize a list of instructions.
     %
-:- pred peephole_optimize(gc_method::in, bool::in, list(instruction)::in,
-    list(instruction)::out, bool::out) is det.
+:- pred peephole_optimize(gc_method::in, maybe_opt_peep_mkword::in,
+    list(instruction)::in, list(instruction)::out, bool::out) is det.
 
 :- pred combine_decr_sp(list(instruction)::in, list(instruction)::out) is det.
 
@@ -725,8 +726,8 @@ replace_tagged_ptr_components_in_rval(OldLval, OldPtag, OldBase,
 
     % Given a GC method, return the list of invalid peephole optimizations.
     %
-:- pred invalid_peephole_opts(gc_method::in, bool::in, list(pattern)::out)
-    is det.
+:- pred invalid_peephole_opts(gc_method::in, maybe_opt_peep_mkword::in,
+    list(pattern)::out) is det.
 
 invalid_peephole_opts(GC_Method, OptPeepMkword, InvalidPatterns) :-
     (
@@ -742,10 +743,10 @@ invalid_peephole_opts(GC_Method, OptPeepMkword, InvalidPatterns) :-
         InvalidPatterns0 = []
     ),
     (
-        OptPeepMkword = yes,
+        OptPeepMkword = opt_peep_mkword,
         InvalidPatterns = InvalidPatterns0
     ;
-        OptPeepMkword = no,
+        OptPeepMkword = do_not_opt_peep_mkword,
         InvalidPatterns = [pattern_mkword | InvalidPatterns0]
     ).
 
diff --git a/compiler/polymorphism.m b/compiler/polymorphism.m
index 6315b63f3..0712a3540 100644
--- a/compiler/polymorphism.m
+++ b/compiler/polymorphism.m
@@ -408,7 +408,7 @@
 :- import_module hlds.status.
 :- import_module libs.
 :- import_module libs.globals.
-:- import_module libs.options.
+:- import_module libs.optimization_options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
 :- import_module mdbcomp.goal_path.
@@ -2991,7 +2991,7 @@ construct_typeclass_info(Constraint, BaseVar, BaseConsId, ArgVarsMCAs,
     poly_info_get_const_struct_db(!.Info, ConstStructDb0),
     const_struct_db_get_poly_enabled(ConstStructDb0, ConstStructEnabled),
     ( if
-        ConstStructEnabled = yes,
+        ConstStructEnabled = enable_const_struct,
         all_are_const_struct_args(ArgVarsMCAs, VarConstArgs)
     then
         poly_info_get_num_reuses(!.Info, NumReuses),
@@ -3424,7 +3424,7 @@ polymorphism_construct_type_info(Type, TypeCtor, TypeArgs, TypeCtorIsVarArity,
         poly_info_get_const_struct_db(!.Info, ConstStructDb0),
         const_struct_db_get_poly_enabled(ConstStructDb0, Enabled),
         ( if
-            Enabled = yes,
+            Enabled = enable_const_struct,
             all_are_const_struct_args(ArgTypeInfoVarsMCAs,
                 ArgTypeInfoConstArgs)
         then
@@ -3987,15 +3987,15 @@ post_copy_polymorphism(ExistsCastPredIds, !ModuleInfo) :-
     map.keys(ClassMap, ClassIds0),
 
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, user_guided_type_specialization,
-        TypeSpec),
+    globals.get_opt_tuple(Globals, OptTuple),
+    UserTypeSpec = OptTuple ^ ot_spec_types_user_guided,
     (
-        TypeSpec = no,
+        UserTypeSpec = do_not_spec_types_user_guided,
         % Don't expand classes from other modules.
         list.filter(class_id_is_from_given_module(ModuleName),
             ClassIds0, ClassIds)
     ;
-        TypeSpec = yes,
+        UserTypeSpec = spec_types_user_guided,
         ClassIds = ClassIds0
     ),
     map.apply_to_list(ClassIds, ClassMap, ClassDefns),
diff --git a/compiler/proc_gen.m b/compiler/proc_gen.m
index 350fd3bbb..589ecebdf 100644
--- a/compiler/proc_gen.m
+++ b/compiler/proc_gen.m
@@ -88,6 +88,7 @@
 :- import_module libs.
 :- import_module libs.file_util.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module ll_backend.code_gen.
@@ -709,9 +710,10 @@ generate_category_code(CodeModel, ProcContext, Goal, ResumePoint,
 generate_det_code(ProcContext, Goal, ResumePoint, TraceSlotInfo,
         Code, MaybeTraceCallLabel, ProcFrameSlots, !CI, !.CLD) :-
     get_globals(!.CI, Globals),
-    globals.lookup_bool_option(Globals, middle_rec, MiddleRec),
+    globals.get_opt_tuple(Globals, OptTuple),
+    MiddleRec = OptTuple ^ ot_opt_middle_rec,
     ( if
-        MiddleRec = yes,
+        MiddleRec = opt_middle_rec,
         middle_rec.match_and_generate(Goal, MiddleRecCode, !CI, !.CLD, _CLD)
     then
         Code = MiddleRecCode,
diff --git a/compiler/simplify_goal_call.m b/compiler/simplify_goal_call.m
index 7677149af..242210309 100644
--- a/compiler/simplify_goal_call.m
+++ b/compiler/simplify_goal_call.m
@@ -83,6 +83,7 @@
 :- import_module libs.
 :- import_module libs.globals.
 :- import_module libs.int_emu.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.builtin_modules.
@@ -244,7 +245,7 @@ simplify_goal_generic_call(GoalExpr0, GoalExpr, GoalInfo, GoalInfo,
                 Common0, Common, !Info),
             ( if
                 MaybeAssignsGoalExpr = yes(AssignsGoalExpr),
-                OptDuplicateCalls = yes
+                OptDuplicateCalls = opt_dup_calls
             then
                 GoalExpr = AssignsGoalExpr
             else
@@ -315,7 +316,7 @@ simplify_goal_foreign_proc(GoalExpr0, GoalExpr, !GoalInfo,
                 GoalExpr1, MaybeAssignsGoalExpr, !Common, !Info),
             ( if
                 MaybeAssignsGoalExpr = yes(AssignsGoalExpr),
-                OptDuplicateCalls = yes
+                OptDuplicateCalls = opt_dup_calls
             then
                 GoalExpr = AssignsGoalExpr
             else
@@ -931,7 +932,7 @@ simplify_look_for_duplicate_call(PredId, ProcId, Args, GoalExpr0, GoalInfo0,
             GoalExpr0, MaybeAssignsGoalExpr0, Common0, Common, !Info),
         ( if
             MaybeAssignsGoalExpr0 = yes(_AssignsGoalExpr0),
-            OptDupCalls = yes
+            OptDupCalls = opt_dup_calls
         then
             MaybeAssignsGoalExpr = MaybeAssignsGoalExpr0
         else
diff --git a/compiler/simplify_goal_scope.m b/compiler/simplify_goal_scope.m
index 3f8b1986d..c7f31168a 100644
--- a/compiler/simplify_goal_scope.m
+++ b/compiler/simplify_goal_scope.m
@@ -51,6 +51,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module mdbcomp.
@@ -77,11 +78,12 @@ simplify_goal_scope(GoalExpr0, GoalExpr, GoalInfo0, GoalInfo,
         const_struct_db_get_ground_term_enabled(ConstStructDb0,
             ConstStructEnabled),
         (
-            ConstStructEnabled = no,
+            ConstStructEnabled = do_not_enable_const_struct,
             module_info_get_globals(ModuleInfo0, Globals),
-            globals.lookup_bool_option(Globals, common_struct, CommonStruct),
+            globals.get_opt_tuple(Globals, OptTuple),
+            CommonStruct = OptTuple ^ ot_opt_common_structs,
             (
-                CommonStruct = yes,
+                CommonStruct = opt_common_structs,
                 % Traversing the construction unifications inside the scope
                 % would allow common.m to
                 %
@@ -106,7 +108,7 @@ simplify_goal_scope(GoalExpr0, GoalExpr, GoalInfo0, GoalInfo,
                 GoalInfo = GoalInfo0,
                 Common = Common0
             ;
-                CommonStruct = no,
+                CommonStruct = do_not_opt_common_structs,
                 % Looking inside the scope may allow us to reduce the number of
                 % memory cells we may need to allocate dynamically. This
                 % improvement in the generated code trumps the cost in compile
@@ -122,7 +124,7 @@ simplify_goal_scope(GoalExpr0, GoalExpr, GoalInfo0, GoalInfo,
                 GoalInfo = GoalInfo0
             )
         ;
-            ConstStructEnabled = yes,
+            ConstStructEnabled = enable_const_struct,
             ( if
                 SubGoal0 = hlds_goal(SubGoalExpr, _),
                 SubGoalExpr = conj(plain_conj, Conjuncts),
diff --git a/compiler/simplify_info.m b/compiler/simplify_info.m
index 9e7c3c956..2f4f6d093 100644
--- a/compiler/simplify_info.m
+++ b/compiler/simplify_info.m
@@ -26,6 +26,7 @@
 :- import_module hlds.hlds_rtti.
 :- import_module hlds.vartypes.
 :- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module libs.trace_params.
 :- import_module parse_tree.
 :- import_module parse_tree.error_util.
@@ -215,8 +216,8 @@
     % Succeed if either warn_duplicate_calls or opt_duplicate_calls is set,
     % and return whether opt_duplicate_calls is set.
     %
-:- pred simplify_do_warn_or_opt_duplicate_calls(simplify_info::in, bool::out)
-    is semidet.
+:- pred simplify_do_warn_or_opt_duplicate_calls(simplify_info::in,
+    maybe_opt_dup_calls::out) is semidet.
 
 %---------------------------------------------------------------------------%
 %---------------------------------------------------------------------------%
@@ -647,22 +648,22 @@ simplify_do_after_front_end(Info) :-
     SimplifyTasks ^ do_after_front_end = yes.
 simplify_do_excess_assign(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
-    SimplifyTasks ^ do_excess_assign = yes.
+    SimplifyTasks ^ do_excess_assign = elim_excess_assigns.
 simplify_do_test_after_switch(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
-    SimplifyTasks ^ do_test_after_switch = yes.
+    SimplifyTasks ^ do_test_after_switch = opt_test_after_switch.
 simplify_do_elim_removable_scopes(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
     SimplifyTasks ^ do_elim_removable_scopes = yes.
 simplify_do_opt_duplicate_calls(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
-    SimplifyTasks ^ do_opt_duplicate_calls = yes.
+    SimplifyTasks ^ do_opt_duplicate_calls = opt_dup_calls.
 simplify_do_const_prop(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
-    SimplifyTasks ^ do_constant_prop = yes.
+    SimplifyTasks ^ do_constant_prop = prop_constants.
 simplify_do_common_struct(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
-    SimplifyTasks ^ do_common_struct = yes.
+    SimplifyTasks ^ do_common_struct = opt_common_structs.
 simplify_do_extra_common_struct(Info) :-
     simplify_info_get_simplify_tasks(Info, SimplifyTasks),
     SimplifyTasks ^ do_extra_common_struct = yes.
@@ -681,7 +682,7 @@ simplify_do_warn_or_opt_duplicate_calls(Info, OptDuplicateCalls) :-
     WarnDuplicateCalls = SimplifyTasks ^ do_warn_duplicate_calls,
     OptDuplicateCalls = SimplifyTasks ^ do_opt_duplicate_calls,
     ( WarnDuplicateCalls = yes
-    ; OptDuplicateCalls = yes
+    ; OptDuplicateCalls = opt_dup_calls
     ).
 
 %---------------------------------------------------------------------------%
diff --git a/compiler/simplify_proc.m b/compiler/simplify_proc.m
index e6f694b61..93f6b3205 100644
--- a/compiler/simplify_proc.m
+++ b/compiler/simplify_proc.m
@@ -102,6 +102,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module parse_tree.prog_data_foreign.
 
@@ -334,7 +335,7 @@ simplify_proc_maybe_vary_parameters(ModuleInfo, PredId, ProcInfo,
         % either the compiler runs out of memory or the user runs out of
         % patience. The fact that we would generate better code if the
         % compilation finished is therefore of limited interest.
-        !SimplifyTasks ^ do_common_struct := no
+        !SimplifyTasks ^ do_common_struct := do_not_opt_common_structs
     else
         true
     ),
@@ -353,7 +354,7 @@ simplify_proc_maybe_vary_parameters(ModuleInfo, PredId, ProcInfo,
             ( if list.member(PredIdInt, CommonStructPredIdInts) then
                 true
             else
-                !SimplifyTasks ^ do_common_struct := no
+                !SimplifyTasks ^ do_common_struct := do_not_opt_common_structs
             )
         else
             true
@@ -514,7 +515,7 @@ simplify_top_level_goal(!Goal, NestedContext0, InstMap0, !Info) :-
             )
         then
             !SimplifyTasks ^ do_mark_code_model_changes := no,
-            !SimplifyTasks ^ do_excess_assign := no,
+            !SimplifyTasks ^ do_excess_assign := do_not_elim_excess_assigns,
             simplify_info_set_simplify_tasks(!.SimplifyTasks, !Info),
 
             do_process_top_level_goal(!Goal, NestedContext0, InstMap0, !Info),
@@ -537,8 +538,8 @@ simplify_top_level_goal(!Goal, NestedContext0, InstMap0, !Info) :-
             % would serve no purpose, since we do nothing in pass 1
             % that would generate new occurrences of the situations
             % that these tasks seek to optimize.
-            !SimplifyTasks ^ do_common_struct := no,
-            !SimplifyTasks ^ do_opt_duplicate_calls := no,
+            !SimplifyTasks ^ do_common_struct := do_not_opt_common_structs,
+            !SimplifyTasks ^ do_opt_duplicate_calls := do_not_opt_dup_calls,
             simplify_info_reinit(!.SimplifyTasks, !Info)
         else
             true
diff --git a/compiler/simplify_tasks.m b/compiler/simplify_tasks.m
index 44ec8f689..d525aa862 100644
--- a/compiler/simplify_tasks.m
+++ b/compiler/simplify_tasks.m
@@ -18,6 +18,7 @@
 
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 
 :- import_module bool.
 :- import_module list.
@@ -113,12 +114,12 @@
                 do_warn_obsolete                :: bool,
                 do_mark_code_model_changes      :: bool,
                 do_after_front_end              :: bool,
-                do_excess_assign                :: bool,
-                do_test_after_switch            :: bool,
+                do_excess_assign                :: maybe_elim_excess_assigns,
+                do_test_after_switch            :: maybe_opt_test_after_switch,
                 do_elim_removable_scopes        :: bool,
-                do_opt_duplicate_calls          :: bool,
-                do_constant_prop                :: bool,
-                do_common_struct                :: bool,
+                do_opt_duplicate_calls          :: maybe_opt_dup_calls,
+                do_constant_prop                :: maybe_prop_constants,
+                do_common_struct                :: maybe_opt_common_structs,
                 do_extra_common_struct          :: bool,
                 do_ignore_par_conjunctions      :: bool,
                 do_warn_suspicious_recursion    :: bool,
@@ -157,13 +158,17 @@ simplify_tasks_to_list(SimplifyTasks) = List :-
         ( MarkCodeModelChanges = yes ->
             [simptask_mark_code_model_changes] ; [] ) ++
         ( AfterFrontEnd = yes -> [simptask_after_front_end] ; [] ) ++
-        ( ExcessAssign = yes -> [simptask_excess_assigns] ; [] ) ++
-        ( TestAfterSwitch = yes -> [simptask_test_after_switch] ; [] ) ++
+        ( ExcessAssign = elim_excess_assigns ->
+            [simptask_excess_assigns] ; [] ) ++
+        ( TestAfterSwitch = opt_test_after_switch ->
+            [simptask_test_after_switch] ; [] ) ++
         ( ElimRemovableScopes = yes ->
             [simptask_elim_removable_scopes] ; [] ) ++
-        ( OptDuplicateCalls = yes -> [simptask_opt_duplicate_calls] ; [] ) ++
-        ( ConstantProp = yes -> [simptask_constant_prop] ; [] ) ++
-        ( CommonStruct = yes -> [simptask_common_struct] ; [] ) ++
+        ( OptDuplicateCalls = opt_dup_calls ->
+            [simptask_opt_duplicate_calls] ; [] ) ++
+        ( ConstantProp = prop_constants -> [simptask_constant_prop] ; [] ) ++
+        ( CommonStruct = opt_common_structs ->
+            [simptask_common_struct] ; [] ) ++
         ( ExtraCommonStruct = yes -> [simptask_extra_common_struct] ; [] ) ++
         ( RemoveParConjunctions = yes -> [simptask_ignore_par_conjs] ; [] ) ++
         ( WarnSuspiciousRecursion = yes ->
@@ -180,12 +185,17 @@ list_to_simplify_tasks(List) =
         ( list.member(simptask_warn_obsolete, List) -> yes ; no ),
         ( list.member(simptask_mark_code_model_changes, List) -> yes ; no ),
         ( list.member(simptask_after_front_end, List) -> yes ; no ),
-        ( list.member(simptask_excess_assigns, List) -> yes ; no ),
-        ( list.member(simptask_test_after_switch, List) -> yes ; no ),
+        ( list.member(simptask_excess_assigns, List) ->
+            elim_excess_assigns ; do_not_elim_excess_assigns ),
+        ( list.member(simptask_test_after_switch, List) ->
+            opt_test_after_switch ; do_not_opt_test_after_switch ),
         ( list.member(simptask_elim_removable_scopes, List) -> yes ; no ),
-        ( list.member(simptask_opt_duplicate_calls, List) -> yes ; no ),
-        ( list.member(simptask_constant_prop, List) -> yes ; no ),
-        ( list.member(simptask_common_struct, List) -> yes ; no ),
+        ( list.member(simptask_opt_duplicate_calls, List) ->
+            opt_dup_calls ; do_not_opt_dup_calls ),
+        ( list.member(simptask_constant_prop, List) ->
+            prop_constants ; do_not_prop_constants ),
+        ( list.member(simptask_common_struct, List) ->
+            opt_common_structs ; do_not_opt_common_structs ),
         ( list.member(simptask_extra_common_struct, List) -> yes ; no ),
         ( list.member(simptask_ignore_par_conjs, List) -> yes ; no ),
         ( list.member(simptask_warn_suspicious_recursion, List) -> yes ; no ),
@@ -201,14 +211,14 @@ find_simplify_tasks(WarnThisPass, Globals, SimplifyTasks) :-
         WarnKnownBadFormat),
     globals.lookup_bool_option(Globals, warn_unknown_format_calls,
         WarnUnknownFormat),
-    globals.lookup_bool_option(Globals, optimize_format_calls,
-        OptFormatCalls),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptFormatCalls = OptTuple ^ ot_opt_format_calls,
     ( if
         (
             WarnThisPass = yes,
             ( WarnKnownBadFormat = yes ; WarnUnknownFormat = yes  )
         ;
-            OptFormatCalls = yes
+            OptFormatCalls = opt_format_calls
         )
     then
         DoFormatCalls = yes
@@ -216,12 +226,11 @@ find_simplify_tasks(WarnThisPass, Globals, SimplifyTasks) :-
         DoFormatCalls = no
     ),
     globals.lookup_bool_option(Globals, warn_obsolete, WarnObsolete),
-    globals.lookup_bool_option(Globals, excess_assign, ExcessAssign),
-    globals.lookup_bool_option(Globals, test_after_switch, TestAfterSwitch),
-    globals.lookup_bool_option(Globals, common_struct, CommonStruct),
-    globals.lookup_bool_option(Globals, optimize_duplicate_calls,
-        OptDuplicateCalls),
-    globals.lookup_bool_option(Globals, constant_propagation, ConstantProp),
+    ExcessAssign = OptTuple ^ ot_elim_excess_assigns,
+    TestAfterSwitch = OptTuple ^ ot_opt_test_after_switch,
+    CommonStruct = OptTuple ^ ot_opt_common_structs,
+    OptDuplicateCalls = OptTuple ^ ot_opt_dup_calls,
+    ConstantProp = OptTuple ^ ot_prop_constants,
     MarkCodeModelChanges = no,
     AfterFrontEnd = no,
     ElimRemovableScopes = no,
diff --git a/compiler/stack_layout.m b/compiler/stack_layout.m
index e0faa0bdb..210e4b46c 100644
--- a/compiler/stack_layout.m
+++ b/compiler/stack_layout.m
@@ -111,6 +111,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module libs.trace_params.
 :- import_module ll_backend.layout_out.
@@ -787,10 +788,10 @@ construct_proc_traversal(Params, EntryLabel, Detism, NumStackSlots,
     ),
     StaticCodeAddr = Params ^ slp_static_code_addresses,
     (
-        StaticCodeAddr = yes,
+        StaticCodeAddr = use_static_code_addresses,
         MaybeEntryLabel = yes(EntryLabel)
     ;
-        StaticCodeAddr = no,
+        StaticCodeAddr = do_not_use_static_code_addresses,
         MaybeEntryLabel = no
     ),
     Traversal = proc_layout_stack_traversal(MaybeEntryLabel,
@@ -2492,7 +2493,7 @@ represent_determinism_rval(Detism,
                 slp_procid_stack_layout     :: bool,
 
                 % Do we have static code addresses or unboxed floats?
-                slp_static_code_addresses   :: bool,
+                slp_static_code_addresses   :: maybe_use_static_code_addresses,
                 slp_unboxed_floats          :: have_unboxed_floats,
 
                 % Do we have unboxed 64-bit integer types?
@@ -2506,24 +2507,24 @@ represent_determinism_rval(Detism,
 
 init_stack_layout_params(ModuleInfo) = Params :-
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals, common_layout_data, CommonLayoutData),
-    globals.lookup_int_option(Globals, layout_compression_limit,
-        CompressLimit),
+    globals.get_opt_tuple(Globals, OptTuple),
+    CommonLayoutData = OptTuple ^ ot_use_common_layout_data,
+    CompressLimit = OptTuple ^ ot_layout_compression_limit,
     globals.get_trace_level(Globals, TraceLevel),
     globals.get_trace_suppress(Globals, TraceSuppress),
     globals.lookup_bool_option(Globals, profile_deep, DeepProfiling),
     globals.lookup_bool_option(Globals, agc_stack_layout, AgcLayout),
     globals.lookup_bool_option(Globals, trace_stack_layout, TraceLayout),
     globals.lookup_bool_option(Globals, procid_stack_layout, ProcIdLayout),
-    globals.lookup_bool_option(Globals, static_code_addresses, StaticCodeAddr),
+    StaticCodeAddr = OptTuple ^ ot_use_static_code_addresses,
     globals.lookup_bool_option(Globals, unboxed_float, UnboxedFloatOpt),
     globals.lookup_bool_option(Globals, unboxed_int64s, UnboxedInt64sOpt),
     globals.lookup_bool_option(Globals, rtti_line_numbers, RttiLineNumbers),
     (
-        CommonLayoutData = no,
+        CommonLayoutData = do_not_use_common_layout_data,
         CompressArrays = no
     ;
-        CommonLayoutData = yes,
+        CommonLayoutData = use_common_layout_data,
         CompressArrays = yes(CompressLimit)
     ),
     (
diff --git a/compiler/stack_opt.m b/compiler/stack_opt.m
index 8448b7d4e..55668faad 100644
--- a/compiler/stack_opt.m
+++ b/compiler/stack_opt.m
@@ -106,6 +106,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.live_vars.
 :- import_module ll_backend.liveness.
@@ -154,9 +155,9 @@
     --->    stack_opt_params(
                 sop_matching_params     :: matching_params,
                 sop_all_path_node_ratio :: int,
-                sop_fixpoint_loop       :: bool,
-                sop_full_path           :: bool,
-                sop_on_stack            :: bool,
+                sop_fixpoint_loop       :: maybe_opt_svcell_loop,
+                sop_full_path           :: maybe_opt_svcell_full_path,
+                sop_on_stack            :: maybe_opt_svcell_on_stack,
                 sop_non_candidate_vars  :: set_of_progvar
             ).
 
@@ -257,14 +258,14 @@ optimize_live_sets(ModuleInfo, OptAlloc, !ProcInfo, Changed, DebugStackOpt,
         InputArgs, OutputArgs, UnusedArgs),
     HeadVars = set.union_list([InputArgs, OutputArgs, UnusedArgs]),
     module_info_get_globals(ModuleInfo, Globals),
-    globals.lookup_bool_option(Globals,
-        optimize_saved_vars_cell_candidate_headvars, CandHeadvars),
+    globals.get_opt_tuple(Globals, OptTuple),
+    CandHeadvars = OptTuple ^ ot_opt_svcell_candidate_headvars,
     (
-        CandHeadvars = no,
+        CandHeadvars = do_not_opt_svcell_candidate_headvars,
         set_of_var.union(set_to_bitset(HeadVars), ParConjOwnSlot,
             NonCandidateVars)
     ;
-        CandHeadvars = yes,
+        CandHeadvars = opt_svcell_candidate_headvars,
         NonCandidateVars = ParConjOwnSlot
     ),
     Counter0 = counter.init(1),
@@ -275,33 +276,20 @@ optimize_live_sets(ModuleInfo, OptAlloc, !ProcInfo, Changed, DebugStackOpt,
     map.init(StartMap0),
     SuccMap0 = map.singleton(CurIntervalId, []),
     VarsMap0 = map.singleton(CurIntervalId, set_to_bitset(OutputArgs)),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_cv_store_cost, CellVarStoreCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_cv_load_cost, CellVarLoadCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_fv_store_cost, FieldVarStoreCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_fv_load_cost, FieldVarLoadCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_op_ratio, OpRatio),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_node_ratio, NodeRatio),
-    globals.lookup_bool_option(Globals,
-        optimize_saved_vars_cell_include_all_candidates, InclAllCand),
+    CellVarStoreCost = OptTuple ^ ot_opt_svcell_cv_store_cost,
+    CellVarLoadCost = OptTuple ^ ot_opt_svcell_cv_load_cost,
+    FieldVarStoreCost = OptTuple ^ ot_opt_svcell_fv_store_cost,
+    FieldVarLoadCost = OptTuple ^ ot_opt_svcell_fv_load_cost,
+    OpRatio = OptTuple ^ ot_opt_svcell_op_ratio,
+    NodeRatio = OptTuple ^ ot_opt_svcell_node_ratio,
+    InclAllCand = OptTuple ^ ot_opt_svcell_all_candidates,
     MatchingParams = matching_params(CellVarStoreCost, CellVarLoadCost,
         FieldVarStoreCost, FieldVarLoadCost, OpRatio, NodeRatio, InclAllCand),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_all_path_node_ratio,
-        AllPathNodeRatio),
-    globals.lookup_bool_option(Globals,
-        optimize_saved_vars_cell_loop, FixpointLoop),
-    globals.lookup_bool_option(Globals,
-        optimize_saved_vars_cell_full_path, FullPath),
-    globals.lookup_bool_option(Globals,
-        optimize_saved_vars_cell_on_stack, OnStack),
-    globals.lookup_bool_option(Globals,
-        opt_no_return_calls, OptNoReturnCalls),
+    AllPathNodeRatio = OptTuple ^ ot_opt_svcell_all_path_node_ratio,
+    FixpointLoop = OptTuple ^ ot_opt_svcell_loop,
+    FullPath = OptTuple ^ ot_opt_svcell_full_path,
+    OnStack = OptTuple ^ ot_opt_svcell_on_stack,
+    globals.lookup_bool_option(Globals, opt_no_return_calls, OptNoReturnCalls),
     IntParams = interval_params(ModuleInfo, VarTypes0, OptNoReturnCalls),
     IntervalInfo0 = interval_info(IntParams,
         set_of_var.init, set_to_bitset(OutputArgs),
@@ -473,14 +461,14 @@ use_cell(CellVar, FieldVarList, ConsId, Goal, !IntervalInfo, !StackOptInfo) :-
                 set_of_var.difference(CandidateArgVars0, RelevantAfterVars,
                     CandidateArgVars),
                 (
-                    OnStack = yes,
+                    OnStack = opt_svcell_on_stack,
                     ( if set_of_var.member(FlushedLater, CellVar) then
                         CellVarFlushedLater = yes
                     else
                         CellVarFlushedLater = no
                     )
                 ;
-                    OnStack = no,
+                    OnStack = do_not_opt_svcell_on_stack,
                     ( if
                         list.member(PathInfo, PathsInfo),
                         PathInfo = match_path_info(_, Segments),
@@ -549,12 +537,12 @@ apply_matching_loop(CellVar, CellVarFlushedLater, IntParams, StackOptParams,
         CandidateArgVars1 = set_of_var.intersect_list(PathViaCellVars),
         FixpointLoop = StackOptParams ^ sop_fixpoint_loop,
         (
-            FixpointLoop = no,
+            FixpointLoop = do_not_opt_svcell_loop,
             BenefitNodeSets = BenefitNodeSets0,
             CostNodeSets = CostNodeSets0,
             ViaCellVars = CandidateArgVars1
         ;
-            FixpointLoop = yes,
+            FixpointLoop = opt_svcell_loop,
             apply_matching_loop(CellVar, CellVarFlushedLater,
                 IntParams, StackOptParams, PathInfos, CandidateArgVars1,
                 BenefitNodeSets, CostNodeSets, ViaCellVars)
@@ -895,7 +883,7 @@ find_all_branches_from(End, RelevantVars, MaybeSearchAnchor0, IntervalInfo,
     StackOptParams = StackOptInfo ^ soi_stack_opt_params,
     FullPath = StackOptParams ^ sop_full_path,
     ( if
-        FullPath = yes,
+        FullPath = opt_svcell_full_path,
         End = anchor_branch_start(branch_disj, EndGoalId)
     then
         MaybeSearchAnchor1 = yes(anchor_branch_end(branch_disj, EndGoalId)),
@@ -907,7 +895,7 @@ find_all_branches_from(End, RelevantVars, MaybeSearchAnchor0, IntervalInfo,
         apply_interval_find_all_branches(RelevantVars, MaybeSearchAnchor0,
             IntervalInfo, StackOptInfo, ContinueId, !AllPaths)
     else if
-        FullPath = yes,
+        FullPath = opt_svcell_full_path,
         End = anchor_branch_start(branch_ite, EndGoalId)
     then
         ( if SuccessorIds = [ElseStartIdPrime, CondStartIdPrime] then
diff --git a/compiler/switch_gen.m b/compiler/switch_gen.m
index b2676c39f..2dd9d1a82 100644
--- a/compiler/switch_gen.m
+++ b/compiler/switch_gen.m
@@ -107,8 +107,8 @@
 :- import_module hlds.hlds_out.
 :- import_module hlds.hlds_out.hlds_out_goal.
 :- import_module libs.
+:- import_module libs.optimization_options.
 :- import_module libs.globals.
-:- import_module libs.options.
 :- import_module ll_backend.code_gen.
 :- import_module ll_backend.dense_switch.
 :- import_module ll_backend.lookup_switch.
@@ -119,7 +119,6 @@
 :- import_module parse_tree.prog_type.
 
 :- import_module assoc_list.
-:- import_module bool.
 :- import_module cord.
 :- import_module int.
 :- import_module maybe.
@@ -170,7 +169,8 @@ generate_switch(CodeModel, SwitchVar, CanFail, Cases, GoalInfo, Code,
         ;
             SwitchCategory = tag_switch,
             num_cons_ids_in_tagged_cases(TaggedCases, NumConsIds, NumArms),
-            globals.lookup_int_option(Globals, tag_switch_size, TagSize),
+            globals.get_opt_tuple(Globals, OptTuple),
+            TagSize = OptTuple ^ ot_tag_switch_size,
             ( if NumConsIds >= TagSize, NumArms > 1 then
                 generate_tag_switch(TaggedCases, SwitchVarRval, SwitchVarType,
                     SwitchVarName, CodeModel, CanFail, GoalInfo, EndLabel,
@@ -210,18 +210,17 @@ generate_atomic_or_int64_switch(ModuleInfo, Globals, CodeModel, CanFail,
         % circumstances, but it would take a pretty complex heuristic
         % to get it right, so, lets just use a simple one - no static
         % ground terms, no lookup switch.
-        globals.lookup_bool_option(Globals, static_ground_cells, yes),
+        globals.get_opt_tuple(Globals, OptTuple),
+        OptTuple ^ ot_use_static_ground_cells = use_static_ground_cells,
 
         % Lookup switches do not generate trace events.
         get_maybe_trace_info(!.CI, MaybeTraceInfo),
         MaybeTraceInfo = no,
 
-        globals.lookup_int_option(Globals, lookup_switch_size,
-            LookupSize),
+        LookupSize = OptTuple ^ ot_lookup_switch_size,
         NumConsIds >= LookupSize,
         NumArms > 1,
-        globals.lookup_int_option(Globals, lookup_switch_req_density,
-            ReqDensity),
+        ReqDensity = OptTuple ^ ot_lookup_switch_req_density,
         filter_out_failing_cases_if_needed(CodeModel,
             TaggedCases, FilteredTaggedCases,
             CanFail, FilteredCanFail),
@@ -244,12 +243,11 @@ generate_atomic_or_int64_switch(ModuleInfo, Globals, CodeModel, CanFail,
     else if
         MaybeIntSwitchInfo = int_switch(IntSwitchInfo),
         IntSwitchInfo = int_switch_info(LowerLimit, UpperLimit, NumValues),
-        globals.lookup_int_option(Globals, dense_switch_size,
-            DenseSize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        DenseSize = OptTuple ^ ot_dense_switch_size,
         NumConsIds >= DenseSize,
         NumArms > 1,
-        globals.lookup_int_option(Globals, dense_switch_req_density,
-            ReqDensity),
+        ReqDensity = OptTuple ^ ot_dense_switch_req_density,
         tagged_case_list_is_dense_switch(!.CI, SwitchVarType,
             TaggedCases, LowerLimit, UpperLimit, NumValues,
             ReqDensity, CanFail, DenseSwitchInfo)
@@ -277,10 +275,9 @@ generate_string_switch(Globals, CodeModel, CanFail,
     num_cons_ids_in_tagged_cases(FilteredTaggedCases,
         NumConsIds, NumArms),
     ( if NumArms > 1 then
-        globals.lookup_int_option(Globals, string_hash_switch_size,
-            StringHashSwitchSize),
-        globals.lookup_int_option(Globals, string_binary_switch_size,
-            StringBinarySwitchSize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        StringHashSwitchSize = OptTuple ^ ot_string_hash_switch_size,
+        StringBinarySwitchSize = OptTuple ^ ot_string_binary_switch_size,
         ( if NumConsIds >= StringHashSwitchSize then
             ( if
                 remember_position(!.CLD, BranchStart),
@@ -486,13 +483,13 @@ order_recursive_cases(Cases0, Cases, CodeModel, CanFail, CI) :-
                 fail
             )
         then
-            globals.lookup_bool_option(Globals, switch_single_rec_base_first,
-                SingleRecBaseFirst),
+            globals.get_opt_tuple(Globals, OptTuple),
+            SingleRecBaseFirst = OptTuple ^ ot_put_base_first_single_rec,
             (
-                SingleRecBaseFirst = yes,
+                SingleRecBaseFirst = put_base_first_single_rec,
                 Cases = [SingleRecCase, BaseCase]
             ;
-                SingleRecBaseFirst = no,
+                SingleRecBaseFirst = do_not_put_base_first_single_rec,
                 Cases = [BaseCase, SingleRecCase]
             )
         else if
@@ -512,13 +509,13 @@ order_recursive_cases(Cases0, Cases, CodeModel, CanFail, CI) :-
                 fail
             )
         then
-            globals.lookup_bool_option(Globals, switch_multi_rec_base_first,
-                MultiRecBaseFirst),
+            globals.get_opt_tuple(Globals, OptTuple),
+            MultiRecBaseFirst = OptTuple ^ ot_put_base_first_multi_rec,
             (
-                MultiRecBaseFirst = yes,
+                MultiRecBaseFirst = put_base_first_multi_rec,
                 Cases = [BaseCase, MultiRecCase]
             ;
-                MultiRecBaseFirst = no,
+                MultiRecBaseFirst = do_not_put_base_first_multi_rec,
                 Cases = [MultiRecCase, BaseCase]
             )
         else
diff --git a/compiler/switch_util.m b/compiler/switch_util.m
index 6f2f07fcb..6c05506cf 100644
--- a/compiler/switch_util.m
+++ b/compiler/switch_util.m
@@ -390,6 +390,7 @@
 :- import_module check_hlds.
 :- import_module check_hlds.type_util.
 :- import_module hlds.hlds_code_util.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 
 :- import_module int.
@@ -851,12 +852,12 @@ find_switch_category(ModuleInfo, SwitchVarType, SwitchCategory,
     SwitchCategory = type_ctor_cat_to_switch_cat(SwitchTypeCtorCat),
 
     module_info_get_globals(ModuleInfo, Globals),
+    globals.get_opt_tuple(Globals, OptTuple),
     ( if
         (
             % We cannot use smart indexing if smart indexing is turned off
             % in general.
-            globals.lookup_bool_option(Globals, smart_indexing, SmartIndexing),
-            SmartIndexing = no
+            OptTuple ^ ot_use_smart_indexing = do_not_use_smart_indexing
         ;
             % We cannot use smart indexing if smart indexing is turned off
             % for this category of switches.
@@ -874,23 +875,39 @@ find_switch_category(ModuleInfo, SwitchVarType, SwitchCategory,
     = bool.
 
 is_smart_indexing_allowed_for_category(Globals, SwitchCategory) = Allowed :-
+    globals.get_opt_tuple(Globals, OptTuple),
     (
         SwitchCategory = atomic_switch,
-        globals.lookup_bool_option(Globals, smart_atomic_indexing, Allowed)
+        Atomic = OptTuple ^ ot_use_smart_indexing_atomic,
+        ( Atomic = use_smart_indexing_atomic, Allowed = yes
+        ; Atomic = do_not_use_smart_indexing_atomic, Allowed = no
+        )
     ;
         SwitchCategory = string_switch,
-        globals.lookup_bool_option(Globals, smart_string_indexing, Allowed)
+        String = OptTuple ^ ot_use_smart_indexing_string,
+        ( String = use_smart_indexing_string, Allowed = yes
+        ; String = do_not_use_smart_indexing_string, Allowed = no
+        )
     ;
         SwitchCategory = tag_switch,
-        globals.lookup_bool_option(Globals, smart_tag_indexing, Allowed)
+        Tag = OptTuple ^ ot_use_smart_indexing_tag,
+        ( Tag = use_smart_indexing_tag, Allowed = yes
+        ; Tag = do_not_use_smart_indexing_tag, Allowed = no
+        )
     ;
         SwitchCategory = float_switch,
-        globals.lookup_bool_option(Globals, smart_float_indexing, Allowed)
+        Float = OptTuple ^ ot_use_smart_indexing_float,
+        ( Float = use_smart_indexing_float, Allowed = yes
+        ; Float = do_not_use_smart_indexing_float, Allowed = no
+        )
     ;
         SwitchCategory = int64_switch,
         % We do not have a separate option for controlling smart indexing
         % of 64-bit integers.
-        globals.lookup_bool_option(Globals, smart_atomic_indexing, Allowed)
+        Gen = OptTuple ^ ot_use_smart_indexing,
+        ( Gen = use_smart_indexing, Allowed = yes
+        ; Gen = do_not_use_smart_indexing, Allowed = no
+        )
     ).
 
 %-----------------------------------------------------------------------------%
diff --git a/compiler/tag_switch.m b/compiler/tag_switch.m
index ce21ad034..bb7809803 100644
--- a/compiler/tag_switch.m
+++ b/compiler/tag_switch.m
@@ -48,6 +48,7 @@
 :- import_module hlds.hlds_llds.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.switch_case.
 
@@ -230,9 +231,10 @@ generate_tag_switch(TaggedCases, VarRval, VarType, VarName, CodeModel, CanFail,
 
     map.count(PtagCaseMap, PtagsUsed),
     get_globals(!.CI, Globals),
-    globals.lookup_int_option(Globals, dense_switch_size, DenseSwitchSize),
-    globals.lookup_int_option(Globals, try_switch_size, TrySwitchSize),
-    globals.lookup_int_option(Globals, binary_switch_size, BinarySwitchSize),
+    globals.get_opt_tuple(Globals, OptTuple),
+    DenseSwitchSize = OptTuple ^ ot_dense_switch_size,
+    TrySwitchSize = OptTuple ^ ot_try_switch_size,
+    BinarySwitchSize = OptTuple ^ ot_binary_switch_size,
     ( if PtagsUsed >= DenseSwitchSize then
         PrimaryMethod = jump_table
     else if PtagsUsed >= BinarySwitchSize then
@@ -731,11 +733,10 @@ generate_primary_tag_code(StagGoalMap, MainPtag, OtherPtags, MaxSecondary,
 
         % There is a secondary tag, so figure out how to switch on it.
         get_globals(!.CI, Globals),
-        globals.lookup_int_option(Globals, dense_switch_size,
-            DenseSwitchSize),
-        globals.lookup_int_option(Globals, binary_switch_size,
-            BinarySwitchSize),
-        globals.lookup_int_option(Globals, try_switch_size, TrySwitchSize),
+        globals.get_opt_tuple(Globals, OptTuple),
+        DenseSwitchSize = OptTuple ^ ot_dense_switch_size,
+        TrySwitchSize = OptTuple ^ ot_try_switch_size,
+        BinarySwitchSize = OptTuple ^ ot_binary_switch_size,
         ( if MaxSecondary >= DenseSwitchSize then
             SecondaryMethod = jump_table
         else if MaxSecondary >= BinarySwitchSize then
diff --git a/compiler/tupling.m b/compiler/tupling.m
index 52ecc9db4..e28538897 100644
--- a/compiler/tupling.m
+++ b/compiler/tupling.m
@@ -119,6 +119,7 @@
 :- import_module libs.compiler_util.
 :- import_module libs.dependency_graph.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.
 :- import_module ll_backend.call_gen.
@@ -165,8 +166,8 @@ tuple_arguments(!ModuleInfo, !IO) :-
     % we have already read in this file, and if we have, then avoid reading
     % it in again.
     module_info_get_globals(!.ModuleInfo, Globals),
-    globals.lookup_string_option(Globals, tuple_trace_counts_file,
-        TraceCountsFile),
+    globals.get_opt_tuple(Globals, OptTuple),
+    TraceCountsFile = OptTuple ^ ot_tuple_trace_counts_file,
     ( if TraceCountsFile = "" then
         report_warning(Globals, "Warning: --tuple requires " ++
             "--tuple-trace-counts-file to work.\n", !IO)
@@ -196,16 +197,13 @@ warn_trace_counts_error(Globals, TraceCountsFile, Reason, !IO) :-
 tuple_arguments_with_trace_counts(!ModuleInfo, TraceCounts0) :-
     module_info_get_globals(!.ModuleInfo, Globals),
     % We use the same cost options as for the stack optimisation.
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_cv_load_cost, CellVarLoadCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_cv_store_cost, CellVarStoreCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_fv_load_cost, FieldVarLoadCost),
-    globals.lookup_int_option(Globals,
-        optimize_saved_vars_cell_fv_store_cost, FieldVarStoreCost),
-    globals.lookup_int_option(Globals, tuple_costs_ratio, CostsRatio),
-    globals.lookup_int_option(Globals, tuple_min_args, MinArgsToTuple),
+    globals.get_opt_tuple(Globals, OptTuple),
+    CellVarLoadCost = OptTuple ^ ot_opt_svcell_cv_load_cost,
+    CellVarStoreCost = OptTuple ^ ot_opt_svcell_cv_store_cost,
+    FieldVarLoadCost = OptTuple ^ ot_opt_svcell_fv_load_cost,
+    FieldVarStoreCost = OptTuple ^ ot_opt_svcell_fv_store_cost,
+    CostsRatio = OptTuple ^ ot_tuple_costs_ratio,
+    MinArgsToTuple = OptTuple ^ ot_tuple_min_args,
     % These are the costs for untupled variables. We just assume it is
     % the lesser of the cell and field variable costs (usually the field
     % variable costs should be smaller).
diff --git a/compiler/unify_gen_construct.m b/compiler/unify_gen_construct.m
index dfe610489..ac7d1d880 100644
--- a/compiler/unify_gen_construct.m
+++ b/compiler/unify_gen_construct.m
@@ -71,6 +71,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module ll_backend.closure_gen.
 :- import_module ll_backend.code_util.
@@ -314,20 +315,20 @@ generate_and_pack_construct_args([RHSVarWidth | RHSVarsWidths],
             (
                 ArgPosWidth = apw_full(_, _),
                 (
-                    LCMCNull = no,
+                    LCMCNull = do_not_opt_lcmc_null,
                     MaybeNull = no
                 ;
-                    LCMCNull = yes,
+                    LCMCNull = opt_lcmc_null,
                     MaybeNull = yes(const(llconst_int(0)))
                 ),
                 HeadCellArgs = [cell_arg_take_addr_one_word(RHSVar, MaybeNull)]
             ;
                 ArgPosWidth = apw_double(_, _, _),
                 (
-                    LCMCNull = no,
+                    LCMCNull = do_not_opt_lcmc_null,
                     MaybeNulls = no
                 ;
-                    LCMCNull = yes,
+                    LCMCNull = opt_lcmc_null,
                     Null = const(llconst_int(0)),
                     MaybeNulls = yes({Null, Null})
                 ),
diff --git a/compiler/unneeded_code.m b/compiler/unneeded_code.m
index ee2decd8f..21dc0fde5 100644
--- a/compiler/unneeded_code.m
+++ b/compiler/unneeded_code.m
@@ -90,6 +90,7 @@
 :- import_module hlds.vartypes.
 :- import_module libs.
 :- import_module libs.globals.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.goal_path.
@@ -332,7 +333,8 @@ unneeded_process_proc(!ProcInfo, !ModuleInfo, PredId, Pass, Successful) :-
     module_info_get_globals(!.ModuleInfo, Globals),
     globals.lookup_bool_option(Globals, reorder_conj, ReorderConj),
     globals.lookup_bool_option(Globals, fully_strict, FullyStrict),
-    globals.lookup_int_option(Globals, unneeded_code_copy_limit, Limit),
+    globals.get_opt_tuple(Globals, OptTuple),
+    Limit = OptTuple ^ ot_opt_unneeded_code_copy_limit,
     globals.lookup_bool_option(Globals, unneeded_code_debug, Debug),
     Options = uc_option_values(FullyStrict, ReorderConj, Debug, Limit),
     (
diff --git a/compiler/unused_args.m b/compiler/unused_args.m
index 680a76aa2..4adf5e34d 100644
--- a/compiler/unused_args.m
+++ b/compiler/unused_args.m
@@ -106,6 +106,7 @@
 :- import_module libs.
 :- import_module libs.globals.
 :- import_module libs.op_mode.
+:- import_module libs.optimization_options.
 :- import_module libs.options.
 :- import_module mdbcomp.
 :- import_module mdbcomp.sym_name.
@@ -292,9 +293,10 @@ unused_args_process_module(!ModuleInfo, Specs, UnusedArgInfos) :-
     else
         true
     ),
-    globals.lookup_bool_option(Globals, optimize_unused_args, DoFixup),
+    globals.get_opt_tuple(Globals, OptTuple),
+    OptUnusedArgs = OptTuple ^ ot_opt_unused_args,
     (
-        DoFixup = yes,
+        OptUnusedArgs = opt_unused_args,
         list.foldl2(unused_args_create_new_pred(UnusedArgInfo),
             PredProcIdsToFix, ProcCallInfo0, ProcCallInfo, !ModuleInfo),
         % maybe_write_string(VeryVerbose, "% Finished new preds.\n",
@@ -310,7 +312,7 @@ unused_args_process_module(!ModuleInfo, Specs, UnusedArgInfos) :-
             module_info_clobber_dependency_info(!ModuleInfo)
         )
     ;
-        DoFixup = no
+        OptUnusedArgs = do_not_opt_unused_args
     ).
 
 %-----------------------------------------------------------------------------%
@@ -435,14 +437,14 @@ setup_proc_args(PredId, ProcId, !VarUsage, !PredProcIds, !OptProcs,
                     initialise_vardep(UnusedVars, !.VarDep, VarDep),
                     PredProcId = proc(PredId, ProcId),
                     map.set(PredProcId, VarDep, !VarUsage),
-                    globals.lookup_bool_option(Globals,
-                        optimize_unused_args, OptimizeUnusedArgs),
+                    globals.get_opt_tuple(Globals, OptTuple),
+                    OptUnusedArgs = OptTuple ^ ot_opt_unused_args,
                     (
-                        OptimizeUnusedArgs = yes,
+                        OptUnusedArgs = opt_unused_args,
                         make_imported_unused_args_pred_info(PredProcId,
                             UnusedArgs, !OptProcs, !ModuleInfo)
                     ;
-                        OptimizeUnusedArgs = no
+                        OptUnusedArgs = do_not_opt_unused_args
                     )
                 ;
                     UnusedArgs = []
diff --git a/library/getopt_template b/library/getopt_template
index a75812a3f..760ea4aab 100644
--- a/library/getopt_template
+++ b/library/getopt_template
@@ -920,8 +920,9 @@ process_arguments(OptionOps, [Arg0 | Args0], NonOptionArgs, !OptionArgs,
 handle_long_option(OptionOps, Option, Flag, OptionData, MaybeOptionArg0,
         Args0, Args, !OptionArgs, MaybeError, !OptionTable,
         !OptionsSet, !UserData, !IO) :-
+    need_arg(OptionData, NeedArg),
     ( if
-        need_arg(OptionData, yes),
+        NeedArg = yes,
         MaybeOptionArg0 = no
     then
         (
@@ -949,7 +950,7 @@ handle_long_option(OptionOps, Option, Flag, OptionData, MaybeOptionArg0,
     ;
         MissingArg = no,
         ( if
-            need_arg(OptionData, no),
+            NeedArg = no,
             MaybeOptionArg = yes(ArgVal)
         then
             Args = Args0,
@@ -1207,15 +1208,21 @@ process_option(OptionOps, OptionData, Option, Flag, MaybeArg,
                         FileMaybeError, !.OptionTable, NewOptionTable,
                         !OptionsSet, !UserData, !IO),
                     (
-                        Args = [],
-                        !:OptionTable = NewOptionTable,
-                        MaybeError = FileMaybeError
+                        FileMaybeError = no,
+                        (
+                            Args = [],
+                            !:OptionTable = NewOptionTable,
+                            MaybeError = no
+                        ;
+                            Args = [_ | _],
+                            Reason =
+                                file_special_contains_non_option_args(FileName),
+                            Error = option_error(Flag, Option, Reason),
+                            MaybeError = yes(Error)
+                        )
                     ;
-                        Args = [_ | _],
-                        Reason =
-                            file_special_contains_non_option_args(FileName),
-                        Error = option_error(Flag, Option, Reason),
-                        MaybeError = yes(Error)
+                        FileMaybeError = yes(_),
+                        MaybeError = FileMaybeError
                     )
                 ;
                     ReadRes = error(_, IO_Error),
diff --git a/tools/make_optimization_options b/tools/make_optimization_options
index e69de29bb..3eb753240 100755
--- a/tools/make_optimization_options
+++ b/tools/make_optimization_options
@@ -0,0 +1,31 @@
+#!/bin/sh
+# vim: ts=4 sw=4 et ft=sh
+#---------------------------------------------------------------------------#
+# Copyright (C) 2020 The Mercury team.
+# This file may only be copied under the terms of the GNU General
+# Public License - see the file COPYING in the Mercury distribution.
+#---------------------------------------------------------------------------#
+#
+# This script constructs (or reconstructs) the contents of
+# compiler/optimization_options.m. The initial and final parts
+# are hand_written, in make_optimization_options_{start,end}
+# respectively, while the middle is generated by
+# make_optimization_options_middle from make_optimization_options_db.
+#
+# make_optimization_options_middle also generates handler_file,
+# which is a code fragment that is intended to comprise part of
+# the code of the special_handler predicate in compiler/options.m.
+#
+
+if test -f ../compiler/optimization_options.m
+then
+    chmod u+w ../compiler/optimization_options.m
+fi
+
+(
+cat make_optimization_options_start;
+awk -f make_optimization_options_middle make_optimization_options_db;
+cat make_optimization_options_end
+) > ../compiler/optimization_options.m
+
+chmod a-w ../compiler/optimization_options.m
diff --git a/tools/make_optimization_options_db b/tools/make_optimization_options_db
index e69de29bb..c7e42374c 100644
--- a/tools/make_optimization_options_db
+++ b/tools/make_optimization_options_db
@@ -0,0 +1,123 @@
+bool   y    allow_inlining
+bool   n    inline_simple                                  
+bool   y    inline_builtins                                
+bool   n    inline_single_use                              
+int    0    inline_call_cost                               
+int    0    inline_compound_threshold                      
+int    5    inline_simple_threshold                        
+int    100  inline_vars_threshold                          
+int    5    intermod_inline_simple_threshold               
+bool   n    inline_linear_tail_rec_sccs                    
+int    0    inline_linear_tail_rec_sccs_max_extra          
+int    5    from_ground_term_threshold                     
+bool   y    enable_const_struct                            
+bool   n    common_struct                                  opt_common_structs
+bool   n    constraint_propagation                         prop_constraints
+bool   n    local_constraint_propagation                   prop_local_constraints
+bool   n    optimize_duplicate_calls                       opt_dup_calls
+bool   n    constant_propagation                           prop_constants
+bool   n    excess_assign                                  elim_excess_assigns
+bool   n    test_after_switch                              opt_test_after_switch
+bool   y    optimize_format_calls                          opt_format_calls
+bool   n    loop_invariants                                opt_loop_invariants
+bool   n    optimize_saved_vars_const                      opt_saved_vars_const
+bool   n    optimize_saved_vars_cell                       opt_svcell
+bool   y    optimize_saved_vars_cell_loop                  opt_svcell_loop
+bool   y    optimize_saved_vars_cell_full_path             opt_svcell_full_path
+bool   y    optimize_saved_vars_cell_on_stack              opt_svcell_on_stack
+bool   y    optimize_saved_vars_cell_candidate_headvars    opt_svcell_candidate_headvars
+int    3    optimize_saved_vars_cell_cv_store_cost         opt_svcell_cv_store_cost
+int    1    optimize_saved_vars_cell_cv_load_cost          opt_svcell_cv_load_cost
+int    1    optimize_saved_vars_cell_fv_store_cost         opt_svcell_fv_store_cost
+int    1    optimize_saved_vars_cell_fv_load_cost          opt_svcell_fv_load_cost
+int    100  optimize_saved_vars_cell_op_ratio              opt_svcell_op_ratio
+int    100  optimize_saved_vars_cell_node_ratio            opt_svcell_node_ratio 
+int    100  optimize_saved_vars_cell_all_path_node_ratio   opt_svcell_all_path_node_ratio
+bool   n    optimize_saved_vars_cell_include_all_candidates opt_svcell_all_candidates
+bool   n    delay_construct                                delay_constructs
+bool   n    follow_code                                    opt_follow_code
+bool   n    optimize_unused_args                           opt_unused_args
+bool   n    intermod_unused_args                           opt_unused_args_intermod
+bool   n    optimize_higher_order                          opt_higher_order
+int    20   higher_order_size_limit                        
+int    10   higher_order_arg_limit                         
+bool   n    unneeded_code                                  opt_unneeded_code
+int    10   unneeded_code_copy_limit                       opt_unneeded_code_copy_limit
+bool   n    type_specialization                            spec_types
+bool   n    user_guided_type_specialization                spec_types_user_guided
+bool   n    introduce_accumulators                         
+bool   n    optimize_constructor_last_call_accumulator     opt_lcmc_accumulator
+bool   n    optimize_constructor_last_call_null            opt_lcmc_null
+bool   n    optimize_constructor_last_call                 opt_lcmc
+bool   n    optimize_dead_procs                            opt_dead_procs
+bool   n    deforestation                                  deforest
+int    4    deforestation_depth_limit                      
+int    1000 deforestation_cost_factor                      
+int    200  deforestation_vars_threshold                   
+int    15   deforestation_size_threshold                   
+bool   n    untuple                                        
+bool   n    tuple                                          
+string -    tuple_trace_counts_file                        
+int    100  tuple_costs_ratio                              
+int    4    tuple_min_args                                 
+bool   n    inline_par_builtins                            
+bool   n    always_specialize_in_dep_par_conjs             spec_in_all_dep_par_conjs
+bool   y    allow_some_paths_only_waits                    
+bool   n    region_analysis                                analyse_regions
+bool   n    smart_indexing                                 use_smart_indexing
+bool   y    smart_atomic_indexing                          use_smart_indexing_atomic
+bool   y    smart_string_indexing                          use_smart_indexing_string
+bool   y    smart_tag_indexing                             use_smart_indexing_tag
+bool   y    smart_float_indexing                           use_smart_indexing_float
+int    25   dense_switch_req_density                       
+int    25   lookup_switch_req_density                      
+int    4    dense_switch_size                              
+int    4    lookup_switch_size                             
+int    16   string_trie_switch_size                        
+int    8    string_hash_switch_size                        
+int    4    string_binary_switch_size                      
+int    3    tag_switch_size                                
+int    3    try_switch_size                                
+int    4    binary_switch_size                             
+bool   n    switch_single_rec_base_first                   put_base_first_single_rec
+bool   y    switch_multi_rec_base_first                    put_base_first_multi_rec
+bool   n    static_ground_cells                            use_static_ground_cells
+bool   n    static_ground_floats                           use_static_ground_floats
+bool   n    static_ground_int64s                           use_static_ground_int64s
+bool   n    static_code_addresses                          use_static_code_addresses
+bool   n    use_atomic_cells                               
+bool   n    middle_rec                                     opt_middle_rec
+bool   n    simple_neg                                     opt_simple_neg
+bool   y    allow_hijacks                                  
+bool   n    optimize_tailcalls                             opt_mlds_tailcalls
+bool   n    optimize_initializations                       opt_initializations
+bool   y    eliminate_unused_mlds_assigns                  elim_unused_mlds_assigns
+bool   n    eliminate_local_vars                           elim_local_vars
+bool   y    generate_trail_ops_inline                      gen_trail_ops_inline
+bool   n    common_data                                    use_common_data
+bool   y    common_layout_data                             use_common_layout_data
+bool   n    optimize                                       
+bool   n    optimize_peep                                  opt_peep
+bool   n    optimize_peep_mkword                           opt_peep_mkword
+bool   n    optimize_jumps                                 opt_jumps
+bool   n    optimize_fulljumps                             opt_fulljumps
+bool   n    pessimize_tailcalls                            
+bool   n    checked_nondet_tailcalls                       opt_checked_nondet_tailcalls
+bool   n    use_local_vars                                 
+int    2    local_var_access_threshold                     
+bool   n    standardize_labels                             
+bool   n    optimize_labels                                opt_labels
+bool   n    optimize_dups                                  opt_dups
+bool   n    optimize_proc_dups                             opt_proc_dups
+bool   n    optimize_frames                                opt_frames
+bool   n    optimize_delay_slot                            opt_delay_slot
+bool   n    optimize_reassign                              opt_reassign
+int    0    optimize_repeat                                opt_repeat
+int    4000 layout_compression_limit                       
+bool   n    use_macro_for_redo_fail                        
+bool   n    emit_c_loops                                   
+int    1    procs_per_c_function                           
+bool   y    local_thread_engine_base                       use_local_thread_engine_base
+bool   n    erlang_switch_on_strings_as_atoms              switch_on_strings_as_atoms
+bool   n    inline_alloc
+bool   n    c_optimize                                     opt_c
diff --git a/tools/make_optimization_options_end b/tools/make_optimization_options_end
index e69de29bb..14a639ff5 100644
--- a/tools/make_optimization_options_end
+++ b/tools/make_optimization_options_end
@@ -0,0 +1,157 @@
+
+:- pred set_opts_upto_level(option_table::in, int::in, int::in,
+    opt_tuple::in, opt_tuple::out,
+    maybe_seen_opt_level::in, maybe_seen_opt_level::out) is det.
+
+set_opts_upto_level(OptionTable, Cur, Max,
+        !OptTuple, !MaybeSeenOptLevel) :-
+    ( if
+        Cur =< Max,
+        opts_enabled_at_level(Cur, LevelOptOptions)
+    then
+        list.foldl2(update_opt_tuple(OptionTable), LevelOptOptions,
+            !OptTuple, !MaybeSeenOptLevel),
+        set_opts_upto_level(OptionTable, Cur + 1, Max,
+            !OptTuple, !MaybeSeenOptLevel)
+    else
+        true
+    ).
+
+:- pred set_opts_for_space(opt_tuple::in, opt_tuple::out) is det.
+
+set_opts_for_space(!OptTuple) :-
+    UnneededCopyLimit = !.OptTuple ^ ot_opt_unneeded_code_copy_limit,
+    !OptTuple ^ ot_opt_unneeded_code_copy_limit :=
+        int.min(UnneededCopyLimit, 1),
+    !OptTuple ^ ot_opt_dead_procs := opt_dead_procs,
+    !OptTuple ^ ot_opt_labels := opt_labels,
+    !OptTuple ^ ot_opt_dups := opt_dups,
+    !OptTuple ^ ot_opt_proc_dups := opt_proc_dups,
+    !OptTuple ^ ot_opt_fulljumps := opt_fulljumps,
+    !OptTuple ^ ot_opt_reassign := opt_reassign,
+    !OptTuple ^ ot_inline_alloc := inline_alloc,
+    !OptTuple ^ ot_use_macro_for_redo_fail := use_macro_for_redo_fail,
+    !OptTuple ^ ot_opt_loop_invariants := do_not_opt_loop_invariants.
+
+:- pred opts_enabled_at_level(int::in, list(optimization_option)::out)
+    is semidet.
+
+opts_enabled_at_level(0, [
+    % Optimization level 0: aim to minimize overall compilation time.
+    oo_use_common_data(yes),
+    oo_optimize(yes),
+    oo_opt_repeat(1),
+    oo_opt_peep(yes),
+    oo_opt_peep_mkword(yes),
+    oo_use_static_ground_cells(yes),
+    oo_use_smart_indexing(yes),
+    oo_opt_jumps(yes),
+    oo_opt_labels(yes),
+    oo_opt_dead_procs(yes),
+    oo_elim_excess_assigns(yes)
+]).
+opts_enabled_at_level(1, [
+    % Optimization level 1: apply optimizations which are cheap and
+    % have a good payoff while still keeping compilation time small.
+    oo_use_local_vars(yes),
+    oo_opt_c(yes),              % XXX We want `gcc -O1'
+    oo_opt_frames(yes),
+    oo_opt_delay_slot(yes),
+    oo_opt_middle_rec(yes),
+    oo_emit_c_loops(yes),
+    oo_opt_mlds_tailcalls(yes)
+]).
+opts_enabled_at_level(2, [
+    % Optimization level 2: apply optimizations which have a good payoff
+    % relative to their cost; but include optimizations which are
+    % more costly than with -O1.
+    oo_opt_fulljumps(yes),
+    oo_opt_repeat(3),
+    oo_opt_dups(yes),
+    oo_opt_follow_code(yes),
+    oo_inline_simple(yes),
+    oo_inline_single_use(yes),
+    oo_inline_compound_threshold(10),
+    oo_opt_common_structs(yes),
+    oo_spec_types_user_guided(yes),
+    oo_opt_simple_neg(yes),
+    oo_opt_test_after_switch(yes),
+    oo_opt_initializations(yes)
+]).
+opts_enabled_at_level(3, [
+    % Optimization level 3: apply optimizations which usually have a good
+    % payoff even if they increase compilation time quite a bit.
+    oo_opt_saved_vars_const(yes),
+    oo_opt_unused_args(yes),
+    oo_opt_higher_order(yes),
+    oo_deforest(yes),
+    oo_prop_constraints(yes),
+    oo_prop_local_constraints(yes),
+    oo_opt_reassign(yes),
+    oo_opt_repeat(4)
+]).
+opts_enabled_at_level(4, [
+    % Optimization level 4: apply optimizations which may have some payoff
+    % even if they increase compilation time quite a bit.
+    %
+    % Currently this enables the use of local variables
+    % and increases the inlining thresholds.
+    oo_inline_simple_threshold(8),
+    oo_inline_compound_threshold(20),
+    oo_higher_order_size_limit(30)
+]).
+opts_enabled_at_level(5, [
+    % Optimization level 5: apply optimizations which may have some
+    % payoff even if they increase compilation time a lot.
+    %
+    % Currently this enables the search for construction unifications that
+    % can be delayed past failing computations, allows more passes of the
+    % low-level optimizations, and increases the inlining thresholds
+    % still further. We also enable eliminate_local_vars only at
+    % this level, because that pass is implemented pretty inefficiently.
+    oo_opt_repeat(5),
+    oo_delay_constructs(yes),
+    oo_inline_compound_threshold(100),
+    oo_higher_order_size_limit(40),
+    oo_elim_local_vars(yes),
+    oo_opt_loop_invariants(yes)
+]).
+opts_enabled_at_level(6, [
+     % Optimization level 6: apply optimizations which may have any payoff
+    % even if they increase compilation time to completely unreasonable
+    % levels.
+    %
+    % Currently this sets `everything_in_one_c_function', which causes
+    % the compiler to put everything in the one C function and treat calls
+    % to predicates in the same module as local. We also enable inlining
+    % of GC_malloc(), redo(), and fail().
+    oo_procs_per_c_function(0),
+    oo_inline_alloc(yes),
+    oo_use_macro_for_redo_fail(yes)
+]).
+
+% The following optimization options are not enabled at any level:
+% XXX This comment is old. Since then, we have added many optimizations
+% that are not enabled automatically at any level :-(
+%
+%   checked_nondet_tailcalls:
+%       This is deliberate, because the transformation might make
+%       code run slower.
+%
+%   unneeded_code:
+%       Because it can cause slowdowns at high optimization levels;
+%       cause unknown
+%
+%   type_specialization:
+%       XXX why not?
+%
+%   introduce_accumulators:
+%       XXX Disabled until a bug in extras/trailed_update/var.m
+%       is resolved.
+%
+%   optimize_constructor_last_call:
+%       Not a speedup in general.
+
+%---------------------------------------------------------------------------%
+:- end_module libs.optimization_options.
+%---------------------------------------------------------------------------%
diff --git a/tools/make_optimization_options_middle b/tools/make_optimization_options_middle
index e69de29bb..ca250ea69 100755
--- a/tools/make_optimization_options_middle
+++ b/tools/make_optimization_options_middle
@@ -0,0 +1,293 @@
+#!/usr/bin/awk -f
+# vim: ts=4 sw=4 et ft=awk
+#---------------------------------------------------------------------------#
+# Copyright (C) 2020 The Mercury team.
+# This file may only be copied under the terms of the GNU General
+# Public License - see the file COPYING in the Mercury distribution.
+#---------------------------------------------------------------------------#
+#
+# The input to this file, make_optimization_options_db, has one line
+# for every optimization option. Each line must have either three
+# or four fields.
+#
+# - The first field indicates what kind of option this is. It must be
+#   one of "bool", "int" and "string".
+#
+# - The second field gives the initial value of the option.
+#
+#   - If the option is boolean, the initial value must be either "n" or "y".
+#
+#   - If the option is an integer, the initial value must be an integer.
+#
+#   - If the option is a strin, the initial value will be set
+#     to the empty string regardless of the contents of the second field.
+#
+# - The third field is the name of the option in options.m.
+#
+# - The fourth field, if it exists, gives the name of the option
+#   in the file this script helps generate, optimization_options.m.
+#   If there is no fourth field, the name is taken to be the same
+#   as the name in the third field.
+#
+#   For boolean options, the name in optimization_options.m should be a verb,
+#   because when we generate a bespoke type for it, its two function symbols
+#   will be named "verb" and "do_not_verb".
+#
+
+BEGIN {
+        default_opt_level = 2;
+        next_bool_opt = 0;
+        next_int_opt = 0;
+        next_string_opt = 0;
+    }
+    {
+        if (NF == 3 || NF == 4) {
+            if ($1 == "bool") {
+                bool_opts[next_bool_opt] = $3;
+                if (NF == 4) {
+                    bool_field_names[next_bool_opt] = $4;
+                } else {
+                    bool_field_names[next_bool_opt] = $3;
+                }
+                if ($2 == "y") {
+                    bool_defaults[next_bool_opt] = bool_field_names[next_bool_opt];
+                } else if ($2 == "n") {
+                    bool_defaults[next_bool_opt] = "do_not_" bool_field_names[next_bool_opt];
+                } else {
+                    printf("ERROR: bad bool default<%s>\n", $0);
+                }
+                next_bool_opt++;
+            } else if ($1 == "int") {
+                int_opts[next_int_opt] = $3;
+                if (NF == 4) {
+                    int_field_names[next_int_opt] = $4;
+                } else {
+                    int_field_names[next_int_opt] = $3;
+                }
+                int_defaults[next_int_opt] = $2 + 0;
+                next_int_opt++;
+            } else if ($1 == "string") {
+                string_opts[next_string_opt] = $3;
+                if (NF == 4) {
+                    string_field_names[next_string_opt] = $4;
+                } else {
+                    string_field_names[next_string_opt] = $3;
+                }
+                # Ignore the default; strings always default to the empty string.
+                next_string_opt++;
+            } else {
+                printf("ERROR: line with unknown type: <%s>\n", $0);
+            }
+        } else {
+            printf("ERROR: line with %d fields: <%s>\n", NF, $0);
+        }
+    }
+END {
+        switch_type_name = "optimization_option";
+        switch_var_name = "OptOption";
+        record_type_name = "opt_tuple";
+        record_var_name = "OptTuple";
+
+        for (i = 0; i < next_bool_opt; i++) {
+            printf(":- type maybe_%s\n", bool_field_names[i]);
+            printf("    --->    %s\n", bool_field_names[i]);
+            printf("    ;       do_not_%s.\n", bool_field_names[i]);
+        }
+
+        printf("\n%%---------------------%%\n\n");
+
+        printf(":- type %s\n", switch_type_name);
+        for (i = 0; i < next_bool_opt; i++) {
+            prefix=";";
+            suffix="";
+
+            if (i == 0)
+                prefix = "--->";
+
+            printf("%-4s%-4s%-4soo_%s(bool)%s\n", "", prefix, "", bool_field_names[i], suffix);
+        }
+        for (i = 0; i < next_int_opt; i++) {
+            prefix=";";
+            suffix="";
+            printf("%-4s%-4s%-4soo_%s(int)%s\n", "", prefix, "", int_field_names[i], suffix);
+        }
+        for (i = 0; i < next_string_opt; i++) {
+            prefix=";";
+            suffix="";
+
+            printf("%-4s%-4s%-4soo_%s(string)%s\n", "", prefix, "", string_field_names[i], suffix);
+        }
+        printf("%-4s%-4s%-4soo_opt_level(int)\n", "", prefix, "");
+        printf("%-4s%-4s%-4soo_opt_for_space.\n", "", prefix, "");
+
+        printf("\n%%---------------------%%\n\n");
+
+        printf(":- type %s\n", record_type_name);
+        printf("%-4s--->%-4s%s(\n", "", "", record_type_name);
+        for (i = 0; i < next_bool_opt; i++) {
+            suffix = ",";
+            printf("%-16sot_%-26s :: maybe_%s%s\n", "", bool_field_names[i], bool_field_names[i], suffix);
+        }
+        for (i = 0; i < next_int_opt; i++) {
+            suffix = ",";
+            printf("%-16sot_%-26s :: int%s\n", "", int_field_names[i], suffix);
+        }
+        for (i = 0; i < next_string_opt; i++) {
+            suffix = ",";
+            if (i == next_string_opt-1)
+                suffix = "";
+
+            printf("%-16sot_%-26s :: string%s\n", "", string_field_names[i], suffix);
+        }
+        printf("%-12s).\n", "");
+
+        printf("\n");
+        printf(":- pred process_optimization_options(option_table::in,\n");
+        printf("%-4slist(%s)::in, %s::out) is det.\n",
+            "", switch_type_name, record_type_name);
+
+        printf("\n");
+        printf("%%---------------------------------------------------------------------------%%\n");
+        printf("\n");
+
+        printf(":- implementation.\n");
+
+        printf("\n");
+        printf(":- import_module int.\n");
+
+        printf("\n%%---------------------%%\n\n");
+
+        printf("process_optimization_options(OptionTable, OptOptions, !:OptTuple) :-\n");
+        printf("%-4s!:OptTuple = init_opt_tuple,\n", "");
+        printf("%-4slist.foldl2(update_opt_tuple(OptionTable), OptOptions,\n", "");
+        printf("%-8s!OptTuple, not_seen_opt_level, MaybeSeenOptLevel),\n", "");
+        printf("%-4s(\n", "");
+        printf("%-8sMaybeSeenOptLevel = not_seen_opt_level,\n", "");
+        printf("%-8sset_opts_upto_level(OptionTable, 0, %d,\n",
+            "", default_opt_level);
+        printf("%-12s!OptTuple, MaybeSeenOptLevel, _)\n", "");
+        printf("%-4s;\n", "");
+        printf("%-8sMaybeSeenOptLevel = seen_opt_level\n", "");
+        printf("%-4s).\n", "");
+
+        printf("\n%%---------------------%%\n");
+
+        printf("\n");
+        printf(":- func init_opt_tuple = %s.\n\n", record_type_name);
+
+        printf("init_opt_tuple = %s(\n", record_type_name);
+        for (i = 0; i < next_bool_opt; i++) {
+            printf("%-8s%s,\n", "", bool_defaults[i]);
+        }
+        for (i = 0; i < next_int_opt; i++) {
+            printf("%-8s%s,\n", "", int_defaults[i]);
+        }
+        for (i = 0; i < next_string_opt; i++) {
+            suffix = ",";
+            if (i == next_string_opt-1) {
+                suffix = "";
+            }
+            printf("%-8s%s%s\n", "", "\"\"", suffix);
+        }
+        printf("%-4s).\n", "");
+
+        printf("\n%%---------------------%%\n");
+
+        printf("\n");
+        printf(":- type maybe_seen_opt_level\n");
+        printf("%-4s%-8snot_seen_opt_level\n", "", "--->");
+        printf("%-4s%-8sseen_opt_level.\n", "", ";");
+
+        printf("\n");
+        printf(":- pred update_opt_tuple(option_table::in, %s::in,\n",
+            switch_type_name);
+        printf("%-4s%s::in, %s::out,\n",
+            "", record_type_name, record_type_name);
+        printf("%-4smaybe_seen_opt_level::in, maybe_seen_opt_level::out) is det.\n", "");
+
+        printf("update_opt_tuple(OptionTable, %s, !%s, !MaybeSeenOptLevel) :-\n",
+            switch_var_name, record_var_name);
+        printf("%-4srequire_complete_switch [%s]\n", "", switch_var_name);
+        lparen_or_semi = "(";
+        for (i = 0; i < next_bool_opt; i++) {
+            printf("%-4s%s\n", "", lparen_or_semi);
+            lparen_or_semi = ";";
+            printf("%-8s%s = oo_%s(Bool),\n", "", switch_var_name, bool_field_names[i]);
+            printf("%-8s( if\n", "");
+            printf("%-12sBool = yes,\n", "");
+            if (bool_field_names[i] == "opt_delay_slot") {
+                printf("%-12sgetopt_io.lookup_bool_option(OptionTable, have_delay_slot, yes),\n",
+                    "");
+            }
+            printf("%-12s!.%s ^ ot_%s = do_not_%s\n",
+                "", record_var_name, bool_field_names[i], bool_field_names[i]);
+            printf("%-8sthen\n", "");
+            printf("%-12s!%s ^ ot_%s := %s\n",
+                "", record_var_name, bool_field_names[i], bool_field_names[i]);
+            printf("%-8selse\n", "");
+            printf("%-12strue\n", "");
+            printf("%-8s)\n", "");
+        }
+        for (i = 0; i < next_int_opt; i++) {
+            printf("%-4s%s\n", "", lparen_or_semi);
+            lparen_or_semi = ";";
+            printf("%-8s%s = oo_%s(N),\n", "", switch_var_name, int_field_names[i]);
+            printf("%-8sOldN = !.%s ^ ot_%s,\n", "", record_var_name, int_field_names[i]);
+            printf("%-8s!%s ^ ot_%s := int.max(OldN, N)\n", "", record_var_name, int_field_names[i]);
+        }
+        for (i = 0; i < next_string_opt; i++) {
+            printf("%-4s%s\n", "", lparen_or_semi);
+            lparen_or_semi = ";";
+            printf("%-8s%s = oo_%s(Str),\n", "", switch_var_name, string_field_names[i]);
+            printf("%-8s!%s ^ ot_%s := Str\n", "", record_var_name, string_field_names[i]);
+        }
+        printf("%-4s%s\n", "", lparen_or_semi);
+        lparen_or_semi = ";";
+        printf("%-8s%s = oo_opt_level(OptLevel),\n", "", switch_var_name);
+        printf("%-8sset_opts_upto_level(OptionTable, 0, OptLevel, !%s, !.MaybeSeenOptLevel, _),\n",
+            "", record_var_name);
+        printf("%-8s!:MaybeSeenOptLevel = seen_opt_level\n", "");
+
+        printf("%-4s%s\n", "", lparen_or_semi);
+        printf("%-8s%s = oo_opt_for_space,\n", "", switch_var_name);
+        printf("%-8sset_opts_for_space(!%s)\n", "", record_var_name);
+        printf("%-4s).\n", "");
+
+        # The code from here to the end of the script generates code
+        # that should be included in the special_handler predicate in options.m.
+
+        handler_file = "handler_file";
+        arm_start = "(";
+        for (i = 0; i < next_bool_opt; i++) {
+            printf("%-8s%s\n", "", arm_start) > handler_file;
+            arm_start = ";";
+            printf("%-12sOption = optopt_%s,\n", "", bool_opts[i]) > handler_file;
+            printf("%-12sSpecialData = bool(Bool),\n", "") > handler_file;
+            printf("%-12sOptOption = oo_%s(Bool)\n", "", bool_field_names[i]) > handler_file;
+        }
+        for (i = 0; i < next_int_opt; i++) {
+            printf("%-8s;\n", "") > handler_file;
+            printf("%-12sOption = optopt_%s,\n", "", int_opts[i]) > handler_file;
+            printf("%-12sSpecialData = int(N),\n", "") > handler_file;
+            printf("%-12sOptOption = oo_%s(N)\n", "", int_field_names[i]) > handler_file;
+        }
+        for (i = 0; i < next_string_opt; i++) {
+            printf("%-8s;\n", "") > handler_file;
+            printf("%-12sOption = optopt_%s,\n", "", string_opts[i]) > handler_file;
+            printf("%-12sSpecialData = string(Str),\n", "") > handler_file;
+            printf("%-12sOptOption = oo_%s(Str)\n", "", string_field_names[i]) > handler_file;
+        }
+
+        printf("%-8s;\n", "") > handler_file;
+        printf("%-12sOption = optopt_level,\n", "") > handler_file;
+        printf("%-12sSpecialData = string(N),\n", "") > handler_file;
+        printf("%-12sOptOption = oo_opt_level(N)\n", "") > handler_file;
+
+        printf("%-8s;\n", "") > handler_file;
+        printf("%-12sOption = optopt_space,\n", "") > handler_file;
+        printf("%-12sSpecialData = none,\n", "") > handler_file;
+        printf("%-12sOptOption = oo_opt_for_space\n", "") > handler_file;
+
+        printf("%-8s),\n", "") > handler_file;
+        printf("%-8scord.snoc(OptOption, !OptOptions)\n", "") > handler_file;
+    }
diff --git a/tools/make_optimization_options_start b/tools/make_optimization_options_start
index e69de29bb..fe2c0b2d6 100644
--- a/tools/make_optimization_options_start
+++ b/tools/make_optimization_options_start
@@ -0,0 +1,28 @@
+%---------------------------------------------------------------------------%
+% vim: ft=mercury ts=4 sw=4 et
+%---------------------------------------------------------------------------%
+% Copyright (C) 2020 The Mercury team.
+% This file may only be copied under the terms of the GNU General
+% Public License - see the file COPYING in the Mercury distribution.
+%---------------------------------------------------------------------------%
+%
+% File: optimization_options.m
+% Main author: zs, via tools/make_optimization_options
+%
+% This module manages optimization options. It defines a type, the opt_tuple,
+% that allows quick access to them, sets their default values, and implements
+% the turning on of optimizations when a given optimization level is selected.
+%
+
+:- module libs.optimization_options.
+:- interface.
+
+:- import_module libs.options.
+
+:- import_module bool.
+:- import_module getopt_io.
+:- import_module list.
+:- import_module map.
+
+%---------------------------------------------------------------------------%
+
